--- Module/script executed from discovery.nse to make next stage actions
--  Used to execute from non Lua 5.1 interpreter and as module for Lua 5.1
--
-- @file    discovery.lua 
-- @author  Vyacheslav Khudan
-- @created 2017/10/05

logger = require"core.logger":init("[WMS 3.0]", 3)

local discovery = {}

-------------------------------------------------------------------------
--- Implementation details
-------------------------------------------------------------------------
--- Builder structure for module lookup and executing rules
local function __initModuleBase()
    -- initialize modules base structure
    local methods = {
         "hints"
        ,"verify"
        ,"provision"
        ,"setuser"
        ,"setfeatures"
    }
    local modules = {
       "wildixair",
       "wildixp",
       "wildixp2",
       "wildixgw",
       "wildixfxs",
       "wildixfxs2",
       "wildixfxo",
       "wildixfxo3",
       "wildixgsm",
       "wildixpa",
       "daysaver",
    }
    local base = {}
    for _, modname in ipairs(modules) do
        local mr = require(modname)
        for i=1,#methods do
            local method          = methods[i]
            base[method]          = base[method] or {}
            base[method][modname] = mr[method] 
        end
    end
    
    return base
end

local function __detectModule(modBase, hintResult, scanData)
    if not modBase or not hintResult or not scanData then
        return nil
    end
    
    local devInfo       = nil
    local devModuleName = nil
    
    local devHintHeaders        = hintResult.headers or {}
          devHintHeaders.status = tostring(hintResult.code)
          devHintHeaders.scode  = hintResult.scode
    
    -- Module detecting by device response headers
    for modName, modHintList in pairs(modBase.hints) do
        for _, modHint in ipairs(modHintList) do
            local totallyMatched = true
            for modHintItem, modHintItemValue in pairs(modHint) do
                if not devHintHeaders[modHintItem] or 
                   (devHintHeaders[modHintItem] ~= modHintItemValue and
                   not devHintHeaders[modHintItem]:match(modHintItemValue)) then
                   totallyMatched = false
                   break
                end
            end
            if totallyMatched then
                devModuleName = modName
                break
            end
        end
        if devModuleName then
            break
        end
    end
    if not devModuleName then
        return nil
    end

    -- processing found module to get detailed information
    _G.fullgconf = require"core.pbxutils".tcopy(gconf)
    _G.gconf = gconf[devModuleName] or {}
    local devInfo = modBase.verify[devModuleName](scanData.host, scanData.port, scanData.authMethod)
    if "table" == type(devInfo) then  -- normalize device information
        devInfo.vendor = devInfo.vendor or "wildix"
        devInfo.model  = devInfo.model or ""
        
        -- update host data by device information
        for k, v in pairs(devInfo) do
            if not scanData.host[k] or "" == scanData.host[k] then
                scanData.host[k] = v
            end
        end
    end  
    
    return devModuleName, devInfo
end

local function __executeModuleProvision(modBase, modName, scanData, deviceInfo)
    local rv = false
    if not modBase or not modName or 
       not scanData or not deviceInfo then
        return rv
    end
    
    repeat
        if not scanData.args.action then
            rv = true   -- skip this procedure for empty action
            break
        end
    
        if not deviceInfo.auth then
            break         -- not found authorization
        end
        
        scanData.args.authMethod = scanData.authMethod
        modBase[scanData.args.action][modName](scanData.host, scanData.port, scanData.args)
        rv = true
    until true
    
    return rv
end

-- WMS-9491
local function __prepareConfigLog(config)
    if "table" ~= type(config) then
        return {}
    end

    config.general = nil

    if config.wildixair and config.wildixair["w-air-sb-pin"] then
        config.wildixair["w-air-sb-pin"] = "XXXX"
    end

    return config
end

local function __makeOutput(scanData, deviceInfo)
    if not scanData or not deviceInfo then
        return nil
    end
    
    -- display device information 
    logger:debug({
        ["token_type"] = deviceInfo["token_type"] or "",
        ["token"]      = deviceInfo["token"] or "",
        ["auth"]       = deviceInfo["auth"] or "",
        ["cookie"]     = deviceInfo["cookie"] or "",
    }, "information")
    
    -- clear authorization data for output
    deviceInfo["token_type"] = nil
    deviceInfo["token"]      = nil
    deviceInfo["cookie"]     = nil
    
    -- processing output
    local output = ""
    if deviceInfo.auth then
        deviceInfo.auth = nil
        output = string.format("\"%s\":%s", scanData.host.ip, require"dkjson".encode(deviceInfo))
    end
    output = 0 ~= #output and output or nil 

    if output then
        if scanData.args.action then
            deviceInfo["host_ip"] = scanData.host.ip
            logger:info(deviceInfo, "information")
            logger:info(__prepareConfigLog(fullgconf), "gconfig")
        else
            logger:debug(deviceInfo, "information")
        end
    end
    
    if output then
        io.stderr:write(output .. "\n" )
    end
    
    return output
end

local function __isDigestAuth(hintResult)
    local devHintHeaders = hintResult.headers or {}
    local auth = "basic"
    if devHintHeaders["www-authenticate"] and devHintHeaders["www-authenticate"]:find("Digest") then
        auth = "digest"
    end
    return auth
end
-------------------------------------------------------------------------
--- Public interface
-------------------------------------------------------------------------
local meta      = { -- meta class
    args = nil  -- scanning results
}

--- Set up arguments
function meta:setArgs(data)
    local args = nil

    -- processing input data
    if "table" == type(data) then
        args = data
    elseif discovery.isShellByNmap() then
        args = _G.arg and _G.arg[1]
        args = args and require"base64".decode(args)
        args = args and require"dkjson".decode(args)
    end
    
    repeat
        -- validate args table
        if "table" ~= type(args) then
            break
        end
        if "table" ~= type(args.host) then
            break
        end
        if not args.host.ip then
            break
        end
        if "table" ~= type(args.port) then
            break
        end
        local port = tonumber(args.port.number) 
        if not port then
            break
        end
        
        --  args normalization
        args.args              = args.args or {}
        args.host.mac_addr_str = args.host.mac_addr_str or "" 
        args.port.number       = port
        
        self.args = args
    until true 
    
    return args
end

--- Executing request to get hint responce
function meta:hintRequest()
    local function _makeRequest(reqURL, attempts)
        local rv = nil
        if not reqURL then
            return rv
        end
        local attempts = attempts or 2
        
        local hc = require"core.httpClient":new()
        for i = 1, attempts do
           rv = hc:get(reqURL, {timeout = 2})
           if rv and rv.body and 0 ~= rv.body:len() then
              break
           end
           rv = nil
        end
        
        return rv
    end
    
    -- make the first request to HTTPS (to avoid load during scan)
    local hostAddress = self.args.host.ip    
    local portNumber  = self.args.port.number
    local reqSchema = 443 == portNumber and "https" or "http"                
    local reqURL    = string.format("%s://%s:%d", reqSchema, hostAddress, portNumber) 
    local rv        = _makeRequest(reqURL, 1)
    
    -- try additional attempts via HTTP if failure basic request
    if not rv then  
        portNumber = 80
        reqURL     = string.format("http://%s:%d", hostAddress, portNumber)
        rv         = _makeRequest(reqURL)
        if rv then  -- normalize scanning data
            self.args.port.number  = portNumber
        end
    end
    
    return rv
end

--- Discovery module runner
function meta:run()
    if not self.args then
        return false
    end
    local hostIPAddr = self.args.host.ip or ""
    
    -- initialize logger for all children modules
    if self.args.args.action then
        logger:info(">> Start NSE")
    end
    
    -- loading global configuration (to use parameters in all modules methods)
    local config, err = require"common.helper".get_provConfig() or {}
    if "table" == type(config) then
        _G.gconf = config
        logger:debug(config, "gconfig")
    else
        logger:warning(err)
    end
    
    -- identify module
    local modBase    = __initModuleBase()
    local hintResult = self:hintRequest()
    if not hintResult then
        logger:error("Device does not answer to the hint request: %s", hostIPAddr)
        return false
    end
    self.args.authMethod = __isDigestAuth(hintResult)

    local modName, deviceInfo = __detectModule(modBase, hintResult, self.args)
    if not modName then
        logger:warning("Not found module for device: %s", hostIPAddr)
        return false
    end
    logger:debug("Using module '%s' for device: %s", modName, hostIPAddr)
    
    if not __executeModuleProvision(modBase, modName, self.args, deviceInfo) then
        logger:warning("Can not start provisioning by wrong authorization")
    end
    
    return __makeOutput(self.args, deviceInfo)
end


-- An object constructor
function discovery:new()
    -- try update package path
    local provModDir = "/var/www/provision/modules/"
    if not package.path:find(provModDir) then
        package.path = package.path .. ";" .. provModDir .. "?.lua"
    end
    
    local obj = setmetatable({}, {__index = meta})
    
    return obj
end

-- An object auto builder and runner 
function discovery:auto()
    local obj = self:new()
          obj:setArgs()
          obj:run()
    
    return obj
end

--- A standalone method to detect the requestor environment 
function discovery.isShellByNmap()
    local argList = _G and _G.arg
    return argList and "lua5.1" == argList[-1] and argList[1]
end

return discovery.isShellByNmap() and discovery:auto() or discovery
