module( ... or "daysaver", package.seeall)

hints = {
    --[[
    ======================
    Wildix DaySaver
    ----------------------
    200 OK
    Server: LuCId-HTTPd/1.0
    Connection: Keep-Alive
    Content-Type: text/html
    Date: Tue, 27 Nov 2018 17:40:15 GMT
    Set-Cookie: devckie=DD08-1109-0000-0002;HttpOnly
    X-Frame-Options: SAMEORIGIN
    Transfer-Encoding: chunked
    Cache-Control: no-cache
    Keep-Alive: timeout=5, max=49
    Expires: 0
    --]]
    {
         ["server"] = 'LuCId%-HTTPd/.*'
        ,["status"] = "200"
    }
}

local policy_template = [[
<?xml version="1.0" encoding="utf-8"?>
<provision version="2.0">
    <param name="product" value="DaySaver-1V"/>
    <config>
        <param name="url" value="{{confurl}}" />
    </config>
    <firmware>
        <param name="url" value="{{fwurl}}" />
    </firmware>
</provision>
]]

local function __encrypt_passwd(passwd, salt)
    local command = "openssl passwd -1 -salt %s %s 2>&1"
    local output = tostring(require"core.pbxutils".run(command, salt, passwd))
    local encrypted = string.match(output, "%$%S+%$%S+%$(%S+)[\r\n]")
    if encrypted then
        return encrypted
    end
    return nil, "password encrypt error"
end

local function __check_response(resp, strict)
    if resp and 200 == resp.code and resp.body then
        if not strict then
            return true
        end
        local lbody = string.lower(resp.body)
        if string.find(lbody, "fail") or string.find(lbody, "invalid") then
            return false
        end
        return true
    end
    return false
end

local function __get_provision_params(url, user, password)
    -- this is important, otherwise device may not to see changes and not start provisioning
    math.randomseed(os.time())
    local provint  = "9" .. tostring(math.random(1000,9999))

    local prov_data = {
        ["provision"] = {
            ["status"] = "on",
            ["check_interval"] = provint,
            ["url"] = url,
            ["url_username"] = user,
            ["url_password"] = password or ""
        }
    }
    return prov_data
end

local daysaver_session = {
    d_host         = nil,
    d_port         = nil,
    hclient        = nil,
    d_cookie       = nil
}

function daysaver_session:login(password)
    local salt  = nil
    local answer = self.hclient:rawget(self.d_host, self.d_port, "/api/login")
    if __check_response(answer, true) then
        salt = answer.body
    else
        return nil, "get salt error"
    end
    local encrypted = __encrypt_passwd(password, salt)
    -- get the token
    answer = self.hclient:rawget(self.d_host, self.d_port, string.format("/api/login?password=%s", encrypted))
    if __check_response(answer, true) then
            -- Cookie: sysauth=e1aa0bc237bbb55f338b2476414e6e62
        local cookie = string.format("sysauth=%s", answer.body)
        self.d_cookie = cookie
    else
        return nil, "get token error"
    end
    return true
end

function daysaver_session:logout()
    local headers = {["Cookie"] = self.d_cookie}
    self.hclient:rawget(self.d_host, self.d_port, "/api/logout", {headers = headers})
end

function daysaver_session:get_system_info()
    local headers = {["Cookie"] = self.d_cookie, ["Content-Type"] = "application/json"}
    local answer = self.hclient:rawpost(self.d_host, self.d_port, "/api/get_status", '{"system"}', {headers = headers})
    if __check_response(answer, true) then
        local result = require"dkjson".decode(answer.body) or {}
        return result["system"]
    else
        return nil, "get device info error"
    end
end

function daysaver_session:get_system_config()
    local headers = {["Cookie"] = self.d_cookie}
    local answer = self.hclient:rawpost(self.d_host, self.d_port, "/api/get_config", "", {headers = headers})
    if __check_response(answer) then
        return answer.body
    else
        return nil, "get device config error"
    end
end

function daysaver_session:start_provision(url, user, passwd)
    local logger     = logger

    local prov_data = __get_provision_params(url, user, passwd)
    local json_data = require"dkjson".encode(prov_data)
    local headers = {["Cookie"] = self.d_cookie, ["Content-Type"] = "application/json"}
    local answer = self.hclient:rawpost(self.d_host, self.d_port, "/api/update_config", json_data, {headers = headers})
    logger:debug("Host %s response: %s", self.d_host, tostring(answer.body))
    if not __check_response(answer, true) then
        return nil, "send provision configuration error"
    end

    headers = {["Cookie"] = self.d_cookie}
    local answer = self.hclient:rawget(self.d_host, self.d_port, "/api/apply_changes", {headers = headers})
    logger:debug("Host %s response: %s", self.d_host, tostring(answer.body))

    -- strange response here starting from 2.53.4.23, be optimistic
    --[[
    if not __check_response(answer, true) then
        return nil, "apply provision configuration error"
    end
    --]]

    return true
end

local function get_session(host, port)
    local obj = {}
    setmetatable(obj, {__index = daysaver_session})
    
    obj.d_host   = host
    obj.d_port   = port
    obj.hclient  = require"core.httpClient":new()
    return obj
end

function verify (host, port)
    local logger     = logger
    local helper     = require"common.helper"
    local connection = require"common.core".connection

    local is_ssl_negotiate = 443 == port.number
    local PORT_NUMBER      = is_ssl_negotiate and 443 or 80
    local t                = {["ssl"] = is_ssl_negotiate}
    -- API passwords, "api" is default
    local auth_table       = {"api"}
    
    local sql, cursor, token_pass, dublicate_ip
    if host.mac_addr_str == "" then
        sql = [[ SELECT token, count(id) FROM autoprov WHERE param=%q AND value LIKE '%%"address":%%"%s"%%' AND value LIKE '%%"model":%%"daysaver"%%' ]]
        cursor = connection:execute(sql, "device", host.ip or "")
        token_pass, dublicate_ip = cursor:fetch(); cursor:close()
        if tonumber(dublicate_ip) > 1 then
            logger:error("Found %s duplicated address %q at provisioning database", dublicate_ip, host.ip or "")
            return t
        end
    else
        sql = "SELECT token FROM autoprov WHERE param=%q AND mac=%q"
        cursor = connection:execute(sql, "device", host.mac_addr_str)
        token_pass = cursor:fetch(); cursor:close()
    end

    t["token_type"] = helper.is_new_token(token_pass)
    if t["token_type"] == "new" then
        table.insert(auth_table, 1, token_pass)
        t["token"] = token_pass
    end

    local session = get_session(host.ip, PORT_NUMBER)
    for _, apipass in pairs(auth_table) do
        if session:login(apipass) then
            t["auth"] = {pass = apipass}
            break
        end
    end

    if not t["auth"] then
        logger:error("Wrong authorization for host %s", host.ip)
        return t
    end

    local sysinfo = session:get_system_info()
    logger:debug(sysinfo)

    t["model"]             = sysinfo["model"]:lower()
    t["fwversion"]         = sysinfo["firmware"]:match("(%S+)%s.+")
    t["vendor:additional"] = "dwg"

    if host.mac_addr_str == "" then
        host.mac_addr_str = sysinfo["mac_address"]:gsub("-","")
    end
    t["mac_addr_str"] = host.mac_addr_str:lower()

    session:logout() -- TODO optimize usage
    return t
end

function provision (host, port, ...)
    local h      = require"common.helper"
    local logger = logger

    local prov_proto  = gconf.HTTPForceUse and "http" or "https"
    local prov_port   = gconf.HTTPForceUse and "80" or "443"

    local PORT_NUMBER = tonumber(prov_port)

    -- first production version does not support https provision
    if "2.53.4.12" == host.fwversion then
        prov_proto  = "http"    
        prov_port   = "80"  
        PORT_NUMBER = 80
    end

    -- processing registrar address
    local registrar = nil
    local case      = "hosts"
    local cfg       = {
             FailoverByDns    = gconf.FailoverByDns
            ,ForceResolveName = gconf.ForceResolveName
            ,ForceResolveIp   = gconf.ForceResolveIp
        }
    local addr_t, err_msg = h.resolve_by_hosts(host.ip_src, nil, cfg)
    if not addr_t then
        if err_msg then -- printing resolving error
            logger:warning("Hosts file was not processed: %s", err_msg)
        end
    else
        registrar = addr_t.domain
    end
    if not registrar then
        registrar, case, err_msg = h.get_registrar(host.ip_src, host.ip) 
    end
    logger:debug("Registrar address %q got from %s%s", registrar, case, err_msg and ". Errors: "..err_msg or "")

    local api_password = host.token or "api"

    -- Provisioning link
    local provlink = string.gsub(
            "$protocol://$registrar:$port/?action=setaps", "%$(%w+)",
            {
                registrar = registrar,
                protocol  = prov_proto,
                port      = prov_port,
            }
        )

    logger:info("Sending http api request to %s for autoprovision now", host.model)
    local session = get_session(host.ip, PORT_NUMBER)
    local res, err = session:login(api_password)
    if res then
        if h.is_exists("/tmp/savemyday") then
            local conf = session:get_system_config()
            logger:debug("Current configuration: %s", conf or "")
        end

        res, err = session:start_provision(provlink, "wmsdevice", host.token)
        if not res then
            logger:error("Cannot start provisioning of device %s, %s", host.model, err or "")
        end
    else
        logger:error("Cannot perform login to device %s, %s", host.model, err or "")
    end
    session:logout()
end

function configuration (agent, request)
    local h      = require"common.helper"
    local core   = require"common.core"
    local json   = require"dkjson"
    local md5    = require"md5"
    local logger = logger
    local connection = core.connection
    local username, err, sql, cursor = ""

    -- getting registrar address
    local registrar       = agent["registrar"]
    local registrar_ip    = false
    local cfg             = {
                FailoverByDns    = gconf.FailoverByDns
               ,ForceResolveName = gconf.ForceResolveName
               ,ForceResolveIp   = gconf.ForceResolveIp
          }
    local addr_t, err_msg = h.resolve_by_hosts(registrar, nil, cfg)
    if not addr_t then
        if err_msg then
            logger:warning("Hosts file was not processed: %s", err_msg)
        end
    else
        logger:debug(addr_t, "hosts_registrar")
        if addr_t.domain then
            registrar = addr_t.domain
        end
        if addr_t.ip then
            registrar_ip = addr_t.ip
        end
    end

    local prov_proto = agent["ssl"] == "on" and not gconf.HTTPForceUse and "https" or "http"

    local prov_port  = gconf.HTTPForceUse and "80" or agent["alternate_port"] or agent["registrar:port"]

    -- first production version does not support https provision
    if "2.53.4.12" == agent["fw:version"] then
        prov_proto = "http"
        prov_port  = "80"
    end

    -- set general url
    local base_url   = string.gsub("$protocol://$registrar:$port/", "%$(%w+)",
                                   {
                                        registrar = registrar,
                                        protocol  = prov_proto,
                                        port      = prov_port,
                                    })

----------------------------------------------------------------
--- Policy file with config url and firmware link
----------------------------------------------------------------
    if "policy" == request["config"] then
        local fw_url  = ""
        local fw_dir = "wildixgsm"
        local firmwares = core.get_firmware_pack({["vendor"] = fw_dir});
        local fw_cfg = firmwares[agent["model"]]
        if fw_cfg then
            if agent["fw:version"] ~= fw_cfg["version"] then
                local firmware_file = fw_cfg["application"]
                fw_url = base_url .. "firmwares/" .. fw_dir .. "/" .. firmware_file
                logger:info("Firmware for device %s %s will be changed from version %s to %s", agent["model"], agent["mac"], agent["fw:version"], fw_cfg["version"])
            end
        else
            logger:warning("Firmware configuration file for DaySaver is not found")
        end
        if "" == fw_url then
            policy_template = string.gsub(policy_template, "<firmware>.+</firmware>", "")
        else
            -- WMS-6236
            policy_template = string.gsub(policy_template, "<config>.+</config>", "")
        end
        local conf_url = base_url .. "$(MAC).cfg/?action=setaps"
        local policy = string.gsub(policy_template, "{{(%w+)}}", {confurl = conf_url, fwurl = fw_url})
        return policy
    else
----------------------------------------------------------------
--- General configuration settings
----------------------------------------------------------------
        -- load default values from resource directory
        local resource = core.load_resource("daysaver")
        local result   = resource.daysaver_defaults
        if not result then
            logger:error("Cannot load DaySaver default config, check that file exists and it's valid JSON")
            return false
        end

        local token_pass = core.proceedDeviceInfo(agent).token

        if request["OldPassword"] then
            if request["OldPassword"] == "" then
               -- first provisioning
                result["user"]["old_password"] = "admin"
            else
               -- reset password
                result["user"]["old_password"] = request["OldPassword"]
            end
            result["user"]["new_password"] = token_pass
            result["user"]["confirm_password"] = token_pass
        else
            result["user"] = nil
        end

        local prov_url = base_url .. "?action=setaps"
        result["provision"] = __get_provision_params(prov_url, "wmsdevice", token_pass).provision
        if not request["OldPassword"] then
            -- avoid provisioning loop
            result["provision"]["check_interval"] = nil
        end

        result["api"]["password"] = token_pass

        -- Codecs Priority
        local netband_str = h.is_wideband(agent["address"]) and "wideband" or "narrowband"
        result["codec"]["1"]["codec"] = resource.codecs[netband_str]


        -- currently only UDP SIP transport supported from provisioning
        local sip_port  = "5060"
        result["sip_trunk"]["1"]["address"] = registrar
        result["sip_trunk"]["1"]["port"] = sip_port
        result["sip_trunk"]["1"]["outboundproxy"] = registrar
        result["sip_trunk"]["1"]["outboundproxy_port"] = sip_port

----------------------------------------------------------------------------------------------
-- Some Wildix style black magic copypasted near as is
        local nports = 1
        local tports = {}
        local tpbx   = {}

        local port_default = {
            ["port_number"] = "",
            ["port_codec_input_gain"] = "0",
            ["port_codec_output_gain"] = "0",
            ["port_context"] = "main",
            ["port_initatcmd"] = "",
        }

        -- get default context
        connection:switch ("wmsdb")
        sql = "select count(id) from dialplan_rulenames where rulename=%q"
        cursor = connection:execute(sql, "main")
        local count = cursor:fetch(); cursor:close()
        port_default.port_context = tonumber(count) > 0 and "main" or "internalcalls"
        local is_upd_wmsdb  = false
        local needSipReload = false

        for k,v in pairs(port_default) do
            if not tports[k] then tports[k] = {} end
            tports[k][1] = v -- only one ports, require for wms compatibility
        end

        -- check trunkid
        connection:switch("devdb")
        sql = "select a1.value, a2.value from autoprov as a1 left join autoprov as a2 on (a1.mac = a2.mac and a2.param=%q) where a1.param=%q and a1.mac=%q"
        cursor = connection:execute(sql, "settings", "trunkid", agent["mac"])
        local trunkid, settings_json = cursor:fetch(); cursor:close()
        if trunkid then -- trunk already present in devdb
            -- get settings
            connection:switch ("wmsdb")
            sql = "select username, settings_mgw, settings_pbx from pstn_trunks where id = %s"
            cursor = connection:execute (sql, trunkid)
            local username, settings_mgw, settings_pbx = cursor:fetch(); cursor:close()

            trunkname = username
            local t = json.decode(settings_mgw)
            for k,v in pairs (t) do
                tports[k] = v
            end

            local t2 = json.decode(settings_pbx)
            for k,v in pairs(t2) do
                tpbx[k] = v
            end
            if not tpbx["0_0_secret"] then
                is_upd_wmsdb = true
            end
            tpbx["0_0_secret"] = h.get_sip_pass(trunkname, trunkid, 0, agent["mac"])
        else -- trunk autocreation
            local lowertech, tnames = "gsm", {}
            is_upd_wmsdb  = true
            needSipReload = true

            connection:switch ("wmsdb")
            sql = "select username from pstn_trunks where technology=%q"
            for name in connection:rows(sql, lowertech) do
                table.insert(tnames, name)
            end
            table.sort(tnames, function(a,b) a = tonumber(a:sub(4)); b = tonumber(b:sub(4));  return a<b end)
            
            for i=1,#tnames do
                if (lowertech .. tostring(i-1)) ~= tnames[i] then
                    trunkname = lowertech .. tostring(i-1)
                    break
                end
            end
            if not trunkname then
                trunkname = lowertech .. tostring(#tnames)
            end
            tpbx["0_0_secret"] = h.generate_pass()

            -- Update db for creating sip-extra.conf
            sql = [[ INSERT or REPLACE INTO pstn_trunks (id,username,ports,technology,settings_pbx,settings_mgw) VALUES(
                    (select id from pstn_trunks where username = %q),%q,%d,%q,'%s','%s') ]]
            connection:execute(sql, trunkname, trunkname, nports, "gsm", json.encode(tpbx), json.encode(tports))

            sql = "SELECT id FROM pstn_trunks WHERE username=%q"
            cursor = connection:execute (sql, trunkname)
            trunkid = cursor:fetch(); cursor:close()

            connection:switch ("devdb") -- set trunk id to devdb
            sql = [[ INSERT or REPLACE INTO autoprov (id,param,value,mac) VALUES(
                    (select id from autoprov where param=%q and mac=%q),%q,'%s',%q) ]]
            connection:execute (sql, "trunkid", agent["mac"], "trunkid", trunkid, agent["mac"])

            -- update kamailio tables
            connection:switch ("htables")
            connection:execute ("delete from dynamic_peers where wms_id = %q", trunkid) -- remove old trunk stuff
            sql = "insert into %s (key_name,key_value,wms_id) values (%q,%q,%q)"
            local username = string.format("%s_%d_%d",trunkname,0,0)
            connection:execute (sql, "dynamic_peers", username .. "::authname", username, trunkid)
            connection:execute (sql, "dynamic_peers", username .. "::secret", tpbx["0_0_secret"], trunkid)
            
            h.create_sip_extra()
        end
        
        -- reload kamailio and sip to update callweaver peers
        if needSipReload then
            os.execute ("kamcmd htable.reload dynamic_peers; callweaver -rx \"sip reload \" > /dev/null")
        end
        
        if is_upd_wmsdb then
            connection:switch ("wmsdb") -- update wmsdb
            sql = [[ INSERT or REPLACE INTO pstn_trunks (id,username,ports,technology,settings_pbx,settings_mgw) VALUES(
                    (select id from pstn_trunks where username = %q),%q,%d,%q,'%s','%s') ]]
            connection:execute(sql, trunkname, trunkname, nports, "gsm", json.encode(tpbx), json.encode(tports))
        end

        local username = string.format ("%s_%d_%d", trunkname, 0, 0)
        local secret   = tpbx["0_0_secret"]

        -- Network settings
        local t = {}
        local update_db_settings = false
        if settings_json then
            t = json.decode(settings_json)
        end

        -- SYSLOG settings
         if t["syslog_server"] and t["syslog_server"] ~= "" then
            local server, port = string.match(t["syslog_server"], "(%S+):(%d+)")
            if not server then
                server = t["syslog_server"]
                port   = "514"
            end
            result["system"]["log"]["syslog_status"] = "on"
            result["system"]["log"]["server_ip"]     = server
            result["system"]["log"]["server_port"]   = port
        end
        
        -- set default DNS and NTP settings if not define
        t["dns_server"], t["second_dns_server"] = h.get_prim_sec_dns_servers(t["dns_server"], t["second_dns_server"])
        t["ntp_server"]                         = t["ntp_server"] or ""
        
        -- use static ip settings
        local is_addr_present = h.is_addr_present(agent["address"], agent["mac"])   -- check duplicate address
        local is_addr_local = h.isLocalIP(agent["address"])     -- device behind NAT can't user static ip
        if t["use_static_ip"] and (is_addr_present or not is_addr_local) then
            if is_addr_present then
                logger:warning("Disable static ip. Other device have the same ip address: %s", agent["address"] or "")
            end
            if not is_addr_local then
                logger:warning("Disable static ip. Device ip address (%s) is placed behind NAT", agent["address"] or "")
            end
            t["use_static_ip"] = nil
            update_db_settings = true
        end
        
        -- get settings for leased device by pbx dhcp
        local is_leased
        if not t["use_static_ip"] or t["dns_server"] == "" or t["ntp_server"] == "" then
            is_leased = h.is_addr_leased(agent["address"])
        end
        if not t["use_static_ip"] and not is_addr_present and is_leased then
            h.set_addr_static(agent["address"], agent["mac"])   -- mark this lease as static
            t["use_static_ip"] = 1
            update_db_settings = true
        end
        if is_leased then
            if "" == t["dns_server"] or "" == t["second_dns_server"] then
                local dns_dhcp1, dns_dhcp2 = h.get_dns_servers()
                if dns_dhcp1 and "" ~= dns_dhcp1 and "" == t["dns_server"] then
                    t["dns_server"] = dns_dhcp1             -- set primary DNS from DHCP config
                    update_db_settings = true
                end
                if dns_dhcp2 and "" ~= dns_dhcp2 and "" == t["second_dns_server"] then
                    t["second_dns_server"] = dns_dhcp2      -- set secondary DNS from DHCP config
                    update_db_settings = true
                end
            end     
            if t["ntp_server"] == "" then -- set NTP from DHCP config
                local ntp_dhcp = h.get_ntp_servers()
                if ntp_dhcp and ntp_dhcp ~= "" then
                    t["ntp_server"] = ntp_dhcp
                    update_db_settings = true
                end
            end
        end

        if t["use_static_ip"] == 1 then -- use static ip
            local _, netmask, gateway = h.host_settings(agent["address"], registrar_ip)
            if netmask and gateway then
                result["network"]["wan"]["proto"]   = "static"
                result["network"]["wan"]["ipaddr"]  = agent["address"]
                result["network"]["wan"]["netmask"] = netmask
                result["network"]["wan"]["gateway"] = gateway
                
                if registrar_ip then
                    if "" == t["dns_server"] then
                        t["dns_server"]    = registrar_ip
                        update_db_settings = true
                    end
                    if "" == t["second_dns_server"] then
                        t["second_dns_server"] = registrar_ip
                        update_db_settings     = true
                    end
                end
                if t["ntp_server"] == "" then
                    t["ntp_server"]    = registrar
                    update_db_settings = true
                end
            else
                logger:warning("For current network configurations is not possible set static ip, disable it")
                t["use_static_ip"] = nil
                update_db_settings = true
            end
        end

        -- update devdb with new settings
        if update_db_settings then
            sql = "INSERT OR REPLACE INTO autoprov (id, param, value, mac) VALUES ((SELECT id FROM autoprov WHERE param=%q AND mac=%q), %q, '%s', %q)"
            connection:execute (sql, "settings", agent["mac"], "settings", json.encode(t), agent["mac"])
        end

        -- DNS settings
        if "" ~= t["dns_server"] then
            result["network"]["wan"]["peerdns"] = "0"
            local dns_list = {t["dns_server"], "" ~= t["second_dns_server"] and t["second_dns_server"] or nil}
            result["network"]["wan"]["dns"] = dns_list
        end
---------------------------------------------------------------------------------------------------------------------------------
        -- NTP settings 
        if t["ntp_server"] ~= "" then
            result["system"]["ntp"]["server"]    = {t["ntp_server"]}
        end
        -- A fail possible here (Etc/UTC), ask vendor about supported
        -- result["system"]["main"]["zonename"] = t["ntp_timezone"] or h.getPbxTimeZone()

        result["mobile"]["volte"]["username"]      = username
        result["mobile"]["volte"]["authuser"]      = username
        result["mobile"]["volte"]["port_password"] = secret

        local pnum = tports["port_number"][1]
        if pnum and "" ~= pnum then
            -- use default here we use only autodial
            -- result["mobile"]["volte"]["number"]   = pnum
            result["mobile"]["volte"]["autodial"] = pnum
        end

        result["mobile"]["volte"]["dsp_output_gain"] = tports["port_codec_output_gain"][1]    -- Tx Gain
        result["mobile"]["volte"]["dsp_input_gain"]  = tports["port_codec_input_gain"][1]     -- Rx Gain

        -- TODO change after web fix
        -- adapt web settings for device
        local txgain = tonumber(tports["voip_output_gain"] or "0") + 1
        local rxgain = tonumber(tports["voip_input_gain"] or "6") + 1
        if txgain > 10 then
            txgain = 10
        end
        if rxgain > 10 then
            rxgain = 10
        end
        result["mobile"]["volte"]["speaker_gain"] = tostring(rxgain)
        result["mobile"]["volte"]["micphone_gain"] = tostring(txgain)

        if tports["port_smsc"] and tports["port_smsc"][1] then
            result["mobile"]["volte"]["at_smsc_number"] = tports["port_smsc"][1]  -- set SMSC
        end

        local function __get_nats_config()
            local config = {}
            local filename = "/rw2/etc/pbx/wildix/gnats.config"
            local file = io.open(filename)
            if not file then
                logger:error("Cannot open nats config file")
                return config
            end
            local content = file:read("*a") or ""
            file:close()
            local pwd = string.match(content, "{[%s\r\n]*user:%s*daysaver,[%s\r\n]*password:%s*(%S+)[%s\r\n]*}")
            config["password"] = pwd
            local param, port = string.match(content, "([%S]*port):%s*(%S+)[%s\r\n]+")
            port = not string.find(param or "", "http") and port
            config["port"] = port

            return config
        end

        local nats_url = "nats://" .. registrar
        local nats_conf = __get_nats_config()
        if nats_conf.port then
            nats_url = nats_url .. ":" .. nats_conf.port
        end
        result["nats_cloud"]["main"]["username"] = nats_conf.user or "daysaver" -- only one is supported currently
        result["nats_cloud"]["main"]["password"] = nats_conf.password or ""
        result["nats_cloud"]["main"]["server"]   = nats_url

        return require"dkjson".encode(result) .. "\n"
    end
end

function setuser (...) end
function setfeatures (...) end
