module( ... or "wildixair", package.seeall)

hints = {
    --[[
    ======================
    Wildix AIR
    ----------------------
    	HTTP/1.1 401 Unauthorized
		WWW-Authenticate: Basic realm=""
		Content-Type: text/html
    ----------------------
    ]]
     {["www-authenticate"] = 'Basic realm=""'}
    --[[
    since fw. version 3.80 B0010 http answer changed to:
    ----------------------
        HTTP/1.1 303 See Other.
        Content-Type: text/html.
        Location: http://10.100.4.74/main.html.
        Content-Type: text/html.
        Content-Length: 92.
    ]]
    ,{["status"] = "303"}
    --[[
        HTTP/1.1 401 Unauthorized
        WWW-Authenticate: Digest realm="", nonce="MDAwMDAwNDggYmUyYzVjOTRiY2Q1NDBkM2YzOTQzMjYwODQyNTZmMDg=", algorithm="MD5", qop="auth"
        Content-Type: text/html
    ]]
    ,{["www-authenticate"] = [[Digest realm="", nonce=".*", algorithm="MD5", qop="auth"]]}
}

local function __parse_firmware(version)
    local model_spec    = nil
    local major_version = nil
    local build_version = nil
    local minor_version = nil
    
    repeat
        if not version then
            break
        end
                
        model_spec, major_version, build_version, minor_version = version:match("(%d+)%.(%d+)%.(%d+)%.(%d+)")
        
        if not minor_version then
            model_spec, major_version = version:match("(%d+)%.(%d+)")
        end
    until true
    
    model_spec    = model_spec    or "00"
    major_version = major_version or "00"
    build_version = build_version or "00"
    minor_version = minor_version or "00"
    return model_spec, major_version, build_version, minor_version
end

local function __prepare_mac_string (mac,...)
    --we need to convert mac from AABBCCDDEEFF to 0xAA,0xBB,0xCC,0xDD,0xEE....
    -- we have 6 parts of 2 symbols each
    local part1 = string.sub(mac, 1, 2)
    local part2 = string.sub(mac, 3, 4)
    local part3 = string.sub(mac, 5, 6)
    local part4 = string.sub(mac, 7, 8)
    local part5 = string.sub(mac, 9, 10)
    local part6 = string.sub(mac, 11, 12)
    local rv = "0x"..part1..",0x"..part2..",0x"..part3..",0x"..part4..",0x"..part5..",0x"..part6
    return rv
end

local _usersTableVerList = {
	 [0] = "STANDART"		-- old standart version
	,[1] = "IPEISUPPORT"	-- processing handset IPUI during setuser
}
local function __getUsersTableVersion(byMACValue)
	local conn = require"common.core".connection
	conn:switch("devdb")
	
	-- extend default table version
	local verList = {}
	for k, v in pairs(_usersTableVerList) do
		verList[k] = v
		verList[v] = k
	end
	_usersTableVerList = verList
	
	local ver = 0
	local dataCursor = conn:execute("SELECT * FROM wair_users LIMIT 1")
	repeat
		if "userdata" ~= type(dataCursor) then
			break
		end
		local colNames = dataCursor:getcolnames()
		if "table" ~= type(colNames) then
			break
		end
		if "ipei" == colNames[4] then	-- IPEI support
			ver = 1
			if byMACValue then
    			local cursor  = conn:execute("SELECT count(ipei) FROM wair_users WHERE mac='%s'", byMACValue)
    			local IPEICnt = cursor:fetch(); cursor:close()
    			if "0" == IPEICnt then
    				ver = 0
    			end
    		end
		end
	until true
	
	if "userdata" == type(dataCursor) then
		dataCursor:close()
	end
		
	return ver 
end

local function __proceed_users(mac, baseType)
	local conn = require"common.core".connection
    local rv   	  = {}
    local err  	  = nil
	local sql  	  = nil
	local sqlTmpl = nil
	
	local limit   = ({
		 ["W-AIR"]	  = 200		-- W-AIR BASE supports 200 concurrent dect phones
		,["W-AIR SB"] = 8		
		,["W-AIR-SP"] = 200
		,["W-AIR-SBR2"] = 20
	})[baseType]
	
	local isIPEISupport = false
	local usersTableVer = __getUsersTableVersion()
	isIPEISupport		= _usersTableVerList["IPEISUPPORT"] <= usersTableVer
	logger:debug("Users table version: %s", _usersTableVerList[usersTableVer])
	if isIPEISupport then
		sql = "SELECT slot_id, uid, mac, ipei FROM wair_users WHERE mac='%s' ORDER BY slot_id"
		sqlTmpl = "INSERT INTO 'wair_users' (slot_id, uid, mac, ipei) VALUES(%s,%s,'%s','FFFFFFFFFF')"
	else
		sql = "SELECT slot_id, uid, mac FROM wair_users WHERE mac='%s' ORDER BY slot_id"
		sqlTmpl = "INSERT INTO 'wair_users' (slot_id, uid, mac) VALUES(%s,%s,'%s')"
	end

	-- select users from table wair_users
	local usersData = conn:rows(sql, mac) or {}
    for slot_id, uid, mac, ipei in usersData do
    	rv[#rv+1] = {slot_id, uid, ipei}
    end
    
    -- if table wair_users is empty initiate new record
    if #rv == 0 then
        for i = 1,limit,1 do
            sql = string.format(sqlTmpl, i, -i, mac)
            conn:execute(sql)
            rv[#rv+1] = {i, -i, isIPEISupport and 'FFFFFFFFFF'}
        end
    end
    return rv
end

--- Get country variant id and country region id by time zone
local function __getCountryVariantRegion(timeZone)
	local variant = nil
	local region  = nil
	local gmtId   = nil
	local continent, location = string.match(timeZone or "", "^([%w%-%+]+)/*(.*)$")
	local t = {
			 ["Africa"]    = {}
			,["America"]   = {
			    ["Bahia"]           = {variant = "2F", region = "04"} -- Brazil
			    ,["Porto_Acre"]     = {variant = "2F", region = "00"} -- Brazil
			    ,["Toronto"]        = {variant = "15", region = "11"} -- Canada, Ontario EST
			    ,["Vancouver"]      = {variant = "15", region = "03"} -- Canada, British Columbia PST
			}
			,["Asia"]      = {
			    ["Bangkok"]         = {variant = "30", region = "00"} -- Thailand
			}
			,["Atlantic"]  = {}
			,["Australia"] = {
			    ["Victoria"]        = {variant = "13", region = "01"}
			    ,["Tasmania"]       = {variant = "13", region = "02"}
			    ,["North"]          = {variant = "13", region = "05"}
			    ,["Queensland"]     = {variant = "13", region = "06"}
			    ,["West"]           = {variant = "13", region = "07"}
			}
			,["Canada"]    = {
			    ["Yukon"]           = {variant = "15", region = "17"}
			    ,["Saskatchewan"]   = {variant = "15", region = "16"}
			}
			,["Etc"]       = {
			     ["GMT-12"] = {gmtId = "0"}
			    ,["GMT-11"] = {gmtId = "1"}
			    ,["GMT-10"] = {gmtId = "2"}
			    ,["GMT-9"]  = {gmtId = "3"}
			    ,["GMT-8"]  = {gmtId = "4"}
			    ,["GMT-7"]  = {gmtId = "5"}
			    ,["GMT-6"]  = {gmtId = "6"}
			    ,["GMT-5"]  = {gmtId = "7"}
			    ,["GMT-4"]  = {gmtId = "9"}
			    ,["GMT-3"]  = {gmtId = "11"}
			    ,["GMT-2"]  = {gmtId = "13"}
			    ,["GMT-1"]  = {gmtId = "14"}
			    ,["GMT-0"]  = {gmtId = "15"}
			    ,["GMT"]    = {gmtId = "15"}
			    ,["GMT0"]   = {gmtId = "15"}
			    ,["GMT+0"]  = {gmtId = "15"}
			    ,["GMT+1"]  = {gmtId = "16"}
			    ,["GMT+2"]  = {gmtId = "17"}
			    ,["GMT+3"]  = {gmtId = "18"}
			    ,["GMT+4"]  = {gmtId = "20"}
			    ,["GMT+5"]  = {gmtId = "22"}
			    ,["GMT+6"]  = {gmtId = "25"}
			    ,["GMT+7"]  = {gmtId = "27"}
			    ,["GMT+8"]  = {gmtId = "28"}
			    ,["GMT+9"]  = {gmtId = "29"}
			    ,["GMT+10"] = {gmtId = "31"}
			    ,["GMT+11"] = {gmtId = "32"}
			    ,["GMT+12"] = {gmtId = "33"}
			    ,["GMT+13"] = {gmtId = "34"}
			}
			,["Europe"]    = {
				 ["Berlin"]     = {variant = "00", region = "00"} -- GERMANY
				,["Luxembourg"] = {variant = "01", region = "00"} -- LUXEMBOURG
				,["Paris"]      = {variant = "02", region = "00"} -- FRANCE
				,["Rome"]       = {variant = "03", region = "00"} -- ITALY
				,["Amsterdam"]  = {variant = "04", region = "00"} -- HOLLAND
				,["Athens"]     = {variant = "05", region = "00"} -- GREECE
				,["Dublin"]     = {variant = "06", region = "00"} -- IRELAND
				,["Stockholm"]  = {variant = "07", region = "00"} -- SWEDEN
				,["Madrid"]     = {variant = "08", region = "00"} -- SPAIN
				,["Copenhagen"] = {variant = "09", region = "00"} -- DENMARK
				,["Helsinki"]   = {variant = "0A", region = "00"} -- FINLAND
				,["Tallinn"]    = {variant = "0A", region = "00"} -- ESTONIA
				,["Kiev"]       = {variant = "0A", region = "00"} -- UKRAINE
				,["Istanbul"]   = {variant = "0B", region = "00"} -- TURK
				,["Zurich"]     = {variant = "0C", region = "00"} -- SWISS
				,["Lisbon"]     = {variant = "0D", region = "00"} -- PORTUGAL
				,["Brussels"]   = {variant = "0E", region = "00"} -- BELGIUM
				,["Vienna"]     = {variant = "0F", region = "00"} -- AUSTRIA
				,["London"]     = {variant = "10", region = "00"} -- ENGLAND
				,["Oslo"]       = {variant = "11", region = "00"} -- NORWAY
				,["Moscow"]     = {variant = "2B", region = "00"} -- RUSSIA
			}
			,["Indian"]    = {}
			,["Pacific"]   = {
			    ["Auckland"]        = {variant = "14", region = "00"} -- New Zealand
            }
			,["SystemV"]   = {}
			,["US"]        = {
				 ["Alaska"]         = {variant = "12", region = "01"}
				,["Aleutian"]       = {variant = "12", region = "0A"}
				,["Arizona"]        = {variant = "12", region = "02"}
				,["Central"]        = {variant = "12", region = "03"}
				,["Eastern"]        = {variant = "12", region = "21"}
				,["East-Indiana"]   = {variant = "12", region = "0E"}
				,["Hawaii"]         = {variant = "12", region = "0A"}
				,["Indiana-Starke"] = {variant = "12", region = "03"}
				,["Michigan"]       = {variant = "12", region = "17"}
				,["Mountain"]       = {variant = "12", region = "05"}
				,["Pacific"]        = {variant = "12", region = "04"}
				,["Pacific-New"]    = {variant = "12", region = "04"}
				,["Samoa"]          = {variant = "12", region = "00"}
			}
	}
	
	if t[continent] and t[continent][location] then
        variant = t[continent][location]["variant"]
        region  = t[continent][location]["region"]
        gmtId   = t[continent][location]["gmtId"]
	end
	
	return variant, region, gmtId
end

--- Verifying device HTML response and retreiving information
-- @param host		A verifying host data
-- @param port		Not used data
-- @return table with verifying device information
function verify (host, port, authMethod)
	local logger 	 = logger 
	local helper 	 = require"common.helper"
    local http 		 = require"core.httpClient":new()
    local xml 		 = require"core.htmlparser"
    local connection = require"common.core".connection
    local pbxutils   = require"core.pbxutils"

    
    local is_ssl_negotiate = 443 == port.number
    local PORT_NUMBER      = is_ssl_negotiate and not gconf.HTTPForceUse and 443 or 80
    
    local t          = {["ssl"] = is_ssl_negotiate}
	local gpassword  = helper.get_gpassword()
	local auth_table = {"admin:admin", "root:root","admin:" .. gpassword}  
	
	local sql, cursor, db_mac, token_pass, dublicate_ip
	if host.mac_addr_str == "" then
		sql = [[ SELECT mac, token, count(id) FROM autoprov WHERE param=%q AND value LIKE '%%"address":%%"%s"%%' AND value LIKE '%%"model":%%"w-air%%"%%' ]]
		cursor = connection:execute(sql, "device", host.ip or "")
		db_mac, token_pass, dublicate_ip = cursor:fetch(); cursor:close()
		if tonumber(dublicate_ip) > 1 then
			logger:error("Found %s duplicated address %q at provisioning database", dublicate_ip, host.ip or "")
			return t
		end
	else
   		sql = "SELECT token FROM autoprov WHERE param=%q AND mac=%q"
    	cursor = connection:execute(sql, "device", host.mac_addr_str)
    	token_pass = cursor:fetch(); cursor:close()
    end
    
    t["token_type"] = helper.is_new_token(token_pass)
    if t["token_type"] == "new" then
    	table.insert(auth_table, 1, "admin:"..token_pass)
    	t["token"] = token_pass
    end
    logger:debug(auth_table, "auth_table")
    
    -- check authorization
    local answer = nil
    for _, authpair in pairs(auth_table) do
        local login, pass = authpair:match("(.*):(.*)")
        answer = http:rawget(host.ip, PORT_NUMBER, "/main.html", { authUser = login, authPass = pass, authMethod = authMethod }) or {}
        if 200 == answer.code then
            t["auth"] = {user = login, pass = pass}
            break
        end
    end
    if not t["auth"] then
    	-- try auth with mac from MASTER of W-AIR network
    	logger:debug("Main list for authorization is wrong. Try password from MASTER of the W-AIR network")
    	sql = [[ SELECT token FROM autoprov WHERE param=%q and mac=(SELECT mac FROM wair_gateways WHERE network_id=
    			(SELECT network_id FROM wair_gateways WHERE mac=%q) and type=1); ]]
    	cursor = connection:execute(sql, "device", db_mac or host.mac_addr_str)
    	token_pass = cursor:fetch(); cursor:close()
    	if token_pass then
    		logger:debug("Gotten auth for MASTER of the W-AIR network: %q", "admin:"..token_pass)
    		t["token_type"] = helper.is_new_token(token_pass)
    		if t["token_type"] == "new" then
    			t["token"] = token_pass
    			answer = http:rawget(host.ip, PORT_NUMBER, "/main.html", { authUser = "admin", authPass = token_pass, authMethod = authMethod }) or {}
    			if 200 == answer.code then
    			    t["auth"] = {user = "admin", pass = token_pass}
    			end
    		end
    	end
    end
    if not t["auth"] then
        logger:error("Wrong authorization for host %s", host.ip)
        return t 
    end
    
    -- processing response body
    answer.body = answer.body or ""
    local doc = xml:new(answer.body)
    
    -- finding W-AIR Base header
    local header = doc:find('//title[contains(text(), "W-AIR")]') -- <title>W-AIR Base Station</title>
    if not header then
        header = doc:find('//div[@id="header" and contains(text(), "W-AIR")]')  -- <div id="header"> W-AIR Base Station </div>
    end
    if not header then
        header = string.find(string.upper(answer.body), "WAIR-BASE", 1, true) -- <td class=normaltext>IPDECT (WAIR-BASE-SB)</td>
    end
    if header then
        t["model"] = "w-air"
    else
    	logger:error("Not found model for host %s", host.ip)
        return t
    end

    --WMS-6021
    local function __remove_tag(txt)
        return string.match(txt or "", ".+>(.-)<.+")
    end

	-- getting mac address    
    if host.mac_addr_str == "" then
    	-- MAC-Address: </td><td class=normaltext>9c7514340002</td>
    	local macAddrStr = pbxutils.get_mac(__remove_tag(doc:find('//div[@id="main"]/table/tr[2]/td[1]/table/tr[8]/td[3]')))
    	if not macAddrStr then  -- second search in case WARNING message is present
    		macAddrStr = pbxutils.get_mac(__remove_tag(doc:find('//div[@id="main"]/table/tr[2]/td[1]/table/tr[9]/td[3]')))
    	end
    	host.mac_addr_str = macAddrStr
    end
    t["mac_addr_str"] = host.mac_addr_str
    
    -- detect W-AIR base type
    t["btype"] = "W-AIR"
    local family = require"common.wdmacs".get_family(host.mac_addr_str)
    if family then
    	t["btype"] = family
    end
    
    -- parsing firmware version
    local firmware = ""
    local fw_str = doc:find('//div[@id="main"]/table/tr[2]/td[1]/table/tr[10]/td[3]')	-- Firmware-Version:</td><td class=normaltext>IPDECT/02.00/B0005/09-Mar-2012 11:38</td>
    fw_str = (fw_str or ""):match("%w+%/(.+%/.+)/.*")
    if fw_str then
    	for num in fw_str:gmatch("%d%d") do
    		firmware = firmware == "" and firmware .. num or firmware .. "." .. num
    	end
    	t["fwversion"] = firmware
    end
    
    -- fixup model for W-AIR SB
    if "W-AIR SB" == t["btype"] and "w-air" == t["model"] then
    	t["model"] = "w-air sb"
    end
     
    return t
end

--- Provision action to updating provision url on device and executing autoprovision
-- @param host	A provision device data
-- @param port	Not used 
function provision (host, port, ...)
    local args        = select(1, ...)
    local logger      = logger
    local http        = require"core.httpClient":new()
    local h           = require"common.helper"
    local connection  = require"common.core".connection
    local sql, cursor = nil, nil

	local httpsSupported = false
	local model_spec, major_version, build_version, minor_version = __parse_firmware(host.fwversion)
	if tonumber(model_spec..major_version..build_version..minor_version) >= 3240024 then
		httpsSupported = true
	end
    
    local prov_proto = "https"
    if not httpsSupported or gconf.HTTPForceUse then
        prov_proto = "http"
    end
    local prov_port   = "https" == prov_proto and "443" or "80"
	local PORT_NUMBER = host.ssl and not gconf.HTTPForceUse and 443 or 80
    
	-- processing registrar address
	local registrar       = nil
	local case            = "hosts"
	local cfg             = {
               FailoverByDns    = gconf.FailoverByDns
              ,ForceResolveName = gconf.ForceResolveName
              ,ForceResolveIp   = gconf.ForceResolveIp
          }
	local addr_t, err_msg = h.resolve_by_hosts(host.ip_src, nil, cfg)
	if not addr_t then
		if err_msg then	-- printing resolving error
			logger:warning("Hosts file was not processed: %s", err_msg)
		end
	else
		registrar = addr_t.domain
	end
	if not registrar then
		registrar, case, err_msg = h.get_registrar(host.ip_src, host.ip)
	end
	logger:debug("Registrar address %q got from %s%s", registrar, case, err_msg and ". Errors: "..err_msg or "")
	
	local options = { authUser = host.auth.user, authPass = host.auth.pass, authMethod = args.authMethod }
    local params = {
        ["LOCAL_HTTP_SERVER_TEMPLATE_TITLE"] = [["W-AIR Base Station"]],
        ["MANAGEMENT_TRANSFER_PROTOCOL"] 	 = "https" == prov_proto and "0x02" or "0x01",
        ["MANAGEMENT_UPLOAD_SCRIPT"] 		 = "",
        ["NETWORK_CONFIGURATION_SERVER"] 	 = registrar..':'..prov_port,
        ["WEB_INPUTS_ALLOWED"] 				 = 1,
        ["FILENAME_BASE_SPECIFIC"] 			 = "?action=setaps",
        ["LOCAL_HTTP_SERVER_AUTH_NAME"] 	 = "admin",
        ["LOCAL_HTTP_SERVER_AUTH_PASS"]      = "admin",
        ["HttpUpdateSystemTimestamp"] 		 = "",
        ["HttpRedirectUrl"] 				 = "reboot.html",
    }
    if "W-AIR SB" == host.btype then -- set params for W-AIR BASE SB or W-AIR BASE SBR2
        params["CONFIGURATION_DOWNLOAD_CTRL"] = 1
        if model_spec and major_version and tonumber(model_spec..major_version) < 355 then
            params["NETWORK_CONFIGURATION_SERVER"] = registrar
        end
    else	-- set params of W-AIR base
        params["CONFIGURATION_DOWNLOAD_CTRL"] = 3
        params["FILENAME_MULTICELL_SPECIFIC"] = "?action=setaps"	    	
    end 

    if host.token_type == "new" then
        params["MANAGEMENT_USERNAME"] = "wmsdevice"
        params["MANAGEMENT_PASSWORD"] = host.token
        params["LOCAL_HTTP_SERVER_AUTH_PASS"] = host.token
    end
	if host.token_type == "old" then
		logger:info("Token for device have old format")
	end

    -- update configuration page
	local answer = http:rawpost(host.ip, PORT_NUMBER, "/update.html", params, options) or {}
	logger:debug("Set autoprovisioning parameters for %q - %s", host.btype or "", answer.code or 408)
	
	-- reboot base station
	answer = http:rawget(host.ip, PORT_NUMBER, "/forcereboot.html", options) or {}
	logger:info("Send http request for autoprovision to %q base station - %s", host.btype or "", answer.code or 408)
end

--- Building and sending provision configuration
-- @param agent		A table with general device information
-- @param mixed		Some parameters. Currently not used
function configuration (agent, ...)
    local json 	 	 = require"dkjson"
    local core	 	 = require"common.core"
    local h 	 	 = require"common.helper"
    local wmsldap    = require"core.wmsldap"
    local pbxutils   = require"core.pbxutils"
    local wair_utils = require"core.wairutils"
    local os	 	 = os
    local logger 	 = logger
    local connection = core.connection
    
    local err 	 = nil
    local sql 	 = nil
    local cursor = nil

    local is_wair_sb = "W-AIR SB" == agent["vendor:additional"]
    local is_wair_sbr2 = "w-air-sbr2" == agent["model"]
	local sip_port 	 = "5060"
    
	local prov_proto = "on" == agent["ssl"] and "https" or "http"
    if gconf.HTTPForceUse then
        prov_proto = "http"
    end
    local prov_port = nil
    if gconf.HTTPForceUse then
        prov_port = "80"
    end
    prov_port = prov_port or h.get_alternate_port()
    prov_port = prov_port or agent["registrar:port"]
    
    -- getting registrar address
    local registrar       = agent["registrar"]
	local registrar_ip 	  = false
	local cfg             = {
               FailoverByDns    = gconf.FailoverByDns
              ,ForceResolveName = gconf.ForceResolveName
              ,ForceResolveIp   = gconf.ForceResolveIp
          }
	local addr_t, err_msg = h.resolve_by_hosts(registrar, nil, cfg)
	if not addr_t then
		if err_msg then
			logger:warning("Hosts file was not processed: %s", err_msg)
		end
	else
		logger:debug(addr_t, "hosts_registrar")
		if addr_t.domain then
			registrar = addr_t.domain
		end
		if addr_t.ip then
			registrar_ip = addr_t.ip
		end
	end
    
    local model_spec, major_version, build_version, minor_version = __parse_firmware(agent["fw:version"])
    local resource = core.load_resource("wildixair")
    local result   = h.tcopy(resource.parameters) or {}	-- get template

    -- retrieve device information
    local token_pass = core.proceedDeviceInfo(agent).token
    
-----------------------------------------------------
--- Provision settings
-----------------------------------------------------

    result["%NETWORK_CONFIGURATION_SERVER%"] = '"' .. registrar .. ':' .. prov_port .. '"'
    result["%MANAGEMENT_TRANSFER_PROTOCOL%"] = "https" == prov_proto and "0x02" or "0x01"
    if not is_wair_sb then
        result["%CONFIGURATION_DOWNLOAD_CTRL%"]  = "0x03"
        result["%FILENAME_MULTICELL_SPECIFIC%"]  = [["?action=setaps"]]
    elseif is_wair_sb and model_spec and major_version and tonumber(model_spec..major_version) < 355 then
        result["%NETWORK_CONFIGURATION_SERVER%"] = '"' .. registrar .. '"'
    end
    sql = "SELECT value FROM autoprov WHERE param=%q AND mac=%q"
    cursor = connection:execute(sql, "description", agent["mac"])
    local description = cursor:fetch(); cursor:close()
    if description and "" ~= description then
        description = string.gsub(description, "[^!@#&_:{}№;'`/<,~|>=%(%)%.%%%+%-%*%?%[%]%^%$%s%w]", "") or "W-AIR" -- removing unsupported characters
        description = string.sub(description, 1, 35)
        result["%LOCAL_HTTP_SERVER_TEMPLATE_TITLE%"] = '"' .. description .. '"'
    end
-----------------------------------------------------
--- Authorization data settings
-----------------------------------------------------
    result["%LOCAL_HTTP_SERVER_AUTH_NAME%"] = '"admin"'
    result["%LOCAL_HTTP_SERVER_AUTH_PASS%"] = string.format([["%s"]], token_pass or "wildix")
    result["%MANAGEMENT_USERNAME%"] = '"wmsdevice"'
    result["%MANAGEMENT_PASSWORD%"] = string.format([["%s"]], token_pass or "wildix")
---------------------------------------------------------
--- Network configuration
---------------------------------------------------------
    sql = "SELECT value FROM autoprov WHERE param=%q AND mac=%q"
    cursor = connection:execute(sql, "settings", agent["mac"])
    local settings = cursor:fetch(); cursor:close()
    
    local t = {}
    local update_db_settings = false
    if settings then
    	t = json.decode(settings)
    end

	-- SYSLOG server (by default syslog server is empty)
    if t["syslog_server"] and t["syslog_server"] ~= "" then
    	result["%SYSLOG_SERVER_IP_ADDRESS%"] = '"'..t["syslog_server"]..'"'
    end
    
    -- VLAN settings (if checkbox disable not send nothing)
    if tonumber(t['vid_voice']) and tonumber(t['vid_voice']) > 0 then
    	result["%NETWORK_VLAN_ID%"] = t['vid_voice']
    end
    if tonumber(t['qos_voice']) and tonumber(t['qos_voice']) > 0 then
	    result["%NETWORK_VLAN_USER_PRIORITY%"] = string.format("0x0%s", t['qos_voice'])
    end
    
    -- set default DNS and NTP settings if not define
	t["dns_server"], t["second_dns_server"] = h.get_prim_sec_dns_servers(t["dns_server"], t["second_dns_server"])
	t["ntp_server"]					   		= t["ntp_server"] or ""
    
    -- use static ip settings
	local is_addr_present = h.is_addr_present(agent["address"], agent["mac"])	-- check duplicate address
	local is_addr_local = h.isLocalIP(agent["address"])		-- device behind NAT can't user static ip
	if t["use_static_ip"] and (is_addr_present or not is_addr_local) then
		if is_addr_present then
			logger:warning("Disable static ip. Other device have the same ip address: %s", agent["address"] or "")
		end
		if not is_addr_local then
			logger:warning("Disable static ip. Device ip address (%s) is placed behind NAT", agent["address"] or "")
		end
		t["use_static_ip"] = nil
		update_db_settings = true
	end
    
    -- get settings for leased device by pbx dhcp
	local is_leased
	if not t["use_static_ip"] or t["dns_server"] == "" or t["ntp_server"] == "" then
		is_leased = h.is_addr_leased(agent["address"])
	end
    if not t["use_static_ip"] and not is_addr_present and is_leased then
		h.set_addr_static(agent["address"], agent["mac"])	-- mark this lease as static
		t["use_static_ip"] = 1
		update_db_settings = true
    end
    if is_leased then
    	if "" == t["dns_server"] or "" == t["second_dns_server"] then
    		local dns_dhcp1, dns_dhcp2 = h.get_dns_servers()
    		if dns_dhcp1 and "" ~= dns_dhcp1 and "" == t["dns_server"] then
    			t["dns_server"] = dns_dhcp1             -- set primary DNS from DHCP config
    			update_db_settings = true
    		end
    		if dns_dhcp2 and "" ~= dns_dhcp2 and "" == t["second_dns_server"] then
    			t["second_dns_server"] = dns_dhcp2      -- set secondary DNS from DHCP config
    			update_db_settings = true
    		end
	    end
	    if t["ntp_server"] == "" then -- set NTP from DHCP config
	        local ntp_dhcp = h.get_ntp_servers()
	        if ntp_dhcp and ntp_dhcp ~= "" then
	            t["ntp_server"] = ntp_dhcp
	            update_db_settings = true
	        end
	    end
	end
    		
	if t["use_static_ip"] == 1 then	-- use static ip
		local _, netmask, gateway = h.host_settings(agent["address"], registrar_ip)
		if netmask and gateway then
			result["%NETWORK_WAN_SETTINGS_DHCP%"] 		 = "0x00"
			result["%NETWORK_WAN_SETTINGS_IP%"] 		 = agent["address"]
			result["%NETWORK_WAN_SETTINGS_SUBNET_MASK%"] = netmask
			result["%NETWORK_WAN_SETTINGS_GATEWAY%"] 	 = gateway
			
			if registrar_ip then
    			if "" == t["dns_server"] then
    				t["dns_server"]    = registrar_ip
    				update_db_settings = true
    			end
    			if "" == t["second_dns_server"] then
    				t["second_dns_server"] = registrar_ip
    				update_db_settings     = true
    			end
    		end
			
			if t["ntp_server"] == "" then
				t["ntp_server"]    = registrar
				update_db_settings = true
			end
		else
			logger:warning("For current network configurations is not possible set static ip, disable it")
			t["use_static_ip"] = nil
			update_db_settings = true
		end
	end
	
	-- update devdb with new settings
	if update_db_settings then
		sql = "INSERT OR REPLACE INTO autoprov (id, param, value, mac) VALUES ((SELECT id FROM autoprov WHERE param=%q AND mac=%q), %q, '%s', %q)"
		connection:execute(sql, "settings", agent["mac"], "settings", json.encode(t), agent["mac"])
	end
	
	-- DNS settings 
	if "" ~= t["dns_server"] then
		result["%NETWORK_WAN_SETTINGS_DNS1%"] = t["dns_server"]
		if "" ~= t["second_dns_server"] then
			result["%NETWORK_WAN_SETTINGS_DNS2%"] = t["second_dns_server"]
		end
	end
	
	-- NTP settings
	if t["ntp_server"] ~= "" then
		result["%NETWORK_SNTP_SERVER%"] = '"'..t["ntp_server"]..'"'
	else
		result["%NETWORK_SNTP_SERVER%"] = '"'..registrar..'"'
	end

	-- Daylight saving time
	if "no" == t["use_dst"] then
		result["%DST_BY_COUNTRY_REGION%"]= "0x00"
		result["%DST_ENABLE%"] = "0x00"
	end
    
    -- Set ToS priority, tonezone
    local tos, toneZone = h.get_sip_settings()
    if tos["tos_sip"] then	-- SIP ToS / DSCP
        result["%SIP_SIP_PRIORITY%"] = string.format("0x%X", tos["tos_sip"])
    end
    if tos["tos"] then	-- RTP / T.38 ToS / DSCP
        result["%SIP_RTP_PRIORITY%"] = string.format("0x%X", tos["tos"])
    end
    
    -- time zone and tone zone (controls by country settings). 
	local timeZone = t["ntp_timezone"]
	if not timeZone or "" == timeZone then
		timeZone = h.getPbxTimeZone()
	end
	local tone     = h.get_dialtone("rtx", toneZone)
	if timeZone and "" ~= timeZone then
		local countryVariantId, countryRegionId, gmtTimeZoneId = __getCountryVariantRegion(timeZone)
		if countryVariantId and countryRegionId then
		    result["%COUNTRY_VARIANT_ID%"] = "0x" .. (countryVariantId or "03")
		    result["%COUNTRY_REGION_ID%"]  = "0x" .. (countryRegionId or "00")
		elseif gmtTimeZoneId then
		    result["%TIMEZONE_BY_COUNTRY_REGION%"] = "0x00"
		    result["%GMT_TIME_ZONE%"] = gmtTimeZoneId
		    result["%DST_BY_COUNTRY_REGION%"]= "0x00"
            result["%DST_ENABLE%"] = "0x00"
		end
	elseif tone and "" ~= tone then
		result["%COUNTRY_VARIANT_ID%"] = "0x" .. tone
	end
    
	-- set SIP transport and SIP port
	local is_addr_remote	= not h.isLocalIP(agent["address"])
	local sip_transport		= "UDP"
	local sip_port			= "5060"
	local gtransp			= nil
	local gsipport			= nil
	if is_addr_remote then
		sip_transport = "TLS"
		sip_port = agent["alternate_port"] or "443"
		gtransp = gconf.SIPTransportRemote
		gsipport = gconf.SIPServerPortRemote
	else
		gtransp	= gconf.SIPTransport
		gsipport = gconf.SIPServerPort
	end
	sip_transport = gtransp or sip_transport
	sip_transport = ("UDP" == sip_transport and "0x01") or ("TCP" == sip_transport and "0x02") or ("TLS" == sip_transport and "0x04")
	sip_port = gsipport or sip_port 
    
    result["%SRV_SIP_TRANSPORT%"]	= sip_transport .. ",0x01,0x01,0x01,0x01,0x01,0x01,0x01,0x01,0x01"

    agent["vendor"] = 'wildixair'
    agent["vendor:additional"] = "w-air-sp" == agent["model"] and agent["model"]:upper() or agent["vendor:additional"]
    agent["vendor:additional"] = "w-air-sbr2" == agent["model"] and agent["model"]:upper() or agent["vendor:additional"]

-----------------------------------------------------
--- Multi cell settings
-----------------------------------------------------
	local devnetworkid = ""
	local ac_code	   = nil
	local autoDectTree = nil
    local syncPlusViaEth = nil
	local sipProxy     = registrar
	local primaryBase  = nil
	local screenNames  = {}
	local users		   = {}
	if not is_wair_sb and not is_wair_sbr2 then
		-- Some static parameters
		result["%NETWORK_SYNC_ENABLE%"] 				= "1"		-- Multi cell system
		result["%NETWORK_AUTO_MULTI_PRIMARY%"]  		= "0x01"	-- Auto create multi primary
		result["%NETWORK_ALLOW_MULTI_PRIMARY%"] 		= "0x01"	-- Allow multi primary
		result["%NETWORK_SYNC_TIME%"] 					= "300"		-- Synchronization time for multi sell
		result["%NETWORK_ROAMING_DEREGISTER%"]			= "0x00"	-- SIP Server support for multiple registrations per account
		
		-- Data Sync for Multi Cell
    	if not is_addr_local then
        	result["%NETWORK_SYNC_DATA_TRANSPORT%"] = "0x00"
    	else
    		result["%NETWORK_SYNC_DATA_TRANSPORT%"] = "0x01"
    	end

        -- Get dev network id wair gateway
        sql = "SELECT network_id, code, sipproxy, primarybase, autodecttree, sync_plus_via_ethernet FROM wair_gateways JOIN wair_networks ON (wair_gateways.network_id == wair_networks.id) WHERE mac=%q"
        cursor = connection:execute(sql, agent["mac"])
        devnetworkid, ac_code, sipProxy, primaryBase, autoDectTree, syncPlusViaEth = cursor:fetch(); cursor:close()
        if not sipProxy or "" == sipProxy then
            sipProxy = registrar
        end
        local currentType = 0
        if devnetworkid then
            result['%NETWORK_SYNC_CHAIN_ID%'] = 238473100 + devnetworkid
            result["%NETWORK_DECT_AUTO_SYNC_TREE_CONFIG%"] = 0 == autoDectTree and "0x00" or "0x01"

            if "W-AIR-SP" == agent["vendor:additional"] then
                result["%IEEE1588_ENABLED%"] = 1 == syncPlusViaEth and "0x01" or "0x00"
            end

            -- get wair_gateways
            local macstring 	= ""
            local treestring	= ""
            local basemaxnum        -- max number of w-air bases in w-air network
            local count
            local devsettings	= nil
            local master_token	= nil
            local current_token = nil

            sql = "SELECT COUNT (*) FROM wair_gateways WHERE network_id=%q"
            cursor = connection:execute(sql, devnetworkid)
            basesInNetwork = tonumber(cursor:fetch()) or 1; cursor:close()
            logger:info("W-AIR network base stations quantity:", basesInNetwork)

            if 1 <= basesInNetwork and 49 >= basesInNetwork then
                result['%NUMBER_OF_BASE_STATIONS%']  = "0x32"
                result['%NUMBER_OF_REPEATER_PER_BASE%'] = "0x03"
                count = 50
                basemaxnum = 50
                logger:info("W-AIR network with up to 50 base stations and up to 3 repeaters connected to each base station")
            elseif 50 <= basesInNetwork and 126 >= basesInNetwork then
                result['%NUMBER_OF_BASE_STATIONS%']  = "0x7F"
                result['%NUMBER_OF_REPEATER_PER_BASE%'] = "0x01"
                count = 127
                basemaxnum = 127
                logger:info("W-AIR network with up to 127 base stations and 1 repeater connected to each base station")
            elseif 127 <= basesInNetwork and 254 >= basesInNetwork then
                result['%NUMBER_OF_BASE_STATIONS%']  = "0xFE"
                result['%NUMBER_OF_REPEATER_PER_BASE%'] = "0x00"
                count = 254
                basemaxnum = 254
                logger:info("W-AIR network with up to 254 base stations and 0 repeaters connected to each base station")
            else
                count = 50
                basemaxnum = 50
                logger:info("Quantity of base stations in W-AIR network can't be calculated, number of base stations/repeaters left as is")
            end

            sql = "SELECT wair_gateways.mac, type, value, token FROM wair_gateways JOIN autoprov ON (autoprov.mac == wair_gateways.mac AND param ='device') WHERE network_id = %q order by type desc;"
            for devmac, devtype, devsettings, devtoken in connection:rows(sql, devnetworkid) do
                devsettings = json.decode(devsettings)
                if devmac == agent["mac"] then
                    currentType = devtype
                    current_token = devtoken
                end
                if 1 == devtype then   -- 1 = base is master in w-air network, 0 = base is slave in w-air network
                    result['%NETWORK_SYNC_PRIMARY_STATIC_IP%']  = devsettings['address']
                    result['%NETWORK_DATA_CONFIG_PRIMARY_MAC%'] = __prepare_mac_string(devmac)
                    users		 = __proceed_users(devmac, agent["vendor:additional"])
                    master_token = devtoken
                end
                macstring  = macstring .. ',' .. __prepare_mac_string(devmac)
                treestring = treestring .. ',' .. '0x00'
                count 	   = count - 1
            end
            result['%NETWORK_SYNC_MAC_CHAIN%'] = string.sub(macstring,2) .. string.rep(",0xFF,0xFF,0xFF,0xFF,0xFF,0xFF",count)
            
            -- Number of SIP accounts before distributed load
            local net_devices = basemaxnum - count
            local accounts = "0x8"
            if net_devices > 0 then
                if gconf.BaseMaxSIPReg then
                    accounts = "0x" .. string.format("%X", gconf.BaseMaxSIPReg)
                elseif net_devices == 1 then
                    accounts = "0x1E"
                end
            end
            result["%NETWORK_SYNC_MAX_SIP_REG_PER_BASE%"] = accounts
            
            if master_token and current_token ~= master_token then
            	local token_type = h.is_new_token(master_token)
            	if token_type == "new" then        	
            		logger:debug("Updating device password to master password: %s", master_token)
            		sql = "UPDATE autoprov SET token=%q WHERE mac=%q AND param=%q"
            		connection:execute(sql, master_token, agent["mac"], "device")
            		result["%LOCAL_HTTP_SERVER_AUTH_PASS%"] = '"'..master_token..'"'
            		result["%MANAGEMENT_PASSWORD%"] = '"'..master_token..'"'
            	else
            		logger:debug("MASTER password found, but it has %s format", token_type)
            	end
            end
            if primaryBase and "" ~= primaryBase then
                result["%NETWORK_SYNC_DATA_TRANSPORT%"] = "0x01"
                result['%NETWORK_SYNC_PRIMARY_STATIC_IP%']  = primaryBase
            end
		end
    else
		result["%NETWORK_ROAMING_DEREGISTER%"] = "0x01"
		users = __proceed_users(agent["mac"], agent["vendor:additional"])
	end

    if is_wair_sbr2 then
        result["%NETWORK_SYNC_ENABLE%"]                         = "0"           -- Disable multi cell system (WMS-10386)
        result["%NETWORK_AUTO_MULTI_PRIMARY%"]                  = "0x00"        -- Don't create multi primary
        result["%NETWORK_ALLOW_MULTI_PRIMARY%"]                 = "0x00"        -- Don't allow multi primary
    end

-----------------------------------------------------
--- Firmware settings
-----------------------------------------------------
    local upgradeFirmwareInfo = wair_utils.getUpgradeFirmwareInfo(agent) or {}
    if upgradeFirmwareInfo["application"] then
    	local appFwVer         = tonumber(upgradeFirmwareInfo["application"]) or tonumber(upgradeFirmwareInfo["application"][1]) or 0
    	local appFwBranch      = tonumber(upgradeFirmwareInfo["application"][2])

        local baseFwVer             = appFwVer
        local baseFwBranch          = appFwBranch
        local wair100FwVer          = appFwVer
        local wair100FwBranch       = appFwBranch
        local wairBasicFwVer        = appFwVer
        local wairBasicFwBranch     = appFwBranch
        local wairLifesaverFwVer    = appFwVer
        local wairLifesaverFwBranch = appFwBranch
        local repeaterFw            = 0
        local repeaterFwBranch      = 0
        local repeater2Fw           = 0 -- new repeater model DECT4027
        local repeater2FwBranch     = 0
        local wairHSFwVer           = appFwVer
        local wairHSFwBranch        = appFwBranch
        local wairMedFwVer          = appFwVer
        local wairMedFwBranch       = appFwBranch
        local wairOfficeFwVer       = appFwVer
        local wairOfficeFwBranch    = appFwBranch

        if is_wair_sb and model_spec and major_version and tonumber(model_spec..major_version) < 355 then -- WMS-2755 wair-sb cannot update firmware
            result["%NETWORK_FWU_SERVER%"] 	 = '"' .. prov_proto .. '://'..registrar .. '"'
        else
            result["%NETWORK_FWU_SERVER%"]   = '"' .. prov_proto .. '://'..registrar..':' .. prov_port .. '"'
        end
        result["%FWU_TFTP_SERVER_PATH%"] = '"/firmwares/wildixair"'
        result["%FWU_TFTP_SERVER_IMAGE_PATH%"] = [["firmwares/wildixair/WAIR-HS-VP"]]

        if upgradeFirmwareInfo["base"] then
            baseFwVer        = tonumber(upgradeFirmwareInfo["base"][1]) or baseFwVer
            baseFwBranch     = tonumber(upgradeFirmwareInfo["base"][2]) or baseFwBranch
        end
        if upgradeFirmwareInfo["wair-100"] then
            wair100FwVer       = tonumber(upgradeFirmwareInfo["wair-100"][1]) or wair100FwVer
            wair100FwBranch    = tonumber(upgradeFirmwareInfo["wair-100"][2]) or wair100FwBranch
        end
        if upgradeFirmwareInfo["wair-basic"] then
            wairBasicFwVer        = tonumber(upgradeFirmwareInfo["wair-basic"][1]) or wairBasicFwVer
            wairBasicFwBranch     = tonumber(upgradeFirmwareInfo["wair-basic"][2]) or wairBasicFwBranch
        end
        if upgradeFirmwareInfo["wair-lifesaver"] then
            wairLifesaverFwVer       = tonumber(upgradeFirmwareInfo["wair-lifesaver"][1]) or wairLifesaverFwVer
            wairLifesaverFwBranch    = tonumber(upgradeFirmwareInfo["wair-lifesaver"][2]) or wairLifesaverFwBranch
        end
        if upgradeFirmwareInfo["repeater"] then
            repeaterFw       = tonumber(upgradeFirmwareInfo["repeater"][1]) or repeaterFw
            repeaterFwBranch = tonumber(upgradeFirmwareInfo["repeater"][2]) or repeaterFwBranch
        end
        if upgradeFirmwareInfo["repeater2"] then
            repeater2Fw       = tonumber(upgradeFirmwareInfo["repeater2"][1]) or repeater2Fw
            repeater2FwBranch = tonumber(upgradeFirmwareInfo["repeater2"][2]) or repeater2FwBranch
        end
        if upgradeFirmwareInfo["wair-hs"] then
            wairHSFwVer      = tonumber(upgradeFirmwareInfo["wair-hs"][1]) or wairHSFwVer
            wairHSFwBranch   = tonumber(upgradeFirmwareInfo["wair-hs"][2]) or wairHSFwBranch
        end
        if upgradeFirmwareInfo["wair-med"] then
            wairMedFwVer      = tonumber(upgradeFirmwareInfo["wair-med"][1]) or wairMedFwVer
            wairMedFwBranch   = tonumber(upgradeFirmwareInfo["wair-med"][2]) or wairMedFwBranch
        end
        if upgradeFirmwareInfo["wair-office"] then
            wairOfficeFwVer      = tonumber(upgradeFirmwareInfo["wair-office"][1]) or wairOfficeFwVer
            wairOfficeFwBranch   = tonumber(upgradeFirmwareInfo["wair-office"][2]) or wairOfficeFwBranch
        end

        -- set repeater and handset version (according to PP_FWU_DEVICE_IDS)
        result["%PP_FWU_SW_VERSIONS%"] 	 = string.format("%s,%s,%s,%s,%s,%s,%s,%s,0,0", wair100FwVer, wairBasicFwVer, wairLifesaverFwVer, repeaterFw, wairHSFwVer, wairMedFwVer, wairOfficeFwVer, repeater2Fw)
        if baseFwBranch then
             -- set repeater and handset branch version.
            result["%PP_FWU_BRANCH_VERSIONS%"]= string.format("%s,%s,%s,%s,%s,%s,%s,%s,0,0", wair100FwBranch, wairBasicFwBranch, wairLifesaverFwBranch, repeaterFwBranch, wairHSFwBranch, wairMedFwBranch, wairOfficeFwBranch, repeater2FwBranch)
        end

        local voice_prompts = upgradeFirmwareInfo["wair-hs-vp"] or ""
        result["%PP_FWU_VOICEPRTS_FILENAMES%"] = string.format([["","","","","%s","","","","",""]], voice_prompts)

        local network_devices_locked = false
        local dev_in_net = devnetworkid and devnetworkid ~= ""
        local allow_network_lock = gconf.DelayDevicesUpgrade ~= "no"
        if dev_in_net then
            if allow_network_lock then
                local lockdata = wair_utils.isLocked(devnetworkid)
                if lockdata then
                    network_devices_locked = true
                    logger:info("Devices upgrade already locked with data: %s", lockdata)
                end
            else
                logger:warning("Unsafe configuration parameter for wair networks 'DelayDevicesUpgrade=no' found")
            end
        end

        if upgradeFirmwareInfo["needUpgrade"] then
        	logger:info("Firmware for device %s %s will be changed from version %s to %s", agent["model"], agent["mac"], agent["fw:version"], upgradeFirmwareInfo["version"] or "")
        	result["%FP_FWU_SW_VERSION%"] = baseFwVer		-- set base version (base load firmware always if this parameter is present)
        	if baseFwBranch then
        		result["%FP_FWU_BRANCH_VERSION%"] = baseFwBranch
        	end

        	-- create devices upgrade lock for wair net
        	if dev_in_net and allow_network_lock and basesInNetwork > 1 and not network_devices_locked then
        		local destver, lmsg = wair_utils.lockNeededVer(devnetworkid, agent["vendor:additional"])
        		if destver then
        			network_devices_locked = true
        			logger:info("Create devices upgrade lock, network: %s", lmsg)
        			wair_utils.devicesLock(devnetworkid, agent["mac"], destver)
        		else
        			logger:info("Devices upgrade lock is not created, reason: %s", lmsg)
        		end
        	end
        else
        	result["%FP_FWU_SW_VERSION%"] = 0 -- disable upgrade
        	result["%FP_FWU_BRANCH_VERSION%"] = 0
        end

        -- disable devices upgrade
        if upgradeFirmwareInfo["needUpgrade"] or network_devices_locked then
            local disable_devices = "0,0,0,0,0,0,0,0,0,0"
            result["%PP_FWU_SW_VERSIONS%"] = disable_devices
            result["%PP_FWU_BRANCH_VERSIONS%"] = disable_devices
        end
    end

---------------------------------------------------------
--- Account settings
---------------------------------------------------------
	if is_wair_sb or is_wair_sbr2 or devnetworkid then
		local screenNames = {}
		local sipNames    = {}
		local authNames   = {}
		local authPasses  = {}
		local voiceMailes = {}
		local ipeis		  = {}
		local ldap_conn = wmsldap:new()
		
        for i, val in pairs(users) do
        	local slot_id    = val[1] or ""
        	local uid	     = val[2]
        	local ipei       = val[3]
        	
            local screenName = nil
            local sipname	 = ""
            local authname	 = ""
            local authpass 	 = ""
            local voicemail	 = ""
            local ldapData 	 = nil
            if tonumber(uid) > 0 then
            	ldapData = ldap_conn:get_lazy("uid",math.abs(uid))
            	if #ldapData ~= 0 then
            		local user_features
            		ldapData   = ldapData[1][2]
            		sipname    = ldapData["telephoneNumber"]
            		authname   = ldapData["telephoneNumber"]
            		authpass   = ldapData["wuserPasswordSip"]
            		screenName = ldapData["cn"]
            		voicemail  = "vmaccess*" .. sipname
            		user_features = require"common.userFeatures.userFeatures":load(uid)
					if user_features["SHAREDVM"] and user_features["SHAREDVM"].status == "1" and user_features["SHAREDVM"].value and user_features["SHAREDVM"].value ~= "" then
						voicemail = "vmaccess*" .. user_features["SHAREDVM"].value
					end
                    local custom_identity = user_features["CURRENT_IDENTITY"]
                    if custom_identity and "" ~= custom_identity then
                        local ldapresult = ldap_conn:get_lazy("telephoneNumber", custom_identity)
                        if ldapresult and 0 ~= #ldapresult then
                            _, customidentity_attrs = unpack(ldapresult[1])
                            if customidentity_attrs.telephoneNumber and customidentity_attrs.cn then
                                screenName = string.format("%s %s", customidentity_attrs.telephoneNumber, customidentity_attrs.cn)
                            end
                        end
                    end
            	end
            end
            if not ldapData or not ldapData.uid then
            	screenName 	= "login"..devnetworkid.."_".. slot_id
                sipname 	= "unknown"..devnetworkid.."_"..slot_id
                authname 	= "unknown"..devnetworkid.. "_".. slot_id
                authpass, err = h.pass_for_unknown(sipname)
                if err then
                    logger:error("password generation error for unknown account:", err)
                end
            end

            -- fix up for user configuration parameters
            -- cut Display name to 29 characters (name="SUBSCR_UA_DATA_DISP_NAME" value="" maxlength="29")
            screenName 	= screenName and '"'..string.sub(screenName, 1, 29)..'"' or ""	
            sipname 	= sipname and '"'..sipname..'"' or ""
            authname 	= authname and '"'..authname..'"' or ""
            authpass 	= authpass and '"'..authpass..'"' or ""
            voicemail 	= voicemail and '"'..voicemail..'"' or ""
            ipei		= ipei and ipei:gsub("(%x%x)", "0x%1,"):sub(1, -2)
            
            -- pack user configuration paramters
            table.insert(screenNames, screenName)
            table.insert(sipNames, sipname)
            table.insert(authNames, authname)
            table.insert(authPasses, authpass)
            table.insert(voiceMailes, voicemail)
            table.insert(ipeis, ipei)
        end
        ldap_conn:close()
        
        -- load additional static parameters from resources.wildixair
        pbxutils.table_concat(result, resource.parameters_subscr or {})
        
        -- load some static parameters only for SB
        if is_wair_sb then
        	pbxutils.table_concat(result, resource.parameters_sb_subscr or {})
        end
        if is_wair_sbr2 then
            pbxutils.table_concat(result, resource.parameters_sbr2_subscr or {})
        end

        result["%SRV_SIP_UA_DATA_DOMAIN%"] 		 = '"' .. sipProxy .. ":" .. sip_port .. '","","","","","","","","",""'
        result["%SRV_SIP_UA_DATA_PROXY_ADDR%"] 	 = '"' .. sipProxy .. ":" .. sip_port .. '","","","","","","","","",""'
        result["%SUBSCR_SIP_UA_DATA_SIP_NAME%"]  = table.concat(sipNames, ',')
        result["%SUBSCR_SIP_UA_DATA_AUTH_NAME%"] = table.concat(authNames, ',')
        result["%SUBSCR_SIP_UA_DATA_AUTH_PASS%"] = table.concat(authPasses, ',')
        result["%SUBSCR_UA_DATA_DISP_NAME%"] 	 = table.concat(screenNames, ',')
        result["%SUBSCR_SIP_UA_DATA_VOICE_MAILBOX_NAME%"]	= table.concat(voiceMailes, ',')
        result["%SUBSCR_SIP_UA_DATA_VOICE_MAIL_NUMBER%"]	= table.concat(voiceMailes, ',')
        -- Set codec priority and codec ordering; 0x00 - G711U, 0x01 - G711A, 0x04 - G729, 0x05 - G722, 0x06 - G726, 0xFF - empty
        if model_spec and major_version and tonumber(model_spec..major_version) >= 303 then
        	if h.is_wideband(agent['address']) then
        		if is_wair_sb and gconf.DisableCodecG729 then
                    -- wideband codecs set (disabled g729 according to WMS-2741) for W-AIR SB: G711A,G711U,G726
                    result["%SRV_SIP_UA_CODEC_PRIORITY%"] = "0x01,0x00,0x06,0xFF,0xFF,0x00,0x01,0x06,0x04,0xFF,0x00,0x01,0x06,0x04,0xFF,0x00,0x01,0x06,0x04,0xFF,0x00,0x01,0x06,0x04,0xFF,0x00,0x01,0x06,0x04,0xFF,0x00,0x01,0x06,0x04,0xFF,0x00,0x01,0x06,0x04,0xFF,0x00,0x01,0x06,0x04,0xFF,0x00,0x01,0x06,0x04,0xFF"
                else
        		    -- wideband codecs set for W-AIR Base: G711A,G711U,G726,G729
        		    result["%SRV_SIP_UA_CODEC_PRIORITY%"] = "0x01,0x00,0x06,0x04,0xFF,0x00,0x01,0x06,0x04,0xFF,0x00,0x01,0x06,0x04,0xFF,0x00,0x01,0x06,0x04,0xFF,0x00,0x01,0x06,0x04,0xFF,0x00,0x01,0x06,0x04,0xFF,0x00,0x01,0x06,0x04,0xFF,0x00,0x01,0x06,0x04,0xFF,0x00,0x01,0x06,0x04,0xFF,0x00,0x01,0x06,0x04,0xFF"
        		end
        		result["%SRV_SIP_USE_OWN_CODEC_PRIORITY%"] = "0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00"
        	else
        		if is_wair_sb and gconf.DisableCodecG729 then
                    -- narrowband codecs set (disabled g729 according to WMS-2741) for W-AIR SB: G726,G711A,G711U 
                    result["%SRV_SIP_UA_CODEC_PRIORITY%"] = "0x06,0x01,0x00,0xFF,0xFF,0x00,0x01,0x06,0x04,0xFF,0x00,0x01,0x06,0x04,0xFF,0x00,0x01,0x06,0x04,0xFF,0x00,0x01,0x06,0x04,0xFF,0x00,0x01,0x06,0x04,0xFF,0x00,0x01,0x06,0x04,0xFF,0x00,0x01,0x06,0x04,0xFF,0x00,0x01,0x06,0x04,0xFF,0x00,0x01,0x06,0x04,0xFF"
                else
        		    -- narrowband codecs set for W-AIR Base: G729,G726,G711A,G711U
        		    result["%SRV_SIP_UA_CODEC_PRIORITY%"] = "0x04,0x06,0x01,0x00,0xFF,0x00,0x01,0x06,0x04,0xFF,0x00,0x01,0x06,0x04,0xFF,0x00,0x01,0x06,0x04,0xFF,0x00,0x01,0x06,0x04,0xFF,0x00,0x01,0x06,0x04,0xFF,0x00,0x01,0x06,0x04,0xFF,0x00,0x01,0x06,0x04,0xFF,0x00,0x01,0x06,0x04,0xFF,0x00,0x01,0x06,0x04,0xFF"
        		end
        		result["%SRV_SIP_USE_OWN_CODEC_PRIORITY%"] = "0x01,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00"
        	end
        	if gconf.g711only then
        	    result["%SRV_SIP_UA_CODEC_PRIORITY%"] = "0x01,0x00,0xFF,0xFF,0xFF,0x00,0x01,0x06,0x04,0xFF,0x00,0x01,0x06,0x04,0xFF,0x00,0x01,0x06,0x04,0xFF,0x00,0x01,0x06,0x04,0xFF,0x00,0x01,0x06,0x04,0xFF,0x00,0x01,0x06,0x04,0xFF,0x00,0x01,0x06,0x04,0xFF,0x00,0x01,0x06,0x04,0xFF,0x00,0x01,0x06,0x04,0xFF"
        	end
        else
	        if h.is_wideband(agent['address']) then
	            result['%SRV_SIP_UA_CODEC_PRIORITY%'] = '0x01,0x00,0x04,0xFF,0xFF,0x00,0x01,0x06,0x04,0xFF,0x00,0x01,0x06,0x04,0xFF,0x00,0x01,0x06,0x04,0xFF,0x00,0x01,0x06,0x04,0xFF,0x00,0x01,0x06,0x04,0xFF,0x00,0x01,0x06,0x04,0xFF,0x00,0x01,0x06,0x04,0xFF,0x00,0x01,0x06,0x04,0xFF,0x00,0x01,0x06,0x04,0xFF'
	        else
	            result['%SRV_SIP_UA_CODEC_PRIORITY%'] = '0x04,0x06,0xFF,0xFF,0xFF,0x00,0x01,0x06,0x04,0xFF,0x00,0x01,0x06,0x04,0xFF,0x00,0x01,0x06,0x04,0xFF,0x00,0x01,0x06,0x04,0xFF,0x00,0x01,0x06,0x04,0xFF,0x00,0x01,0x06,0x04,0xFF,0x00,0x01,0x06,0x04,0xFF,0x00,0x01,0x06,0x04,0xFF,0x00,0x01,0x06,0x04,0xFF'
	        end
	    end
		
        -- Enabling support for XML Phonebooks on fw >=02.55
        if model_spec and major_version and tonumber(model_spec..major_version) >= 450 then
            -- WMS-6162
            result['%PHONEBOOK_SERVER_LOCATION%'] = "0x02"
            result['%XSI_PHONEBOOK_ENABLED%'] = "0x01"
            result['%LDAP_SERVER_ENABLED%'] = "0x00"
            result['%XSI_SERVER%']            = '"' .. sipProxy .. ":" .. prov_port .. '"'
            result["%SUBSCR_XSI_UA_DATA_AUTH_NAME%"] = table.concat(authNames, ',')
            result["%SRV_USE_SIP_FOR_XSI_LOGIN%"] = "0x01,0x01,0x01,0x01,0x01,0x01,0x01,0x01,0x01,0x01"
        elseif model_spec and major_version and tonumber(model_spec..major_version) >= 255 then
            result['%PHONEBOOK_SERVER_LOCATION%'] = "0x02"
            result['%PHONEBOOK_LOCATION%'] 		  = '"' .. sipProxy .. ":" .. prov_port .. '"'
            result['%PHONEBOOK_FILENAME%'] 		  = '""'
            result['%LDAP_ENABLED%'] 			  = "0x01"
            result['%LDAP_SERVER%'] 			  = '"' .. sipProxy .. ":" .. prov_port .. '"'
        else
            result['%PHONEBOOK_SERVER_LOCATION%'] = "0x00"
            result['%PHONEBOOK_LOCATION%'] 		  = '"http://' .. registrar .. '/?action=getphonebook&phonebook=directory&t="'
            result['%PHONEBOOK_FILENAME%'] 	 	  = '"phonebooks"'
        end

        if ac_code and string.len(ac_code) == 4 then
            result['%AC_CODE%'] = string.gsub(ac_code, "(%d%d)(%d%d)", "0x%1,0x%2")
        elseif (is_wair_sb or is_wair_sbr2) and gconf["w-air-sb-pin"] then
            logger:debug("AC Code for W-AIR SB was found in provision.conf")
            result['%AC_CODE%'] = string.gsub(gconf["w-air-sb-pin"], "(%d%d)(%d%d)", "0x%1,0x%2")
        else
            result['%AC_CODE%'] = '0x00,0x00'
        end
    end
    
-----------------------------------------------------
--- Result serialization
-----------------------------------------------------    
    local output = {}
    for section, v in pairs(result) do
        output[#output+1] = section ..":" .. v
    end
    return output, err
end

--- Update W-AIR users database and reprovision base
-- @param host
-- @param port
-- @param mixed	
function setuser (host, port, ...)
    local args = select(1, ...)	-- expect username, callername
    if args["username"] then
    	local logger = require"core.logger":init("[WMS 3.0]", 3)	-- init module for use via login service in dialplan
    	
    	-- check already include modules in discovery.nse or require its (in case execute from dialplan)
    	local h 		 = require"common.helper"
    	local core 		 = require"common.core"
    	local astman 	 = require"core.astman"
    	local pbxutils	 = require"core.pbxutils"
    	local wmsldap    = require"core.wmsldap"
        local connection = core.connection
		
		-- dst attributes (user used for LOGIN)
        local dst_name = args["username"] or ""
        local dst_uid = args["userid"] or ""
        
        -- src attributes (user which call to LOGIN)
        local src_name 	= args["callername"]:gsub("_@_", " ") or ""
        local src_num 	= args["cid"] or ""
        local src_ua 	= args["useragent"] or ""
        local uatable 	= pbxutils.split(src_ua, "_")
        local src_mac 	= uatable[4] or ""
        local src_fmw 	= uatable[3] or ""
        local src_ipei  = (uatable[5] or ""):gsub(":", "")
        
        local btype		 = require"common.wdmacs".get_family(src_mac)
        local is_wair_sb = "W-AIR SB" == btype
        local is_wair_sbr2 = nil

        if uatable[2]:find("W-AIR") and uatable[2]:find("/") then
            model, _ = string.match(uatable[2], "(.*)/(.*)")
            is_wair_sbr2 = "W-AIR-SBR2" == model
        end
        
       	local function __syncExten (sync_exten, removecommand)
       	    local astman = require"core.astman"
            logger:debug("Execute \"sip notify check-sync %s\"", sync_exten)
            local res, err = astman.send_command("sip notify check-sync " .. sync_exten)
            if not res then -- show logs from astman module
                logger:error(err)
            else
                logger:debug(res)
            end
            if removecommand then -- remove old registration
                os.execute(removecommand)
            end
        end
       	logger:debug({
       		["dst_name"] 	  = dst_name, 
       		["dst_uid"]  	  = dst_uid, 
       		["src_name"] 	  = src_name, 
       		["src_num"] 	  = src_num,
       		["src_useragent"] = src_ua, 
       		["src_mac"] 	  = src_mac,
       		["src_ipei"]	  = src_ipei, 
       		["src_fmw"] 	  = src_fmw,
       		["btype"]		  = btype,
       	}, "input arguments")
		
        if src_name:find("login%d") and not src_name:find("login%d+_%d") then
            local old_name = src_name
            local old_slotid = tonumber(src_name:match("(%d+)"))
            src_name = "login0X_" .. (old_slotid + 1) 
            logger:debug("Fix callername: %s --> %s", old_name, src_name)
        end
        
        -- set uid which will be set in wair_users as destination
		local to_uid = dst_uid	
		if dst_name:find("unknown%d") then
			to_uid = "-slot_id"
		end
		logger:info("Calculate to_uid: %s", to_uid or "")
		
		local sync_exten = nil	-- sync_exten used for check-sync		
		local from_uid 	 = nil
        if src_name:find("login") then
        	local network_id = nil
        	local slot_id 	 = src_name:sub(6)
        	slot_id 		 = pbxutils.split(slot_id, "_")
        	if slot_id[2] then
        		network_id 	 = slot_id[1]
        		slot_id    	 = slot_id[2]
        	else
        		slot_id 	 = slot_id[1]
        	end
        	if is_wair_sb or is_wair_sbr2 then			-- set user from W-AIR BASE SB
        		sync_exten = "unknown_" .. slot_id
        	elseif src_name:find("login0X_") then -- set user from loginX
        		sync_exten = "unknown"..(slot_id - 1)
        	else						-- set user from W-AIR BASE
                sync_exten = "unknown".. network_id .. "_" .. slot_id
            end
            from_uid = -slot_id
            logger:info("This unknown user, then (from_uid = -slotid)")
        else
            -- getting from LDAP uid and extension for src (caller)
            local extenNumber = h.get_extenNumber_str()
            local src_res 	  = wmsldap:get(extenNumber, src_num)
            if 0 ~= #src_res then
            	from_uid 	  = src_res[1][2]["uid"]
            	sync_exten 	  = src_res[1][2][extenNumber]
            end
        end
        if not from_uid then
        	logger:error("Source UID was not found. Exiting ...")
        	return false
        end
        
        logger:info("Calculate from_uid: %s", from_uid or "")
        if from_uid == to_uid then
        	logger:info("Calculated from_uid equal to_uid. Do nothing")
        	return false
        end
        
        connection:switch("devdb")
        local sql		= nil
        local cursor    = nil
        local mastermac = src_mac
        if not is_wair_sb and not is_wair_sbr2 then
            -- we got src_name, now we need to find a network id
            sql    = "SELECT network_id FROM wair_gateways WHERE mac=%q"
            cursor = connection:execute(sql, src_mac)
            local network_id = cursor:fetch(); cursor:close()
            logger:info("Getting from DB network id by mac from useragent: %s", network_id or "")
            
            if not network_id or "" == network_id then -- W-AIR network located on the master pbx (multisite W-AIR installation) - removed api request in WMS-8551
                local err = "W-AIR network id not found!"
                logger:error(err)
                return false, err
            end
    		
            -- we got network_id , now need to get a master
            sql    = "SELECT mac FROM wair_gateways WHERE network_id=%q and type=1"
            cursor = connection:execute(sql, network_id)
            mastermac = cursor:fetch(); cursor:close()
            logger:info("Getting from DB mac(mastermac) of the base by network id: %s", mastermac or "")
        end

        -- searching related uid for source handset IPEI
        local usersTableVer = __getUsersTableVersion(mastermac)
        logger:debug("Users table version by mac: %s", _usersTableVerList[usersTableVer])
        if _usersTableVerList["IPEISUPPORT"] <= usersTableVer then		-- XXX: fixing double IPEI by handsets deregister/register actions
        	sql    		   = "SELECT uid FROM wair_users WHERE mac='%s' AND ipei='%s'"
        	cursor 		   = connection:execute(sql, mastermac, src_ipei)
        	local ipei_uid = cursor:fetch(); cursor:close()
        	if ipei_uid and ipei_uid ~= from_uid then
        		logger:debug("Found used slot by ipei %q with uid %q. Try to use this slot ...", src_ipei or "", ipei_uid)
        		from_uid = ipei_uid
        	end 
        end
        
        local provConfData = h.get_provConfig() or {}
        local resetHandset = provConfData.wildixair and provConfData.wildixair.ResetHandset
		if 0 > tonumber(from_uid) then	-- this unknown user
			local slot_id = -from_uid
			if _usersTableVerList["IPEISUPPORT"] <= usersTableVer then
				logger:info("Update DB: user to_uid - %q for slot_id - %q and mac(mastermac) - %s and ipei - %s", to_uid or "", slot_id or "", mastermac or "", src_ipei or "")
				sql 		  = "UPDATE wair_users SET uid=%s, ipei='%s' WHERE slot_id=%s AND mac='%s'"
				local sqlIpei = dst_name:find("unknown%d") and "FFFFFFFFFF" or src_ipei
				connection:execute(sql, to_uid, sqlIpei, slot_id, mastermac)
			else
				logger:info("Update DB: user to_uid - %q for slot_id - %q and mac(mastermac) - %s", to_uid or "", slot_id or "", mastermac or "")
				sql = "UPDATE wair_users SET uid=%s WHERE slot_id=%s AND mac='%s'"
				connection:execute(sql, to_uid, slot_id, mastermac)
			end
		else						-- this user now logged in system
            if _usersTableVerList["IPEISUPPORT"] <= usersTableVer then
                logger:info("Update DB: user to_uid - %q for from_uid - %q and mac(mastermac) - %s and ipei - %q", to_uid or "", from_uid or "", mastermac or "", src_ipei or "")
                local sqlIpei = dst_name:find("unknown%d") and ",ipei='FFFFFFFFFF'" or ""
                sql = "UPDATE wair_users SET uid=%s %s WHERE uid=%s AND mac='%s'"
                -- do not add "WHERE src ipei" to the "SQL UPDATE" in case restetHandset flag is enabled - all users slots will be cleared
                sql = not resetHandset and sql.."AND (ipei='"..src_ipei.."' OR ipei='FFFFFFFFFF')" or sql
                connection:execute(sql, to_uid, sqlIpei, from_uid, mastermac)
			else
    			logger:info("Update DB: user to_uid - %q for from_uid - %q and mac(mastermac) - %s", to_uid or "", from_uid or "", mastermac or "")
            	sql			  = "UPDATE wair_users SET uid=%s WHERE uid=%s AND mac='%s'"
            	connection:execute(sql, to_uid, from_uid, mastermac)
            end
        end
        connection:close()
        if sync_exten then
            __syncExten(sync_exten, args["removecommand"])
        end
    end
	return true
end

function setfeatures (...) end
