module( ... or "wildixfxo", package.seeall)

hints = {
	--[[
	======================
	Wildix FXO
	----------------------
		HTTP/1.1 401 Unauthorized
		Server: GoAhead-Webs
		Date: Tue Jun 14 09:43:54 2011
		WWW-Authenticate: Basic realm="Wildix Device Management"
		Pragma: no-cache
		Cache-Control: no-cache
		Content-Type: text/html
	]]
    {
         ['server']           = 'GoAhead-Webs'
        ,['www-authenticate'] = 'Basic realm="Wildix Device Management"'
    }
}

--- Convert the impedance index to parameter value
--  @param idx - index of the impedance.
--  @return  impedance parameter value for device
local function __getImpedance(idx)
	if not idx then
		return false, "Impedance index not found"
	end
	local rv = 15
	--[[
		value="0">600 R
        value="1">900 R
        value="2">270 R + 150nf
        value="3">220 R + 120 nf
        value="4">370 R + 310 nF
        value="5">320 R + 230 nF
        value="6">370 R + 110 nF
        value="7">275 R + 115 nF
        value="8">120 R + 110 nF
        value="9">350 R + 210 nF
        value="10">200 R + 100 nF
        value="11">600 R + 2.16 μF
        value="12">900 R + 1 μF
        value="13">900 R + 2.16 μF
        value="14">600 R + 1 μF
        value="15">Global
    ]]--
	local impedance = {
		[1]  = 2, -- 270+750//150nF (Italy/Sweden)
		[2]  = 3, -- 220+820//115nF (Germany)
		[3]  = 4, -- 370+620//310nF (UK)
		[4]  = 3, -- 180+910//150nF (France)
		[5]  = 8, -- 120+820//110nF (Norway)
		[6]  = 8, -- 150+830//72nF (Belgium)
		[7]  = 4, -- 400+500//330nF (Denmark)
		[8]  = 0, -- 600 R
		[9]  = 1, -- 900 R
		[10] = 5, -- 320 R + 230 nF
		[11] = 6, -- 370 R + 110 nF
		[12] = 7, -- 275 R + 115 nF
		[13] = 9, -- 350 R + 210 nF
		[14] = 10, -- 200 R + 100 nF
		[15] = 11, -- 600 R + 2.16 μF
		[16] = 12, -- 900 R + 1 μF
		[17] = 13, -- 900 R + 2.16 μF
		[18] = 14, -- 600 R + 1 μF
		[19] = 15  -- Global
		
	}
	if impedance[tonumber(idx)] then
		rv = impedance[tonumber(idx)]
	end
	
	return rv
end

--- Check device version for skipping authorization or not
-- @note From 3.08(q - FXO and o - FXS)
function is_skip_authorization(agent)
	if not agent or "table" ~= type(agent) then
		return false
	end
	
	local function fw_parse(fw_ver)
		local num1, num2, char = (fw_ver or ""):match("(%d+)%.(%d+)(%a)")
		local num = (num1 or "") .. (num2 or "")
		return tonumber(num) or 0, char or ""
	end
	
	local num_ver, char_ver = fw_parse(agent["fw:version"])
	if num_ver < 308 then
		return true
	elseif agent["vendor:additional"] == "FXO" and num_ver == 308 and char_ver < "q" then
		return true
	elseif agent["vendor:additional"] == "W04FXS" and num_ver == 308 and char_ver < "o" then
		return true
	end
	
	return false
end

function verify (host, port)
    local logger     = logger
    local xml        = require"core.htmlparser"
    local http       = require"core.httpClient":new()
	local helper     = require"common.helper"
	local connection = require"common.core".connection
	
	
	local is_ssl_negotiate = 443 == port.number
	local PORT_NUMBER      = is_ssl_negotiate and not gconf.HTTPForceUse and 443 or 80
	local t                = {["ssl"] = is_ssl_negotiate}
	
	local gpassword  = helper.get_gpassword()
	local auth_table = {"root:wil!superroot!dix", "admin:admin", "admin:" .. gpassword}

	local sql, cursor, token_pass, dublicate_ip
	if host.mac_addr_str == "" then
		sql = [[ SELECT token, count(id) FROM autoprov WHERE param=%q AND value LIKE '%%"address":%%"%s"%%' AND value LIKE '%%"model":%%"w04fx%%"%%' ]]
		cursor = connection:execute(sql, "device", host.ip or "")
		token_pass, dublicate_ip = cursor:fetch(); cursor:close()
		if tonumber(dublicate_ip) > 1 then
			logger:error("Found %s duplicated address %q at provisioning database", dublicate_ip, host.ip or "")
			return t
		end
	else
   		sql = "SELECT token FROM autoprov WHERE param=%q AND mac=%q"
    	cursor = connection:execute(sql, "device", host.mac_addr_str)
    	token_pass = cursor:fetch(); cursor:close()
    end
    
    t["token_type"] = helper.is_new_token(token_pass)
    if t["token_type"] == "new" then
    	table.insert(auth_table, 1, "root:wil--dix"..token_pass)
    	t["token"] = token_pass
    end
    logger:debug(auth_table, "auth_table")
    
    -- check authorization
    local answer = nil
    for _, authpair in pairs(auth_table) do
        local login, pass = authpair:match("(.*):(.*)")
        answer = http:rawget(host.ip, PORT_NUMBER, "/Device_status.asp", { authUser = login, authPass = pass }) or {}
        if 200 == answer.code then
            t["auth"] = {user = login, pass = pass}
            break
        end
    end
    if not t["auth"] then
		logger:error("Wrong authorization for host %s", host.ip)
		return t
	end

	local doc      = xml:new(answer.body)
	local model    = doc:find('//script[contains(text(),"var mModel")]'):match('mModel = "([^%"]+)"')
	local firmware = doc:find('//script[contains(text(),"var mFirmware")]'):match('mFirmware = ".*[fxo|fxs]%.(.+)%.bin"')
	t["model"]     = model or ""
	t["fwversion"] = firmware or ""
	
	if host.mac_addr_str == "" then
		host.mac_addr_str = doc:find('//script[contains(text(),"var mMac")]'):match('mMac = "([^%"]+)"'):lower()
	end
	t["mac_addr_str"] = host.mac_addr_str:lower()
	
	return t
end

function provision (host, port, ...)
	local args = select(1, ...)
	local http = require"core.httpClient":new()
	local h    = require"common.helper"
	
	local PORT_NUMBER = host.ssl and 443 or 80
	local http_protocol, http_port = "http", 443
	
	-- processing registrar address
	local registrar       = nil
	local case            = "hosts"
    local cfg             = {
                FailoverByDns    = gconf.FailoverByDns
               ,ForceResolveName = gconf.ForceResolveName
               ,ForceResolveIp   = gconf.ForceResolveIp
          }
	local addr_t, err_msg = h.resolve_by_hosts(host.ip_src, nil, cfg)
	if not addr_t then
		if err_msg then	-- printing resolving error
			logger:warning("Hosts file was not processed: %s", err_msg)
		end
	else
		registrar = addr_t.domain
	end
	if not registrar then
		registrar, case, err_msg = h.get_registrar(host.ip_src, host.ip) 
	end
	logger:debug("Registrar address %q got from %s%s", registrar, case, err_msg and ". Errors: "..err_msg or "")
	
	local gpassword
	if host.token_type == "new" then
		gpassword = host.token or h.get_gpassword()
	else
		gpassword = h.get_gpassword()
		if host.token_type == "old" then
			logger:info("Token has old format")
		end
	end
	
	local provlink = string.gsub(
		"$protocol://$registrar:$port/?action=setaps", "%$(%w+)",
		{
			protocol  = http_protocol,
			registrar = registrar,
			port      = http_port,
		}
	)

	local options = {
         authUser = host.auth.user
        ,authPass = host.auth.pass
        ,timeout  = 2
	}
    
    local data   = {
        ["SendSignal"] = "YES",
        ["provtype"] = "3",
        ["ems"] = "0",
        ["emsdp"] = "61005",
        ["emssri"] = "3600",
        ["emsci"] = "",
        ["IPAddress1"] = "",
        ["IPAddress2"] = "",
        ["IPAddress3"] = "",
        ["IPAddress4"] = "",
        ["emsp"] = "61003",
        ["denc"] = "1",
        ["httpcfg"] = "1",
        ["httpurl"] = provlink,
        ["httpcri"] = "14400",
        ["httpcu"] = "wmsdevice",
        ["httpcp"] = gpassword,
    }
	local answer = http:rawpost(host.ip, PORT_NUMBER, "/ProvisioningSQL.asp", data, options) or {}
	logger:debug("Updating provisioning settings at %s - %s", host.model, answer.code or 408)

	-- Reboot
	answer = http:rawget(host.ip, PORT_NUMBER, "/ExecuteCommand.asp?cmdid=reboot", options) or {}
	logger:info("Sending http request for autoprovision to %s - %s", host.model, answer.code or 408)
end

function configuration (agent, ...)
	local json        = require"dkjson"
	local md5         = require"md5"
	local h           = require"common.helper"
	local core        = require"common.core"
	local wmsldap     = require"core.wmsldap"
	local os, logger  = os, logger
	local connection  = core.connection
	local err, sql, cursor
	local secret      = {}

	local http_protocol, http_port = 'http', h.get_alternate_port() or '443'
	local tlsSupported  = false
	local devFwInfo	  = h.parse_firmware(agent["fw:version"], "sht_v1")
	
	--the firmware version since 5.01 supports tls
	if devFwInfo and 501 <= devFwInfo then
		tlsSupported = true
	end
	
	-- getting registrar address
	local registrar       = agent["registrar"]
	local registrar_ip 	  = false
    local cfg             = {
                FailoverByDns    = gconf.FailoverByDns
               ,ForceResolveName = gconf.ForceResolveName
               ,ForceResolveIp   = gconf.ForceResolveIp
          }
	local addr_t, err_msg = h.resolve_by_hosts(registrar, nil, cfg)
	if not addr_t then
		if err_msg then
			logger:warning("Hosts file was not processed: %s", err_msg)
		end
	else
		logger:debug(addr_t, "hosts_registrar")
		if addr_t.domain then
			registrar = addr_t.domain
		end
		if addr_t.ip then
			registrar_ip = addr_t.ip
		end
	end
	
	local is_addr_remote = not h.isLocalIP(agent["address"])

	local resource
	if agent["vendor:additional"] == "W04FXS" then
		resource = core.load_resource("wildixfxs4")
	else
		resource = core.load_resource("wildixfxo")
	end
	local result = h.tcopy(resource.parameters) or {}

    -- retrieve device information
    local token_pass = core.proceedDeviceInfo(agent).token

	local nports = tonumber(resource.models[agent["model"]])
	local tports, trunkname = {} -- data for FXO settings

	-- Failover
    local pbxFailover     = nil
    local failoverSipPort = nil
    local failoverObj = require"common.failover":new()
    if failoverObj and "1" == failoverObj:isEnabled() then
        local failoverCfg = failoverObj:getConfig()
        pbxFailover = failoverCfg and failoverCfg.pbxFailoverDomain
        failoverSipPort = failoverCfg and failoverCfg.sipPort
        logger:debug(failoverCfg, "failoverCfg")
    end

	--Set SIP transport and SIP port
	local tos_data, defaultToneZone = h.get_sip_settings()
	local transport = "udp"
	local sip_port = "5060"
	if tlsSupported then
		transport = "TLS"
		sip_port  = agent.alternate_port or "443"
	end
	local gtransp = nil
	local gsipport = nil
	if is_addr_remote then
		gtransp = gconf.SIPTransportRemote
		gsipport = gconf.SIPServerPortRemote
	else
		gtransp = gconf.SIPTransport
		gsipport =  gconf.SIPServerPort
	end
	transport = gtransp or transport
	sip_port = gsipport or sip_port
    if pbxFailover then
            transport       = "TLS"
            sip_port        = agent["alternate_port"] or "443"
            failoverSipPort = failoverSipPort or sip_port
    end
	transport = (transport == "TLS" and "3") or "0"

	-- FXO configuration 
	if agent["vendor:additional"] == "FXO" then
		-- Get default port settings 
		connection:switch ("wmsdb")
		sql = "select count(id) from dialplan_rulenames where rulename=%q"
		cursor = connection:execute(sql, "main")
		local count = cursor:fetch(); cursor:close()
		resource.port_default.port_context = tonumber(count) > 0 and "main" or "internalcalls"
		local is_upd_wmsdb = false

		for i=1, nports do
			for k,v in pairs(resource.port_default) do
				if not tports[k] then tports[k] = {} end
				tports[k][i] = v
			end
		end

		-- Check trunkid
		connection:switch ("devdb")
		sql = "select value from autoprov where param = %q and mac = %q"
		cursor = connection:execute(sql, "trunkid", agent["mac"])
		local trunkid = cursor:fetch(); cursor:close()
		
		if trunkid then		-- we have a trunk
			connection:switch ("wmsdb")
			sql = "select username, settings_mgw from pstn_trunks where id = %s"
			cursor = connection:execute(sql, trunkid)
			local username, settings_mgw = cursor:fetch(); cursor:close()
			trunkname = username
			
			local settings_mgw = json.decode(settings_mgw or {});
			for k,v in pairs(settings_mgw) do
				tports[k] = v
			end

            for i=1, nports do
                secret[i] = h.get_sip_pass(trunkname, trunkid, i-1, agent["mac"])
            end

			if "yes" == tports["dc_impedance"] then
				result["Devconf"]["tbr21"] = "1"
			end
		else -- Trunk autocreation
			local lowertech, tnames = "fxo", {}
			is_upd_wmsdb = true

			connection:switch ("wmsdb")
			sql = "select username from pstn_trunks where technology=%q"
			for name in connection:rows(sql, lowertech) do
				table.insert(tnames, name)
			end
			table.sort(tnames, function(a,b) a = tonumber(a:sub(4)); b = tonumber(b:sub(4));  return a<b end)

			for i=1,#tnames do
				if (lowertech .. tostring(i-1)) ~= tnames[i] then
					trunkname = lowertech .. tostring(i-1)
					break
				end
			end
			if not trunkname then
				trunkname = lowertech .. tostring(#tnames)
			end
			
			-- Update db for creating sip-extra.conf
			sql = [[ insert or replace into pstn_trunks (id,username,ports,technology,settings_pbx,settings_mgw) values(
					(select id from pstn_trunks where username = %q),%q,%d,%q,%q,'%s') ]]
			connection:execute(sql, trunkname, trunkname, nports, "fxo", "{}", json.encode(tports))
			
			sql = "select id from pstn_trunks where username = %q"
			cursor = connection:execute(sql, trunkname)
			local trunkid = cursor:fetch(); cursor:close()
			connection:switch ("devdb")		-- Set trunk id
			sql = [[ insert or replace into autoprov (id,param,value,mac) values(
					(select id from autoprov where param=%q and mac=%q),%q,'%s',%q) ]]
			connection:execute(sql, "trunkid", agent["mac"], "trunkid", trunkid, agent["mac"]) 
			
			connection:switch("htables") -- update kamailio tables
			connection:execute("delete from dynamic_peers where wms_id = %q", trunkid)	-- remove old trunk stuff
			local username, key
			sql = "insert into %s (key_name,key_value,wms_id) values (%q,%q,%q)"	-- dynamic_peers
			for i=1, nports do
				username = string.format("%s_%d", trunkname, i-1)
				secret[i] = h.generate_pass()
				key = string.format ("%s:%s", agent["address"], tostring (5059 + i))
				connection:execute (sql, "dynamic_peers", username .. "::authname", username, trunkid)
				connection:execute (sql, "dynamic_peers", username .. "::secret", secret[i], trunkid)
			end
			
			h.create_sip_extra()
			os.execute "sudo kamcmd htable.reload dynamic_peers"	-- reload kamailio
			os.execute "sudo callweaver -rx \"sip reload \" > /dev/null"	-- reload to update callweaver peers

		end
		
		if is_upd_wmsdb then
			connection:switch ("wmsdb")	-- update wmsdb
			sql = [[ insert or replace into pstn_trunks (id,username,ports,technology,settings_pbx,settings_mgw) values(
					(select id from pstn_trunks where username = %q),%q,%d,%q,%q,'%s') ]]
			connection:execute(sql, trunkname, trunkname, nports, "fxo", "{}", json.encode(tports))
		end
		
		-- Generate ports
		local tmatchedinclist, tusr = {"1","2","4","8"}, {}
		local username, siptk, fxo, route
		for i=1, nports do
			result["Devconf"]["line" .. i] = { ["active"] = "1" }

			username = string.format ("%s_%d",trunkname,i-1)
			tusr[#tusr+1] = {username, secret[i]}

			siptk = "siptk" .. i
			result["Devconf"][siptk] = {
				["regtype"] = transport, -- UDP:0, predefine:1, TLS:3 default is 1
				["domain"] = registrar,
				["proxy"] = registrar,
				["proxyport"] = sip_port,
				["outboundproxy"] = registrar,
				["outboundproxyport"] = sip_port,
				["ttl"] = "300", -- Register time to live (sec)
				["telno"] = username, -- fxo0_0 .. fxo3_0
				["userid"] = username,
				["password"] = secret[i],
				["displayname"] = "FXO Line" .. i,
				["rejectanonymcall"] = "0",
				["outcallerid_displayname"] = "4",
				["outcallerid_userid"] = "1",
				["twostagedial"] = "0",
				["keepalive"] = "1",
				["keepsec"] = "50",
			}

			fxo = "fxo" .. i
			result["Devconf"][fxo] = {
				["curdropfordisconn"]         = "1",
				["incomingcallhandling"]      = "0",
				["polarityreversaldetection"] = tports["port_preversal_detection"][i],
				["FlashTime"]                 = tports["port_flash_time"][i],
				["hotlinetel"]                = tports["port_hotline_number"][i],
				["telno"]                     = tports["port_hotline_number"][i],
				["faxrelay"]                  = tports["port_fax_type"][i],
				["voicevolume"]               = tports["port_codec_input_gain"][i],
				["remotercvvolume"]           = tports["port_codec_output_gain"][i],
				["dialanswertime"]            = tports["port_answer_time"][i],
				["answerringcnt"]             = (tonumber(tports["port_pstn_ring_count"][i]) > 0 and "2" or "0"),
				["calleridtype"]              = tports["port_cid_type"][i],
			}
				
			-- Generate incoming routes
			route = "route" .. i
			result["Devconf"][route] = {
				["enable"] = "1",
				["incomingcalltype"] = "1", -- Line -> FXO (incoming)
				["matchprefix"] = tports["port_hotline_number"][i],
				["matchedincominglist"] = tmatchedinclist[i],
				["matchlen"] = string.len (tports["port_hotline_number"][i]),
				["noanstimeout"] = "120",
				["primary"] = {
					["outgoingtype"] = "0", -- FXO -> Voip
					["huntingtype"] = "0", -- Priority Ring:0,Cyclic Ring:1,Simultaneous Ring:2
					["routinglist"] = (i-1), -- none:-1, trunk1: 0, trunk2: 1, trunk3: 2, trunk4: 3
					["huntingcycle"] = "1",
				},
				["priext"] = {
					["dmgroup"] = "0",
				},
				["bakext"] = {
					["bakrouteactive"] = pbxFailover and 1 or "0", -- disabled:0, enabled:1
					["reroutedmgroup"] = "0",
				}
			}
			
			if pbxFailover then
                local siptkFailover = "siptk"..(nports+i)
                -- sip trunks to failover PBX
                result["Devconf"][siptkFailover] = {
                    ["regtype"] = transport, -- UDP:0, predefine:1, TLS:3 default is 1
                    ["domain"] = pbxFailover,
                    ["proxy"] = pbxFailover,
                    ["proxyport"] = failoverSipPort,
                    ["outboundproxy"] = pbxFailover,
                    ["outboundproxyport"] = failoverSipPort,
                    ["ttl"] = "300", -- Register time to live (sec)
                    ["telno"] = username, -- fxo0_0 .. fxo3_0
                    ["userid"] = username,
                    ["password"] = secret[i],
                    ["displayname"] = "FXO Line" .. i,
                    ["rejectanonymcall"] = "0",
                    ["outcallerid_displayname"] = "1",
                    ["outcallerid_userid"] = "1",
                    ["twostagedial"] = "0",
                    ["keepalive"] = "1",
                    ["keepsec"] = "50",
                }
                -- backup route to failover PBX
                result["Devconf"][route]["bak"] = {
                    ["outgoingtype"] = "0", -- FXO -> Voip
                    ["huntingtype"] = "0", -- Priority Ring:0,Cyclic Ring:1,Simultaneous Ring:2
                    ["routinglist"] = (nports+i-1), -- none:-1, trunk1: 0, trunk2: 1, trunk3: 2, trunk4: 3
                    ["huntingcycle"] = "1",
                }
            end
		end
		result["Devconf"]["impedance"] = __getImpedance(tports["impedance"]) or 15 -- 15 - default "Global" value. 
	end
	
	-- FXS configuration -- Generate lines (set account, line properties)
	if agent["vendor:additional"] == "W04FXS" then
		local account_data = {}
		local res, person, unknown_user
		sql = "select value, flag from autoprov where param=%q and mac=%q"
		for username, port in connection:rows (sql, "username", agent["mac"]) do
			if username:find("unknown") then
				unknown_user = username		-- save old unknown account for current configuration
			else
				res = wmsldap:get(h.get_extenNumber_str(), username)
				person = {}
				if res and res[1] and username then
					_, person = unpack(res[1])
					local user_features = person.uid and require"common.userFeatures.userFeatures":load(person.uid) or {}
					account_data[tonumber(port)] = {
						["login"]       = username,
						["password"]    = person["wuserPasswordSip"],
						["displayName"] = person["cn"] or username --primary display name
					}
					if user_features.HOTLINE then
						account_data[tonumber(port)]["hotline"] = {
							enable = user_features.HOTLINE.status or 0,
							number = user_features.HOTLINE.value or ""
						}
					else
						account_data[tonumber(port)]["hotline"] = {enable = 0, number = ""}
					end
				elseif
					not res or not res[1] then
					logger:error("ldap is not ok")
					return {}, ""
				end
			end
		end
		
		if not unknown_user then	-- get new unknown user
			local unknown_ports = {}	-- list port without account
			for i=1, nports do
				if not account_data[i] then
					table.insert (unknown_ports, i)
				end
			end
			if unknown_ports[1] then	-- write to autoprov used unknown account
				unknown_user = core.register(agent["mac"], nil, unknown_ports) 
			end
		end
		
		local fxs
		for i=1, nports do
			result["Devconf"]["line" .. i] = { ["active"] = "1" }
			if account_data[i] == nil then
				local passwd, err = h.pass_for_unknown(unknown_user)
				if err then
					logger:error("password generation error for unknown account:", err)
				end
				account_data[i] = {
					["login"] = unknown_user,
					["password"] = passwd,
					["displayName"] = "port",
					["hotline"] = {enable = 0, number = ""}
				}
			end
			fxs = "fxs" .. i
			result["Devconf"][fxs] = {
				["reg"] = transport, -- Register type (UDP:0, predefine:1, internal:2, TLS:3, default 1)
				["telno"] = account_data[i]["login"],
				["userid"] = account_data[i]["login"],
				["password"] = account_data[i]["password"],
				["displayname"] = account_data[i]["displayName"] .. "-" .. i, --display name like "User name-1", where '-1' - port number (used in service 99 LOGIN to identify FXS port number)
				["forwardreason"] = "0", -- Forward type (null:0, unconditional:1, busy:2, no answer:4, busy+no answer:6, default is 0)
				["forwardtel"] = "",
				["noanstimeout"] = "120", -- No answer time out (seconds) (default 120 sec)
				["callwaiting"] = "3", -- Call waiting (enabled:3, disabled:2, default 3)
				["rejectanonymcall"] = "0", -- Reject anonymous calls (disabled:0, enabled:1, default 0)
				["hotline"] = account_data[i]["hotline"]["enable"], -- Hotline type(disabled:0, enabled:1, default 0)
				["hotlinetel"] = account_data[i]["hotline"]["number"], -- Hotline number
				["polarityreversal"] = "0", -- Polarity reversal generation (disabled:0, enabled:1, default is 0)
				["curdropfordisconn"] = "1", -- Current drop generation (disabled:0, enabled:1,default is 1)
				["voicevolume"] = "32", -- Input (encode) Gain  (interval 22~37 (-10db ~ +5db), default 32(0db)
				["remotercvvolume"] = "32", -- Output (decode) Gain (interval 22~37 (-10 ~ +5db), default 32(0db))
				["faxrelay"] = "2", -- disabled:0, bypass:1, T38:2, default 2
				["vms"] = "0", -- Voice Mail subscription (disabled:0, enabled:1, default is 0)
				["calleridtype"] = "1", -- Caller ID (inhibit:0, transparent:1, default 0)
				["calleridtovoip"] = "1" -- Caller ID SIP (inhibit:0, transparent:1, default 0)
			}
		end

		-- SIP proxy options
		result["Devconf"]["px"] = {
			["registrardomain"] = registrar,
			["primary"] = {
				["registrarpx"] = registrar, -- proxy server
				["registrarport"] = sip_port, -- proxy server port
				["outxpx"] = registrar, -- Outbound sever proxy
				["outxpxport"] = sip_port, -- outbound proxy sever port
				["keepalive"] = pbxFailover and "1" or "0", -- Primary Proxy Server Keep Alive: (disabled:0, enabled:1, default 0)
				["keepsec"] = "10" -- Keep Alive Time (sec): default 30
			},
			["havesecondpx"] = pbxFailover and "1" or "0", -- secondary proxy server status (disabled:0, enabled:1, default 0)
			["second"] = {
				["registrarpx"] = pbxFailover or "", -- proxy server
				["registrarport"] = failoverSipPort, -- proxy server port
				["outxpx"] = pbxFailover or "", -- Outbound sever proxy
				["outxpxport"] = failoverSipPort, -- outbound proxy sever port
				["keepalive"] = pbxFailover and "1" or "0", -- Primary Proxy Server Keep Alive: (disabled:0, enabled:1, default 0)
                ["keepsec"] = "10" -- Keep Alive Time (sec): default 30
			},
			["registerexpires"] = "300" -- SIP Register Expires (seconds) (default 120sec)
		}
	end
	
	-- Common configuration 	
	-- Firmware 
	local firmwares = core.get_firmware_pack({ vendor = "wildixfxx" })
	if firmwares and firmwares [agent["model"]] then
		local tcfg = firmwares [agent["model"]]
		local filename = tcfg["application"]

		local pack_version = tcfg["version"]
		local agent_version = agent["fw:version"]

		local dir = "/var/www/firmwares/wildixfxx/"
		local preventUpgrade = false

		if pack_version and (pack_version == "2.07l" or pack_version == "2.07q") then  -- we should prevent upgrade/downgrade for this firmwares versions
			preventUpgrade  =  true
		end

		if agent_version ~= pack_version and h.is_exists (dir .. filename) and not preventUpgrade then
			result["Firmware"]["frmurl"] = filename
			result["Firmware"]["fm_updatetime"] = "99:99"
		end
	end

	-- Set codec priority
	local netband_str = h.is_wideband(agent["address"]) and "wideband" or "narrowband"
	local codecs = resource.codecs[netband_str]
	for param, value in pairs (codecs) do
		result["Devconf"][param] = value
	end
	if netband_str == "wideband" then                                                                                                                                                                                                   
		result["Devconf"]["l1acpr"] = "0"                                                                                                                                                                                           
	else                                                                                                                                                                                                                                
		result["Devconf"]["l1acpr"] = "1"                                                                                                                                                                                           
	end         
	
	-- Network settings
	connection:switch ("devdb")
	sql = "SELECT value FROM autoprov WHERE param=%q AND mac=%q"
	cursor = connection:execute (sql, "settings", agent["mac"])
	local settings_json = cursor:fetch(); cursor:close()
	
	local t = {}
	local update_db_settings = false
	if settings_json then
		t = json.decode(settings_json)
	end
	
	-- SYSLOG settings
	if require"core.pbxutils".get_ipv4(t["syslog_server"]) then
		result["Devconf"]["slog"] = "1"
		result["Devconf"]["slogs"] = t["syslog_server"]
		result["Devconf"]["dbl"] = "phoneMgr=7,SipMgr=7,sntp=7,devmgr=7,emsclt=7"
		result["Devconf"]["dbm"] = "phoneMgr=15,SipMgr=15,sntp=15,devmgr=15,emsclt=15"	
		result["Devconf"]["DSPDebug"] = "1"
		result["Devconf"]["DSPLogServer"] = t["syslog_server"]
		result["Devconf"]["slogen"] = "1"
		result["Devconf"]["enslogs"] = t["syslog_server"]
	elseif t["syslog_server"] and t["syslog_server"] ~= "" then
		logger:warning("The syslog server does not support domain name!")
	end

	-- VLAN settings
	if tonumber(t["vid_voice"]) and tonumber(t["vid_voice"]) > 1 then
		result["Devconf"]["vlan"] = t["vid_voice"]
		result["Devconf"]["vlanprio"] = t["qos_voice"] or "7"
	end
	
	-- set default DNS and NTP settings if not define
	t["dns_server"], t["second_dns_server"] = h.get_prim_sec_dns_servers(t["dns_server"], t["second_dns_server"])
	t["ntp_server"] = t["ntp_server"] or ""
	
	-- use static ip settings
	local is_addr_present = h.is_addr_present (agent["address"], agent["mac"])	-- check dublicate address
	local is_addr_local = h.isLocalIP(agent["address"])		-- device behind NAT can't user static ip
	if t["use_static_ip"] and (is_addr_present or not is_addr_local) then
		if is_addr_present then
			logger:warning("Disable static ip. Other device have the same ip address: %s", agent["address"] or "")
		end
		if not is_addr_local then
			logger:warning("Disable static ip. Device ip address (%s) is placed behind NAT", agent["address"] or "")
		end
		t["use_static_ip"] = nil
		update_db_settings = true
	end

	-- get settings for leased device by pbx dhcp
	local is_leased
	if not t["use_static_ip"] or t["dns_server"] == "" or t["ntp_server"] == "" then
		is_leased = h.is_addr_leased(agent["address"])
	end
	if not t["use_static_ip"] and not is_addr_present and is_leased then
		h.set_addr_static (agent["address"], agent["mac"])	-- mark this lease as static
		t["use_static_ip"] = 1
		update_db_settings = true
	end
    if is_leased then
    	if "" == t["dns_server"] or "" == t["second_dns_server"] then
	    	local dns_dhcp1, dns_dhcp2 = h.get_dns_servers()
    		if dns_dhcp1 and "" ~= dns_dhcp1 and "" == t["dns_server"] then
    			t["dns_server"] = dns_dhcp1             -- set primary DNS from DHCP config
    			update_db_settings = true
    		end
    		if dns_dhcp2 and "" ~= dns_dhcp2 and "" == t["second_dns_server"] then
    			t["second_dns_server"] = dns_dhcp2      -- set secondary DNS from DHCP config
    			update_db_settings = true
    		end
	    end	    
	    if t["ntp_server"] == "" then -- set NTP from DHCP config
	        local ntp_dhcp = h.get_ntp_servers()
	        if ntp_dhcp and ntp_dhcp ~= "" then
	            t["ntp_server"] = ntp_dhcp
	            update_db_settings = true
	        end
	    end
	end

	if t["use_static_ip"] == 1 then	-- use static ip
		local _, netmask, gateway = h.host_settings(agent["address"], registrar_ip)
		if netmask and gateway then
			result["Devconf"]["ipa"] = agent["address"]
			result["Devconf"]["ipm"] = netmask
			result["Devconf"]["ipgw"] = gateway
			result["Devconf"]["dhcp"] = "0"
			
			if registrar_ip then
    			if "" == t["dns_server"] then
    				t["dns_server"]    = registrar_ip
    				update_db_settings = true
    			end
    			if "" == t["second_dns_server"] then
    				t["second_dns_server"] = registrar_ip
    				update_db_settings 	   = true
    			end
    		end
			
			if t["ntp_server"] == "" then
				t["ntp_server"]    = registrar
				update_db_settings = true
			end
		else
			logger:warning("For current network configurations is not possible set static ip, disable it")
			t["use_static_ip"] = nil
			update_db_settings = true
		end
	end
	
	-- processing flag for disabling SIP firewall
	if "1" == tostring(t["disable_sip_firewall"]) and "FXO" == agent["vendor:additional"] then
		result["Devconf"]["pxonly"] = "0"			
	end
	
	-- update devdb with new settings
	if update_db_settings then
		connection:switch("devdb")
		local sql = "INSERT OR REPLACE INTO autoprov (id, param, value, mac) VALUES ((SELECT id FROM autoprov WHERE param=%q AND mac=%q), %q, '%s', %q)"
		connection:execute(sql, "settings", agent["mac"], "settings", json.encode(t), agent["mac"])
	end

	-- DNS settings 
	if "" ~= t["dns_server"] then
		result["Devconf"]["dns1"] = t["dns_server"]
		if "" ~= t["second_dns_server"] then
			result["Devconf"]["dns2"] = t["second_dns_server"]
		end
	end
	
	-- NTP settings 
	if t["ntp_server"] ~= "" then
		result["Devconf"]["ntps"] = t["ntp_server"]
	end
	-- time zone offset
	local timeZone = t["ntp_timezone"]
	if not timeZone or "" == timeZone then
	    timeZone = h.getPbxTimeZone()
	end
	local td, errMsg = require"core.timedate":new(os.time(), timeZone)
	if td then
		td:toTimeZone()
		local tzoffset, errMsg = td:getTzOffset()
		if tzoffset and tzoffset.sign and tzoffset.dec then
			result["Devconf"]["tz"] = tzoffset.sign .. tzoffset.dec --"Time Zone" like "-6.5"
		end
		if errMsg then
    		logger:debug(errMsg)
		end
		-- Daylight saving time
		local dst = td:getDstData(timeZone)
		if "no" ~= t["use_dst"] and dst then
			local startMonth   = dst[1].month < 10 and "0"..dst[1].month or dst[1].month 
			local startWeek    = "0"..math.floor(dst[1].day/7)     -- 0 is first week 4 is fifth week
			local startWeekDay = "0"..(tonumber(dst[1].wday) - 1) -- 0 is sunday 6 is saturday
			local startHour    = dst[1].hour < 10 and "0"..dst[1].hour or dst[1].hour
			local stopMonth    = dst[2].month < 10 and "0"..dst[2].month or dst[2].month
			local stopWeek     = "0"..math.floor(dst[2].day/7)
			local stopWeekDay  = "0"..(tonumber(dst[2].wday) - 1)
			local stopHour     = dst[2].hour < 10 and "0"..dst[2].hour or dst[2].hour
			result["Devconf"]["dlt"]  = 1
			result["Devconf"]["dltbs"]= dst.offsetmin -- Daylight Bias 60min
			result["Devconf"]["dltb"] = startMonth..","..startWeek..","..startWeekDay..","..startHour -- default is "03,04,00,02" - last sunday of march at 02 o'clock
			result["Devconf"]["dlte"] = stopMonth..","..stopWeek..","..stopWeekDay..","..stopHour     -- default is "10,04,00,03" - last sunday of october at 03 o'clock
		end
    end
    if errMsg then
    	logger:debug(errMsg)
    end

	-- jitter buffer
	local tjb = h.get_jitter_buffer(is_addr_remote)
	result["Devconf"]["ijitb"] = tonumber(tjb["min"]) > 280 and "280" or tjb["min"]
	result["Devconf"]["xjitb"] = tjb["max"]

	-- QOS
	result["Devconf"]["rqos"] = tos_data["tos"] / 4 -- Voice QoS
	result["Devconf"]["sqos"] = tos_data["tos_sip"] / 4 -- SIP QoS

	-- EOD
	result["Devconf"]["eod"] = h.get_eod()

	-- Silence detection
	if not tports["silence_detection_enabled"] then
		tports["silence_detection_enabled"] = false
	end
	if not tonumber(tports["silence_detection"]) then
		tports["silence_detection"] = "60"
	end
	result["Devconf"]["silc"] = tports["silence_detection_enabled"] == true and "1" or "0"
	result["Devconf"]["tSilc"] = tports["silence_detection"]

	-- {{ Provisioning
	local provlink = string.gsub(
		"$protocol://$registrar:$port/?action=setaps", "%$(%w+)",
		{
			protocol  = http_protocol,
			registrar = registrar,
			port      = http_port,
		}
	)
	result["Devconf"]["httpurl"] = provlink
	result["Devconf"]["httpcu"] = "wmsdevice"
	result["Devconf"]["httpcp"] = token_pass or h.get_gpassword()
	-- }}

	-- Secret passwords for all logins
	result["Devconf"]["u1p"] = "wil--dix"..token_pass or "wil!superroot!dix"
	result["Devconf"]["u2p"] = token_pass or h.get_gpassword()
	result["Devconf"]["u3p"] = token_pass or h.get_gpassword()

	-- {{ Set country tones
	local dt = tports["tonezone"] or t["tonezone"] or defaultToneZone
	local tone = h.get_dialtone("welltech04", dt)	
	if tone then
		result["Devconf"]["tone"] = tone 
	end
	-- }}

	-- make output
	local output = {}
	local function get_configuration (t, p)
		local output = {}
		local parent = p and p .. "." or ""
		for param, value in pairs(t) do
			if type(value) == "table" then
				local tcfg = get_configuration (value, parent .. param)
				for _, v in pairs (tcfg) do
					output[#output+1] = v
				end
			else
				output[#output+1] = parent .. param .. "=" .. value
			end
		end
		return output
	end

	-- Important thing
	output[#output+1] = "FmtNo=_wildix_wgw04xx"

	for section, v in pairs (result) do
		output[#output+1] = "[".. section .."]"
		for _, v in pairs (get_configuration (v)) do
			output[#output+1] = v
		end
	end
	return output, err
end

function setuser (...) end

function setfeatures (...) end
