--- Provision module for 04FXO revision 3
-- 
-- @file    wildixfxo3.lua 
-- @author  
-- @created 2019/10/28

module( ... or "wildixfxo3", package.seeall)

----------------------------------------------------------------
-- Private interface
----------------------------------------------------------------

--- Push section configuration
local function __add_section_cfg(cfg_t, name, root_elm)
    if not name or "" == name then
        return false
    end
    
    local sect_cfg = cfg_t[name] 
    if "table" ~= type(sect_cfg) then
        return false
    end
    
    -- pushing configuration
    local elm = root_elm:add_elm(name)
    for pname, pvalue in pairs(sect_cfg) do
        if pname and pvalue then
            if "table" == type(pvalue) then
                __add_section_cfg(sect_cfg, pname, elm)
            else
                local param_elm = elm:add_elm("param")
                param_elm:add_attr({name = pname, value = pvalue})
            end
        end
    end
end

--- Create XML table from configuration table
local function __create_xml_table(cfg_t, cfg_type)
    if "table" ~= type(cfg_t) then
        return false
    end
    
    local wxml_obj = require"core.wxml":new()
    local root_elm = wxml_obj:elm_tbl()
    
    if "policy" == cfg_type then
        local prov_elm = root_elm:add_elm("provision")
        prov_elm:add_attr({version = "2.0"})
        
        if cfg_t["firmware"] then
            -- push firmware parameters
            local prod_elm = prov_elm:add_elm("product")
            prod_elm:add_attr(cfg_t["product"])
            prod_elm:add_attr(cfg_t["firmware"]["product"])
            
            local pack_elm = prod_elm:add_elm("package")
            pack_elm:add_attr(cfg_t["firmware"])
        end
        
        -- push config parameters
        local prod_elm = prov_elm:add_elm("product")
        prod_elm:add_attr(cfg_t["product"])
        prod_elm:add_attr({url = cfg_t["config"]["product"]["url"]})
        
        local conf_elm = prod_elm:add_elm("ConfigFile")
        conf_elm:add_attr(cfg_t["config"])
    else
        local conf_elm = root_elm:add_elm("config")
        conf_elm:add_attr({
            version = cfg_t["version"],
            md5     = cfg_t["md5"]
        })  
        __add_section_cfg(cfg_t, "provision", conf_elm)             -- push provision parameters
        __add_section_cfg(cfg_t, "syslog", conf_elm)                -- push syslog parameters
        __add_section_cfg(cfg_t, "network", conf_elm)               -- push network parameters
        __add_section_cfg(cfg_t, "fxs_fxo", conf_elm)               -- push FXS/FXO parameters
        __add_section_cfg(cfg_t, "feature_code", conf_elm)          -- push feature codes parameters
        __add_section_cfg(cfg_t, "sipserver", conf_elm)             -- push SIP servers parameters
        __add_section_cfg(cfg_t, "media", conf_elm)                 -- push media parameters
        __add_section_cfg(cfg_t, "telprofile", conf_elm)            -- push telprofile settings
        __add_section_cfg(cfg_t, "sip", conf_elm)                   -- push SIP parameters
        __add_section_cfg(cfg_t, "sipacc", conf_elm)                -- push SIP account settings
        __add_section_cfg(cfg_t, "ports", conf_elm)                 -- push ports settings
        __add_section_cfg(cfg_t, "fax", conf_elm)                   -- push fax settings
        __add_section_cfg(cfg_t, "digitmap", conf_elm)              -- push digitmap settings
        __add_section_cfg(cfg_t, "system", conf_elm)                -- push system settings
        __add_section_cfg(cfg_t, "iptrunk", conf_elm)               -- push iptrunk settings
        __add_section_cfg(cfg_t, "route", conf_elm)                 -- push route settings
        __add_section_cfg(cfg_t, "manipulation", conf_elm)          -- push manipulation settings
        __add_section_cfg(cfg_t, "simserver", conf_elm)             -- push simserver settings
        __add_section_cfg(cfg_t, "filelog", conf_elm)               -- push filelog settings
        __add_section_cfg(cfg_t, "serverlog", conf_elm)             -- push serverlog settings
        __add_section_cfg(cfg_t, "snmp", conf_elm)                  -- push snmp settings
        __add_section_cfg(cfg_t, "tr069", conf_elm)                 -- push tr069 settings
        __add_section_cfg(cfg_t, "actionurl", conf_elm)             -- push actionurl settings
        __add_section_cfg(cfg_t, "websclassify", conf_elm)          -- push websclassify settings
        __add_section_cfg(cfg_t, "relayserver", conf_elm)           -- push relayserver settings
    end
    
    return root_elm
end

local function __makeToneString(tones)
    local rv = nil
    if tones and "table" == type(tones[1]) then
        local freq1 = tones[1]["freq1"] or "0"
        local gain1 = "0" ~= freq1 and "260" or "0"
        local dur1  = tones[1]["duration"] or "0"
        local freq2 = tones[2] and tones[2]["freq1"] or "0"
              freq2 = "0" ~= freq2 and freq2 or freq1
        local gain2 = "0" ~= freq2 and "630" or "0"
        local dur2  = tones[2] and tones[2]["duration"] or "0"
              rv    = freq1..","..gain1..","..freq2..","..gain2..","..dur1..","..dur2..",0,0"
    end
    return rv
end

----------------------------------------------------------------
-- Public interface
----------------------------------------------------------------
hints = {
    --[[
    ======================
    Wildix 04FXOR3
    ----------------------
        Server: Web Server/2.1.0.
        Date: Fri Apr 10 14:28:51 2015.
        WWW-Authenticate: Basic realm="Web Server".
        Set-Cookie: devckie=da00-0040-af00-0450.
        Pragma: no-cache.
        Cache-Control: no-cache.
        Content-Type: text/html.
    --]]
    {
         ["server"]           = "Web Server%/.*"
        ,["www-authenticate"] = 'Basic realm="Web Server"'
    }   
}

--- Convert the impedance index to parameter value
--  @param idx - index of the impedance.
--  @return  impedance parameter value for device
local function __getImpedance(idx)
    if not idx then
        return false, "Impedance index not found"
    end
    local rv = 15
        
        -- W04FXOR3 parameters
        --[[
        value="0">600 R
        value="1">900 R
        value="2">270 R + (750 R||150nf) and 275 R + (780 R||150nf)
        value="3">220 R + (820 R||120nf) and 220 R + (820 R||115nf)
        value="4">370 R + (620 R||310 nF)
        value="5">320 R + (1050 R||230 nF)
        value="6">370 R + (820 R||110 nF)
        value="7">275 R + (780 R||115 nF)
        value="8">120 R + (820 R||110 nF)
        value="9">350 R + (1000 R||210 nF)
        value="10">200 R + (680 R||100 nF)
        value="11">600 R + 2.16 μF
        value="12">900 R + 1 μF
        value="13">900 R + 2.16 μF
        value="14">600 R + 1 μF
        value="15">Global
    ]]--

    local impedance = {
        [1]  = 2, -- 270+750//150nF (Italy/Sweden)
        [2]  = 3, -- 220+820//115nF (Germany)
        [3]  = 4, -- 370+620//310nF (UK)
        [4]  = 3, -- 180+910//150nF (France)
        [5]  = 8, -- 120+820//110nF (Norway)
        [6]  = 8, -- 150+830//72nF (Belgium)
        [7]  = 4, -- 400+500//330nF (Denmark)
        [8]  = 0, -- 600 R
        [9]  = 1, -- 900 R
        [10] = 5, -- 320 R + 230 nF
        [11] = 6, -- 370 R + 110 nF
        [12] = 7, -- 275 R + 115 nF
        [13] = 9, -- 350 R + 210 nF
        [14] = 10, -- 200 R + 100 nF
        [15] = 11, -- 600 R + 2.16 μF
        [16] = 12, -- 900 R + 1 μF
        [17] = 13, -- 900 R + 2.16 μF
        [18] = 14, -- 600 R + 1 μF
        [19] = 15  -- Global
        
    }
    if impedance[tonumber(idx)] then
        rv = impedance[tonumber(idx)]
    end
    
    return rv
end

--- Verifying device HTML request and retreiving information
-- @param host      A verifying host data
-- @param port      Not used data
-- @return table with verifying device information
function verify (host, port)
    local logger     = logger
    local http       = require"core.httpClient":new()
    local helper     = require"common.helper"
    local connection = require"common.core".connection
    
    local is_ssl_negotiate  =  443 == port.number
    local PORT_NUMBER       = is_ssl_negotiate and not gconf.HTTPForceUse and 443 or 80
    local t                 = {["ssl"] = is_ssl_negotiate}  
    
    local gpassword  = helper.get_gpassword()
    local auth_table = {"admin:admin", "admin:" .. gpassword}

    local sql           = nil
    local cursor        = nil
    local token_pass    = nil
    local dublicate_ip  = nil
    if "" == host.mac_addr_str then
        sql    = [[ SELECT token, count(id) FROM autoprov WHERE param=%q AND value LIKE '%%"address":%%"%s"%%' AND value LIKE '%%"model":%%"w%%fxor_"%%' ]]
        cursor = connection:execute(sql, "device", host.ip or "")
        token_pass, dublicate_ip = cursor:fetch(); cursor:close()
        if 1 < tonumber(dublicate_ip) then
            logger:error("Found %s duplicated address %q at provisioning database", dublicate_ip, host.ip or "")
            return t
        end
    else
        sql        = "SELECT token FROM autoprov WHERE param=%q AND mac=%q"
        cursor     = connection:execute(sql, "device", host.mac_addr_str)
        token_pass = cursor:fetch(); cursor:close()
    end
    
    t["token_type"] = helper.is_new_token(token_pass)
    if "new" == t["token_type"] then
        table.insert(auth_table, 1, "admin:"..token_pass)
        t["token"] = token_pass
    end
    logger:debug(auth_table, "auth_table")
    
    -- Get device status information via GET request
    -- check authorization
    local answer = nil
    for _, authpair in pairs(auth_table) do
        local login, pass = authpair:match("(.*):(.*)")
        answer = http:rawget(host.ip, PORT_NUMBER, "/enSysInfo.htm", { authUser = login, authPass = pass }) or {}
        if 200 == answer.code then
            t["auth"] = {user = login, pass = pass}
            break
        end
    end
    if not t["auth"] then
        logger:error("Wrong authorization for host %s", host.ip)
        return t
    end
    
    local model, fw_ver = answer.body:match("Current Software Version</td>\r?\n%s*<td colspan=\"%d\">\r?\n%s*(W%d+FX[SO]R%d)%s(.-)%s")
    if not model or "" == model then
        logger:error("Not found model for host %s", host.ip)
        return t
    end

    t["model"]             = model:lower() 
    t["fwversion"]         = fw_ver
    t["vendor:additional"] = string.match(model, "FXOR%d")

    if "" == host.mac_addr_str then
        local web_mac = answer.body:match("MAC Address </td>\r?\n%s*<td colspan=\"%d\">\r?\n%s*([0-9a-fA-F%-]*)\r?\n")
        if web_mac then
            host.mac_addr_str = web_mac:gsub("-","")
        end
    end
    t["mac_addr_str"] = host.mac_addr_str:lower()
    
    return t
end

--- Provision action to updating provision url on device and executing autoprovision
-- @param host  A provision device data
-- @param port  Not used 
function provision(host, port, ...)
    local args       = select(1, ...)
    local http       = require"core.httpClient":new()
    local h          = require"common.helper"
    local core       = require"common.core"
    local connection = core.connection
    
    local model      = host.model:upper()
    local devFwInfo  = h.parse_firmware(host.fwversion)
    
    local prov_proto  = gconf.HTTPForceUse and "http" or "https"    
    local prov_port   = gconf.HTTPForceUse and "80" or "443"            
    local PORT_NUMBER = host.ssl and not gconf.HTTPForceUse and 443 or 80
    
    --the firmware version less than 27.18.01.05 does not support ssl
    local ssl_supported = 105 <= devFwInfo.sht_v2
    if not ssl_supported then
        prov_proto  = "http"    
        prov_port   = "80"  
        PORT_NUMBER = 80
    end
    
    local registrar, case, err_msg = h.get_registrar(host.ip_src, host.ip) 
    logger:debug("Registrar ip %q get from %s%s", registrar, case, err_msg and ". Errors: "..err_msg or "")
    
    local gpassword = nil
    if "new" == host.token_type then
        gpassword = host.token or h.get_gpassword()
    else
        gpassword = h.get_gpassword()
        if "old" == host.token_type then
            logger:info("Token has old format")
        end
    end
    
    math.randomseed(os.time())
    local provint  = "6048"..math.random(10,99) -- hack: set random "ProvisonInterval" to differ from device value, otherwise provision via wmsbind doesn't work since fw. ver 27.18.01.11  
    local provlink = string.gsub(
        "$protocol://$registrar:$port/?action=setaps", "%$(%w+)",
        {
            protocol  = prov_proto,
            registrar = registrar,
            port      = prov_port,
        }
    )

    local options = {
         authUser = host.auth.user
        ,authPass = host.auth.pass
        ,timeout  = 2
    }
    
    -- Configuring provisioning
    local config_data = {
        "ProvisonURL=" .. http.urlEncode(provlink),
        "ProvisonInterval=" .. provint,
        "ProvisonAcc=wmsdevice",
        "ProvisonPwd=" .. gpassword,
        "ProxyDomain=",
        "ProxyPort=",
        "ProxyAcc=",
        "ProxyPwd=",
        "ok=Save",
        "ModFlag=1",
        "ProxyModFlag=0"
    }
    local data = table.concat(config_data, "&")
    local answer = http:rawpost(host.ip, PORT_NUMBER, "/goform/Provison", data, options) or {}  -- Save options
    logger:info("Send http request for autoprovision to with provision configuration to %s - %s", model, answer.code or 408)
end

--- Building and sending provision configuration
-- @param agent         A table with general device information
-- @param request       A parameters got from provision request     
function configuration(agent, request)
    if "table" ~= type(agent) or "table" ~= type(request) then
        return false
    end
    
    local h           = require"common.helper"
    local core        = require"common.core"
    local logger      = logger
    local connection  = core.connection
    local json        = require"dkjson"
    local md5         = require"md5"
    local wmsldap     = require"core.wmsldap"
    
    local callerid    = "enable"
    local devFwInfo   = h.parse_firmware(agent["fw:version"])
    local cfg_t       = nil -- configuration table
    local prov_proto  = nil
    local prov_port   = nil
    local secret      = {}
    
    -- retreive registrar URL parameters
    local registrar  = agent["registrar"]
    
    --the firmware version less than 27.18.01.05 does not support ssl
    local ssl_supported = 105 <= devFwInfo.sht_v2
    
    local prov_proto = agent["ssl"] == "on" and not gconf.HTTPForceUse and "https" or "http"
    local prov_port  = gconf.HTTPForceUse and "80" or agent["alternate_port"] or agent["registrar:port"]
    if not ssl_supported then
        prov_proto = "http"
        prov_port  = "80"
        callerid    = "disable" -- invert the value if firmware version less than 27.18.01.05
    end
    
    -- set general url
    local base_url   = string.gsub("$protocol://$registrar:$port/", "%$(%w+)",
                                   {
                                        registrar = registrar,
                                        protocol  = prov_proto,
                                        port      = prov_port,
                                    })
    
    -- retreiving information about device firmware
    local fw_cfg    = nil
    local fw_name   = "wildixfxs2"  -- directory of the firmware for this devices
    local firmwares = core.get_firmware_pack({["vendor"] = fw_name}) -- need to recheck get_firmware_pack - checked, OK
    local model = agent["model"]    
    if firmwares then
        fw_cfg = firmwares[model]
    end
    if not fw_cfg then
        logger:error("Firmware pack not found for device %q", model)
        return false
    end
    
    -- retrieve device information from provision DB
    local sql    = "SELECT value, token FROM autoprov WHERE param=%q AND mac=%q"
    local cursor = connection:execute(sql, "device", agent["mac"])
    local device_json, token_pass = cursor:fetch(); cursor:close()
    
    token_pass = token_pass or h.get_gpassword()

    local db_device_data = json.decode(device_json or "") or {}
    local reboot_needed = db_device_data.reboot == 1

----------------------------------------------------------------
--- Policy file with provision url and firmware links
----------------------------------------------------------------
    if "policy" == request["config"] then
        cfg_t        = {} -- buffer for policy configuration file
        
        -- set general product parameters
        cfg_t["product"] = {
            url       = {},   -- firmware and configuration urls
            id        = tostring(devFwInfo.major), -- id is major FW version
            oemid     = "27", -- 27: Wildix, 1:Dinstar 
            model     = "",   -- provision settings will be used by all devices
--          macfilter = "",     --TODO: upgrade only Wildix devices
        }
        
        -- firmware settings
        local fmw_file      = nil
        if fw_cfg and fw_cfg["version"] ~= agent["fw:version"] then 
            local fw_filename = fw_cfg["application"]
            local fw_dir      = "/var/www/firmwares/" .. fw_name .. "/"
            local fw_file = fw_dir .. fw_filename
            if h.is_exists(fw_file) then
                local fmw_url  = base_url .. "firmwares/" .. fw_name .. "/"
                local fmw_ver  = fw_cfg["version"]:gsub("%.", "")
                fmw_file       = fw_filename
                
                -- set firmware parameters
                cfg_t["firmware"] = {
                    ver       = fmw_ver,
                    name      = fmw_file,
                    reboot    = "delay",    -- Apply setting (wait when active call will be ended)
                }
                local pbxFwInfo = h.parse_firmware(fw_cfg["version"])
                local downgrade = "false"
                if devFwInfo.sht_v2 > pbxFwInfo.sht_v2 then
                    downgrade = "true"
                end
                cfg_t["firmware"]["product"] = {
                    url   = fmw_url,
                    force = downgrade,  -- true allowing downgrade firmware version
                }
                
                logger:info("Found new firmware version - %s", fw_cfg["version"])
                logger:info("Firmware for device %s %s will be changed from version %s to %s", agent["model"], agent["mac"], agent["fw:version"], fw_cfg["version"])
            else
                logger:warning("Firmware file %q couldn't be loaded", fw_file)
            end
        end
        
        -- set general configuration parameters
        cfg_t["config"] = {
            name   = "$(MAE).cfg/?action=setaps",   -- Provision url
            action = "replace",                 -- Applying configuration action
--          action = "merge",           --XXX: must be moved to replace afeter vendor fix                   
            reboot = reboot_needed and "always" or "never", -- Apply configuration with/without reboot
        }
        cfg_t["config"]["product"] = {url = base_url}   -- Provision init url
    else
    
----------------------------------------------------------------
--- General configuration settings
----------------------------------------------------------------
        local resource   = core.load_resource("wildixfxo3")
        cfg_t            = resource.cfg_def or {}
        cfg_t["version"] = fw_cfg["config"] -- confiugration version
        
        local fxo_ports  = resource.models[agent["model"]:upper()]
        local tports, trunkname = {} -- data for FXO settings

        -- update database by device information
        local agent_json = json.encode(agent)
        if not device_json or md5.digest(device_json) ~= md5.digest(agent_json) then
            sql = "INSERT OR REPLACE INTO autoprov (id, param, value, mac, token) VALUES ((SELECT id FROM autoprov WHERE param=%q AND mac=%q), %q, '%s', %q, %q)"
            connection:execute(sql, "device", agent["mac"], "device", agent_json, agent["mac"], token_pass or '')
        end

        -- Get default port settings 
        connection:switch ("wmsdb")
        sql = "select count(id) from dialplan_rulenames where rulename=%q"
        cursor = connection:execute(sql, "main")
        local count = cursor:fetch(); cursor:close()
        resource.port_default.port_context = tonumber(count) > 0 and "main" or "internalcalls"
        local is_upd_wmsdb = false

        for i=1, fxo_ports do
            for k,v in pairs(resource.port_default) do
                if not tports[k] then tports[k] = {} end
                tports[k][i] = v
            end
        end

        -- Check trunkid
        connection:switch ("devdb")
        sql = "select value from autoprov where param = %q and mac = %q"
        cursor = connection:execute(sql, "trunkid", agent["mac"])
        local trunkid = cursor:fetch(); cursor:close()
            
        if trunkid then     -- we have a trunk
            connection:switch ("wmsdb")
            sql = "select username, settings_mgw from pstn_trunks where id = %s"
            cursor = connection:execute(sql, trunkid)
            local username, settings_mgw = cursor:fetch(); cursor:close()
            trunkname = username
                
            local settings_mgw = json.decode(settings_mgw or {});
            for k,v in pairs(settings_mgw) do
                tports[k] = v
            end

            for i=1, fxo_ports do
                secret[i] = h.get_sip_pass(trunkname, trunkid, i-1, agent["mac"])
            end

            -- DC Impedance
            if "yes" == tports["dc_impedance"] then
                cfg_t["fxs_fxo"]["fxo"]["dc_impedance"] = "1"
            end
        else -- Trunk autocreation
            local lowertech, tnames = "fxo", {}
            is_upd_wmsdb = true

            connection:switch ("wmsdb")
            sql = "select username from pstn_trunks where technology=%q"
            for name in connection:rows(sql, lowertech) do
                table.insert(tnames, name)
            end
            table.sort(tnames, function(a,b) a = tonumber(a:sub(4)); b = tonumber(b:sub(4));  return a<b end)

            for i=1,#tnames do
                if (lowertech .. tostring(i-1)) ~= tnames[i] then
                    trunkname = lowertech .. tostring(i-1)
                    break
                end
            end
            if not trunkname then
                trunkname = lowertech .. tostring(#tnames)
            end
                
            -- Update db for creating sip-extra.conf
            sql = [[ insert or replace into pstn_trunks (id,username,ports,technology,settings_pbx,settings_mgw) values(
                    (select id from pstn_trunks where username = %q),%q,%d,%q,%q,'%s') ]]
            connection:execute(sql, trunkname, trunkname, fxo_ports, "fxo", "{}", json.encode(tports))
                
            sql = "select id from pstn_trunks where username = %q"
            cursor = connection:execute(sql, trunkname)
            local trunkid = cursor:fetch(); cursor:close()
            connection:switch ("devdb")     -- Set trunk id
            sql = [[ insert or replace into autoprov (id,param,value,mac) values(
                    (select id from autoprov where param=%q and mac=%q),%q,'%s',%q) ]]
            connection:execute(sql, "trunkid", agent["mac"], "trunkid", trunkid, agent["mac"]) 
                
            connection:switch("htables") -- update kamailio tables
            connection:execute("delete from dynamic_peers where wms_id = %q", trunkid)  -- remove old trunk stuff
            local username, key
            sql = "insert into %s (key_name,key_value,wms_id) values (%q,%q,%q)"    -- dynamic_peers
            for i=1, fxo_ports do
                username = string.format("%s_%d", trunkname, i-1)
                secret[i] = h.generate_pass()
                key = string.format ("%s:%s", agent["address"], tostring (5059 + i))
                connection:execute (sql, "dynamic_peers", username .. "::authname", username, trunkid)
                connection:execute (sql, "dynamic_peers", username .. "::secret", secret[i], trunkid)
            end
                
            h.create_sip_extra()
            os.execute "sudo kamcmd htable.reload dynamic_peers"    -- reload kamailio
            os.execute "sudo callweaver -rx \"sip reload \" > /dev/null"    -- reload to update callweaver peers
        end

        if is_upd_wmsdb then
            connection:switch ("wmsdb") -- update wmsdb
            sql = [[ insert or replace into pstn_trunks (id,username,ports,technology,settings_pbx,settings_mgw) values(
                    (select id from pstn_trunks where username = %q),%q,%d,%q,%q,'%s') ]]
            connection:execute(sql, trunkname, trunkname, fxo_ports, "fxo", "{}", json.encode(tports))
        end

        -- Generate ports
        local tusr = {}
        cfg_t["ports"] = {}

        local username, telprofile, sipacc, fxo, route, ports
        for i=1, fxo_ports do

            username = string.format("%s_%d", trunkname, i-1)
            tusr[#tusr+1] = {username, secret[i]}

            sipacc = "sipacc" .. i-1
            cfg_t["sipacc"][sipacc] = {
                ["isregister"] = "enable",
                ["serverid"] = "0",
                ["auth_id"] = username, -- fxo0_0 .. fxo3_0
                ["user_id"] = username,
                ["auth_pwd"] = secret[i],
                ["display"] = "FXO_Line" .. i,
            }

            ports = "port" .. i-1
            cfg_t["ports"][ports] = {
                ["enable"] = "enable",
                ["primary_sipacc_id"] = i-1,
                ["auto_dial"] = tports["port_hotline_number"][i], -- hotline number
                ["auto_dial_timeout"] = tports["port_answer_time"][i],  -- hotline timeout
                ["dnd"] = "disable",
                ["callerid"] = callerid,
                ["cfu"] = "",
                ["cfb"] = "",
                ["cfnry"] = "",
                ["function"] = "",
                ["ipprofileindex"]  = "0",
                ["telprofileindex"] = i,
            }

            telprofile = "telprofile" .. i
            cfg_t["telprofile"][telprofile]["tx_gain"] = tports["port_codec_output_gain"][i]
            cfg_t["telprofile"][telprofile]["rx_gain"] = tports["port_codec_input_gain"][i]

            -- Fax type. mode=0 - T.38, mode=4 - pass through(vbd), mode=3 - adaptive
            if "1" == tports["port_fax_type"][i] then
                cfg_t["telprofile"][telprofile]["fax"]["mode"] = "4"
                cfg_t["telprofile"][telprofile]["fax"]["rate"] = tports["port_fax_speed"][i]
            elseif "2" == tports["port_fax_type"][i] then
                cfg_t["telprofile"][telprofile]["fax"]["mode"] = "0"
                cfg_t["telprofile"][telprofile]["fax"]["rate"] = tports["port_fax_speed"][i]
            end 

        end 
            -- polarity reversal detection.
            if "1" == tports["port_preversal_detection"][1] then 
                cfg_t["fxs_fxo"]["fxo"]["outgoing"]["answer_caller_when_polarity_reversal"] = "enable"
            else    
                cfg_t["fxs_fxo"]["fxo"]["outgoing"]["answer_caller_when_polarity_reversal"] = "disable"
            end
            -- WMS-9978 Obtain FSK CID from Num(value 0) or from Callee(value 1)
            if "0" == tports["fsk_cid_field"] then
                cfg_t["fxs_fxo"]["fxo"]["incoming"]["fsk_cid_field"] = "0"
            end

            -- WMS-10949
            if "0" == tports["busy_tone_detected"] then
                cfg_t["fxs_fxo"]["fxo"]["onhook_when_busy_detect"] = "disable"
            end

            if "1" == tports["current_detected"] then
                cfg_t["fxs_fxo"]["fxo"]["onhook_when_no_current"] = "enable"
            end

            local no_srtp = "no" == gconf.SRTPForceUse
            cfg_t["media"]["srtp_mode"] = no_srtp and "0" or "2" -- "srtp_mode" 0-disable, 1-auto, 2-forced

            local disconnect_threshold = tonumber(tports["current_disconnect_threshold"])
            if disconnect_threshold and disconnect_threshold >= 100 and disconnect_threshold <= 2000 then
                cfg_t["fxs_fxo"]["fxo"]["current_disconnect_threshold"] = tports["current_disconnect_threshold"]
            end

            if string.find(model, "fxor3") then
                if 1013 <= h.parse_firmware(agent["fw:version"]).sht_v2 then
                    cfg_t["fxs_fxo"]["fxo"]["polarity_detection"] = "disable"
                    if "1" == tports["polarity_normal_detected"] then
                        cfg_t["fxs_fxo"]["fxo"]["polarity_detection"] = "enable"
                    end
                elseif "1" == tports["polarity_normal_detected"] then
                    logger:warning("Incorrect device's firmware version. Polarity reversal function won't work.")
                end
            end

            -- caller id before ring or after ring
--            if "1" == tports["port_pstn_ring_count"][1] then
--                cfg_t["fxs_fxo"]["fxo"]["incoming"]["cid_detect"] = "1"
--            else
--                cfg_t["fxs_fxo"]["fxo"]["incoming"]["cid_detect"] = "0"
--            end

            -- AC Impedance
            cfg_t["fxs_fxo"]["fxo"]["ac_impedance"] = __getImpedance(tports["impedance"]) or 15 -- 15 - default "Global" value.
        
----------------------------------------------------------------
------- Provision settings
----------------------------------------------------------------
        cfg_t["provision"]["url"]               = base_url .. "?action=setaps"
        cfg_t["provision"]["provison_account"]  = "wmsdevice"
        cfg_t["provision"]["provison_password"] = token_pass

----------------------------------------------------------------
------- Access settings
----------------------------------------------------------------
        if ssl_supported then
            cfg_t["system"]["access_ctrl"]["web"]["ssl_port"]                   = "443"
            cfg_t["system"]["access_ctrl"]["cli"]["username_aes"]               = "admin"
            local k = "1234567890123456"
            if "r3" == string.sub(model, -2) then
                local hex_token = string.upper(require"core.strConverter".toHex(token_pass))
                local hex_token_aes, err = core.encrypt_cipher(hex_token, "aes", k)
                cfg_t["system"]["access_ctrl"]["cli"]["hex_pwd_aes"]                = hex_token_aes
                cfg_t["websclassify"]["users"]["user0"]["hex_pwd_aes"]              = hex_token_aes
            else
                local token_aes, err = core.encrypt_cipher(token_pass, "aes", k)
                cfg_t["system"]["access_ctrl"]["cli"]["password_aes"]               = token_aes
                cfg_t["system"]["access_ctrl"]["web"]["users"]["user1"]["password"] = token_aes
                cfg_t["system"]["access_ctrl"]["cli"]["users"]["user1"]["password"] = token_aes
                cfg_t["websclassify"]["users"]["user0"]["pwd"]                      = token_aes
                cfg_t["websclassify"]["users"]["user0"]["pwd_aes"]                  = token_aes
            end
            if err then
                logger:debug("Cipher generation is incorrect :" .. err)
            end
        else
            cfg_t["system"]["access_ctrl"]["web"]["users"]["user1"]["password"] = token_pass
            cfg_t["system"]["access_ctrl"]["cli"]["users"]["user1"]["password"] = token_pass
            cfg_t["websclassify"]["users"]["user0"]["pwd"]                      = "@5XZ.PW]K[28#;ZD=H#]!E+UUNEW#[0*&amp;H^T[U52`NI!er &lt;"
        end

----------------------------------------------------------------
------- Network configuration
----------------------------------------------------------------
        local net_cfg            = {}
        local update_db_settings = false
        local is_addr_remote     = not h.isLocalIP(agent["address"])
        -- get device setting from provision DB
        connection:switch("devdb")
        sql    = "SELECT value FROM autoprov WHERE param='%s' AND mac='%s'"
        cursor = connection:execute(sql, "settings", agent["mac"])
        local settings_json = cursor:fetch(); cursor:close()
        if settings_json then
            net_cfg = json.decode(settings_json)
        end
        
        -- SYSLOG server (by default syslog server is empty)
        if net_cfg["syslog_server"] and "" ~= net_cfg["syslog_server"] then
            cfg_t["syslog"]["flag"] = "on"
            cfg_t["syslog"]["ip"]   = net_cfg["syslog_server"]
        end
        
        --TODO: SNMP acess server
            
        -- VLAN settings (set only Voice vlan)
        if tonumber(net_cfg["vid_voice"]) and 0 < tonumber(net_cfg["vid_voice"]) then
            cfg_t["network"]["vlan"]["vlan1_enabled"]     = "enable"
            cfg_t["network"]["vlan"]["vlan1_group"]       = "273" -- 273 mean that Data/Voice/Mngmt VLAN will be enabled, 16 mean that only Voice VLAN will be enabled
            cfg_t["network"]["vlan"]["vlan1_id"]          = net_cfg["vid_voice"]
            cfg_t["network"]["vlan"]["vlan1_priid"]       = net_cfg["qos_voice"] or "7"
            cfg_t["network"]["vlan"]["wan_vlanid_router"] = "1"
            cfg_t["network"]["vlan"]["lan_vlanid_router"] = "1"
        end

        -- set DNS and NTP settings by pbx network configuration and device setttings
        --TODO: secondary DNS settings
        net_cfg["dns_server"] = net_cfg["dns_server"] or ""
        net_cfg["ntp_server"] = net_cfg["ntp_server"] or ""
        
        -- checking static option with device location
        local is_addr_present = h.is_addr_present(agent["address"], agent["mac"])   -- check dublicate address
        if net_cfg["use_static_ip"] and (is_addr_present or is_addr_remote) then
            if is_addr_present then
                logger:warning("Disable static ip. Other device have the same ip address: %s", agent["address"] or "")
            end
            if is_addr_remote then  -- device behind NAT can't user static ip
                logger:warning("Disable static ip. Device ip address (%s) is placed behind NAT", agent["address"] or "")
            end
            net_cfg["use_static_ip"] = nil
            update_db_settings       = true
        end
        
        -- get settings for leased device by pbx dhcp
        local is_leased = false
        if not net_cfg["use_static_ip"] or "" == net_cfg["dns_server"] or "" == net_cfg["ntp_server"] then
            is_leased = h.is_addr_leased(agent["address"])
        end
        if not net_cfg["use_static_ip"] and not is_addr_present and is_leased then
            h.set_addr_static(agent["address"], agent["mac"])   -- mark this lease as static
            net_cfg["use_static_ip"] = 1
            update_db_settings       = true
        end
        if is_leased then
            if "" == net_cfg["dns_server"] then -- set DNS from DHCP config
                local dns_dhcp = h.get_dns_servers()
                if dns_dhcp and "" ~= dns_dhcp then
                    net_cfg["dns_server"] = dns_dhcp
                    update_db_settings    = true
                end
            end
            if "" == net_cfg["ntp_server"] then -- set NTP from DHCP config
                local ntp_dhcp = h.get_ntp_servers()
                if ntp_dhcp and "" ~= ntp_dhcp then
                    net_cfg["ntp_server"] = ntp_dhcp
                    update_db_settings    = true
                end
            end
        end
        
        if 1 == net_cfg["use_static_ip"] then   -- use static ip settings
            local _, netmask, gateway = h.host_settings(agent["address"], registrar)
            if netmask and gateway then
                cfg_t["network"]["local_network"]["net_mode"]    = "1"  
                cfg_t["network"]["local_network"]["wan_ip"]      = agent["address"]
                cfg_t["network"]["local_network"]["wan_mask"]    = netmask
                cfg_t["network"]["local_network"]["wan_gateway"] = gateway
                
                if "" == net_cfg["dns_server"] then
                    net_cfg["dns_server"] = registrar
                    update_db_settings    = true
                end
        
                if "" == net_cfg["ntp_server"] then
                    net_cfg["ntp_server"] = registrar
                    update_db_settings    = true
                end
            else
                logger:warning("For current network configurations is not possible set static ip, disable it")
                net_cfg["use_static_ip"] = nil
                update_db_settings       = true
            end
        end
        -- set tonezone to FXO
        local tos_data, defaulttoneZone = h.get_sip_settings()
        
        local tone = tports["tonezone"]
        if not tone or "" == tone then
            tone = defaulttoneZone or ""
        end
        if resource.country_tones[tone] then -- WMS-12638
            cfg_t["fxs_fxo"]["call_process_tone"] = "255"
            cfg_t["fxs_fxo"]["busy_tone"] = resource.country_tones[tone]["busy"] or "" -- 425,180,425,630,500,500,0,0
            cfg_t["fxs_fxo"]["dial_tone"] = resource.country_tones[tone]["dial"] or "" -- 425,180,425,630,200,200,600,1000
            cfg_t["fxs_fxo"]["ring_tone"] = resource.country_tones[tone]["ring"] or "" -- 425,180,425,630,1000,4000,0,0
            if 1015 <= devFwInfo.sht_v2 then
                cfg_t["fxs_fxo"]["disconnect_tone"] = resource.country_tones[tone]["disconnect"] or ""
            end
        elseif "number" ~= type(tonumber(tone)) then
            tone = h.get_dialtone("dinstar3", tone)
        end
        if "table" == type(tone) then -- custom tones
            cfg_t["fxs_fxo"]["call_process_tone"] = tone.country
            cfg_t["fxs_fxo"]["busy_tone"] = __makeToneString(tone.busy) or "" -- 425,180,425,630,500,500,0,0
            cfg_t["fxs_fxo"]["dial_tone"] = __makeToneString(tone.dial) or "" -- 425,180,425,630,200,200,600,1000
            cfg_t["fxs_fxo"]["ring_tone"] = __makeToneString(tone.ring) or "" -- 425,180,425,630,1000,4000,0,0
            -- WMS-10374 disconnect tone will be set as busy tone only for R3 devices.
            if 1015 <= devFwInfo.sht_v2 then
                cfg_t["fxs_fxo"]["disconnect_tone"] = __makeToneString(tone.busy) or ""
            end
        elseif "number" == type(tonumber(tone)) then -- predefined country tone set
            cfg_t["fxs_fxo"]["call_process_tone"] = "7" -- Italy, in case if something went wrong.
        end
        
        -- set QoS (need to recheck, the same on FXS)
        cfg_t["network"]["qos"] = {}
        cfg_t["network"]["qos"]["lan_aos_enable"] = "enable"
        cfg_t["network"]["qos"]["lan_qos0"] = "0"
        cfg_t["network"]["qos"]["lan_qos1"] = tos_data["tos"]/4     -- Voice QoS
        cfg_t["network"]["qos"]["lan_qos2"] = tos_data["tos_sip"]/4 -- SIP QoS

        -- update devdb with new settings (need to recheck)
        if update_db_settings then
--          sql = "INSERT OR REPLACE INTO autoprov (id, param, value, mac) VALUES ((SELECT id FROM autoprov WHERE param=%q AND mac=%q), %q, '%s', %q)"
            connection:execute (sql, "settings", agent["mac"], "settings", json.encode(net_cfg), agent["mac"])
        end
        
        -- DNS settings 
        net_cfg["dns_server"], net_cfg["second_dns_server"] = h.get_prim_sec_dns_servers(net_cfg["dns_server"], net_cfg["second_dns_server"])
        if "" ~= net_cfg["dns_server"] then
            cfg_t["network"]["local_network"]["wan_dns1"]     = net_cfg["dns_server"]
            if "" ~= net_cfg["second_dns_server"] then
                cfg_t["network"]["local_network"]["wan_dns2"]     = net_cfg["second_dns_server"]
            end
        end

        -- NTP settings
        local timeZone = net_cfg["ntp_timezone"]
        if not timeZone or "" == timeZone then
            timeZone = h.getPbxTimeZone()
        end
        if "" ~= net_cfg["ntp_server"] or "no" ~= net_cfg["use_dst"] then
            -- enable flag
            cfg_t["system"]["ntp"]["flag"] = "on"
            -- NTP server
            if "" ~= net_cfg["ntp_server"] then
                cfg_t["system"]["ntp"]["server1"]["domain"] = net_cfg["ntp_server"]
            end
            -- time zone settings
            local td, errMsg = require"core.timedate":new(os.time(), timeZone)
            if td then
                td:toTimeZone()
                local tzoffset = td:getTzOffset()
                if tzoffset and tzoffset.sign and tzoffset.hour and tzoffset.min then
                    cfg_t["system"]["ntp"]["zone hour"]   = tzoffset.sign .. tzoffset.hour
                    cfg_t["system"]["ntp"]["zone minute"] = tzoffset.min
                end
                -- Daylight saving time
                local dst = "no" ~= net_cfg["use_dst"] and td:getDstData(timeZone)
                if dst then
                    cfg_t["system"]["ntp"]["DaylightSavingTime"]["flag"]          = 1
                    cfg_t["system"]["ntp"]["DaylightSavingTime"]["save_time"]     = dst.offsetmin
                    cfg_t["system"]["ntp"]["DaylightSavingTime"]["start_month"]   = dst[1].month
                    cfg_t["system"]["ntp"]["DaylightSavingTime"]["start_day"]     = dst[1].day
                    cfg_t["system"]["ntp"]["DaylightSavingTime"]["start_hour"]    = dst[1].hour
                    cfg_t["system"]["ntp"]["DaylightSavingTime"]["start_minute"]  = dst[1].min
                    cfg_t["system"]["ntp"]["DaylightSavingTime"]["end_month"]     = dst[2].month
                    cfg_t["system"]["ntp"]["DaylightSavingTime"]["end_day"]       = dst[2].day
                    cfg_t["system"]["ntp"]["DaylightSavingTime"]["end_hour"]      = dst[2].hour
                    cfg_t["system"]["ntp"]["DaylightSavingTime"]["end_minute"]    = dst[2].min
                end
            end
            if errMsg then
                logger:debug(errMsg)
            end
        end

        -- ToS settings
        --XXX: wait parameters description from vendor
        
        -- Jitter buffer settings
        --TODO: need checking
        
        -- Silence detection
        --TODO: need checking
        
        -- End digit settings
        --TODO: need checking
        
        -- Interdigit timeout
        local inter_digit_time = h.get_interdigit_timeout() -- (the same on FXSR3, web-service parameter-timeout for dialing) 
        if 3 > tonumber(inter_digit_time) then  --Timeout must be between 3 and 10 sec
            inter_digit_time = 3
        elseif 10 < tonumber(inter_digit_time) then
            inter_digit_time = 10
        end
        cfg_t["fxs_fxo"]["dial_timeout"] = inter_digit_time 

        if gconf.DelayAfterOffhook then
            cfg_t["fxs_fxo"]["fxo"]["outgoing"]["answer_caller_delay"] = gconf.DelayAfterOffhook
        end

--      no rtp detect. the same setting on FXS, so disable it
--      if "w04fxor3" == model then
--          cfg_t["fxs_fxo"]["no_rtp_detect"]             = "disable" -- "enable" value drops hold (WMS7067, WMS4570)
--      end

        -- Silence detection
        if "on" == tports["silence_detection_enabled"] then
            cfg_t["fxs_fxo"]["no_rtp_detect"] = "enable"
            cfg_t["fxs_fxo"]["no_rtp_period"] = tports["silence_detection"]
        else
            cfg_t["fxs_fxo"]["no_rtp_detect"] = "disable"
        end

----------------------------------------------------------------
------- Accounts and SIP settings
----------------------------------------------------------------
        -- inherited from existing FXO module
        -- Failover
        local pbxFailover     = nil
        local failoverSipPort = nil
        local failoverObj = require"common.failover":new()
        if failoverObj and "1" == failoverObj:isEnabled() then
            local failoverCfg = failoverObj:getConfig()
            pbxFailover = failoverCfg and failoverCfg.pbxFailoverDomain
            failoverSipPort = failoverCfg and failoverCfg.sipPort
            logger:debug(pbxFailover, "failoverCfg")
        end

        --Set SIP transport and SIP port
        local transport = "TLS"
        local sip_port = "5061"
        local gtransp = nil
        local gsipport = nil
        if is_addr_remote then
            sip_port  = agent["alternate_port"] or "443"
            gtransp = gconf.SIPTransportRemote
            gsipport = gconf.SIPServerPortRemote
        else
            gtransp = gconf.SIPTransport
            gsipport =  gconf.SIPServerPort
        end
        transport = gtransp or transport
        sip_port = gsipport or sip_port
        if pbxFailover then
            transport       = "TCP"
            sip_port        = agent["alternate_port"] or "443"
            failoverSipPort = failoverSipPort or sip_port
        end

        -- SIP server options
        cfg_t["sipserver"]["transport"]  = transport
        if pbxFailover then
                -- support of 2 sip servers removed for new devices
                cfg_t["sipserver"]["outbound"]["domain"]     = registrar
                cfg_t["sipserver"]["outbound"]["port"]       = sip_port
                cfg_t["sipserver"]["outbound"]["sec_domain"] = pbxFailover
                cfg_t["sipserver"]["outbound"]["sec_port"]   = failoverSipPort
                -- due to strange reasons we need also settings below
                cfg_t["sipserver"]["server0"]["port"]        = sip_port
                cfg_t["sip"]["only_accept_calls_from_ACL"]   = "disable" -- ???????

        else
            cfg_t["sipserver"]["server0"]["domain"] = registrar
            cfg_t["sipserver"]["server0"]["port"]   = sip_port
        end

        -- inherited from existing FXO module 
        -- Set codec priority
        local netband   = h.is_wideband(agent["address"]) and "wideband" or "narrowband"
        local codecs    = resource.codecs[netband]
        local codec_idx = 0
        for param, value in pairs(codecs) do
            cfg_t["media"][param] = value
            codec_idx = codec_idx + 1 
        end
        
        -- set unused codecs to disable state
        for idx = codec_idx, 15 do
            cfg_t["media"]["codec" .. idx] = resource.codecs["dummy"]["disable"]
        end
        
        -- set codec priority
        if "wideband" == netband then
            cfg_t["media"]["codecs_prefered"] = "1"     -- remote                                                                                                                                                                                      
        else
            cfg_t["media"]["codecs_prefered"] = "0"     -- local                                                                                                                                                                                                                                
        end

    end
    
----------------------------------------------------------------
--- Result serialization
----------------------------------------------------------------
    local root_elm     = __create_xml_table(cfg_t, request["config"]) -- build XML table for configuration file
    
    local wxml_obj     = require"core.wxml":new()
    local base_xml_str = '<?xml version="1.0" encoding="UTF-8"?>'
    local xml_cfg      = wxml_obj:generate(root_elm, base_xml_str)

    local extra_lines = reboot_needed and "\n\n" or "\n"
    return xml_cfg and xml_cfg .. extra_lines
end

function setuser (...) end
function setfeatures (...) end
