module( ... or "wildixfxs", package.seeall)

hints = {
	--[[
	======================
	Wildix FXS
	----------------------
		HTTP/1.0 200 OK.
		Cache-Control: no-cache.
		Content-Type:text/html.
	]]
    {
         ['cache-control'] = 'no-cache'
        ,['status']        = "200"
        ,['scode']         = "HTTP/1%.0.*"
    }
}


function verify(host, port)
    local logger     = logger
    local http       = require"core.httpClient":new()
    local xml        = require"core.htmlparser"
	local helper     = require"common.helper"
	local connection = require"common.core".connection
	
	local is_ssl_negotiate  = 443 == port.number
	local PORT_NUMBER       = is_ssl_negotiate and not gconf.HTTPForceUse and 443 or 80
	local t                 = {["ssl"] = is_ssl_negotiate}
	
	local function __remove_tag(txt)
		if txt then
		    return string.match(txt, ".+>.+>(.-)<.+<.+")
		end	
	end
	
	-- indentify device family
	local title  = nil
	local answer = http:rawget(host.ip, PORT_NUMBER, "/") or {}
	if answer.body then
	    local doc = xml:new(answer.body)    -- pre-verify index page
	    title     = string.match(doc:find("//title") or "", ".+>(.-)<.+")
	end
	if title ~= "Wildix W01FXS" and not title:find("W0[1-2]FXSR2") then
        logger:error("Not found verified index page for host %s", host.ip or "")
        return false
    end
	
	local gpassword         = helper.get_gpassword()
	local auth_table        = {"admin:admin", "admin:" .. gpassword}
	if title == "Wildix W01FXS" then
		table.insert(auth_table, 2, "administrator:84117341")
	end

	local sql, cursor, token_pass, dublicate_ip
	if host.mac_addr_str == "" then
		sql = [[ SELECT token, count(id) FROM autoprov WHERE param=%q AND value LIKE '%%"address":%%"%s"%%' AND value LIKE '%%"model":%%"w0_fxs%%"%%' ]]
		cursor = connection:execute(sql, "device", host.ip or "")
		token_pass, dublicate_ip = cursor:fetch(); cursor:close()
		if tonumber(dublicate_ip) > 1 then
			logger:error("Found %s duplicated address %q at provisioning database", dublicate_ip, host.ip or "")
			return t
		end
	else
   		sql = "SELECT token FROM autoprov WHERE param=%q AND mac=%q"
    	cursor = connection:execute(sql, "device", host.mac_addr_str)
    	token_pass = cursor:fetch(); cursor:close()
    end
    
    t["token_type"] = helper.is_new_token(token_pass)
    if t["token_type"] == "new" then
    	table.insert(auth_table, 1, "admin:"..token_pass)
    	t["token"] = token_pass
    end
    logger:debug(auth_table, "auth_table")	

	local answer = nil
	for _, entry in pairs(auth_table) do
	   local login, passwd = entry:match("(.*):(.*)")
	   local data          = string.format("page=%%2F&user=%s&pass=%s", login, passwd)
	   answer              = http:rawpost(host.ip, PORT_NUMBER, "/login.cgi", data) or {}
	   if answer.body and answer.body:find("frameset") then
		  t["auth"] = {user = login, pass = passwd}
		  break
	   end
	end
	if not t["auth"] then
		logger:error("Wrong authorization for host %s", host.ip)
		return t
	end
	
	answer          = http:rawget(host.ip, PORT_NUMBER, "/info.htm") or {}
	local doc       = xml:new(answer.body)
	local old_model = true
	local version_str = doc:find('//table[@id="table2"]/tr[2]/td[2]')
	local version_str = __remove_tag(version_str)
	if not version_str then
		version_str = doc:find('//table[@id="tb1"]/tr[14]/td[2]')
		if not version_str then
			logger:error("Not found version during gateway verififcation")
			return t
		end
		old_model = false
	end

	local fwversion, fwminor, model
	if old_model then		-- verify for planet firmware (version 1)
		model, fwversion = version_str:match("^%s*(.-)%s*$"):gsub("%-",""):lower():match("(.*)%sv(%d+.-%d+)")
		fwminor = doc:find('//table[@id="table2"]/tr[3]/td[2]'):match("^%s*(.-)%s*$"):lower()
		fwminor = __remove_tag(fwminor)
		if fwminor == "1008030" then -- check minor version to other ways
			local msign = require"md5".digest(answer.body)
			local resource = require"resources.wildixfxs"
			if type(resource.minor_version) == "table" and resource.minor_version[msign] then
				fwminor = resource.minor_version[msign]
			end
		end
		if host.mac_addr_str == "" then
			answer            = http:rawget(host.ip, PORT_NUMBER, "/WAN.htm") or {}
			doc               = xml:new(answer.body)
			host.mac_addr_str = (doc:find('//input[@name="MAC1"]/@value') or ""):lower()
		end	
	else	-- verify for welltech firmware (version 2)
		fwversion, fwminor = version_str:match("^%s*(.*)%-(%d*)%s*$")
		if fwversion and fwminor then
			model = title:lower()
		end
		if host.mac_addr_str == "" then
			host.mac_addr_str = (doc:find('//table[@id="tb1"]/tr[5]/td[4]') or ""):match("^%s*(.-)%s*$"):gsub(":", ""):lower()
		end
	end
	t["model"]     = model
	t["fwversion"] = fwversion
	t["fwminor"]   = fwminor
	
	if host.mac_addr_str then
	   t["mac_addr_str"] = host.mac_addr_str:lower()
	end
	
	return t
end 

function provision (host, port, ...)
	local args = select(1, ...)

	local http = require"core.httpClient":new()
	local h    = require"common.helper"
	
	local PORT_NUMBER = host.ssl and not gconf.HTTPForceUse and 443 or 80
	
	-- processing registrar address
	local registrar, case, err_msg = h.get_registrar(host.ip_src, host.ip) 
	logger:debug("Registrar address %q got from %s%s", registrar, case, err_msg and ". Errors: "..err_msg or "")
	
	local gpassword
	if host.token_type == "new" then
		gpassword = host.token or h.get_gpassword()
	else
		gpassword = h.get_gpassword()
		if host.token_type == "old" then
			logger:info("Token has old format")
		end
	end
	
	local config_data, data = {
		 ["registrar"] = registrar,
		 ["login"]     = "wmsdevice",
		 ["password"]  = gpassword,
	}
	
	local options = { timeout  = 2 }
	local answer  = nil
	if host.model then 
		if host.model == "w01fxs" then -- provision for planet (version 1)
			config_data["provlink"] = http.urlEncode("/wfxs/" .. host.model .. "%20v" .. host.fwversion .. "%20" .. host.fwminor .. "/" .. host.mac_addr_str .. "/")
			data = string.gsub("ACONF=3&DHCPF=0&TFTPS=&TFTPP=&HTTPS=$registrar&HTTPT=80&HUSER=$login&HPASS=$password&HTTPP=$provlink&FTPS1=&FUSER=&FPASS=&FPATH=&T1=7&submit=Submit", "%$(%w+)", config_data)
			answer = http:rawpost(host.ip, PORT_NUMBER, "/autoConf.cgi", data, options) or {}
			logger:debug("Setting autoprovisioning parameters for '%s' - %s", host.model, answer.code or 408)
			answer = http:rawpost(host.ip, PORT_NUMBER, "/save.cgi", "submit=Save", options) or {} -- save and reboot
			logger:info("Sending http request to '%s' for autoprovision now", host.model)
		else	-- provision for welltech (version 2)
			data = string.gsub("ACONF=03&CONFS=00&DHCPF=00&CNFSH=00&TFTPS=&TFTPP=&HTTPS=$registrar&HTTPP=&FFTPS=&FUSER=$login&FPASS=$password&FPATH=", "%$(%w+)", config_data)
			answer = http:rawpost(host.ip, PORT_NUMBER, "/autoConf.cgi", data, options) or {}
			logger:debug("Setting autoprovisioning parameters for '%s' - %s", host.model, answer.code or 408)
			answer = http:rawget(host.ip, PORT_NUMBER, "/saveChange.htm", options) or {} -- save and reboot
			logger:info("Sending http request to '%s' for autoprovision now - %s", host.model, answer.code or 408)
		end
	end
end

-- This configuration for Planet send only upgrade parameters. Used fully for Welltech W01FXS
function configuration(agent, ...)
	local json        = require"dkjson"
	local md5         = require"md5"
	local wmsldap     = require"core.wmsldap"
	local pbxutils    = require"core.pbxutils"
	local core        = require"common.core"
	local h           = require"common.helper"
	local logger      = logger
	local connection  = core.connection 
	local err, sql, cursor
	local registrar = agent["registrar"]
	local old_model = agent["model"] == "w01fxs"
    
	-- retrieve device information
    local devInfo    = core.updDeviceInfo(agent)
    local token_pass = devInfo.token
    if devInfo.needUpgrade then
		-- Chech firmware version from db
		if devInfo["fw:additional"] == "revalidate" then
            -- retrieve firmware version from device WEB side
            agent["fw:additional"] = "revalidate"
            local scan_data = h.run("sudo wmsbind -so scan %q", agent["address"])
            if scan_data and "" ~= scan_data then
    			local json_str = scan_data:match(".*(%{[^}]*%})")
    			if json_str then
    				json_str                 = json.decode(json_str)
    				agent["fw:version"]      = json_str["fwversion"]
    				agent["fw:additional"]   = json_str["fwminor"]
                    devInfo["fw:version"]    = agent["fw:version"]
                    devInfo["fw:additional"] = agent["fw:additional"]
    			end
    		end
		end
		if "revalidate" ~= agent["fw:additional"] then -- if we can get FW info from WEB
            core.saveDeviceInfo(devInfo)    -- Update db with new agent values
		else
			logger:error("Couldn't get firmware info from WEB interface %q. DB was not updated.", agent["address"])
		end
	end
	
	-- Check new version
	local is_new_fw = false
	if "revalidate" ~= agent["fw:additional"] then -- if we can get FW info from WEB
    	local firmwares = core.get_firmware_pack({vendor = "wildixata"})
    	if firmwares and firmwares[agent["model"]] then
    		local tcfg = firmwares[agent["model"]]
    		if tcfg["version"] ~= agent["fw:version"] or tcfg["code"] ~= agent["fw:additional"] then
    			logger:info("Present new firmware for %s: %s.%s",agent.model, tcfg["version"] or "", tcfg["code"] or "")
    			is_new_fw = true
    		end
    	end
    end
	
	local resource	-- Load default values
	if old_model then
		resource = core.load_resource("wildixfxs") or {}
	else
		resource = core.load_resource("wildixfxsr2") or {}
	end
	local result = resource.parameters_init or {}
	
	-- Provisioning and firmware
	result["HTTPServer"]  = registrar
	result["UHTTPServer"] = registrar
	if old_model then
		result["HTTPServerPort"] = agent["alternate_port"] or agent["registrar:port"]
		result["HTTPPassword"] = token_pass or h.get_gpassword()
		if "revalidate" ~= agent["fw:additional"] then
    		local path = string.format("/wfxs/w01fxs%%20v%s%%20%s/%s/", agent["fw:version"], agent["fw:additional"], agent["mac"])
    		result["HTTPPath"] = path
    		result["UHTTPPath"] = path
    	end
	else
		result["FTPPassword"]	= token_pass or h.get_gpassword()	-- Username for provision via HTTP or FTP
		result["HTTPPort"]	= agent["alternate_port"] or agent["registrar:port"]
	end
	
	-- Access for device login/password
	if old_model then
		result["AdminUserPass"] = token_pass or h.get_gpassword()
		result["NormalUserPass"] = token_pass or h.get_gpassword()
	else
		result["AdminUserPass"] = "wil--dix"..token_pass or "84117341"
		result["SystemUserPass"] = token_pass or h.get_gpassword()
		result["NormalUserPass"] = token_pass or h.get_gpassword()
	end
	
	if not is_new_fw then
		result["AutoUpdateMode"] = "0" -- disable auto upgrade
		pbxutils.table_concat(result, resource.parameters or {}) -- merge default values

----------------------------------------------------------------
------- Assigned user data 
----------------------------------------------------------------
		local account_data = {}
		local unknown_user = nil
		local fxs_ports    = resource.models[agent["model"]:upper()]
		
		sql = "SELECT value, flag FROM autoprov WHERE param=%q AND mac=%q"
		local sql_rows = connection:rows(sql, "username", agent["mac"])
		local sqlFixPortNumberStr = nil
		if not sql_rows then
			return {}, "", "Database request error"
		end
		for username, port in sql_rows do
			if username:find("unknown") then
				unknown_user = username		-- save old unknown account for current configuration
			elseif username and "" ~= username then
				local port = tonumber(port)
				if "w01fxsr2" == agent["model"] and (not port or "" == port) then
					port = 1
					connection:execute("delete from autoprov where param='username' and mac=%q and flag='1'", agent["mac"])
				end
				if port and "" ~= port then
					local res    = wmsldap:get(h.get_extenNumber_str(), username)
					local person = {}
					if res and res[1] then
						_, person = unpack(res[1])
						local user_features = person.uid and require"common.userFeatures.userFeatures":load(person.uid) or {}
						account_data[port] = {
							["login"]		= username,
							["password"]	= person["wuserPasswordSip"],
							["userKind"]	= person["userKind"],
							["display"]		= (person["cn"] or username) .. "-" .. port --display name like "User name-1", where '-1' - port number (used in service 99 LOGIN to identify FXS port number)
						}
						if user_features.HOTLINE then
							local hotlineDelay = tonumber(user_features.HOTLINE.delay) or 0
							hotlineDelay = hotlineDelay < 0 and 0 or hotlineDelay > 6 and 6 or hotlineDelay -- hotline delay have to be in range 0 - 10 sec.
							account_data[port]["hotline"] = {
							enable = user_features.HOTLINE.status or 0,
							number = user_features.HOTLINE.value or "",
							delay = hotlineDelay
							}
						else
							account_data[port]["hotline"] = {enable = 0, number = "", delay = 0}
						end
					else
						return {}, "", "LDAP connection error"
					end
				end
			end
		end
		
		-- fill not used account data by unknown user
		if not unknown_user then	-- get new unknown user
			local unknown_ports = {}	-- list port without account
			for idx = 1, fxs_ports do
				if not account_data[idx] then
					table.insert(unknown_ports, idx)
				end
			end
			if unknown_ports[1] then	-- write to autoprov used unknown account
				unknown_user = core.register(agent["mac"], nil, unknown_ports) 
			end
		end
		for idx = 1, fxs_ports do
			if not account_data[idx] then
				local passwd, err = h.pass_for_unknown(unknown_user)
				if err then
					logger:error("password generation error for unknown account:", err)
				end
				account_data[idx] = {
					["login"]		= unknown_user,
					["password"]	= passwd,
					["display"]		= "port-"..idx, --display name like "port-1", where '-1' - port number (used in service 99 LOGIN to identify FXS port number)
					["hotline"]     = {enable = 1, number = "99@"..registrar or "", delay = 0}
				}		
			end
		end
		

        -- QOS
        local tos_data, toneZone = h.get_sip_settings()
        result["VoiceQoS"] = tos_data["tos"]/4 -- Voice QoS
        result["SIPQoS"] = tos_data["tos_sip"]/4 -- SIP QoS
        local sipTransport = "UDP"
        local gtransp      = nil
        local sipPort      = "5060"
        local gsipport     = nil
        -- WMS-7159 Temporary workaround to fix problem with sip registration and avoid SIPRTPEnCryption
        -- tls_srtp variable has been deleted in WMS-9041! Change the logics before uncomment!
        --[[ 
        local sipSecurPort = h.getSipRedirectPort() or sipPort
        if h.isLocalIP(agent["address"]) then
            if tls_srtp then
                sipTransport = "TLS"
                sipPort      = sipSecurPort
            end
            gtransp  = gconf.SIPTransport
            gsipport = gconf.SIPServerPort
        else
            sipTransport = "TLS"
            sipPort      = sipSecurPort
            gtransp = gconf.SIPTransportRemote
            gsipport = gconf.SIPServerPortRemote
        end ]]--
        sipTransport = gtransp or sipTransport
        sipPort      = gsipport or sipPort
        local sipEncrypt = "TLS" == sipTransport and "1" or "0"
        local sipEncryptKey = h.getWelltechEncryptionKey()
        if sipEncrypt and sipEncryptKey then
            result["SIPRTPEnCrypt"] = sipEncrypt
            --TODO check "sip encrypt key" parameter name after vendor relese firmware with support custom sip encrypt key.
--          result["SIPRTPEnCryptKey"] = sipEncryptKey
        end

		-- Failover
		local pbxFailover = nil
		local failoverObj = require"common.failover":new()
        if failoverObj and "1" == failoverObj:isEnabled() then
            local failoverCfg = failoverObj:getConfig()
            pbxFailover = failoverCfg and failoverCfg.pbxFailoverDomain or pbxFailover
            logger:debug(pbxFailover, "failoverCfg")
        end

		-- Processing account settings
		for idx, data in ipairs(account_data) do
			if old_model then
				result["ph"..idx.."HotLineEnable"] = account_data[idx].hotline.enable
				result["ph"..idx.."HotLineNumber"] = account_data[idx].hotline.number
			else
				result["ph"..idx.."HotLine"] = account_data[idx].hotline.enable..","..account_data[idx].hotline.number..","..account_data[idx].hotline.delay
			end
			result["ph"..idx.."Realm1DisplayName"] 	= data.display
    		result["ph"..idx.."Realm1UserName"]		= data.login
    		result["ph"..idx.."Realm1RegisterName"]	= data.login
    		result["ph"..idx.."Realm1RegisterPass"]	= data.password
    		result["ph"..idx.."Realm1DomainServ"]	= registrar
    		result["ph"..idx.."Realm1ProxyServ"]	= registrar
    		result["ph"..idx.."Realm1Outbound"]		= registrar .. ":" .. sipPort
    		result["ph"..idx.."Realm1Active"]       = "1"
    		if not data.login:find("unknown%d") and pbxFailover then
    			result["ph"..idx.."Realm2DisplayName"] 	= data.display
        		result["ph"..idx.."Realm2UserName"]		= data.login
        		result["ph"..idx.."Realm2RegisterName"]	= data.login
        		result["ph"..idx.."Realm2RegisterPass"]	= data.password
        		result["ph"..idx.."Realm2DomainServ"]	= pbxFailover
        		result["ph"..idx.."Realm2ProxyServ"]	= pbxFailover
        		result["ph"..idx.."Realm2Outbound"]		= pbxFailover
        		result["ph"..idx.."Realm2Active"]       = "1"
        	end
		end
	
		-- Set codec priority
		local netband_str = h.is_wideband(agent["address"]) and "wideband" or "narrowband"
		if old_model then
			pbxutils.table_concat(result, resource.codecs[netband_str] or {})
		else
			if netband_str == "wideband" then
				result["CodecPri"] = "ALAW,ULAW,G729"
			else
				result["CodecPri"] = "G729,ALAW,ULAW"
			end
		end
		
		-- Get network settings
		sql = "SELECT value FROM autoprov WHERE param=%q AND mac=%q"
		cursor = connection:execute(sql, "settings", agent["mac"])
		local settings_json = cursor:fetch(); cursor:close()
		local t = {}
		local update_db_settings = false
		if settings_json then			 
            t = json.decode(settings_json)
        end
        
        -- SYSLOG settings
        if t["syslog_server"] and t["syslog_server"] ~= "" then
        	result["SysLogSrv"] = t["syslog_server"]
        	result["SysLogType"] = "7"
        end
        
        -- VLAN settings
		if tonumber(t["vid_voice"]) and tonumber(t["vid_voice"]) > 1 then
			result["VLANEnable"] = "1"
			result["VLANTag"] = t["vid_voice"]
			result["VLANPriority"] = t["qos_voice"]
		end
	
		-- set default DNS and NTP settings if not define
		t["dns_server"], t["second_dns_server"] = h.get_prim_sec_dns_servers(t["dns_server"], t["second_dns_server"])
		t["ntp_server"] 						= t["ntp_server"] or ""
	
		-- use static ip settings
		local is_addr_present = h.is_addr_present(agent["address"], agent["mac"])	-- check dublicate address
		local is_addr_local = h.isLocalIP(agent["address"])		-- device behind NAT can't user static ip
		if t["use_static_ip"] and (is_addr_present or not is_addr_local) then
			if is_addr_present then
				logger:warning("Disable static ip. Other device have the same ip address: %s", agent["address"] or "")
			end
			if not is_addr_local then
				logger:warning("Disable static ip. Device ip address (%s) is placed behind NAT", agent["address"] or "")
			end
			t["use_static_ip"] = nil
			update_db_settings = true
		end

		-- get settings for leased device by pbx dhcp
		local is_leased
		if not t["use_static_ip"] or t["dns_server"] == "" or t["ntp_server"] == "" then
			is_leased = h.is_addr_leased(agent["address"])
		end
		if not t["use_static_ip"] and not is_addr_present and is_leased then
			h.set_addr_static(agent["address"], agent["mac"])	-- mark this lease as static
			t["use_static_ip"] = 1
			update_db_settings = true
		end
		if is_leased then
    		if "" == t["dns_server"] or "" == t["second_dns_server"] then
	    		local dns_dhcp1, dns_dhcp2 = h.get_dns_servers()
        		if dns_dhcp1 and "" ~= dns_dhcp1 and "" == t["dns_server"] then
        			t["dns_server"] = dns_dhcp1             -- set primary DNS from DHCP config
        			update_db_settings = true
        		end
        		if dns_dhcp2 and "" ~= dns_dhcp2 and "" == t["second_dns_server"] then
        			t["second_dns_server"] = dns_dhcp2      -- set secondary DNS from DHCP config
        			update_db_settings = true
        		end
	    	end			
		    if t["ntp_server"] == "" then -- set NTP from DHCP config
		        local ntp_dhcp = h.get_ntp_servers()
		        if ntp_dhcp and ntp_dhcp ~= "" then
		            t["ntp_server"] = ntp_dhcp
		            update_db_settings = true
		        end
		    end
		end

		if t["use_static_ip"] == 1 then	-- use static ip
			local _, netmask, gateway = h.host_settings(agent["address"], registrar)
			if netmask and gateway then
				result["IPType"] = old_model and "0" or "1" 	-- for new fmw 1 - static ip
				result["IPAddress"] = agent["address"]
				result["IPMask"] = netmask
				result["IPGateway"] = gateway
				
				if "" == t["dns_server"] then
					t["dns_server"]    = registrar
					update_db_settings = true
				end
				if "" == t["second_dns_server"] then
					t["second_dns_server"] = registrar
					update_db_settings     = true
				end
				
				if t["ntp_server"] == "" then
					t["ntp_server"]    = registrar
					update_db_settings = true
				end
			else
				logger:warning("For current network configurations is not possible set static ip, disable it")
				t["use_static_ip"] = nil
				update_db_settings = true
			end
		end
	
		-- update devdb with new settings
		if update_db_settings then
			connection:switch("devdb")
			local sql = "INSERT OR REPLACE INTO autoprov (id, param, value, mac) VALUES ((SELECT id FROM autoprov WHERE param=%q AND mac=%q), %q, '%s', %q)"
			connection:execute(sql, "settings", agent["mac"], "settings", json.encode(t), agent["mac"])
		end
	
		-- DNS settings 
		if "" ~= t["dns_server"] then
			result["DnsType"] = old_model and "0" or "1"
			result["IPDNS0"] = t["dns_server"]
			if "" ~= t["second_dns_server"] then
				result["IPDNS1"] = t["second_dns_server"]
			end
		end
	
		-- NTP settings
		local timeZone = t["ntp_timezone"]
		if not timeZone or "" == timeZone then
		    timeZone = h.getPbxTimeZone()
		end
		if "" ~= t["ntp_server"] or "no" ~= t["use_dst"] then
			local td, errMsg = require"core.timedate":new(os.time(), timeZone)
            if td then
            	td:toTimeZone()
    			-- NTP server and offset (time zone)
    			if t["ntp_server"] ~= "" then
    				result["SNTPEnable"]  = 1
    				result["SNTP1stServ"] = t["ntp_server"]
    				result["SNTP2ndServ"] = "pool.ntp.org"
    				-- time zone offset
				end
				local tzoffset = td:getTzOffset()
    			if tzoffset and tzoffset.sign and tzoffset.hour and tzoffset.min then
                	local sign = ("" == tzoffset.sign and "+") or tzoffset.sign
                	result["SNTPTimeZone"] = sign .. tzoffset.hour .. ":" .. tzoffset.min -- time zone offset, "+01:00" for Italy 	
    			end
    			-- Daylight saving time
    			local dst = "no" ~= t["use_dst"] and td:getDstData(timeZone)
    			if dst and dst.offsetsign and dst.offsethour then
    				result["SNTPEnable"]      = 1
    				result["DayLightEnable"]  = 1
    				result["DSTOffset"]       = dst.offsetsign .. dst.offsethour
    				result["DSTStartDate"]    = dst[1].month .. ",1," .. dst[1].day .. ",1,7," .. dst[1].hour
    				result["DSTEndDate"]      = dst[2].month .. ",1," .. dst[2].day .. ",1,7," .. dst[2].hour
    			end
    		end
            if errMsg then
            	logger:debug(errMsg)
            end
		end
	
		-- Jitter buffer
		local tjb = h.get_jitter_buffer(is_addr_local)
		if old_model then
			result["JitterBuffer"] = math.floor(tjb["max"]/20 + 0.5)
		else
			-- Note: JitterBufferMax & JitterBuffer must different (this condition observes WMS)
			local jit_buf = math.floor(tjb["min"]/20 + 0.5)
			result["JitterBuffer"] = jit_buf > 64 and 64 or jit_buf
			
			jit_buf = math.floor(tjb["max"]/20 + 0.5)
			result["JitterBufferMax"] = jit_buf > 64 and 64 or jit_buf
		end

		local toneZone = t.tonezone or toneZone
		local tone = h.get_dialtone("welltech", toneZone)
		local tonesGain = resource.tonesGain
		
		if tone and "table" == type(tone) and tonesGain and "table" == type(tonesGain) then
    		for k,v in pairs({
    			["DialTone"]     = "dial",
    			["RingBackTone"] = "ring",
    			["BusyTone"]     = "busy",
    			["CongesTone"]   = "congestion",
    			["RingTone"]     = "ring",
    			["CallWaitTone"] = "callwaiting"
    		}) do
    			if tone[v] and "table" == type(tone[v]) and tonesGain[k] and "table" == tonesGain[k] then
        			local set1 = tone[v][1]
        			local set2 = tone[v][2]
        			local set3 = tone[v][3]
        			local set4 = tone[v][4]
                    local set5 = tone[v][5]
                    local set6 = tone[v][6]
                    local CadenceOn  = "dial" == v and "0" or "1"
        			local freq1      = set1 and set1["freq1"] or "425"
        			local freq2      = set1 and set1["freq2"] or "0"
        			local hiToneGain = tonesGain[k]["hiToneGain"]
        			local loToneGain = tonesGain[k]["loToneGain"]
        			local onTime1    = (set1 and set1["duration"] or "0")/10
        			local offTime1   = (set2 and set2["duration"] or "0")/10
        			local onTime2    = (set3 and set3["duration"] or "0")/10
        			local offTime2   = (set4 and set4["duration"] or "0")/10
        			local onTime3    = (set5 and set5["duration"] or "0")/10
        			local offTime3   = (set6 and set6["duration"] or "0")/10
        			result[k] = CadenceOn..","..freq1..","..freq2..","..hiToneGain..","..loToneGain..","..onTime1..","..offTime1..","..onTime2..","..offTime2..","..onTime3..","..offTime3
        		end
    		end
    	end
	end

	-- make output
	local output
	if old_model then	-- auto provision tools running start key word
		output = {"line0Name="}
	else
		output = {"$TXT", "line0Name="}
	end	
	for key, value in pairs(result) do
		output[#output+1] = key .. "=" .. value
	end
	return output, username, err
end

function setuser (...) 	end

function setfeatures (...) end
