--- Provision module for 24FXS revision 2
-- 
-- @file	wildixfxs2.lua 
-- @author	Vyacheslav Khudan
-- @created 2015/04/02

module( ... or "wildixfxs2", package.seeall)

----------------------------------------------------------------
-- Private interface
----------------------------------------------------------------

--- Push section configuration
local function __add_section_cfg(cfg_t, name, root_elm)
	if not name or "" == name then
		return false
	end
	
	local sect_cfg = cfg_t[name] 
	if "table" ~= type(sect_cfg) then
		return false
	end
	
	-- pushing configuration
	local elm = root_elm:add_elm(name)
	for pname, pvalue in pairs(sect_cfg) do
		if pname and pvalue then
			if "table" == type(pvalue) then
				__add_section_cfg(sect_cfg, pname, elm)
			else
				local param_elm = elm:add_elm("param")
				param_elm:add_attr({name = pname, value = pvalue})
			end
		end
	end
end

--- Create XML table from configuration table
local function __create_xml_table(cfg_t, cfg_type)
	if "table" ~= type(cfg_t) then
		return false
	end
	
	local wxml_obj = require"core.wxml":new()
	local root_elm = wxml_obj:elm_tbl()
	
	if "policy" == cfg_type then
		local prov_elm = root_elm:add_elm("provision")
		prov_elm:add_attr({version = "2.0"})
		
    	if cfg_t["firmware"] then
    		-- push firmware parameters
    		local prod_elm = prov_elm:add_elm("product")
    		prod_elm:add_attr(cfg_t["product"])
    		prod_elm:add_attr(cfg_t["firmware"]["product"])
    		
    		local pack_elm = prod_elm:add_elm("package")
    		pack_elm:add_attr(cfg_t["firmware"])
    	end
    	
    	-- push config parameters
    	local prod_elm = prov_elm:add_elm("product")
    	prod_elm:add_attr(cfg_t["product"])
    	prod_elm:add_attr({url = cfg_t["config"]["product"]["url"]})
    	
    	local conf_elm = prod_elm:add_elm("ConfigFile")
    	conf_elm:add_attr(cfg_t["config"])
	else
		local conf_elm = root_elm:add_elm("config")
		conf_elm:add_attr({
			version = cfg_t["version"],
			md5     = cfg_t["md5"]
		})	
		__add_section_cfg(cfg_t, "provision", conf_elm)				-- push provision parameters
		__add_section_cfg(cfg_t, "syslog", conf_elm)				-- push syslog parameters
		__add_section_cfg(cfg_t, "network", conf_elm)				-- push network parameters
		__add_section_cfg(cfg_t, "fxs_fxo", conf_elm)				-- push FXS/FXO parameters
		__add_section_cfg(cfg_t, "feature_code", conf_elm)			-- push feature codes parameters
		__add_section_cfg(cfg_t, "sipserver", conf_elm)				-- push SIP servers parameters
		__add_section_cfg(cfg_t, "media", conf_elm)					-- push media parameters
		__add_section_cfg(cfg_t, "sip", conf_elm)					-- push SIP parameters
		__add_section_cfg(cfg_t, "sipacc", conf_elm)				-- push SIP account settings
		__add_section_cfg(cfg_t, "ports", conf_elm)					-- push ports settings
		__add_section_cfg(cfg_t, "fax", conf_elm)					-- push fax settings
		__add_section_cfg(cfg_t, "digitmap", conf_elm)				-- push digitmap settings
		__add_section_cfg(cfg_t, "system", conf_elm)				-- push system settings
		__add_section_cfg(cfg_t, "iptrunk", conf_elm)				-- push iptrunk settings
		__add_section_cfg(cfg_t, "route", conf_elm)					-- push route settings
		__add_section_cfg(cfg_t, "manipulation", conf_elm)			-- push manipulation settings
		__add_section_cfg(cfg_t, "simserver", conf_elm)				-- push simserver settings
		__add_section_cfg(cfg_t, "filelog", conf_elm)				-- push filelog settings
		__add_section_cfg(cfg_t, "serverlog", conf_elm)				-- push serverlog settings
		__add_section_cfg(cfg_t, "snmp", conf_elm)					-- push snmp settings
		__add_section_cfg(cfg_t, "tr069", conf_elm)					-- push tr069 settings
		__add_section_cfg(cfg_t, "actionurl", conf_elm)				-- push actionurl settings
		__add_section_cfg(cfg_t, "websclassify", conf_elm)			-- push websclassify settings
		__add_section_cfg(cfg_t, "relayserver", conf_elm)			-- push relayserver settings
	end
	
	return root_elm
end

local function __makeToneString(tones)
    local rv = nil
    if tones and "table" == type(tones[1]) then
        local freq1 = tones[1]["freq1"] or "0"
        local gain1 = "0" ~= freq1 and "260" or "0"
        local dur1  = tones[1]["duration"] or "0"
        local freq2 = tones[2] and tones[2]["freq1"] or "0"
              freq2 = "0" ~= freq2 and freq2 or freq1
        local gain2 = "0" ~= freq2 and "630" or "0"
        local dur2  = tones[2] and tones[2]["duration"] or "0"
              rv    = freq1..","..gain1..","..freq2..","..gain2..","..dur1..","..dur2..",0,0"
    end
    return rv
end

----------------------------------------------------------------
-- Public interface
----------------------------------------------------------------
hints = {
	--[[
	======================
	Wildix 24FXSR2
	----------------------
		Server: Web Server/2.1.0.
        Date: Fri Apr 10 14:28:51 2015.
        WWW-Authenticate: Basic realm="Web Server".
        Set-Cookie: devckie=da00-0040-af00-0450.
        Pragma: no-cache.
        Cache-Control: no-cache.
        Content-Type: text/html.
	--]]
    {
         ["server"]           = "Web Server%/.*"
        ,["www-authenticate"] = 'Basic realm="Web Server"'
    }	
}

--- Verifying device HTML request and retreiving information
-- @param host		A verifying host data
-- @param port		Not used data
-- @return table with verifying device information
function verify (host, port)
	local logger     = logger
	local http 		 = require"core.httpClient":new()
	local helper 	 = require"common.helper"
	local connection = require"common.core".connection
	
	local is_ssl_negotiate 	=  443 == port.number
	local PORT_NUMBER 		= is_ssl_negotiate and not gconf.HTTPForceUse and 443 or 80
	local t 		  		= {["ssl"] = is_ssl_negotiate}	
	
	local gpassword  = helper.get_gpassword()
	local auth_table = {"admin:admin", "admin:" .. gpassword}

	local sql			= nil
	local cursor		= nil
	local token_pass	= nil
	local dublicate_ip	= nil
	if "" == host.mac_addr_str then
		sql    = [[ SELECT token, count(id) FROM autoprov WHERE param=%q AND value LIKE '%%"address":%%"%s"%%' AND value LIKE '%%"model":%%"w%%fxsr_"%%' ]]
		cursor = connection:execute(sql, "device", host.ip or "")
		token_pass, dublicate_ip = cursor:fetch(); cursor:close()
		if 1 < tonumber(dublicate_ip) then
			logger:error("Found %s duplicated address %q at provisioning database", dublicate_ip, host.ip or "")
			return t
		end
	else
   		sql 	   = "SELECT token FROM autoprov WHERE param=%q AND mac=%q"
    	cursor 	   = connection:execute(sql, "device", host.mac_addr_str)
    	token_pass = cursor:fetch(); cursor:close()
    end
    
    t["token_type"] = helper.is_new_token(token_pass)
    if "new" == t["token_type"] then
    	table.insert(auth_table, 1, "admin:"..token_pass)
    	t["token"] = token_pass
    end
    logger:debug(auth_table, "auth_table")
    
    -- Get device status information via GET request
    -- check authorization
    local answer = nil
    for _, authpair in pairs(auth_table) do
        local login, pass = authpair:match("(.*):(.*)")
        answer = http:rawget(host.ip, PORT_NUMBER, "/enSysInfo.htm", { authUser = login, authPass = pass }) or {}
        if 200 == answer.code then
            t["auth"] = {user = login, pass = pass}
            break
        end
    end
    if not t["auth"] then
		logger:error("Wrong authorization for host %s", host.ip)
		return t
	end
	
	local model, fw_ver = answer.body:match("Current Software Version</td>\r?\n%s*<td colspan=\"%d\">\r?\n%s*(W%d+FX[SO]R%d)%s(.-)%s")
	if not model or "" == model then
    	logger:error("Not found model for host %s", host.ip)
        return t
    end

	t["model"] 			   = model:lower() 
	t["fwversion"] 		   = fw_ver
	t["vendor:additional"] = string.match(model, "FXSR%d")

	if "" == host.mac_addr_str then
		local web_mac = answer.body:match("MAC Address </td>\r?\n%s*<td colspan=\"%d\">\r?\n%s*([0-9a-fA-F%-]*)\r?\n")
		if web_mac then
			host.mac_addr_str = web_mac:gsub("-","")
		end
	end
	t["mac_addr_str"] = host.mac_addr_str:lower()
	
	return t
end

--- Provision action to updating provision url on device and executing autoprovision
-- @param host	A provision device data
-- @param port	Not used 
function provision(host, port, ...)
	local args 		 = select(1, ...)
	local http       = require"core.httpClient":new()
	local h			 = require"common.helper"
	local core 		 = require"common.core"
	local connection = core.connection
	
	local model      = host.model:upper()
	local devFwInfo	 = h.parse_firmware(host.fwversion)
	
	local prov_proto  = gconf.HTTPForceUse and "http" or "https" 	
	local prov_port   = gconf.HTTPForceUse and "80" or "443"			
	local PORT_NUMBER = host.ssl and not gconf.HTTPForceUse and 443 or 80
	
	--the firmware version less than 27.18.01.05 does not support ssl
	local ssl_supported	= 105 <= devFwInfo.sht_v2
	if not ssl_supported then
		prov_proto  = "http" 	
		prov_port   = "80"	
		PORT_NUMBER = 80
	end
	
	local registrar, case, err_msg = h.get_registrar(host.ip_src, host.ip) 
	logger:debug("Registrar ip %q get from %s%s", registrar, case, err_msg and ". Errors: "..err_msg or "")
	
	local gpassword = nil
	if "new" == host.token_type then
		gpassword = host.token or h.get_gpassword()
	else
		gpassword = h.get_gpassword()
		if "old" == host.token_type then
			logger:info("Token has old format")
		end
	end
	
	math.randomseed(os.time())
	local provint  = "6048"..math.random(10,99) -- hack: set random	"ProvisonInterval" to differ from device value, otherwise provision via wmsbind doesn't work since fw. ver 27.18.01.11 	
	local provlink = string.gsub(
		"$protocol://$registrar:$port/?action=setaps", "%$(%w+)",
		{
			protocol  = prov_proto,
			registrar = registrar,
			port 	  = prov_port,
		}
	)

	local options = {
	     authUser = host.auth.user
        ,authPass = host.auth.pass
        ,timeout  = 2
	}
	
	-- Configuring provisioning
	local config_data = {
		"ProvisonURL=" .. http.urlEncode(provlink),
		"ProvisonInterval=" .. provint,
		"ProvisonAcc=wmsdevice",
		"ProvisonPwd=" .. gpassword,
		"ProxyDomain=",
		"ProxyPort=",
		"ProxyAcc=",
		"ProxyPwd=",
		"ok=Save",
		"ModFlag=1",
		"ProxyModFlag=0"
	}
	local data = table.concat(config_data, "&")
	local answer = http:rawpost(host.ip, PORT_NUMBER, "/goform/Provison", data, options) or {}	-- Save options
	logger:info("Send http request for autoprovision to with provision configuration to %s - %s", model, answer.code or 408)
end

--- Building and sending provision configuration
-- @param agent			A table with general device information
-- @param request		A parameters got from provision request		
function configuration(agent, request)
	if "table" ~= type(agent) or "table" ~= type(request) then
		return false
	end
	
	local h           = require"common.helper"
	local core        = require"common.core"
	local logger      = logger
	local connection  = core.connection
	local json        = require"dkjson"
	local md5         = require"md5"
	local wmsldap     = require"core.wmsldap"
	
	local callerid	  = "enable"
	local callwait	  = "enable"
	local devFwInfo	  = h.parse_firmware(agent["fw:version"])
	local cfg_t       = nil -- configuration table
	local prov_proto  = nil
	local prov_port   = nil
	
	-- retreive registrar URL parameters
	local registrar  = agent["registrar"]
	
	--the firmware version less than 27.18.01.05 does not support ssl
	local ssl_supported	= 105 <= devFwInfo.sht_v2
	
	local prov_proto = agent["ssl"] == "on" and not gconf.HTTPForceUse and "https" or "http"
	local prov_port  = gconf.HTTPForceUse and "80" or agent["alternate_port"] or agent["registrar:port"]
	if not ssl_supported then
		prov_proto = "http"
		prov_port  = "80"
		callerid	= "disable" -- invert the value if firmware version less than 27.18.01.05
		callwait	= "disable"	-- invert the value if firmware version less than 27.18.01.05
	end
	
	-- set general url
	local base_url 	 = string.gsub("$protocol://$registrar:$port/", "%$(%w+)",
								   {
										registrar = registrar,
                					   	protocol  = prov_proto,
                					   	port 	  = prov_port,
                				    })
	
	-- retreiving information about device firmware
	local fw_cfg    = nil
	local fw_name   = "wildixfxs2" 	-- directory of the firmware for this devices
	local firmwares = core.get_firmware_pack({["vendor"] = fw_name})
	local model = agent["model"]
	if firmwares then
		fw_cfg = firmwares[model]
	end
	if not fw_cfg then
		logger:error("Firmware pack not found for device %q", model)
		return false
	end
	
	-- retrieve device information from provision DB
	local sql 	 = "SELECT value, token FROM autoprov WHERE param=%q AND mac=%q"
	local cursor = connection:execute(sql, "device", agent["mac"])
	local device_json, token_pass = cursor:fetch(); cursor:close()
	
	token_pass = token_pass or h.get_gpassword()

	local db_device_data = json.decode(device_json or "") or {}
	local reboot_needed = db_device_data.reboot == 1

----------------------------------------------------------------
--- Policy file with provision url and firmware links
----------------------------------------------------------------
	if "policy" == request["config"] then
		cfg_t		 = {} -- buffer for policy configuration file
		
		-- set general product parameters
		cfg_t["product"] = {
			url       = {},	  -- firmware and configuration urls
			id    	  = tostring(devFwInfo.major), -- id is major FW version
    		oemid 	  = "27", -- 27: Wildix, 1:Dinstar 
    		model 	  = "",	  -- provision settings will be used by all devices
--    		macfilter = "",		--TODO: upgrade only Wildix devices
		}
		
		-- firmware settings
    	local fmw_file 		= nil
    	if fw_cfg and fw_cfg["version"] ~= agent["fw:version"] then 
			local fw_filename = fw_cfg["application"]
			local fw_dir  	  = "/var/www/firmwares/" .. fw_name .. "/"
			local fw_file = fw_dir .. fw_filename
			if h.is_exists(fw_file) then
				local fmw_url  = base_url .. "firmwares/" .. fw_name .. "/"
				local fmw_ver  = fw_cfg["version"]:gsub("%.", "")
				fmw_file 	   = fw_filename
				
				-- set firmware parameters
				cfg_t["firmware"] = {
   					ver  	  = fmw_ver,
    				name 	  = fmw_file,
    				reboot    = "delay",	-- Apply setting (wait when active call will be ended)
				}
				local pbxFwInfo	= h.parse_firmware(fw_cfg["version"])
				local downgrade = "false"
				if devFwInfo.sht_v2 > pbxFwInfo.sht_v2 then
					downgrade = "true"
				end
				cfg_t["firmware"]["product"] = {
					url   = fmw_url,
					force = downgrade,	-- true allowing downgrade firmware version
				}
				
				logger:info("Found new firmware version - %s", fw_cfg["version"])
				logger:info("Firmware for device %s %s will be changed from version %s to %s", agent["model"], agent["mac"], agent["fw:version"], fw_cfg["version"])
			else
				logger:warning("Firmware file %q couldn't be loaded", fw_file)
			end
		end
    	
		-- set general configuration parameters
		cfg_t["config"] = {
			name   = "$(MAE).cfg/?action=setaps",	-- Provision url
    		action = "replace",					-- Applying configuration action
--			action = "merge",			--XXX: must be moved to replace afeter vendor fix 					
    		reboot = reboot_needed and "always" or "never",	-- Apply configuration with/without reboot
		}
		cfg_t["config"]["product"] = {url = base_url}	-- Provision init url
	else
	
----------------------------------------------------------------
--- General configuration settings
----------------------------------------------------------------
		local resource   = core.load_resource("wildixfxs2")
		cfg_t 			 = resource.cfg_def or {}
		cfg_t["version"] = fw_cfg["config"] -- confiugration version

		-- update database by device information
		local agent_json = json.encode(agent)
    	if not device_json or md5.digest(device_json) ~= md5.digest(agent_json) then
    		sql = "INSERT OR REPLACE INTO autoprov (id, param, value, mac, token) VALUES ((SELECT id FROM autoprov WHERE param=%q AND mac=%q), %q, '%s', %q, %q)"
    		connection:execute(sql, "device", agent["mac"], "device", agent_json, agent["mac"], token_pass or '')
    	end
    	
----------------------------------------------------------------
------- Assigned user data 
----------------------------------------------------------------
		local account_data = {}
		local unknown_user = nil
		local mviUserList  = ""
		local fxs_ports    = resource.models[agent["model"]:upper()]
		
		sql = "SELECT value, flag FROM autoprov WHERE param=%q AND mac=%q"
		local sql_rows = connection:rows(sql, "username", agent["mac"])
		if not sql_rows then
			return {}, "", "Database request error"
		end
		for username, port in sql_rows do
			if username:find("unknown") then
				unknown_user = username		-- save old unknown account for current configuration
			elseif username and "" ~= username then
				local port = tonumber(port)
				if port and "" ~= port then
					local res    = wmsldap:get(h.get_extenNumber_str(), username)
					local person = {}
					if res and res[1] then
						_, person = unpack(res[1])
						local user_features = person.uid and require"common.userFeatures.userFeatures":load(person.uid) or {}
						account_data[port] = {
							["login"]		= username,
							["password"]	= person["wuserPasswordSip"],
							["userKind"]	= person["userKind"],
							["display"]		= (person["cn"] or username) .. "-" .. port --display name like "User name-1", where '-1' - port number (used in service 99 LOGIN to identify FXS port number)
						}
						if user_features.HOTLINE then
							local hotlineDelay = tonumber(user_features.HOTLINE.delay) or 0
							hotlineDelay = hotlineDelay < 0 and 0 or hotlineDelay > 10 and 10 or hotlineDelay -- hotline delay have to be in range 0 - 10 sec.
							account_data[port]["hotline"] = {
							enable = user_features.HOTLINE.status or 0,
							number = user_features.HOTLINE.value or "",
							delay = hotlineDelay
							}
						else
							account_data[port]["hotline"] = {enable = 0, number = "", delay = 0}
						end
					else
						return {}, "", "LDAP connection error"
					end
				end
				mviUserList = mviUserList .. ("" ~= mviUserList and "," or "") .. username
			end
		end
		
		-- fill not used account data by unknown user
		if not unknown_user then	-- get new unknown user
			local unknown_ports = {}	-- list port without account
			for idx = 1, fxs_ports do
				if not account_data[idx] then
					table.insert(unknown_ports, idx)
				end
			end
			if unknown_ports[1] then	-- write to autoprov used unknown account
				unknown_user = core.register(agent["mac"], nil, unknown_ports) 
			end 
		end
		for idx = 1, fxs_ports do
			if not account_data[idx] then
				local passwd, err = h.pass_for_unknown(unknown_user)
				if err then
					logger:error("password generation error for unknown account:", err)
				end
				account_data[idx] = {
					["login"]		= unknown_user,
					["password"]	= passwd,
					["display"]		= "port-" .. idx, --display name like "port-1", where '-1' - port number (used in service 99 LOGIN to identify FXS port number)
					["hotline"]     = {enable = 0, number = "", delay = 0}
				}		
			end
		end
		
----------------------------------------------------------------
------- Provision settings
----------------------------------------------------------------
		cfg_t["provision"]["url"] 				= base_url .. "?action=setaps"
		cfg_t["provision"]["provison_account"]  = "wmsdevice"
		cfg_t["provision"]["provison_password"] = token_pass

----------------------------------------------------------------
------- Access settings
----------------------------------------------------------------
		if ssl_supported then
			cfg_t["system"]["access_ctrl"]["web"]["ssl_port"]					= "443"
			cfg_t["system"]["access_ctrl"]["cli"]["username_aes"]				= "admin"
			local k = "1234567890123456"
			local err = nil
			if "r3" == string.sub(model, -2) then
				local hex_token = string.upper(require"core.strConverter".toHex(token_pass))
				local hex_token_aes, err = core.encrypt_cipher(hex_token, "aes", k)
				cfg_t["system"]["access_ctrl"]["cli"]["hex_pwd_aes"]				= hex_token_aes
				cfg_t["websclassify"]["users"]["user0"]["hex_pwd_aes"]				= hex_token_aes
			else
				local token_aes, err = core.encrypt_cipher(token_pass, "aes", k)
				cfg_t["system"]["access_ctrl"]["cli"]["password_aes"]				= token_aes
				cfg_t["system"]["access_ctrl"]["web"]["users"]["user1"]["password"] = token_aes
				cfg_t["system"]["access_ctrl"]["cli"]["users"]["user1"]["password"] = token_aes
				cfg_t["websclassify"]["users"]["user0"]["pwd"]						= token_aes
				cfg_t["websclassify"]["users"]["user0"]["pwd_aes"]					= token_aes
			end
			if err then
				logger:debug("Cipher generation is incorrect :" .. err)
			end
		else
			cfg_t["system"]["access_ctrl"]["web"]["users"]["user1"]["password"] = token_pass
			cfg_t["system"]["access_ctrl"]["cli"]["users"]["user1"]["password"] = token_pass
			cfg_t["websclassify"]["users"]["user0"]["pwd"]						= "@5XZ.PW]K[28#;ZD=H#]!E+UUNEW#[0*&amp;H^T[U52`NI!er &lt;"
		end

----------------------------------------------------------------
------- Network configuration
----------------------------------------------------------------
		local net_cfg			 = {}
        local update_db_settings = false
        local is_addr_remote 	 = not h.isLocalIP(agent["address"])
        
        -- get device setting from provision DB
    	sql    = "SELECT value FROM autoprov WHERE param='%s' AND mac='%s'"
    	cursor = connection:execute(sql, "settings", agent["mac"])
    	local settings_json = cursor:fetch(); cursor:close()
        if settings_json then
    		net_cfg = json.decode(settings_json)
    	end
    	
    	-- SYSLOG server (by default syslog server is empty)
    	if net_cfg["syslog_server"] and "" ~= net_cfg["syslog_server"] then
    		cfg_t["syslog"]["flag"] = "on"
    		cfg_t["syslog"]["ip"]   = net_cfg["syslog_server"]
    	end
    	
    	--TODO: SNMP acess server
        	
        -- VLAN settings (set only Voice vlan)
        if tonumber(net_cfg["vid_voice"]) and 0 < tonumber(net_cfg["vid_voice"]) then
            cfg_t["network"]["vlan"]["vlan1_enabled"]     = "enable"
            cfg_t["network"]["vlan"]["vlan1_group"]       = "273"
            cfg_t["network"]["vlan"]["vlan1_id"]          = net_cfg["vid_voice"]
            cfg_t["network"]["vlan"]["vlan1_priid"]       = net_cfg["qos_voice"] or "7"
            cfg_t["network"]["vlan"]["vlan1_ip_mode"]     = "2"
            cfg_t["network"]["vlan"]["wan_vlanid_router"] = "1"
            cfg_t["network"]["vlan"]["lan_vlanid_router"] = "1"
        end

        -- set DNS and NTP settings by pbx network configuration and device setttings
        --TODO: secondary DNS settings
    	net_cfg["dns_server"] = net_cfg["dns_server"] or ""
    	net_cfg["ntp_server"] = net_cfg["ntp_server"] or ""
    	
    	-- checking static option with device location
    	local is_addr_present = h.is_addr_present(agent["address"], agent["mac"])	-- check dublicate address
    	if net_cfg["use_static_ip"] and (is_addr_present or is_addr_remote) then
    		if is_addr_present then
    			logger:warning("Disable static ip. Other device have the same ip address: %s", agent["address"] or "")
    		end
    		if is_addr_remote then	-- device behind NAT can't user static ip
    			logger:warning("Disable static ip. Device ip address (%s) is placed behind NAT", agent["address"] or "")
    		end
    		net_cfg["use_static_ip"] = nil
    		update_db_settings 		 = true
    	end
    	
    	-- get settings for leased device by pbx dhcp
    	local is_leased	= false
    	if not net_cfg["use_static_ip"] or "" == net_cfg["dns_server"] or "" == net_cfg["ntp_server"] then
    		is_leased = h.is_addr_leased(agent["address"])
    	end
    	if not net_cfg["use_static_ip"] and not is_addr_present and is_leased then
    		h.set_addr_static(agent["address"], agent["mac"])	-- mark this lease as static
    		net_cfg["use_static_ip"] = 1
    		update_db_settings 		 = true
    	end
    	if is_leased then
    		if "" == net_cfg["dns_server"] then -- set DNS from DHCP config
    	        local dns_dhcp = h.get_dns_servers()
    	        if dns_dhcp and "" ~= dns_dhcp then
    	            net_cfg["dns_server"] = dns_dhcp
    	            update_db_settings 	  = true
    	        end
    	    end
    	    if "" == net_cfg["ntp_server"] then -- set NTP from DHCP config
    	        local ntp_dhcp = h.get_ntp_servers()
    	        if ntp_dhcp and "" ~= ntp_dhcp then
    	            net_cfg["ntp_server"] = ntp_dhcp
    	            update_db_settings    = true
    	        end
    	    end
    	end
    	
    	if 1 == net_cfg["use_static_ip"] then	-- use static ip settings
    		local _, netmask, gateway = h.host_settings(agent["address"], registrar)
    		if netmask and gateway then
    			cfg_t["network"]["local_network"]["net_mode"] 	 = "1"	
    			cfg_t["network"]["local_network"]["wan_ip"]   	 = agent["address"]
    			cfg_t["network"]["local_network"]["wan_mask"] 	 = netmask
    			cfg_t["network"]["local_network"]["wan_gateway"] = gateway
    			
    			if "" == net_cfg["dns_server"] then
    				net_cfg["dns_server"] = registrar
    				update_db_settings 	  = true
    			end
    	
    			if "" == net_cfg["ntp_server"] then
    				net_cfg["ntp_server"] = registrar
    				update_db_settings    = true
    			end
    		else
    			logger:warning("For current network configurations is not possible set static ip, disable it")
    			net_cfg["use_static_ip"] = nil
    			update_db_settings       = true
    		end
    	end
    	
    	-- set dialtone to FXS
    	local tos_data, defaulttoneZone = h.get_sip_settings()
        local tone = net_cfg["tonezone"]
        if not tone or "" == tone then
            tone = defaulttoneZone or ""
        end
        local isR3 = require"core.pbxutils".in_table(devFwInfo.major or "", {19, 81}) -- model version is R3 (w02fxsr3, w04fxsr3, w24fxsr3)
        local toneParams = isR3 and "dinstar3" or "dinstar2"
        if isR3 and resource.country_tones[tone] then -- WMS-12638
            cfg_t["fxs_fxo"]["call_process_tone"] = "255"
            cfg_t["fxs_fxo"]["busy_tone"] = resource.country_tones[tone]["busy"] or "" -- 425,180,425,630,500,500,0,0
            cfg_t["fxs_fxo"]["dial_tone"] = resource.country_tones[tone]["dial"] or "" -- 425,180,425,630,200,200,600,1000
            cfg_t["fxs_fxo"]["ring_tone"] = resource.country_tones[tone]["ring"] or "" -- 425,180,425,630,1000,4000,0,0
            if 1015 <= devFwInfo.sht_v2 then
                cfg_t["fxs_fxo"]["disconnect_tone"] = resource.country_tones[tone]["disconnect"] or ""
            end
        elseif "number" ~= type(tonumber(tone)) then
            tone = h.get_dialtone(toneParams, tone)
        end
        if "table" == type(tone) then -- custom tones
            cfg_t["fxs_fxo"]["call_process_tone"] = tone.country
            cfg_t["fxs_fxo"]["busy_tone"] = __makeToneString(tone.busy) or "" -- 425,180,425,630,500,500,0,0
            cfg_t["fxs_fxo"]["dial_tone"] = __makeToneString(tone.dial) or "" -- 425,180,425,630,200,200,600,1000
            cfg_t["fxs_fxo"]["ring_tone"] = __makeToneString(tone.ring) or "" -- 425,180,425,630,1000,4000,0,0
            -- WMS-10374 disconnect tone will be set as busy tone only for R3 devices.
            if isR3 and 1015 <= devFwInfo.sht_v2 then
                -- we've decided to set busy tone as disconnect tone, but it might be better for customers to set disconnect tone (tone.congestion). We don't really know.
                cfg_t["fxs_fxo"]["disconnect_tone"] = __makeToneString(tone.busy) or ""
            end
        elseif "number" == type(tonumber(tone)) and not isR3 then -- predefined country tone set for R2 devices
            cfg_t["fxs_fxo"]["call_process_tone"] = tone
        end
        
        -- set QoS
        cfg_t["network"]["qos"] = {}
        cfg_t["network"]["qos"]["lan_aos_enable"] = "enable"
        cfg_t["network"]["qos"]["lan_qos0"] = "0"
        cfg_t["network"]["qos"]["lan_qos1"] = tos_data["tos"]/4     -- Voice QoS
        cfg_t["network"]["qos"]["lan_qos2"] = tos_data["tos_sip"]/4 -- SIP QoS

    	-- update devdb with new settings
    	if update_db_settings then
--    		sql = "INSERT OR REPLACE INTO autoprov (id, param, value, mac) VALUES ((SELECT id FROM autoprov WHERE param=%q AND mac=%q), %q, '%s', %q)"
    		connection:execute (sql, "settings", agent["mac"], "settings", json.encode(net_cfg), agent["mac"])
    	end
        
        -- DNS settings 
        net_cfg["dns_server"], net_cfg["second_dns_server"] = h.get_prim_sec_dns_servers(net_cfg["dns_server"], net_cfg["second_dns_server"])
    	if "" ~= net_cfg["dns_server"] then
    		cfg_t["network"]["local_network"]["wan_dns1"] 	  = net_cfg["dns_server"]
    		if "" ~= net_cfg["second_dns_server"] then
    			cfg_t["network"]["local_network"]["wan_dns2"] 	  = net_cfg["second_dns_server"]
			end
    	end

		-- NTP settings
		local timeZone = net_cfg["ntp_timezone"]
		if not timeZone or "" == timeZone then
		    timeZone = h.getPbxTimeZone()
		end
		if "" ~= net_cfg["ntp_server"] or "no" ~= net_cfg["use_dst"] then
			-- enable flag
			cfg_t["system"]["ntp"]["flag"] = "on"
        	-- NTP server
        	if "" ~= net_cfg["ntp_server"] then
        		cfg_t["system"]["ntp"]["server1"]["domain"] = net_cfg["ntp_server"]
        	end
        	-- time zone settings
            local td, errMsg = require"core.timedate":new(os.time(), timeZone)
            if td then
            	td:toTimeZone()
    			local tzoffset = td:getTzOffset()
    			if tzoffset and tzoffset.sign and tzoffset.hour and tzoffset.min then
                 	cfg_t["system"]["ntp"]["zone hour"]   = tzoffset.sign .. tzoffset.hour
            		cfg_t["system"]["ntp"]["zone minute"] = tzoffset.min
    			end
    			-- Daylight saving time
    			local dst = "no" ~= net_cfg["use_dst"] and td:getDstData(timeZone)
				if dst then
            		cfg_t["system"]["ntp"]["DaylightSavingTime"]["flag"]          = 1
            		cfg_t["system"]["ntp"]["DaylightSavingTime"]["save_time"]     = dst.offsetmin
            		cfg_t["system"]["ntp"]["DaylightSavingTime"]["start_month"]   = dst[1].month
            		cfg_t["system"]["ntp"]["DaylightSavingTime"]["start_day"]     = dst[1].day
            		cfg_t["system"]["ntp"]["DaylightSavingTime"]["start_hour"]    = dst[1].hour
            		cfg_t["system"]["ntp"]["DaylightSavingTime"]["start_minute"]  = dst[1].min
            		cfg_t["system"]["ntp"]["DaylightSavingTime"]["end_month"]     = dst[2].month
            		cfg_t["system"]["ntp"]["DaylightSavingTime"]["end_day"]       = dst[2].day
            		cfg_t["system"]["ntp"]["DaylightSavingTime"]["end_hour"]      = dst[2].hour
            		cfg_t["system"]["ntp"]["DaylightSavingTime"]["end_minute"]    = dst[2].min
            	end
            end
            if errMsg then
            	logger:debug(errMsg)
            end
		end

    	-- ToS settings
    	--XXX: wait parameters description from vendor
    	
    	-- Jitter buffer settings
    	--TODO: need checking
    	
    	-- Silence detection
    	--TODO: need checking
    	
    	-- End digit settings
    	--TODO: need checking
    	
    	-- Interdigit timeout
    	local inter_digit_time = h.get_interdigit_timeout()
    	if 3 > tonumber(inter_digit_time) then	--Timeout must be between 3 and 10 sec
    		inter_digit_time = 3
    	elseif 10 < tonumber(inter_digit_time) then
    		inter_digit_time = 10
    	end
    	cfg_t["fxs_fxo"]["dial_timeout"] = inter_digit_time

    	if "w02fxsr3" == model or "w04fxsr3" == model or "w24fxsr3" == model then
    		cfg_t["fxs_fxo"]["no_rtp_detect"]             = "disable" -- "enable" value drops hold (WMS7067, WMS4570)
    	end

   		if "w24fxsr2" == model then
    		for idx = 0, fxs_ports-1 do
    			cfg_t["fxs_fxo"]["lines"]["line" .. idx]["gain"]["rx_gain"] = "2"
    			cfg_t["fxs_fxo"]["lines"]["line" .. idx]["gain"]["tx_gain"] = "2"
    		end
    	end

----------------------------------------------------------------
------- Accounts and SIP settings
----------------------------------------------------------------
		-- Failover
        local pbxFailover     = nil
        local failoverSipPort = nil
        local failoverObj = require"common.failover":new()
        if failoverObj and "1" == failoverObj:isEnabled() then
            local failoverCfg = failoverObj:getConfig()
            pbxFailover = failoverCfg and failoverCfg.pbxFailoverDomain
            failoverSipPort = failoverCfg and failoverCfg.sipPort
            logger:debug(pbxFailover, "failoverCfg")
        end

		--Set SIP transport and SIP port
		local transport = "udp"
		local sip_port = "5060"
		local gtransp = nil
		local gsipport = nil
		if is_addr_remote then
			if ssl_supported then
				transport = "TLS"
				sip_port  = agent["alternate_port"] or "443"
			end
			gtransp = gconf.SIPTransportRemote
			gsipport = gconf.SIPServerPortRemote
		else
			if ssl_supported then
				transport = "TLS"
				sip_port = "5061"
			end
			gtransp = gconf.SIPTransport
			gsipport =  gconf.SIPServerPort
		end
		transport = gtransp or transport
		sip_port = gsipport or sip_port

        if "w02fxsr3" == model or "w04fxsr3" == model then
    		local no_srtp = "no" == gconf.SRTPForceUse
    		cfg_t["media"]["srtp_mode"] = no_srtp and "0" or "2" -- "srtp_mode" 0-disable, 1-auto, 2-forced (WMS8243)
    	end
        if pbxFailover then
            transport       = "TCP"
            sip_port        = agent["alternate_port"] or "443"
            failoverSipPort = failoverSipPort or sip_port
        end

		-- SIP server options
		cfg_t["sipserver"]["transport"]  = transport
		if pbxFailover then
			if "r3" == string.sub(model, -2) then
				-- support of 2 sip servers removed for new devices
				cfg_t["sipserver"]["outbound"]["domain"]     = registrar
				cfg_t["sipserver"]["outbound"]["port"]       = sip_port
				cfg_t["sipserver"]["outbound"]["sec_domain"] = pbxFailover
				cfg_t["sipserver"]["outbound"]["sec_port"]   = failoverSipPort
				-- due to strange reasons we need also settings below
				cfg_t["sipserver"]["server0"]["port"]        = sip_port
				cfg_t["sip"]["only_accept_calls_from_ACL"]   = "disable"
			else
				cfg_t["sipserver"]["server0"]["domain"] = registrar
				cfg_t["sipserver"]["server0"]["port"]   = sip_port
				cfg_t["sipserver"]["server1"]["domain"] = pbxFailover
				cfg_t["sipserver"]["server1"]["port"]   = failoverSipPort
			end
		else
			cfg_t["sipserver"]["server0"]["domain"] = registrar
			cfg_t["sipserver"]["server0"]["port"]   = sip_port
		end

		-- Set codec priority
		local netband   = h.is_wideband(agent["address"]) and "wideband" or "narrowband"
    	local codecs    = resource.codecs[netband]
    	local codec_idx = 0
    	for param, value in pairs(codecs) do
    		cfg_t["media"][param] = value
    		codec_idx = codec_idx + 1 
    	end
    	
    	-- set unused codecs to disable state
    	for idx = codec_idx, 15 do
    		cfg_t["media"]["codec" .. idx] = resource.codecs["dummy"]["disable"]
    	end
    	
    	-- set codec priority
    	if "wideband" == netband then
    		cfg_t["media"]["codecs_prefered"] = "1"		-- remote                                                                                                                                                                                      
    	else
    		cfg_t["media"]["codecs_prefered"] = "0"		-- local                                                                                                                                                                                                                                
    	end
		
		-- Processing account settings
		cfg_t["sipacc"] 	  = {}
		cfg_t["ports"]  	  = {}
		local unknown_idx = false
		for idx, data in ipairs(account_data) do
			local acc_idx        = idx - 1
			local cw             = callwait
			local secondSipAccId = nil
			if  "fax" == data.userKind then
				cw = ("enable" == callwait and "disable") or "enable"
			end
    		cfg_t["sipacc"]["sipacc" .. acc_idx] = {
    			isregister 	= "enable",
    			serverid 	= "0",
    			user_id  	= data.login,
    			auth_id  	= data.login,
    			auth_pwd 	= data.password,
    			display		= data.display,
    		}
    		if pbxFailover then
    		    secondSipAccId = fxs_ports + acc_idx
    		    cfg_t["sipacc"]["sipacc" .. secondSipAccId] = {
                isregister  = "enable",
                serverid    = "1",
                user_id     = data.login,
                auth_id     = data.login,
                auth_pwd    = data.password,
                display     = data.display,
            }
    		end
    		cfg_t["ports"]["port" .. acc_idx] = {
    			enable 			  = "enable",
    			primary_sipacc_id = acc_idx,
    			second_sipacc_id  = secondSipAccId or 65535,
    			auto_dial         = "1" == data.hotline.enable and data.hotline.number or "", -- hotline number
    			auto_dial_timeout = data.hotline.delay, -- hotline timeout
    			dnd				  = "disable",
    			callerid		  = callerid,
    			cfu				  = "",
    			cfb				  = "",
    			cfnry			  = "",
    			call_waiting	  = cw,
    			call_waiting_tone = cw,
    		}
		end
		
		-- set MWI Subscription
		if gconf.EnableMWISubscribe then
			cfg_t.sip.mwi = "enable"
		end

		--set MWI Subscription user list
		cfg_t.sip.voicemail = mviUserList
		
		-- set lines configuration 		
		--TODO: lines configuration

----------------------------------------------------------------
------- FXS feature codes settings
----------------------------------------------------------------
--		cfg_t["feature_code"] = {}	-- Disable all feature codes (it is used only before first provision)
--		for idx = 0, 42 do
--			cfg_t["feature_code"]["feature" .. idx] = {
--				["current"]	= "",
--				["default"]	= "",
--				["enabled"]    = "disable",
--				["usedefault"] = "disable"
--			}
--		end
	end
	
----------------------------------------------------------------
--- Result serialization
----------------------------------------------------------------
	local root_elm     = __create_xml_table(cfg_t, request["config"]) -- build XML table for configuration file
	
	local wxml_obj     = require"core.wxml":new()
	local base_xml_str = '<?xml version="1.0" encoding="UTF-8"?>'
	local xml_cfg 	   = wxml_obj:generate(root_elm, base_xml_str)
	
	local extra_lines = reboot_needed and "\n\n" or "\n"
	return xml_cfg and xml_cfg .. extra_lines
end

function setuser (...) end
function setfeatures (...) end
