module( ... or "wildixgsm", package.seeall)

hints = {
	--[[
	======================
	Wildix GSM
	----------------------
		HTTP/1.1 401 Unauthorized
		Server: GoAhead-Webs
		Date: Mon, 26 Jul 2020 05:00:00 GMT
		WWW-Authenticate: Basic realm="GoAhead"
		Pragma: no-cache
		Cache-Control: no-cache
		Content-Type: text/html
	]]
     {
         ['server'] = 'GoAhead-Webs'
        ,['www-authenticate'] = 'Basic realm="GoAhead"'
     }
    --[[
        ----------------------
        HTTP/1.1 401 Unauthorized
        Server: GoAhead-Webs
        Date: Mon, 26 Jul 2020 05:00:00 GMT
        WWW-Authenticate: Basic realm="Wildix"
        Pragma: no-cache
        Cache-Control: no-cache
        Content-Type: text/html
    --]]
    ,{
         ['server'] = 'GoAhead-Webs'
        ,['www-authenticate'] = 'Basic realm="Wildix"'
     }
}

-- Function for parsing firmware version
-- @return model_spec_version, major_version, build_version, version
-- TODO: use function from helper.lua
local function parse_firmware(version)
	local model_spec, major_version, build_version, minor_version = string.match(version, "(%d+)%.(%d+)%.(%d+)%.(%d+)")
	return model_spec, major_version, build_version, minor_version
end

function verify (host, port)
    local logger     = logger
	local http       = require"core.httpClient":new()
	local helper     = require"common.helper"
	local connection = require"common.core".connection
	
	local gpassword        = helper.get_gpassword()
	local is_ssl_negotiate = 443 == port.number
	local PORT_NUMBER      = is_ssl_negotiate and 443 or 80
	local t                = {["ssl"] = is_ssl_negotiate}
	local auth_table       = {"admin:admin", "root:root", "admin:" .. gpassword}
	
	local sql, cursor, token_pass, dublicate_ip
	if host.mac_addr_str == "" then
		sql = [[ SELECT token, count(id) FROM autoprov WHERE param=%q AND value LIKE '%%"address":%%"%s"%%' AND value LIKE '%%"model":%%"w01gsm"%%' ]]
		cursor = connection:execute(sql, "device", host.ip or "")
		token_pass, dublicate_ip = cursor:fetch(); cursor:close()
		if tonumber(dublicate_ip) > 1 then
			logger:error("Found %s duplicated address %q at provisioning database", dublicate_ip, host.ip or "")
			return t
		end
	else
   		sql = "SELECT token FROM autoprov WHERE param=%q AND mac=%q"
    	cursor = connection:execute(sql, "device", host.mac_addr_str)
    	token_pass = cursor:fetch(); cursor:close()
    end
    
    t["token_type"] = helper.is_new_token(token_pass)
    if t["token_type"] == "new" then
    	table.insert(auth_table, 1, "admin:"..token_pass)
    	t["token"] = token_pass
    end
    logger:debug(auth_table, "auth_table")
    
    local answer = nil
    for _, authpair in pairs(auth_table) do
        local login, pass = authpair:match("(.*):(.*)")
        answer = http:rawget(host.ip, PORT_NUMBER, "/enSysInfo.htm", { authUser = login, authPass = pass })
        if 200 == answer.code then
            t["auth"] = {user = login, pass = pass}
            break
        end
    end
    if not t["auth"] then
        logger:error("Wrong authorization for host %s", host.ip)
        return t
    end
	
	local fwversion = answer.body:match("td colspan=\"4\">Wildix W01GSM%s(.-)%s"):lower()
	local model = fwversion and "w01gsm" or ""
	if model == "" then
    	logger:error("Not found model for host %s", host.ip)
        return t
    end
	t["model"]             = model
	t["fwversion"]         = fwversion
	t["vendor:additional"] = "dwg"

	if host.mac_addr_str == "" then
		host.mac_addr_str = answer.body:match("MAC Address.-(9C.-)\r?\n"):gsub("-","")
	end
	t["mac_addr_str"] = host.mac_addr_str:lower()
	
	return t
end

function provision (host, port, ...)
	local args = select(1, ...)
	local http = require"core.httpClient":new()
	local h    = require"common.helper"
	local core = require"common.core"
	
	local connection = core.connection
	
	local http_protocol, http_port = "http", 80
	local PORT_NUMBER = host.ssl and 443 or 80
	
	-- processing registrar address
	local registrar       = nil
	local case            = "hosts"
    local cfg             = {
            FailoverByDns    = gconf.FailoverByDns
           ,ForceResolveName = gconf.ForceResolveName
           ,ForceResolveIp   = gconf.ForceResolveIp
          }
	local addr_t, err_msg = h.resolve_by_hosts(host.ip_src, nil, cfg)
	if not addr_t then
		if err_msg then	-- printing resolving error
			logger:warning("Hosts file was not processed: %s", err_msg)
		end
	else
		registrar = addr_t.domain
	end
	if not registrar then
		registrar, case, err_msg = h.get_registrar(host.ip_src, host.ip) 
	end
	logger:debug("Registrar address %q got from %s%s", registrar, case, err_msg and ". Errors: "..err_msg or "")
	
	local gpassword
	if host.token_type == "new" then
		gpassword = host.token or h.get_gpassword()
	else
		gpassword = h.get_gpassword()
		if host.token_type == "old" then
			logger:info("Token has old format")
		end
	end
	
	local provlink = string.gsub(
		"$protocol://$registrar:$port/?action=setaps&config=xml", "%$(%w+)",
		{
			protocol  = http_protocol,
			registrar = registrar,
			port 	  = http_port,
		}
	)

	local options = {
	     authUser = host.auth.user
	    ,authPass = host.auth.pass
		,timeout  = 2
	}
	
	-- Configuring provisioning
	local config_data = {"HintLanguage=1","DisplayLanguage=0","MainProvsServIP="..http.urlEncode(provlink),"SlaveProvsServIP=","Username=wmsdevice","Password="..gpassword,"IsEnableNTP=0","MainNTPServIP=","SlaveNTPServIP=","TimeZone=6","ok=Save"}
	local data        = table.concat(config_data, "&")
	local answer      = http:rawpost(host.ip, PORT_NUMBER, "/goform/SystemCfg", data, options) or {}	-- Save options
	logger:debug("Setting autoprovisioning parameters for W01GSM - %s", answer.code or 408)
	answer = http:rawpost(host.ip, PORT_NUMBER, "/goform/ProvisionStart", data, options) or {}	-- Starting provision
	logger:info("Send http request for autoprovision to W01GSM - %s", answer.code or 408)
end

function configuration (agent, ...)
	local json = require"dkjson"
	local md5  = require"md5"
	local core = require"common.core"
	local h    = require"common.helper"
	local os, logger = os, logger
    
    local wmsldap    = require"core.wmsldap"
	local connection = core.connection
	local username, err, sql, cursor = ""
	
	local resource = core.load_resource("wildixgsm")
	local result = h.tcopy(resource.parameters) or {}

	local request = select(1, ...) or {}
	local output = {}

	local default_lines, trunkname = 1
	local nports = tonumber(resource.models[agent["model"]])
	local tports = {}
	local tpbx   = {}
	
	local sip_port  = "5060"
	
	-- getting registrar address
	local registrar       = agent["registrar"]
	local registrar_ip 	  = false
    local cfg             = {
                FailoverByDns    = gconf.FailoverByDns
               ,ForceResolveName = gconf.ForceResolveName
               ,ForceResolveIp   = gconf.ForceResolveIp
          }
	local addr_t, err_msg = h.resolve_by_hosts(registrar, nil, cfg)
	if not addr_t then
		if err_msg then
			logger:warning("Hosts file was not processed: %s", err_msg)
		end
	else
		logger:debug(addr_t, "hosts_registrar")
		if addr_t.domain then
			registrar = addr_t.domain
		end
		if addr_t.ip then
			registrar_ip = addr_t.ip
		end
	end

	if request["config"] ~= "xml" then
		-- retrieve device information
        local token_pass = core.proceedDeviceInfo(agent).token

		-- get default context
		connection:switch ("wmsdb")
		sql = "select count(id) from dialplan_rulenames where rulename=%q"
		cursor = connection:execute(sql, "main")
		local count = cursor:fetch(); cursor:close()
		resource.port_default.port_context = tonumber(count) > 0 and "main" or "internalcalls"
		local is_upd_wmsdb  = false
		local needSipReload = false
		
		for i=1, nports do
			for k,v in pairs(resource.port_default) do
				if not tports[k] then tports[k] = {} end
				tports[k][i] = v
			end
		end

		-- check trunkid
		connection:switch("devdb")
		sql = "select a1.value, a2.value from autoprov as a1 left join autoprov as a2 on (a1.mac = a2.mac and a2.param=%q) where a1.param=%q and a1.mac=%q"
		cursor = connection:execute(sql, "settings", "trunkid", agent["mac"])
		local trunkid, settings_json = cursor:fetch(); cursor:close()
		if trunkid then	-- trunk already present in devdb
			-- get settings
			connection:switch ("wmsdb")
			sql = "select username, settings_mgw, settings_pbx from pstn_trunks where id = %s"
			cursor = connection:execute (sql, trunkid)
			local username, settings_mgw, settings_pbx = cursor:fetch(); cursor:close()

			trunkname = username
			local t = json.decode(settings_mgw)
			for k,v in pairs (t) do
				tports[k] = v
			end

			if not tports["dialtone"] then	-- Set dialtone
				tports["dialtone"] = h.get_dialtone("dinstar")
				is_upd_wmsdb = true
			end
			local t2 = json.decode(settings_pbx)
			for k,v in pairs(t2) do
				tpbx[k] = v
			end
            if not tpbx["0_0_secret"] then
                is_upd_wmsdb = true
            end
            tpbx["0_0_secret"] = h.get_sip_pass(trunkname, trunkid, 0, agent["mac"])
		else -- trunk autocreation
			local lowertech, tnames = "gsm", {}
			is_upd_wmsdb  = true
			needSipReload = true

			connection:switch ("wmsdb")
			sql = "select username from pstn_trunks where technology=%q"
			for name in connection:rows(sql, lowertech) do
				table.insert(tnames, name)
			end
			table.sort(tnames, function(a,b) a = tonumber(a:sub(4)); b = tonumber(b:sub(4));  return a<b end)
			
			for i=1,#tnames do
				if (lowertech .. tostring(i-1)) ~= tnames[i] then
					trunkname = lowertech .. tostring(i-1)
					break
				end
			end
			if not trunkname then
				trunkname = lowertech .. tostring(#tnames)
			end
			tpbx["0_0_secret"] = h.generate_pass()

			-- Update db for creating sip-extra.conf
			sql = [[ INSERT or REPLACE INTO pstn_trunks (id,username,ports,technology,settings_pbx,settings_mgw) VALUES(
					(select id from pstn_trunks where username = %q),%q,%d,%q,'%s','%s') ]]
			connection:execute(sql, trunkname, trunkname, nports, "gsm", json.encode(tpbx), json.encode(tports))

			sql = "SELECT id FROM pstn_trunks WHERE username=%q"
			cursor = connection:execute (sql, trunkname)
			trunkid = cursor:fetch(); cursor:close()

			connection:switch ("devdb")	-- set trunk id to devdb
			sql = [[ INSERT or REPLACE INTO autoprov (id,param,value,mac) VALUES(
					(select id from autoprov where param=%q and mac=%q),%q,'%s',%q) ]]
			connection:execute (sql, "trunkid", agent["mac"], "trunkid", trunkid, agent["mac"])

			-- update kamailio tables
			connection:switch ("htables")
			connection:execute ("delete from dynamic_peers where wms_id = %q", trunkid)	-- remove old trunk stuff
			sql = "insert into %s (key_name,key_value,wms_id) values (%q,%q,%q)"
			for i=1, nports do
				local username = string.format("%s_%d_%d",trunkname,i-1,0)
				local key = string.format("%s:%s", agent["address"], tostring (5059 + i))
				connection:execute (sql, "dynamic_peers", username .. "::authname", username, trunkid)
				connection:execute (sql, "dynamic_peers", username .. "::secret", tpbx["0_0_secret"], trunkid)
			end
			
			h.create_sip_extra()
			tports["dialtone"] = h.get_dialtone("dinstar") -- append dialtone after creation sip-extra
		end
		
		-- reload kamailio and sip to update callweaver peers
		if needSipReload then
			os.execute ("kamcmd htable.reload dynamic_peers; callweaver -rx \"sip reload \" > /dev/null")
		end
		
		if is_upd_wmsdb then
			connection:switch ("wmsdb")	-- update wmsdb
			sql = [[ INSERT or REPLACE INTO pstn_trunks (id,username,ports,technology,settings_pbx,settings_mgw) VALUES(
					(select id from pstn_trunks where username = %q),%q,%d,%q,'%s','%s') ]]
			connection:execute(sql, trunkname, trunkname, nports, "gsm", json.encode(tpbx), json.encode(tports))
		end

		local username = string.format ("%s_%d_%d", trunkname, 0, 0)
		local secret   = tpbx["0_0_secret"]

		-- Network settings
		local t = {}
		local update_db_settings = false
		if settings_json then
			t = json.decode(settings_json)
		end
		
		-- SYSLOG settings
		if t["syslog_server"] and t["syslog_server"] ~= "" then
			result["P810"] = "1"						-- Enable syslog
			result["P811"] = t["syslog_server"]	-- Syslog server address
		end
		
		-- set default DNS and NTP settings if not define
		t["dns_server"], t["second_dns_server"] = h.get_prim_sec_dns_servers(t["dns_server"], t["second_dns_server"])
		t["ntp_server"] 						= t["ntp_server"] or ""
		
		-- use static ip settings
		local is_addr_present = h.is_addr_present(agent["address"], agent["mac"])	-- check duplicate address
		local is_addr_local = h.isLocalIP(agent["address"])		-- device behind NAT can't user static ip
		if t["use_static_ip"] and (is_addr_present or not is_addr_local) then
			if is_addr_present then
				logger:warning("Disable static ip. Other device have the same ip address: %s", agent["address"] or "")
			end
			if not is_addr_local then
				logger:warning("Disable static ip. Device ip address (%s) is placed behind NAT", agent["address"] or "")
			end
			t["use_static_ip"] = nil
			update_db_settings = true
		end
		
		-- get settings for leased device by pbx dhcp
		local is_leased
		if not t["use_static_ip"] or t["dns_server"] == "" or t["ntp_server"] == "" then
			is_leased = h.is_addr_leased(agent["address"])
		end
		if not t["use_static_ip"] and not is_addr_present and is_leased then
			h.set_addr_static(agent["address"], agent["mac"])	-- mark this lease as static
			t["use_static_ip"] = 1
			update_db_settings = true
		end
		if is_leased then
    		if "" == t["dns_server"] or "" == t["second_dns_server"] then
	    		local dns_dhcp1, dns_dhcp2 = h.get_dns_servers()
        		if dns_dhcp1 and "" ~= dns_dhcp1 and "" == t["dns_server"] then
        			t["dns_server"] = dns_dhcp1             -- set primary DNS from DHCP config
        			update_db_settings = true
        		end
        		if dns_dhcp2 and "" ~= dns_dhcp2 and "" == t["second_dns_server"] then
        			t["second_dns_server"] = dns_dhcp2      -- set secondary DNS from DHCP config
        			update_db_settings = true
        		end
	    	end		
		    if t["ntp_server"] == "" then -- set NTP from DHCP config
		        local ntp_dhcp = h.get_ntp_servers()
		        if ntp_dhcp and ntp_dhcp ~= "" then
		            t["ntp_server"] = ntp_dhcp
		            update_db_settings = true
		        end
		    end
		end

		if t["use_static_ip"] == 1 then	-- use static ip
			local _, netmask, gateway = h.host_settings(agent["address"], registrar_ip)
			if netmask and gateway then
				result["P3000"] = "1"				-- Network mode (1 - static, 2 - DHCP)
				result["P3001"] = agent["address"]	-- Static ip
				result["P3002"] = netmask			-- Mask of the network
				result["P3003"] = gateway			-- Default gateway
				
				if registrar_ip then
    				if "" == t["dns_server"] then
    					t["dns_server"]    = registrar_ip
    					update_db_settings = true
    				end
    				if "" == t["second_dns_server"] then
    					t["second_dns_server"] = registrar_ip
    					update_db_settings     = true
    				end
    			end
								
				if t["ntp_server"] == "" then
					t["ntp_server"]    = registrar
					update_db_settings = true
				end
			else
				logger:warning("For current network configurations is not possible set static ip, disable it")
				t["use_static_ip"] = nil
				update_db_settings = true
			end
		end
		
		-- processing flag for disabling SIP firewall
		if "1" == tostring(t["disable_sip_firewall"]) then
			result["P90"] = "0"
		end
		
		-- update devdb with new settings
		if update_db_settings then
			sql = "INSERT OR REPLACE INTO autoprov (id, param, value, mac) VALUES ((SELECT id FROM autoprov WHERE param=%q AND mac=%q), %q, '%s', %q)"
			connection:execute (sql, "settings", agent["mac"], "settings", json.encode(t), agent["mac"])
		end
		
		-- DNS settings 
		if "" ~= t["dns_server"] then
			result["P3004"] = t["dns_server"]	-- Primary dns server
			if "" ~= t["second_dns_server"] then
				result["P3005"] = t["second_dns_server"]	-- Secondary dns server
			end
		end
		
		-- NTP settings 
		if t["ntp_server"] ~= "" then
			result["P3400"] = "1"              -- enable auto reboot
			result["P7"]    = "1"		       -- default at resources.wildixgsm - disable
			result["P8"]    = t["ntp_server"]  -- NTP server 1
			-- time zone offset    	
			local timeZone = t["ntp_timezone"]
			if not timeZone or "" == timeZone then
			    timeZone = h.getPbxTimeZone()
			end
			local td, errMsg = require"core.timedate":new(os.time(), timeZone)
			if td then
				td:toTimeZone()
				local tzoffset = td:getTzOffset()
				if tzoffset and tzoffset.sign and tzoffset.hour and tzoffset.min then
					
					local timezones = {['-1200']=0,['-1100']=1,['-1000']=2,['-0900']=3,['-0800']=4,['-0700']=5,['-0600']=6,['-0500']=7,['-0430']=8,['-0400']=8,
        				['-0330']=26,['-0300']=9,['-0230']=10,['-0200']=10,['-0100']=11,['0000']=12,['0100']=13,['0200']=14,['0300']=15,['0330']=16,
        				['0400']=16,['0430']=17,['0500']=17,['+0530']=27,['0545']=28,['0600']=18,['0630']=29,['0700']=19,['0800']=20,['0900']=21,
        				['0930']=30,['1000']=22,['1100']=23,['1200']=24,['1300']=25 }
        			result["P10"] = timezones[tzoffset.sign .. tzoffset.hour .. tzoffset.min]
				end
			end
			if errMsg then
            	logger:debug(errMsg)
            end
		else
			logger:warning("NTP server was not set. Auto reboot is disabled.")
		end
		
		---SIP Settings
		result["P100"] = registrar	-- SIP server URL
		result["P101"] = sip_port	-- SIP server PORT
		result["P102"] = registrar	-- SIP outbound proxy URL
		result["P103"] = sip_port	-- SIP outbound proxy PORT
		result["P150"] = username	-- SIP User ID
		result["P200"] = username	-- SIP Authenticate ID
		result["P250"] = secret 	-- SIP Authenticate pass

		---Provisioning settings
		local http_protocol, http_port = "http", (h.get_alternate_port() or 80)
		local provlink = string.gsub(
			"$protocol://$registrar:$port/?action=setaps&config=xml", "%$(%w+)",
			{
				protocol  = http_protocol,
				registrar = registrar,
				port 	  = http_port,
			}
		)
		result["P814"] = provlink							-- Primary Profile URL
		result["P816"] = "wmsdevice"						-- Provision Username
		result["P817"] = token_pass or h.get_gpassword()	-- Provision Password
		
		-- Password settings
		result["P3100"] = "admin"							-- Web login
		result["P3101"] = token_pass or h.get_gpassword()	-- Web password
		result["P3102"] = "admin"							-- Telnet username
		result["P3103"] = token_pass or h.get_gpassword()	-- Telnet password

		-- SMS API Config
		result["P1015"]  = registrar						-- SMS API Server
		result["P1013"]  = "wmsdevice"						-- SMS API user name
		result ["P1014"] = token_pass or h.get_gpassword()	-- SMS API password

		-- Codecs Priority
		local netband_str = h.is_wideband (agent["address"]) and "wideband" or "narrowband"
		if netband_str == 'wideband' then
			result ["P2218"] = 0 -- Local Priority off
		else
			result ["P2218"] = 1 -- Local Priority on
		end
		local codecs = resource.codecs[netband_str]
		for key, value in pairs (codecs) do
			result[key] = value
		end

		-- Prepare settings from trunk page config
		--    {"country_code":"380",
		-- -- "busy_response":"503",
		-- -- "sms_delivery_tries":"5",
		-- -- "dtmf_debounce":"80",
		-- -- "voip_input_gain":"11",
		-- -- "voip_output_gain":"9",
		-- -- "port_pricelist":["0"],
		-- -- "port_answer_type":["1"],
		-- -- "port_destination_code":[""],
		-- -- "port_context":["main"],
		-- -- "port_sim_pin":[""],
		-- -- "port_number":["380930802790"],
		-- -- "port_sms_read_enable":["0"],
		-- -- "port_sms_read":["300"],
		-- -- "port_email":[""],
		-- -- "port_initatcmd":[""],
		-- -- "port_codec_input_gain":["6"],
		-- -- "port_codec_output_gain":["6"]}'

		local txGainRangeCheck = tports["voip_output_gain"] or "0"
		local rxGainRangeCheck = tports["voip_input_gain"] or "6"
		if tonumber(txGainRangeCheck) and tonumber(txGainRangeCheck) > 10  then
			txGainRangeCheck = 10
		end
		if tonumber(txGainRangeCheck) and tonumber(rxGainRangeCheck) > 10  then
			rxGainRangeCheck = 10
		end
		result["P1050"] = string.gsub(result["P1050"],"%$(%w+)",	-- parsed from resources.wildixgsm
			{
				portNumber= tports["port_number"][1],
				txGain = txGainRangeCheck,
				rxGain = rxGainRangeCheck
			})

		-- Values range fix
		if tonumber(tports["port_codec_output_gain"][1]) and tonumber(tports["port_codec_output_gain"][1]) > 6  then
			tports["port_codec_output_gain"][1] = 6
		else
			tports["port_codec_output_gain"][1] =  -6 + (2 * tonumber(tports["port_codec_output_gain"][1]))
		end
		
		if tonumber(tports["port_codec_input_gain"][1]) and tonumber(tports["port_codec_input_gain"][1]) > 6 then
			tports["port_codec_input_gain"][1] = 6
		else
			tports["port_codec_input_gain"][1] =  -6 + (2 * tonumber(tports["port_codec_input_gain"][1]))
		end
		
		result["P300"] = tports["port_codec_output_gain"][1]	-- Tx Gain
		result["P350"] = tports["port_codec_input_gain"][1]		-- Rx Gain
		result["P400"] = tports["port_number"][1]				-- VOIP Hotline

		if tports['dialtone'] then
			local dialtones = resource.dialtones
			result["P52"] = dialtones[tports['dialtone']]
		end
		
		if tports["port_smsc"] and tports["port_smsc"][1] then
			result["P3600"] = tports["port_smsc"][1]  -- set SMSC
		end

		local config_data = ""
		for key, value in pairs(result) do
			config_data = config_data .. key .. "=" .. value .. ";"
		end
		output = config_data
	else
		local model_spec, major_version, build_version, minor_version = parse_firmware(agent["fw:version"])
		local firmwares = core.get_firmware_pack({["vendor"] = "wildixgsm"});
		local firmware_file = firmwares[agent["model"]]["application"]
		-- fix for correct upgrade without loop reboot
		if firmwares then
			if agent["fw:version"] == firmwares[agent["model"]]["version"] then
				if model_spec and major_version and minor_version and tonumber(model_spec..major_version..minor_version) <= 272005 then
					firmware_file = 'none.ldf'
				end
			else
				logger:info("Firmware will be upgrade to version %s", firmwares[agent["model"]]["version"] or "")
			end
		end

		output[#output+1] = '<?xml version="1.0" encoding="UTF-8" standalone="no"?>'
		output[#output+1] = "   <Provisioning>"
		output[#output+1] = "       <FirmWare>"
		output[#output+1] = "<SipV7>firmwares/wildixgsm/".. firmware_file .."</SipV7>"
		output[#output+1] = "           <SipV7></SipV7>"
		output[#output+1] = "       </FirmWare>"
		output[#output+1] = "       <ConfigFile>"
		output[#output+1] = "           <Favorite></Favorite>"
		output[#output+1] = "           <General>/?action=setaps</General>"
		output[#output+1] = "       </ConfigFile>"
		output[#output+1] = "   </Provisioning>"
	end
	return output, username, err
end

function setuser (...) end

function setfeatures (...) end
