module( ... or "wildixgw", package.seeall)

hints = {
	--[[
	======================
	Mediatrix
	----------------------
	4108:
		HTTP/1.1 302 Moved Temporarily
		Date: Mon, 25 May 2009 15:38:23 GMT
		Server: Mbedthis-Appweb/2.4.0
		Cache-Control: no-cache
		ETag: 68d-93-1
		Content-length: 5
		Connection: keep-alive
		Keep-Alive: timeout=60, max=100
		Location: http://172.19.4.209/system_info.esp
	]]
	{['server'] = 'Mbedthis.*Appweb.*'},
	{['location'] = 'login.esp'}
}

function verify (host, port)
    local xml        = require"core.htmlparser"
	local http       = require"core.httpClient":new()
	local helper     = require"common.helper"
	local connection = require"common.core".connection
	
	local gpassword        = helper.get_gpassword()
	local is_ssl_negotiate = 443 == port.number
	local PORT_NUMBER      = is_ssl_negotiate and not gconf.HTTPForceUse and 443 or 80
	local t = {["ssl"]     = is_ssl_negotiate}
	local auth_table       = {"public:","admin:" .. gpassword}
	
	local sql, cursor, token_pass, dublicate_ip
	if host.mac_addr_str == "" then
		sql = [[ SELECT token, count(id) FROM autoprov WHERE param=%q AND value LIKE '%%"address":%%"%s"%%' AND value LIKE '%%"model":%%"MGW"%%' ]]
		cursor = connection:execute(sql, "device", host.ip or "")
		token_pass, dublicate_ip = cursor:fetch(); cursor:close()
		if tonumber(dublicate_ip) > 1 then
			logger:error("Found %s duplicated address %q at provisioning database", dublicate_ip, host.ip or "")
			return t
		end
	else
   		sql = "SELECT token FROM autoprov WHERE param=%q AND mac=%q"
    	cursor = connection:execute(sql, "device", host.mac_addr_str)
    	token_pass = cursor:fetch(); cursor:close()
    end
    
    t["token_type"] = helper.is_new_token(token_pass)
    if t["token_type"] == "new" then
    	table.insert(auth_table, 1, "admin:"..token_pass)
    	t["token"] = token_pass
    end
    logger:debug(auth_table, "auth_table")
	
	local answer = nil
	for _, entry in pairs(auth_table) do
		local login, passwd = entry:match("(.*):(.*)")
		      answer        = http:rawpost(host.ip, PORT_NUMBER, "/login.esp?r=system_info.esp", {["username"] = login,["passwd"] = passwd }) or {}
		                    
        if 302 == answer.code and answer.headers["set-cookie"] then
            local cookie = answer.headers["set-cookie"] 
            answer       = http:rawget(host.ip, PORT_NUMBER, '/system_info.esp', { headers = { ['Cookie'] = cookie }}) or {}
            if 200 == answer.code and answer.body:match("(%w+)") then
                t["auth"]   = {user = login, pass = passwd}
                t["cookie"] = cookie
                break
            end
        end
	end
	if not t["auth"] then
		logger:error("Wrong authorization for host %s", host.ip)
		return t
	end

	-- WMS-5803
	local function __remove_tag(txt)
		return string.match(txt or "", "<td>(.-)</td>")
	end

	local doc = xml:new(answer.body)
	if host.mac_addr_str == "" then
		host.mac_addr_str = __remove_tag(doc:find('//tr[@class="Row1"]/td[3 and contains(text(),"0090f8")]')):match("(%w+)"):lower()
	end
	if host.mac_addr_str then
		t["mac_addr_str"] = host.mac_addr_str:lower()
	end

	-- fix model
	local model = __remove_tag(doc:find(string.format('//tr[@class="%s"]/td[3 and contains(text(),"W")]',
														string.find(answer.body, "AltRow") and "AltRow" or "Row0")))
	local is_pri = model:find("Wildix 3000")
	if is_pri then
		local str = __remove_tag(doc:find('//table[@class="MainBorder"]/tr[@class="Row1"]/td[2 and contains(text(),"33")]'))
		if (str or ""):find("Wildix%s330(%d)%-") then
			model = "W2PRI - 3302"	-- convert to w1pri or w2pri
		end
	end
	if model:find("Wildix%s(%d+)%splus") then	-- fix for "plus" models
		model = model:match("Wildix%s(%d+)%splus")
	end
	t["model"] = model or ""

	-- firmware version
	local firmware = __remove_tag(doc:find(string.format('//tr[@class="%s"]/td[3 and contains(text(),"Dgw")]',
														string.find(answer.body, "AltRow") and "AltRow" or "Row1"))):match("Dgw%s(.+)")
	t["fwversion"] = firmware
	
	return t
end

function provision (host, port, ...)
	local args = select(1, ...)
	local http = require"core.httpClient":new()
	local h    = require"common.helper"
	local core = require"common.core"
	local connection = core.connection
	
	local http_protocol, http_port = "http", 80
	local PORT_NUMBER = host.ssl and not gconf.HTTPForceUse and 443 or 80
	
	-- processing registrar address
	local registrar       = nil
	local case            = "hosts"
    local cfg             = {
                FailoverByDns    = gconf.FailoverByDns
               ,ForceResolveName = gconf.ForceResolveName
               ,ForceResolveIp   = gconf.ForceResolveIp
          }
	local addr_t, err_msg = h.resolve_by_hosts(host.ip_src, nil, cfg)
	if not addr_t then
		if err_msg then	-- printing resolving error
			logger:warning("Hosts file was not processed: %s", err_msg)
		end
	else
		registrar = addr_t.domain
	end
	if not registrar then
		registrar, case, err_msg = h.get_registrar(host.ip_src, host.ip) 
	end
	logger:debug("Registrar address %q got from %s%s", registrar, case, err_msg and ". Errors: "..err_msg or "")
	
	local gpassword = nil
	if host.token_type == "new" then
		gpassword   = host.token or h.get_gpassword()
	else
		gpassword   = h.get_gpassword()
		if host.token_type == "old" then
			logger:info("Token has old format")
		end
	end
	
	-- check firmware version 29.531+
	-- get firmware version number
	local fw_ver = h.parse_firmware(host.fwversion, "2" ~= string.sub(host.fwversion or "", 1, 1) and "v_sfu")  -- v_sfu = version single file upgrade (WMS-6966)
	if fw_ver and 29531 <= fw_ver.sht_v2 then
	   -- indicate that FW version 29.531+
	   logger:debug("Gateway %q has firmware version 29.531+", host.model)
	end
	
	-- set provision link for old versions
	local provlink = nil
	if fw_ver and fw_ver.sht_v2 < 29531 then
    	--TODO: remove after 6 mount (from 20.05.13) hack for Mediatrix with empty useragent
    	local fix_model = host.model:match("W%d+BRI") or host.model:match("W%d+PRI") or host.model:match("W%d+FXS")
    	if host.model:match("Wildix%s+3000") then -- remove space from provlink
    		fix_model = "Wildix3000"
    	end
    	if fix_model then
    		provlink = "?action=setaps&useragent=Wildix_"..fix_model.."_"..host.fwversion.."_"..host.mac_addr_str
        else
            provlink = "?action=setaps"
        end
	end
	
	local provProto = not gconf.HTTPForceUse and host.ssl and 200 or 100 -- 100 - http, 200 - https
	local provPort  = gconf.HTTPForceUse and "80" or "443"
	if fw_ver and 32599 > fw_ver.sht_v2 then
		provProto = 100
		provPort  = "80"
	end

	local options = {
		headers	= {
		   ['Cookie'] = host.cookie
		  ,['Connection'] = 'keep-alive' 
	    }
		,timeout  = 2000
	}
    
    -- build and request for autoprovision 	
    local reqData = nil
    local reqURL  = nil
	if fw_ver and fw_ver.sht_v2 < 29531 then
	    reqURL  = "/mgmt_scripts.esp?"
	    reqData = {
            ['bPasswordChanged']                  = 'true',
            ['bSecretKeyChanged']                 = 'false',
            ['bExportSecretKeyChanged']           = 'false',
            ['bUploadAndRunScript']               = 'false',
            ['bDoRunScript']                      = 'true',
            ['bDoExportScript']                   = 'false',
            ['bDoDownloadScript']                 = 'false',
            ['szRandom']                          = '',
            ['scriptExportContent']               = '200',
            ['scriptExportServiceName']           = 'All',
            ['scriptExportUrl']                   = '',
            ['scriptExportSecretKey']             = '',
            ['scriptGenerericFileName']           = '',
            ['localGenericFiles']                 = '---',
            ['scriptSpecificFileName']            = provlink,
            ['localSpecificFiles']                = '---',
            ['scriptTransferProtocol']            = provProto,
            ['scriptTransferServerHostname']      = registrar .. ':' ..provPort,
            ['scriptLocation']                    = '',
            ['scriptTransferUsername']            = 'wmsdevice',
            ['scriptTransferPassword']            = gpassword,
            ['scriptSecretKey']                   = '',
            ['scriptsAllowRepeatedExecution']     = '0',
            ['scriptAutoTransferOnRestartEnable'] = '1',
            ['scriptAutoTransferPeriodicEnable']  = '0',
            ['scriptsDhcpDownloadEnable']         = '0',
	    }
    elseif fw_ver and fw_ver.sht_v2 <= 431264 then
        reqURL  = "/mgmt_scripts.esp"
        reqData = {
            ['bPasswordChanged']                  = 'true',
            ['bSecretKeyChanged']                 = 'false',
            ['bExportSecretKeyChanged']           = 'false',
            ['bUploadAndRunScript']               = 'false',
            ['bDoRunScript']                      = 'true',
            ['bDoExportScript']                   = 'false',
            ['bDoDownloadScript']                 = 'false',
            ['szRandom']                          = '',
            ['ScriptExportContent']               = '200',
            ['ScriptExportServiceName']           = 'All',
            ['ScriptExportUrl']                   = '',
            ['ScriptExportSecretKey']             = '',
            ['ScriptGenericFileName']             = '',
            ['localGenericFiles']                 = '---',
            ['ScriptSpecificFileName']            = '?action=setaps',
            ['localSpecificFiles']                = '---',
            ['ScriptsTransferProtocol']           = provProto,
            ['ScriptsTransferSrvHostname']        = registrar .. ':' ..provPort,
            ['ScriptsLocation']                   = '',
            ['ScriptsTransferUsername']           = 'wmsdevice',
            ['ScriptsTransferPassword']           = gpassword,
            ['ScriptsSecretKey']                  = '',
            ['ScriptsAllowRepeatedExecution']     = '0',
            ['ScriptsTransferOnRestartEnable']    = '1',
            ['ScriptsTransferPeriodicEnable']     = '0',
            ['ScriptsDhcpDownloadEnable']         = '0',
        }
    else
    	reqURL  = "/mgmt_scripts.esp"
        reqData = {
            ['bPasswordChanged']                  = 'true',
            ['bSecretKeyChanged']                 = 'false',
            ['bUploadAndRunScript']               = 'false',
            ['bDoRunScript']                      = 'true',
            ['bDoInlineScript']                   = 'false',
            ['ScriptGenericFileName']             = '',
            ['localGenericFiles']                 = '---',
            ['ScriptSpecificFileName']            = '?action=setaps',
            ['localSpecificFiles']                = '---',
            ['ScriptsTransferProtocol']           = provProto,
            ['ScriptsTransferSrvHostname']        = registrar .. ':' ..provPort,
            ['ScriptsLocation']                   = '',
            ['ScriptsTransferUsername']           = 'wmsdevice',
            ['ScriptsTransferPassword']           = gpassword,
            ['ScriptsSecretKey']                  = '',
            ['ScriptsAllowRepeatedExecution']     = '0',
            ['ScriptsTransferOnRestartEnable']    = '1',
            ['ScriptsTransferPeriodicEnable']     = '0',
            ['ScriptsDhcpDownloadEnable']         = '0',
        }
    end

    local answer = http:rawpost(host.ip, PORT_NUMBER, reqURL, reqData, options) or {}
	logger:info("Send http request to '%s' for autoprovision now - %s", host.model:upper(), answer.code or 408)
end

function configuration (agent, ...)
	local md5        = require"md5"
	local json       = require"dkjson"
	local core       = require"common.core"
	local h          = require"common.helper"
	local pbxutils   = require"core.pbxutils"
	local logger, os = logger, os
	local connection = core.connection
	local wmsldap    = require"core.wmsldap"
	local err, sql, cursor

	local secret    = {}
	local data 		= {}
	local resource 	= core.load_resource("wildixgw")

	-- getting registrar address
	local registrar       = agent["registrar"]
	local registrar_ip 	  = false
    local cfg             = {
                FailoverByDns    = gconf.FailoverByDns
               ,ForceResolveName = gconf.ForceResolveName
               ,ForceResolveIp   = gconf.ForceResolveIp
          }
	local addr_t, err_msg = h.resolve_by_hosts(registrar, nil, cfg)
	if not addr_t then
		if err_msg then
			logger:warning("Hosts file was not processed: %s", err_msg)
		end
	else
		logger:debug(addr_t, "hosts_registrar")
		if addr_t.domain then
			registrar = addr_t.domain
		end
		if addr_t.ip then
			registrar_ip = addr_t.ip
		end
	end
	
	-- get gateway technology
	local m, technology, nports = agent["model"]:upper(), "", 1
	for k, v in pairs(resource.models) do
		if v[m] ~= nil then
			technology, nports = k, tonumber(v[m])
			break
		end
	end

	local default_lines = 1
	local cfgModel = technology	
	if "G7" == agent["model:additional"] and cfgModel then
	    cfgModel = cfgModel.." ".. agent["model:additional"]
	end
	local content = resource.templates[cfgModel]
	
	agent["technology"] = technology  -- save technology to provisioning db

    -- retrieve device information
    local token_pass = core.proceedDeviceInfo(agent).token
	
	-- get firmware version number
	local fw_ver = h.parse_firmware(agent["fw:version"], "2" ~= string.sub(agent["fw:version"] or "", 1, 1) and "v_sfu")  -- v_sfu = version single file upgrade (WMS-6966)
    
    sql    = "SELECT value,token FROM autoprov WHERE param='%s' AND mac='%s'"
	cursor = connection:execute(sql, "upgrade", agent["mac"])
	local is_remote_upgrade = (cursor:fetch() or "") == "remupgrade"; cursor:close()
	if is_remote_upgrade then
		local fwResult, err = h.getRemoteFirmwaresInfo("MGW", technology)
		if not fwResult.data then
		    logger:warning("%s %s", err, fwResult.tag)
	    end
	    local remoteFwPack = fwResult.data
		if remoteFwPack and agent["fw:version"] ~= remoteFwPack["version"] then
			logger:info("Firmware for device %s %s will be changed from version %s to %s", agent["model"], agent["mac"], agent["fw:version"] or "unknown", remoteFwPack["version"])

			local provlink = "?action=setaps"
			local content = string.gsub(resource.templates["UPGRADE"], "%<(.-)>", {
				pack_location = string.format("%s/%s", h.getFirmwareUpgradeHost(), remoteFwPack["path"]),
				provlink      = provlink
			})

			-- set upgrade to os.time()
			sql = [[ INSERT OR REPLACE INTO autoprov (id,param,value,mac) VALUES(
					(SELECT id FROM autoprov WHERE param=%q AND mac=%q),%q,%q,%q) ]]
			connection:execute(sql, "upgrade", agent["mac"], "upgrade", os.time(), agent["mac"])
			return content, err
		end
	end
	-- remove upgrade entry
	connection:execute("DELETE FROM autoprov WHERE param=%q AND mac=%q","upgrade", agent["mac"])

	local tv, usernames, trunkname = {}, {}
	if technology == "FXS" then		-- generate accounts for FXS
		local res, person, unknown_user
		local escaped_passwd = ""
		sql = "select value, flag from autoprov where param=%q and mac=%q"
		for username, port in connection:rows(sql, "username", agent["mac"]) do
			if username:find("unknown") then
				unknown_user = username		-- save old unknown account for current configuration
			else
				res = wmsldap:get(h.get_extenNumber_str(), username)
				person = {}
				if res and res[1] and username then
					_, person = unpack(res[1])
					if person["wuserPasswordSip"] then
						escaped_passwd = person["wuserPasswordSip"]:gsub("\\", "\\\\")
					end
					-- load user features
					local user_features = person.uid and require"common.userFeatures.userFeatures":load(person.uid) or {}
					local hotline = user_features.HOTLINE and "1" == user_features.HOTLINE.status and user_features.HOTLINE.value or "" -- hotline number
					if port then
						tv["username"..(port-1)]     = username
						tv["password"..(port-1)]     = escaped_passwd
						tv["onhook"..(port-1)]       = hotline
						tv["FriendlyName"..(port-1)] = person["cn"] or username -- display name
						tv["FriendlyName"..(port-1)] = string.format("%s-%s", tv["FriendlyName" .. (port-1)], port) -- display name
					else
						tv["username"] = username
						tv["password"] = escaped_passwd
						tv["onhook"] = hotline
						tv["FriendlyName"] = person["cn"] or username -- display name
					end
				elseif
					not res or not res[1] then
					logger:error("ldap is not ok")
					return {}, ""
				end
			end
		end
		
		if not unknown_user then	-- get new unknown user
			local unknown_ports = {}	-- list port without account
			for i = 1, nports do
				if not tv["username"..(i-1)] then
					table.insert(unknown_ports, i)
				end
			end
			if unknown_ports[1] then	-- write to autoprov used unknown account
				unknown_user = core.register(agent["mac"], nil, unknown_ports) 
			end
		end

		for i = 1, nports do	-- set unknown users
			if not tv["username" .. (i-1)] then
				local passwd, err = h.pass_for_unknown(unknown_user)
				if err then
					logger:error("password generation error for unknown account:", err)
				end
				tv["username" .. (i-1)]     = unknown_user
				tv["password" .. (i-1)]     = passwd
				tv["onhook" .. (i-1)]       = "sip:99@" .. registrar
				tv["FriendlyName" .. (i-1)] = "port-" .. i -- display name
			end
		end
	else -- BRI/PRI
		-- Get default context
		connection:switch("wmsdb")
		sql = "select count(id) from dialplan_rulenames where rulename=%q"
		cursor = connection:execute(sql, "main")
		local count = cursor:fetch(); cursor:close()
		resource.port_default.port_context = tonumber(count) > 0 and "main" or "internalcalls"
		local is_upd_wmsdb = false
		local lowertech = technology:lower()

		-- port_link_establishment OnDemand for PRI
		if technology == "PRI" then
			resource.port_default.port_link_establishment = 100
		end
		
		local tports = {}
		for i=1, nports do
			for k,v in pairs(resource.port_default) do
				if not tports[k] then tports[k] = {} end
				tports[k][i] = v
			end
		end

		-- Check trunkid
		connection:switch("devdb")
		local sql = "select value from autoprov where param = %q and mac = %q"
		local cursor = connection:execute(sql, "trunkid", agent["mac"]) -- devdb
		local trunkid = cursor:fetch(); cursor:close()
		if trunkid then
			-- Get settings
			connection:switch("wmsdb")
			sql = "select username, settings_mgw from pstn_trunks where id = %s"
			cursor = connection:execute(sql, trunkid) -- wmsdb
			local username, settings_mgw = cursor:fetch(); cursor:close()
			trunkname = username

			-- we have a trunk
			local settings_mgw = json.decode(settings_mgw or {});
			for k,v in pairs(settings_mgw) do
				if k:find("^port_") then
					tports[k] = v
				else
					data[k] = v
				end
			end

			for i=1, nports do
				secret[i] = h.get_sip_pass(trunkname, trunkid, i-1, agent["mac"])
			end

			-- fixup to default calling TON/NPI for Italy 
			if data["country_code"] == "39" then
			    for key in pairs(tports.port_calling_ton) do
			        tports.port_calling_ton[key] = "Unknown"
        		end
        		for key in pairs(tports.port_calling_npi) do
        		    tports.port_calling_npi[key] = "Unknown"
        		end
        	end
			
			if not data["dialtone"] then	-- Set dialtone
				data["dialtone"] = h.get_dialtone("media5")
				is_upd_wmsdb = true
			end
			for i=1, nports do
				if tports["port_facility_services"][i] == "1" then
					tports["port_calling_name_delivery"][i] = "FacilityIe"
				end
			end
			for i=1, nports do
				if "500" == tports["port_signal_protocol"][i] then  -- signaling protocol QSIG (WMS-6967)
					tports["port_facility_services"][i] = "1"
					tports["port_connected_line_identification_presentation"][i] = "1"
					tports["port_connected_name_identification_presentation"][i] = "1"
					tports["port_connected_line_identification_restriction"][i] = "0"
				end
			end

		else -- Trunk autocreaton
			local tnames = {}
			is_upd_wmsdb = true

			connection:switch("wmsdb")
			sql = "select username from pstn_trunks where technology=%q"
			for name in connection:rows(sql, lowertech) do -- wmsdb
				table.insert(tnames, name)
			end
			table.sort(tnames, function(a,b) a = tonumber(a:sub(4)); b = tonumber(b:sub(4));  return a<b end)

			for i=1,#tnames do
				if (lowertech .. tostring(i-1)) ~= tnames[i] then
					trunkname = lowertech .. tostring(i-1)
					break
				end
			end
			if not trunkname then
				trunkname = lowertech .. tostring(#tnames)
			end

			if "39" == tostring(pbxutils.getDialplanGeneralSettings("GCOUNTRY_CODE")) then
				tports["preserve_orig_cid"] = "no"
			end
			
			-- Update db for creating sip-extra.conf
			sql = [[ insert or replace into pstn_trunks (id,username,ports,technology,settings_pbx,settings_mgw) values(
					(select id from pstn_trunks where username = %q),%q,%d,%q,%q,'%s') ]]
			connection:execute(sql, trunkname, trunkname, nports, lowertech, "{}", json.encode(tports)) -- wmsdb

			sql = "select id from pstn_trunks where username = %q"
			cursor = connection:execute(sql, trunkname) -- wmsdb
			local trunkid = cursor:fetch(); cursor:close()

			-- Set trunk id
			connection:switch("devdb")
			sql = [[ insert or replace into autoprov (id,param,value,mac) values(
					(select id from autoprov where param=%q and mac=%q),%q,'%s',%q) ]]
			connection:execute(sql, "trunkid", agent["mac"], "trunkid", trunkid, agent["mac"]) -- devdb

			connection:switch("htables")	-- Update kamailio tables
			connection:execute("delete from dynamic_peers where wms_id = %q", trunkid)	-- Remove old trunk stuff

			local username_short, key, username
			sql = "insert into %s (key_name,key_value,wms_id) values (%q,%q,%q)"
			for i=1, nports do
				username_short = string.format("%s_%d",trunkname,i-1)
				secret[i] = h.generate_pass()
				key = string.format("%s:%s", agent["address"], tostring(5059 + i))

				for j=1, default_lines do
					username = username_short .. "_" .. (j-1)
					connection:execute(sql, "dynamic_peers", username .. "::authname", username, trunkid)
					connection:execute(sql, "dynamic_peers", username .. "::secret", secret[i], trunkid)
				end
			end
			
			h.create_sip_extra()
			os.execute "sudo kamcmd htable.reload dynamic_peers"	-- reload kamailio peers
			os.execute "sudo callweaver -rx \"sip reload \" > /dev/null" --reload sip to update callweaver peers
			
			data["dialtone"] = h.get_dialtone("media5")		-- append dialtone after creation sip-extra
		end
		data["ports"] = tports

		if is_upd_wmsdb then
			tports["dialtone"] = data["dialtone"]
			connection:switch("wmsdb")	-- update wmsdb
			sql = [[ insert or replace into pstn_trunks (id,username,ports,technology,settings_pbx,settings_mgw) values(
					(select id from pstn_trunks where username = %q),%q,%d,%q,%q,'%s') ]]
			connection:execute(sql, trunkname, trunkname, nports, lowertech, "{}", json.encode(tports)) -- wmsdb
			tports["dialtone"] = nil -- for correct parsing ports settings
		end
	end

	-- Defines for firewall rules 
	local firewall_src_addr  = registrar
	local firewall_dest_addr = agent["address"]
	local RtpPortRange = 2*("BRI" == technology and 2*nports or 0) + 2*("PRI" == technology and 30*nports or 0)
	local t38PortRange = ("BRI" == technology and 2*nports or 0) + ("PRI" == technology and 30*nports or 0)
	if RtpPortRange < 1 then RtpPortRange = 1 end
	if t38PortRange < 1 then t38PortRange = 1 end
	
	-- Linkname value dependeds on technology and fw version
	local linkname = "eth1" -- > 2.0.10.182 (fxs, bri); eth5 - pri
	if technology == "PRI" then
		linkname = "eth5"
	end
		
	-- Get device settings
	local is_addr_remote 	 = not h.isLocalIP(agent["address"])
	local update_db_settings = false
	local tskip				 = {}
	local t 				 = {}
	
	connection:switch("devdb")
	sql    = "select value from autoprov where param=%q and mac=%q"
	cursor = connection:execute(sql, "settings", agent["mac"])
	local settings_json = cursor:fetch(); cursor:close()
	if settings_json then
		t = json.decode(settings_json)
	end
	
	-- SYSLOG server
	local syslog_server = ""
	if t["syslog_server"] and t["syslog_server"] ~= "" then
		syslog_server = t["syslog_server"]
	end
	
	-- SNMP access
	local snmpstate				  = "Disable"
	local snmp_access 			  = ""
	local firewall_snmp_dest_addr = ""
	if t["snmp_access"] and "" ~= t["snmp_access"] then
		snmpstate	= "Enable"
		snmp_access = t["snmp_access"]
		if agent["address"] and agent["address"] ~= "" then
			firewall_snmp_dest_addr = agent["address"]
		end
	end

	-- VLAN settings
	if t["vid_voice"] and t["vid_voice"] ~= "" then
		data["vid"] = t["vid_voice"]
		data["qos"] = t["qos_voice"]
	else -- if not found vid_voice than not send vlan configuration (for clear vlan at device need reset)
		tskip[#tskip+1] = "Eth%.Vlan";
		tskip[#tskip+1] = "Eth%.AddVlan";
		tskip[#tskip+1] = "Bni%.NetworkInterfaces[^\n]*%.LinkName";
	end
	
	-- set default DNS and NTP settings if not define
	t["dns_server"], t["second_dns_server"] = h.get_prim_sec_dns_servers(t["dns_server"], t["second_dns_server"])
	t["ntp_server"] 						= t["ntp_server"] or ""
	
	-- use static ip settings
	local is_addr_present = h.is_addr_present(agent["address"], agent["mac"])	-- check dublicate address
	if t["use_static_ip"] and (is_addr_present or is_addr_remote) then
		if is_addr_present then
			logger:warning("Disable static ip. Other device have the same ip address: %s", agent["address"] or "")
		end
		if is_addr_remote then	-- device behind NAT can't user static ip
			logger:warning("Disable static ip. Device ip address (%s) is placed behind NAT", agent["address"] or "")
		end
		t["use_static_ip"] = nil
		update_db_settings = true
	end
	
	-- get settings for leased device by pbx dhcp
	local is_leased
	if not t["use_static_ip"] or t["dns_server"] == "" or t["ntp_server"] == "" then
		is_leased = h.is_addr_leased(agent["address"])
	end
	if not t["use_static_ip"] and not is_addr_present and is_leased then
		h.set_addr_static(agent["address"], agent["mac"]) -- mark this lease as static
		t["use_static_ip"] = 1
		update_db_settings = true
	end
	if is_leased then
    	if "" == t["dns_server"] or "" == t["second_dns_server"] then
	    	local dns_dhcp1, dns_dhcp2 = h.get_dns_servers()
    		if dns_dhcp1 and "" ~= dns_dhcp1 and "" == t["dns_server"] then
    			t["dns_server"] = dns_dhcp1             -- set primary DNS from DHCP config
    			update_db_settings = true
    		end
    		if dns_dhcp2 and "" ~= dns_dhcp2 and "" == t["second_dns_server"] then
    			t["second_dns_server"] = dns_dhcp2      -- set secondary DNS from DHCP config
    			update_db_settings = true
    		end
	    end	
	    if t["ntp_server"] == "" then -- set NTP from DHCP config
	        local ntp_dhcp = h.get_ntp_servers()
	        if ntp_dhcp and ntp_dhcp ~= "" then
	            t["ntp_server"] = ntp_dhcp
	            update_db_settings = true
	        end
	    end
	end
	
	if t["use_static_ip"] == 1 then	-- use static ip
		local _, _, gateway, bitmask = h.host_settings(agent["address"], registrar_ip)
		if bitmask and gateway then
			data["bitmask"] = bitmask
			data["gateway"] = gateway
			
			if registrar_ip then
    			if "" == t["dns_server"] then
    				t["dns_server"]    = registrar_ip
    				update_db_settings = true
    			end
    			if "" == t["second_dns_server"] then
    				t["second_dns_server"] = registrar_ip
    				update_db_settings     = true
    			end
    		end
	
			if t["ntp_server"] == "" then
				t["ntp_server"]    = registrar
				update_db_settings = true
			end
		else
			logger:warning("For current network configurations is not possible set static ip, disable it")
			t["use_static_ip"] = nil
			update_db_settings = true
		end
	end
	
	-- set dialtone to FXS
	if technology == "FXS" then
		if not t["dialtone"] then
			t["dialtone"] = h.get_dialtone("media5")
			update_db_settings = true
		else
			data["dialtone"] = t["dialtone"]
		end
	end

	-- Failover
    local pbxFailover = nil
    local failoverSipPort = nil
    -- Failover supported by W01-02BRI and WXPRI gateways
    local deviceSupportFailover = "PRI" == technology or (h.in_table(agent["model"], {"w01bri","w02bri","4401","4402"})) 
    local failoverObj = require"common.failover":new()
    if deviceSupportFailover and failoverObj and "1" == failoverObj:isEnabled() then
        local failoverCfg = failoverObj:getConfig()
        pbxFailover     = failoverCfg and failoverCfg.pbxFailoverDomain
        failoverSipPort = failoverCfg and failoverCfg.sipPort
        logger:debug(failoverCfg, "failoverCfg")
    end

		-- TODO: add after fix from vendor problem with direct rtp (CONVENTION-1876)
--		security_rtp_mode = "SecureWithFallback"
	--Set SIP transport and SIP port
	local tos_data = h.get_sip_settings()
	local transport = "TLS"
	local sip_port = "5061"
	local gtransp = nil
	local gsipport = nil
	local security_rtp_mode  = "PRI" == technology and "secureWithFallback" or "Unsecure"
	if is_addr_remote then
		sip_port = agent["alternate_port"] or "443"
		gtransp = gconf.SIPTransportRemote
		gsipport = gconf.SIPServerPortRemote
	else
		gtransp = gconf.SIPTransport
		gsipport =  gconf.SIPServerPort
	end
	transport = gtransp or transport
	sip_port = gsipport or sip_port

	if "TLS" == transport and "" == t["ntp_server"] then
		local def_ntp_srv  = "pool.ntp.org"
		t["ntp_server"]    = def_ntp_srv
		update_db_settings = true
		logger:warning("Gateway will use TLS transport. NTP settings was not found. Set NTP server to %q!", def_ntp_srv)
	end
	
	-- update devdb with new settings
	if update_db_settings then
		local sql = "insert or replace into autoprov (id,param,value,mac) values((select id from autoprov where param=%q and mac=%q),%q,'%s',%q)"
		connection:execute(sql, "settings", agent["mac"], "settings", json.encode(t), agent["mac"])
	end
	connection:close()	-- close database connection
	
	-- if not found bitmask and gateway then not send network configuration
	if not data["bitmask"] and not data["gateway"] then
		if "" == t["dns_server"] then
			tskip[#tskip+1] = "Hoc%.StaticDnsServers%[Priority=1%]"
			t["dns_server"] = ""
		end
		if "" == t["second_dns_server"] then
			tskip[#tskip+1] = "Hoc%.StaticDnsServers%[Priority=2%]"
		end
		if t["ntp_server"] == "" then
			tskip[#tskip+1] = "Hoc%.SntpConfigSource"
			tskip[#tskip+1] = "Hoc%.StaticSntpServerHost"
		end
		tskip[#tskip+1] = "Hoc%.DefaultRouterConfigSource"
		tskip[#tskip+1] = "Hoc%.StaticDefaultRouter"
		tskip[#tskip+1] = "Bni%.NetworkInterfaces[^\n]*%.ConnectionType"
		tskip[#tskip+1] = "Bni%.NetworkInterfaces[^\n]*%.StaticIpAddr"
		tskip[#tskip+1] = "Bni%.NetworkInterfaces[^\n]*%.Activation"
		firewall_dest_addr = ""	-- set destination address <empty>
	end

	local timezone_posix = "WEST-1DWEST-2,M3.5.0/02:00:00,M10.5.0/03:00:00"
	local timezone     = t["ntp_timezone"]
	if not timezone or "" == timezone then
	    timezone = h.getPbxTimeZone()
	end
	local td           = require"core.timedate":new(os.time(), timezone)
	if td then
		local zs = td:getTZString()
		if zs and "" ~= zs then
			timezone_posix = zs
		end
	end

	-- DNS configuration source
	local dnsstate = "100" -- Automatic
	if "" ~= t["dns_server"] then
		dnsstate = "200" -- Static
	end
	
	-- QOS
	local tos = tos_data["tos"] or "184"	 			-- Voice QoS
	local tos_sip = tos_data["tos_sip"] or "184"		-- SIP QoS
	

	-- allow all firewall rule, if registrar is not domain name (devices not accept domain name in fw rules)
	local firewall_all_state = pbxutils.get_ipv4(registrar) and "Enable" or "Disable"

	-- jitter buffer
	local tjb = h.get_jitter_buffer(is_addr_remote);
	local jitter_template = "300"
	local jitter_min = tjb["min"]
	local jitter_nom = tjb["nom"]
	local jitter_max = tjb["max"]
	if "PRI G7" ~= cfgModel then
		jitter_max = tonumber(jitter_max) > 135 and "135" or jitter_max
	else
		jitter_template = "500"
	end

	if agent["fw:version"] ~= "REVALIDATE" then
		tskip[#tskip+1] = "Conf%.ConfiguredScriptsTransferAndRun"
	end

	if fw_ver and 22320 >= fw_ver.sht_v2 then
		tskip[#tskip+1] ="SipEp.GwEventHandling[^\n]*%.Reboot"
		tskip[#tskip+1] ="SipEp.GwEventHandling[^\n]*%.CheckSync"
	end
	
	if "FXS" == technology or (fw_ver and 22303 > fw_ver.sht_v2) then
		--ISUP
		tskip[#tskip+1] = "SipEp%.ReasonHeaderSupport"
	end

	-- Session Timers
	local tsestimers = h.getSessionTimer()
	local session_timers = "180"
	if next(tsestimers) then
		session_timers = tsestimers["expires"] or session_timers
		if tonumber(session_timers) < 180 then
			session_timers = "180"
		end
	end
	if not tsestimers["timers"] then
		tskip[#tskip+1] = "SipEp%.DefaultSessionTimerMinimumExpirationDelay"
		tskip[#tskip+1] = "SipEp%.DefaultSessionTimerMaximumExpirationDelay"
	end
	local is_session_enabled = tsestimers["timers"] ~= nil
	
	local firmware_loc  = nil 

	-- WMS-8517 new BRI C7
	local firmware_model = cfgModel
	if "C7" == agent["model:additional"] and cfgModel then
		firmware_model = cfgModel.." ".. agent["model:additional"]
	end

	local fwResult, err = h.getRemoteFirmwaresInfo("MGW", firmware_model)
    if not fwResult.data then
        logger:warning("%s %s", err, fwResult.tag)
    end
    local remoteFwPack = fwResult.data
	if remoteFwPack and agent["fw:version"] ~= remoteFwPack["version"] then
		logger:info("Firmware (%s) upgrade to %s %s", agent["fw:version"] or "", remoteFwPack["version"], fwResult.tag)
		firmware_loc = string.format("%s/%s", h.getFirmwareUpgradeHost(), remoteFwPack["path"])
	else
		tskip[#tskip+1] = "Fpu%.MfpUrl"
		tskip[#tskip+1] = "Fpu%.AutomaticRestartEnable"
		tskip[#tskip+1] = "Fpu%.Install"  
	end

	-- skip paramters
	for i=1, #tskip do
		content = string.gsub(content, "(" .. tskip[i] .. "[^\n]*)", "")
	end

	local is_zport_format = not h.in_table(agent["model"], {"w02fxs","w04fxs","4102","4104"})
	local gpassword = token_pass or h.get_gpassword()
	local interdigits_timeout = h.get_interdigit_timeout()
	
	-- wms access to get configuration
	local login       = "wmsdevice"
	local password    = token_pass or h.get_gpassword()
	local provlink    = "?action=setaps"
	local provProto = not gconf.HTTPForceUse and agent.ssl == "on" and 200 or 100 -- 100 - http, 200 - https
	local provPort  = gconf.HTTPForceUse and "80" or agent.alternate_port or agent["registrar:port"]
	if fw_ver and 32599 > fw_ver.sht_v2 then
		provProto = 100
		provPort  = "80"
	end

	-- Codecs right order for corresponding network band
	local is_wideband = h.is_wideband(agent["address"])
	local netband_str = is_wideband and "wideband" or "narrowband"
	local codecs_part = resource.codecs[netband_str]
	local preffered_codec = ""                                                                                                                                                                                                          
   	if is_wideband then                                                                                                                                                                                                                 
        preffered_codec = "AllCommonPeerPriority"                                                                                                                                                                                   
    else                                                                                                                                                                                                                                
        preffered_codec = "AllCommonLocalPriority"                                                                                                                                                                                  
    end
	local bearercap_g711_index = is_wideband and "1" or "2"
	local bearercap_g729_index = is_wideband and "2" or "1"

	-- Line Type settings for PRi gateways
	local line_type               = nil
	local max_active_calls        = nil
	local channel_range           = nil
	local PreferredEncodingScheme = nil
	local FallbackEncodingScheme  = nil
    if technology == "PRI" then
        line_type = data["line_type"] or "E1" -- System -> Hardware -> Line Type - E1 or T1
        if "E1" == line_type then
            max_active_calls  = "30"
        elseif "T1" == line_type then 
            max_active_calls  = "23"
            for port = 1, nports do
                data.ports.port_preferred_encoding_scheme[port] = "200" -- G.711 u-law
                data.ports.port_fallback_encoding_scheme[port] = "200"  -- G.711 u-law
            end
        end
        channel_range		= "1-" .. max_active_calls
    end

    local sipKeepAliveMethod = "None"
    if pbxFailover then
        sipKeepAliveMethod = "Ping"
        transport          = "UDP" == transport and "TCP" or transport
        if gsipport then  -- WMS-7271
            sip_port = gsipport
            failoverSipPort = gsipport
        else
            sip_port           = agent["alternate_port"] or "443"
            failoverSipPort    = failoverSipPort or sip_port
        end
        content            = string.gsub(content, "#opt_?", "")
    else
        content = string.gsub(content, "#opt_.-#opt", "")
    end
    local isQsigEnable = false
	if technology == "PRI" or technology == "BRI" then
		for i=1, nports do
			if "500" == data["ports"]["port_signal_protocol"][i] then  -- signaling protocol QSIG (WMS-6967)
				isQsigEnable = true
				break
			end
		end
	end
	-- WMS-7176, WMS-9023
	local use_srtp = true
	if "FXS" == technology or "no" == gconf.SRTPForceUse then
		use_srtp = false
	end

	-- substituded at once
	content = string.gsub(content, "%<(.-)>", {
		prov_srv				 = registrar .. ":" .. provPort,
		provProto                = provProto,
		username                 = trunkname,
		adminpass                = gpassword,
		login                    = login,
		password                 = password,
		ports                    = nports,
		lines                    = nports * default_lines,
		registrar                = registrar,
		registrar_port           = sip_port,
		udp_enable               = ("UDP" == transport and "Enable") or "Disable",
		tcp_enable               = ("TCP" == transport and "Enable") or "Disable",
		tls_enable               = ("TLS" == transport and "Enable") or "Disable",
		security_rtp_mode		 = security_rtp_mode,
		syslog_server			 = syslog_server,
		snmpstate				 = snmpstate,
		snmpaccess				 = snmp_access,
		fwsnmpdstaddr			 = firewall_snmp_dest_addr,
		siprefresh               = "60",
		sipexpire                = "180",
		ipaddress                = agent["address"],
		bitmask                  = data["bitmask"] or "",
		router                   = data["gateway"] or "",
		dns                      = data["gateway"] or "",
		linkname                 = linkname,
		fwsrcaddr                = firewall_src_addr,
		fwdstaddr                = firewall_dest_addr,
		fwDstEndRtpPort          = 10004 + RtpPortRange - 1,
		fwDstEndT38Port          = 6004 + t38PortRange - 1,
		fwallstate				 = firewall_all_state,
		clocksource              = "",
		dialtone                 = data["dialtone"] or "Italy1",
		nationalprefix           = "0",
		internationalprefix      = "00",
		dtmfrelay                = "3",
		no_modem_detection       = "1",
		fax_max_bitrate          = "33",
		timeout                  = "20000",
		complete_with_sharp      = "1",
		interdigits_timeout      = tonumber(interdigits_timeout) * 1000,
		ecm_mode                 = "",
		model                    = agent["model"],
		vlan_id                  = data["vid"],
		vlan_qos                 = data["qos"],
		pound_support            = data["pound_support"] or "0",
		detection_threshold      = "300",
		dtmfmode                 = "400",
		clock_reference_port     = data["clock_reference_port"] or "100",
		bearercap_g711           = data["bearercap_g711"] or "200",
		bearercap_g729           = data["bearercap_g729"] or "200",
		bearercap_g711_index     = bearercap_g711_index,
		bearercap_g729_index     = bearercap_g729_index,
		provlink                 = provlink,
		dnsstate                 = dnsstate,
		dnsserver                = t["dns_server"],
		secdnsserver             = t["second_dns_server"],
		ntpserver                = t["ntp_server"],
		timezone                 = timezone_posix,
		tos                      = tos,
		tos_sip                  = tos_sip,
		jitter_template          = jitter_template,
		jitter_min               = jitter_min,
		jitter_nom               = jitter_nom,
		jitter_max               = jitter_max,
		session_timers           = session_timers,
		preffered_codec          = preffered_codec,
		session_enabled          = is_session_enabled and "1" or "0",
		firmware_location		 = firmware_loc,
		firstport                = is_zport_format and "01" or "1",
		CODECS_PART              = codecs_part,
		line_type				 = line_type,
		max_active_calls		 = max_active_calls,
		channel_range			 = channel_range,
		pbxFailover              = pbxFailover,
		failoverSipPort          = failoverSipPort,
		sipKeepAliveMethod       = sipKeepAliveMethod,
		sip_privacy_headers      = isQsigEnable and "200" or "100",
		sigproperties_desthost   = isQsigEnable and registrar or [[""]],
		use_srtp                 = use_srtp and "1" or "0",
	})

	content = string.gsub(content, "#BEGIN_REPEAT:([^\#%s]*):\n?%s+([^\#]+)\#END_REPEAT\n",

		-- We'll get a block of template which should be repeated
		function (pattern, block)
			local variables = h.split(pattern, "|")
			local portinc = 5059
			local portinc_sec = portinc + (nports * default_lines)
			local portinc_lines = portinc + nports
			local counter = 0
			local index, expression = 1, pattern:match("${index:(.-)}")
			
			if expression then
				index = loadstring("return (" .. expression .. ")")()
			end

			local sblock, tblock = "", {}
			for port = 0, nports-1 do
				local t = {}
				local zport = tostring(port < 10 and "0" .. port or port) -- ${zport}
				local zport_plus = tostring((port+1) < 10 and "0" .. (port+1) or (port+1)) -- ${zport+1}

				for _, variable in ipairs(variables) do
					-- remove brackets ${...}
					local k = string.gsub(variable,"${(.-)}","%1")

					t[k] = string.gsub(variable, "%${(.-)}",
						{
							["port"]    = port,
							["zport"]   = is_zport_format and zport or port,
							["zport+1"] = is_zport_format and zport_plus or port+1,
							["port+1"]  = port+1,
							["port+2"]  = port+2,
							["port+n"]  = port+1+nports,   -- index started from 2, 3 etc for failover registration and authentication index
							["port+s"]  = port+1+nports*2, -- index started from 3, 5 etc for failover route index
							["fport"]   = portinc + port+1+nports*2, -- failover gateway port 5064 for PRI
							["fsport"]  = portinc + port+2+nports*2, -- failover gateway port 5065 for PRI
						}
					)
				end
				
				-- set cid number
				if data["ports"] then
					for pkey in pairs(data["ports"]) do
						local pvalue = data["ports"][pkey][port+1]
						if pkey == "port_correct_incoming_calling_numbers" then
							if data["country_code"] ~= "1" then
								if data["ports"]["port_calling_name_max_length"][port+1] == "0" then
									pvalue = tonumber(pvalue) ~= 1 and "" or "setcidnumber, setcidname2cidnum"
								else
									pvalue = tonumber(pvalue) ~= 1 and "" or "setcidnumber"
								end
							else
								pvalue = ""
							end
						end

						t[pkey] = pvalue
					end
				end

				if technology ~= "FXS" then
					for i=0, default_lines-1 do
						local suser = string.format("%s_%d_%d", trunkname, port, i)
						local vuser = string.format("%s%d_%d", "username", port, i)
						local vpassword = string.format("%s%d_%d", "password", port, i)

						t[vuser] = suser
						t[vpassword] = secret[port+1]

						if not h.in_table("sip-" .. suser, usernames) then
							table.insert(usernames, "sip-" .. suser)
						end
					end
				end

				if string.find(block, "${line}") ~= nil or
					string.find(block, "${portincrement}") ~= nil or
					string.find(block, "${secportincrement}") ~= nil then

					local to = {}
					for i=1, default_lines do
						counter = counter + 1

						if string.find(block, "${portincrement+lines}") ~= nil then
							portinc_lines = portinc_lines + 1
						end
						

						to[#to+1] = string.gsub(block, "%${(.-)}",
							{
								["line"]                = i-1,
								["index"]               = "${index+" .. tostring(counter) .. "}",
								["portincrement"]       = "${portincrement+" .. tostring(counter) .. "}",
								["secportincrement"]    = "${secportincrement+" .. tostring(counter) .. "}",
								["portincrement+lines"] = portinc_lines
							}
						)
					end

					block = table.concat(to)
				end

				t["index"] = index
				for i=1,5 do
					t["index+" .. tostring(i-1)] = index + i - 1
					t["portincrement+" .. tostring(i-1)] = portinc + i - 1
					t["secportincrement+" .. tostring(i-1)] = portinc_sec + i - 1
				end

				local g = block
				g = string.gsub(g, "%${(.-)}", t)
				g = string.gsub(g, "%<(.-)>", t)

				tblock[#tblock+1] = g
				sblock = table.concat(tblock)
				sblock = string.gsub(sblock, "%<(.-)>", tv)

				index = index + default_lines
				portinc = portinc + default_lines
				portinc_sec = portinc_sec + default_lines
			end
			return sblock
		end
	)
	-- post cleanup --
	if data["ports"] then
		-- data["ports"] port_connection_type 100 - PTP, 200 - PTMP
		for pkey in pairs(data["ports"]) do
			if pkey == "port_connection_type" then
				for i=1, nports do
					if data["ports"][pkey][i] == "100" then
						content = string.gsub(content, "Isdn%.SignalingChannel[^\n]*".. i .."%]%.TeiNegociation=200", "")
					end
				end
				break;
			end
		end
	end
	content = string.gsub(content, "\n(#[^\n]*)", "")		-- remove comments
	content = string.gsub(content, "(\r?\n)%s*\r?\n", "%1")	-- remove empty lines
	return content, err
end

function setuser (...) end

function setfeatures (...) end
