--- Provision module for new Wp480 phone
-- 
-- @file	wildixp2.lua 
-- @author	Vyacheslav Khudan
-- @created 2014/10/30

module( ... or "wildixp2", package.seeall)

hints = {
	--[[
	======================
	Wildix phone (WARN: answer is 200 OK by request to /)
	----------------------
		HTTP/1.1 200 OK.
		Content-Type: text/html.
		Accept-Ranges: bytes.
		ETag: "-1173437454".
		Last-Modified: Sat, 11 Oct 2014 13:10:04 GMT.
		Content-Length: 467.
		Connection: close.
		Date: Tue, 04 Nov 2014 16:46:48 GMT.
		Server: lighttpd/1.4.30-devel-10132M.
	--]]
    {
         ["server"] = "lighttpd/.*"
        ,["status"] = "200"
    }
}

--- Convert user keys to phone supported keys
-- @param user_keys		Table with user keys settings
-- @param agent			Device agent information table
-- @param user_group    	Decsribe user group ACL
-- @return phone keys table, otherwise false value
local function __keys_convert(user_keys, agent, user_group, registrar, prov_port, prov_protocol, custom_identities)
	if not user_keys or "table" ~= type(user_keys) or not agent or not agent["model"] then
		return false
	end
	
	-- check BLF property allowed by ACL (like 'can view calls')
    local function _isAllowByACL(blf_prop, blf_type, dst)
        local rv         = true
        local user_group = user_group or ""
        if "eye" == blf_prop then
            local dst_group = nil
            local licenseType = nil
            if "EXT" == blf_type and dst and "" ~= dst then
                local res = require"core.wmsldap":get(require"common.helper".get_extenNumber_str(), dst)
                if 0 ~= #res then
                    local _, person = unpack(res[1])
                    dst_group = person["owner"] or ""
                    licenseType = person["licenseType"] or ""
                end
            end
            if dst_group then
                rv = "yes" == require"common.permissions":check(user_group, "yesno", "SEE_CALLS", dst_group, licenseType)
            end
        end
        return rv
    end

	--- Try get extension label from LDAP and wmsdb
    -- @param label_type    Marker for getting label
    -- @param value         An extension value
    -- @param label         A current extension label
    -- @return label value
    local function get_label(label_t, value, label)
        if not label_t or "" == label_t then
            return ""
        end
        
        local label = label or ""
        if not value then
            return label
        end
        
        local cursor     = nil
        local wmsldap    = require"core.wmsldap"
        local connection = require"common.core".connection
        connection:switch("wmsdb")
        
        if "EXT" == label_t or "SD" == label_t or "VM" == label_t then
            local res = wmsldap:get(require"common.helper".get_extenNumber_str(), value)
            if 0 ~= #res then
                local _, person = unpack(res[1])
                label = person["cn"]
                label = label and string.format("%s %s", label, value) or label
            end
         elseif "GPICKUP" == label_t then
            local res = wmsldap:get("wcgrp", value, "wGroup")
            if 0 ~= #res then
                local _, group = unpack(res[1])
                label          = group["cn"]
            end
        elseif "TRUNK" == label_t then
            local technology, gw_id, gw_port = value:match("^(%a%a%a)(%d)_(%d)")
            if "bri" == technology or "pri" == technology and gw_id and gw_port then
                label = string.format("ISDN %d %d", gw_id, gw_port)
            elseif technology and gw_id and gw_port then
                label = string.format("%s %d %d", technology:upper(), gw_id, gw_port)
            else
                label = value
            end
        elseif "STIME" == label_t then
            cursor = connection:execute("SELECT timetable FROM timetable_main WHERE id='%s'", value)
            label  = cursor:fetch(); cursor:close()
        elseif "SWITCH" == label_t or "3SWITCH" == label_t then
            cursor = connection:execute("SELECT switchname FROM wmsswitches WHERE id='%s'", value)
            label  = cursor:fetch(); cursor:close()
        elseif "QCALL" == label_t then
            cursor = connection:execute("SELECT call_groupname FROM call_groups WHERE id='%s'", value)
            label  = cursor:fetch(); cursor:close()
        end
        connection:close()
        
        return label
    end
    
    --- Format key value by format
    -- @param value     A value of the key
    -- @param fmt       A formatting string
    -- @return formatted key value
    local function fmt_key(value, fmt)
        local value = value or ""
        if not fmt or "" == fmt then
            return ""
        end
        
        return fmt:format(value)
    end
	
	--- Build phone key table from user key
	-- @param user_key		An user key (like 'EXT 200 user_200')
	-- @return phone supported table
	local function build_pkey(user_key)
		if not user_key then
			return false
		end

		local phone_key    = {}
		local pkey_defines = {
			-- 10 - Speed Dial
			-- 14 - Account (used as default)
			-- 16 - BLF
			-- 22 - DTMF
			["EMPTY"]		= { ["type"] = "0",		["value"] = "", 			["extension"] = "", 		["mode"] = "0",	   ["line"] = "1"},
			["LINE"] 		= { ["type"] = "14",	["value"] = "%s", 			["extension"] = "", 		["mode"] = "0",    ["line"] = "1"},
			["EXT"]  		= { ["type"] = "16",	["value"] = "%s", 			["extension"] = "pickup*",	["mode"] = "0",    ["line"] = "1"},
			["SD"]	 		= { ["type"] = "10",	["value"] = "%s", 			["extension"] = "", 		["mode"] = "0",    ["line"] = "1"},
			["TRUNK"]		= { ["type"] = "16",	["value"] = "%s", 			["extension"] = "pickup*",	["mode"] = "0",    ["line"] = "1"},
			["DTMF"] 		= { ["type"] = "22",	["value"] = "%s", 			["extension"] = "", 		["mode"] = "0",    ["line"] = "1"},
			["STIME"]		= { ["type"] = "16",	["value"] = "stime*%s*1", 	["extension"] = "",			["mode"] = "0",    ["line"] = "1"},
			["SWITCH"]		= { ["type"] = "16",	["value"] = "switch*%s*1", 	["extension"] = "",			["mode"] = "0",    ["line"] = "1"},
			["3SWITCH"] 	= { ["type"] = "16",	["value"] = "3switch*%s*1", ["extension"] = "", 		["mode"] = "0",    ["line"] = "1"},
			["LOCALGROUP"]	= { ["type"] = "14",	["value"] = "", 			["extension"] = "",			["mode"] = "0",    ["line"] = "1"},
			["GPICKUP"]		= { ["type"] = "10",	["value"] = "gpickup*%s",	["extension"] = "",			["mode"] = "0",    ["line"] = "1"},
			["QCALL"]		= { ["type"] = "16",	["value"] = "cg_%s", 		["extension"] = "pickup*",	["mode"] = "0",    ["line"] = "1"},
			["PORBIT"]		= { ["type"] = "16",	["value"] = "%s",	 		["extension"] = "pickup*",	["mode"] = "0",    ["line"] = "1"},
            ["CONTACTCENTER"] = { ["type"] = "21",  ["value"] = string.format("%s://%s:%s/?action=getcallgroups", prov_protocol, registrar, prov_port), ["extension"] = "", ["mode"] = "0", ["line"] = "1"},
            ["VM"]          = { ["type"] = "26",    ["value"] = "vmaccess*%s",  ["extension"] = "", ["mode"] = "0", ["line"] = "1", ["xmlurl"] = string.format("%s://%s:%s/?action=getxmlbrowser&get=vmaccess&menu=main&user=", prov_protocol, registrar, prov_port) .. "%s"},
            ["IDENTITIES"]    = { ["type"] = "21",    ["value"] = string.format("%s://%s:%s/?action=getidentities", prov_protocol, registrar, prov_port), ["extension"] = "", ["mode"] = "0", ["line"] = "1"},
		}
		pkey_defines["QPICKUP"] = pkey_defines["QCALL"]	-- Not used keys and saved only for compatibility
		
		local split_key = require"common.helper".split(user_key, " ")
		local type       = split_key[1] or ""
		local value      = split_key[2] or ""
		local key_label  = ""
		local transfer   = nil
		local sound      = nil
		local eye        = nil
        local melody     = gconf.MelodyBLF
		local blfModes   = {
			 ["00"] = 0
			,["01"] = 1
			,["10"] = melody and 4 or 2
			,["11"] = melody and 4 or 3
		}
		if "QCALL" == type then
			transfer   = split_key[3] or 0
			sound      = split_key[4] or 0
			eye        = split_key[5] or 0
		else
			key_label  = split_key[3] or ""
			if "CONTACTCENTER" == type then
				key_label = "Contact center"
			elseif "IDENTITIES" == type then
				key_label = "Identities"
			elseif "LINE" == type and "" ~= value then
				if require"core.pbxutils".in_table(value, custom_identities) then
					if "" == key_label then
						key_label = value
					end
					value = value .. "."
				else
					value = ""
				end
			end
			transfer   = split_key[4] or 0
			sound      = split_key[5] or 0
			eye        = _isAllowByACL("eye", type, value) and split_key[6] or 0
		end
		local blfMode    = blfModes[tostring(sound..eye)] or 0
        if "VM" == type then  -- WMS-5592 limitation, only xmlbrowser
            blfMode = 1
        end

		-- restore label separator changed by wms interface
		local function _normalize_label(label)
			local label_types = {["EXT"] = true, ["PORBIT"] = true, ["SD"] = true, ["DTMF"] = true, ["LINE"] = true}
			return label_types[type] and string.gsub(label, "_", " ") or label
		end

		-- set phone key configuration
		if pkey_defines[type] then
    		phone_key["Type"]   = pkey_defines[type]["type"] or "14"
    		phone_key["Param1"] = fmt_key(value, pkey_defines[type]["value"]) or ""		-- value field
    		phone_key["Param2"] = pkey_defines[type]["line"] or "1"						-- user account number
    		phone_key["Param3"] = fmt_key(value, pkey_defines[type]["extension"]) or ""	-- extension field
    		phone_key["Param4"] = blfMode                                               -- blf mode (0 - originail blf LED,1 - dialog+led，2 - beep+led，3 - led+label+beep)
    		phone_key["Param5"] = fmt_key(value, pkey_defines[type]["xmlurl"]) or ""    -- WMS-5592: shared voicemail
    		phone_key["Param6"] = (0 == tonumber(transfer) and 2) or (1 == tonumber(transfer) and 1) -- blf direct transfer (1 - blind, 2 - attended)
			    		
    		if "" == key_label then
    			key_label    = get_label(type, value, phone_key["Param1"])
                local klt    = require"core.pbxutils".split(key_label, "%s+")
                local lblStr = {
                    (klt[#klt-3] or "") .. (klt[#klt-2] or "")      -- person name
--                    ,klt[#klt-1] or ""                            -- person surname
--                    ," " .. (klt[#klt] or "")                     -- person number
                }
                if "" == lblStr[1] then
                    lblStr = { key_label }                          -- full string
                end
                if "VM" == type then
                    if "vmaccess*" == lblStr[1]:sub(1, 9) then
                        lblStr = { value }
                    end
                    key_label = "VM " .. table.concat(lblStr, " ")
                else
    				key_label = require"xmlbrowser".fitString(lblStr, "BLF_LABEL", agent.model)
    			end
    		else
    			key_label = _normalize_label(key_label)
    		end
    		phone_key["Label"]  = key_label
    	end
    	
		return phone_key
	end
	
	local pkeys_limits = {
		["wp410r2v1"] = {
			["max_exp"]	 = 0,
			["BLFKeys"]  = {},
			["funcKeys"] = {}
		},
		["wp410r2v2"] = { --model with blf func keys
			["max_exp"]	 = 0,
			["BLFKeys"]  = {
				[1]  = "FUNC1",
				[2]  = "FUNC2",
				[3]  = "FUNC3",
				[4]  = "FUNC4"
			},
			["funcKeys"] = {
				[1] = "TRANSFER",
				[3] = "VOICEMESSAGE",
			}
		},	
		["wp410r3"] = {
			["max_exp"]	 = 0,
			["BLFKeys"]  = {
				[1]  = "FUNC1",
				[2]  = "FUNC2",
				[3]  = "FUNC3",
				[4]  = "FUNC4"
			},
			["funcKeys"] = {
				[1] = "TRANSFER",
				[3] = "VOICEMESSAGE",
			}
		},
		["wp480r2"] = {
			["max_exp"]	  = 0,
			["BLFKeys"] = {
				[1] = "LINEKEY1",
				[2] = "LINEKEY2",
				[3] = "LINEKEY3"
			},
			["funcKeys"] = {}
		},
		["wp480gr2"] = {
			["max_exp"]	= 0,
			["BLFKeys"] = {
    			[1]  = "LINEKEY1",
    			[2]  = "LINEKEY2",
    			[3]  = "LINEKEY3",
    			[4]  = "FUNC1",
    			[5]  = "FUNC2",
    			[6]  = "FUNC3",
    			[7]  = "FUNC4",
    			[8]  = "LINEKEY4",
    			[9]  = "LINEKEY5",
    			[10] = "LINEKEY6",
    			[11] = "FUNC5",
    			[12] = "FUNC6",
    			[13] = "FUNC7",
    			[14] = "FUNC8",
    			[15] = "LINEKEY7",
    			[16] = "LINEKEY8",
    			[17] = "LINEKEY09"
    		},
    		["funcKeys"] = {
    			[4] = "TRANSFER",
    			[5] = "REDIAL",
    			[6] = "VOICEMESSAGE",
    			[7] = "BOOK"
    		},
    	},
    	["wp480gr3"] = {
            ["max_exp"] = 0,
            ["BLFKeys"] = {
                [1]  = "LINEKEY1",
                [2]  = "LINEKEY2",
                [3]  = "LINEKEY3",
                [4]  = "LINEKEY4",
                [5]  = "LINEKEY5",
                [6]  = "LINEKEY6",
                [7]  = "LINEKEY7",
                [8]  = "LINEKEY8",
                [9]  = "LINEKEY09",
                [10] = "LINEKEY10",
                [11] = "LINEKEY11",
                [12] = "LINEKEY12",
                [13] = "LINEKEY13",
                [14] = "LINEKEY14",
                [15] = "LINEKEY15",
                [16] = "LINEKEY16"
            },
            ["funcKeys"] = {}
        },
		["wp490r2"] = {
			["max_exp"]	  = 20,
			["BLFKeys"] = {
    			[1]  = "LINEKEY1",
    			[2]  = "LINEKEY2",
    			[3]  = "LINEKEY3",
    			[4]  = "LINEKEY4",
    			[5]  = "LINEKEY5",
    			[6]  = "LINEKEY6",
    			[7]  = "LINEKEY7",
    			[8]  = "LINEKEY8",
    			[9]  = "DSSKEY01",
    			[10] = "DSSKEY02",
    			[11] = "DSSKEY03",
    			[12] = "DSSKEY04",
    			[13] = "DSSKEY05",
    			[14] = "DSSKEY06",
    			[15] = "DSSKEY07",
    			[16] = "DSSKEY08",
    			[17] = "DSSKEY09",
    			[18] = "DSSKEY10"
			},
			["funcKeys"] = {}
		},
		["wp490gr2"] = {
			["max_exp"]	  = 20,
			["BLFKeys"] = {
    			[1]  = "LINEKEY1",
    			[2]  = "LINEKEY2",
    			[3]  = "LINEKEY3",
    			[4]  = "LINEKEY4",
    			[5]  = "LINEKEY5",
    			[6]  = "LINEKEY6",
    			[7]  = "LINEKEY7",
    			[8]  = "LINEKEY8",
    			[9]  = "DSSKEY01",
    			[10] = "DSSKEY02",
    			[11] = "DSSKEY03",
    			[12] = "DSSKEY04",
    			[13] = "DSSKEY05",
    			[14] = "DSSKEY06",
    			[15] = "DSSKEY07",
    			[16] = "DSSKEY08",
    			[17] = "DSSKEY09",
    			[18] = "DSSKEY10"
			},
			["funcKeys"] = {}
		},
		["wp490gcr2"] = {
			["max_exp"]	= 20,
			["BLFKeys"] = {
    			[1] = "LINEKEY1",
    			[2] = "LINEKEY2",
    			[3] = "LINEKEY3",
    			[4] = "LINEKEY4",
    			[5] = "LINEKEY5",
    			[6] = "LINEKEY6",
    			[7] = "LINEKEY7",
    			[8] = "LINEKEY8"
			},
			["funcKeys"] = {}
		},
		["wp490gr3"] = {
            ["max_exp"]   = 40,
            ["BLFKeys"] = {
                [1]  = "LINEKEY1",
                [2]  = "LINEKEY2",
                [3]  = "LINEKEY3",
                [4]  = "LINEKEY4",
                [5]  = "LINEKEY5",
                [6]  = "LINEKEY6",
                [7]  = "LINEKEY7",
                [8]  = "LINEKEY8",
                [9]  = "LINEKEY09",
                [10] = "LINEKEY10",
                [11] = "LINEKEY11",
                [12] = "LINEKEY12",
                [13] = "LINEKEY13",
                [14] = "LINEKEY14",
                [15] = "LINEKEY15",
                [16] = "LINEKEY16",
                [17] = "LINEKEY17",
                [18] = "LINEKEY18",
                [19] = "LINEKEY19",
                [20] = "LINEKEY20",
                [21] = "DSSKEY01",
                [22] = "DSSKEY02",
                [23] = "DSSKEY03",
                [24] = "DSSKEY04",
                [25] = "DSSKEY05",
                [26] = "DSSKEY06",
                [27] = "DSSKEY07",
                [28] = "DSSKEY08",
                [29] = "DSSKEY09",
                [30] = "DSSKEY10"
            },
            ["funcKeys"] = {}
        },
        ["wp490gr4"] = {
            ["max_exp"]   = 40,
            ["BLFKeys"] = {
                [1]  = "LINEKEY1",
                [2]  = "LINEKEY2",
                [3]  = "LINEKEY3",
                [4]  = "LINEKEY4",
                [5]  = "LINEKEY5",
                [6]  = "LINEKEY6",
                [7]  = "LINEKEY7",
                [8]  = "LINEKEY8",
                [9]  = "LINEKEY09",
                [10] = "LINEKEY10",
                [11] = "LINEKEY11",
                [12] = "LINEKEY12",
                [13] = "LINEKEY13",
                [14] = "LINEKEY14",
                [15] = "LINEKEY15",
                [16] = "LINEKEY16",
                [17] = "LINEKEY17",
                [18] = "LINEKEY18",
                [19] = "LINEKEY19",
                [20] = "LINEKEY20"
            },
            ["funcKeys"] = {}
        },
        ["wp480gr4"] = {
            ["max_exp"] = 0,
            ["BLFKeys"] = {
                [1]  = "LINEKEY1",
                [2]  = "LINEKEY2",
                [3]  = "LINEKEY3",
                [4]  = "LINEKEY4",
                [5]  = "LINEKEY5",
                [6]  = "LINEKEY6",
                [7]  = "LINEKEY7",
                [8]  = "LINEKEY8",
                [9]  = "LINEKEY09",
                [10] = "LINEKEY10",
                [11] = "LINEKEY11",
                [12] = "LINEKEY12",
                [13] = "LINEKEY13",
                [14] = "LINEKEY14",
                [15] = "LINEKEY15",
                [16] = "LINEKEY16"
            },
            ["funcKeys"] = {}
        },
	}
	
	local model   = agent["model"]..(agent["model:additional"] or "")
	local mlimits = pkeys_limits[model]
	if not mlimits then
		logger:debug("Not found phone keys limits for %q", model)
		return false
	end
	
	local keys_result 	 = {}
	local lineNum = 0
	local Em ={}
	for expMod in pairs(user_keys["expKeys"]) do
		if expMod then
			Em[expMod] = expMod:match("^expMod(%d+)$")
		end
	end
	
	-- Set unused BLF keys as EMPTY. And fill phone BLF keys
	local isModelR3R4color = require"core.pbxutils".in_table(model, {"wp480gr3", "wp480gr4", "wp490gr3", "wp490gr4"})
	for BLFKey, value in ipairs(mlimits.BLFKeys) do
		local defaultLabel = ""
		local userkey = user_keys.stdKeys[BLFKey] or ""
		userkey = userkey:gsub("%s", "")
		if "LINE" == userkey then
			if "LINE  " ~= user_keys.stdKeys[BLFKey] then
				user_keys.stdKeys[BLFKey] = "LINE  "
			end
			lineNum = lineNum + 1
			defaultLabel = "Line_" .. lineNum -- to set line key label as Line_1, Line_2 etc.
		elseif userkey:find("LINE") then
			if not isModelR3R4color then
				user_keys.stdKeys[BLFKey] = "LINE  "
			end
		end
        if "VM" == string.sub(userkey, 1, 2) and not isModelR3R4color then -- WMS-5592 limitation: r3-4 phones only
            keys_result[value] = build_pkey("EMPTY")
        elseif "IDENTITIES" == string.sub(userkey, 1, 10) and not (isModelR3R4color or "wp410r3" == model) then
            keys_result[value] = build_pkey("EMPTY")
        else
            keys_result[value] = build_pkey((user_keys["stdKeys"][BLFKey] or "EMPTY")  .. defaultLabel)
        end
	end
	
	-- Set unused functional keys as EMPTY. And fill phone functional keys.
	for funcKey, value in pairs(mlimits.funcKeys) do
		keys_result[value] = build_pkey(user_keys["stdKeys"][funcKey] or "EMPTY")
	end
	
	-- Fill phone expansion module keys
	local expkeys_result = {}
	for i = 1, mlimits["max_exp"] do
		for _, exp_name in ipairs({"expMod1", "expMod2", "expMod3"}) do
			if not user_keys["expKeys"][exp_name] then
				user_keys["expKeys"][exp_name] = {}
			end
			if not user_keys["expKeys"][exp_name][i] then
				user_keys["expKeys"][exp_name][i] = "EMPTY"
			end
		end
	end
	for exp_name in pairs(user_keys["expKeys"]) do
		if exp_name and "" ~= exp_name then
			local idx = exp_name:match("^expMod(%d+)$")
    		expkeys_result["Em" .. idx] = {}
        	for i = 1, mlimits["max_exp"] do
        		local ek = 10 > i and "0" .. i or i
        		local defaultLabel = ""
        		local userkey = user_keys["expKeys"][exp_name][i] or ""
        		userkey = userkey:gsub("%s", "")
        		if "LINE" == userkey then
					if "LINE  " ~= user_keys["expKeys"][exp_name][i] then
						user_keys["expKeys"][exp_name][i] = "LINE  "
					end
					lineNum = lineNum + 1
					defaultLabel = "Line_" .. lineNum -- to set line key label as Line_1, Line_2 etc.
				end
				expkeys_result["Em" .. idx]["EXTKEY" .. ek] = build_pkey(user_keys["expKeys"][exp_name][i] .. defaultLabel)
        	end
        end
    end
	for _, idx in pairs(Em) do
		if idx then	expkeys_result["Em" .. idx]["SETTING"] = {["BackLightTime"] = "0"}	end 
	end

	return keys_result, expkeys_result
end

--- Verifying device HTML request and retreiving information
-- @param host		A verifying host data
-- @param port		Not used data
-- @return table with verifying device information
function verify (host, port)
	local logger     = logger
	local http       = require"core.httpClient":new()
	local xml        = require"core.htmlparser"
	local helper 	 = require"common.helper"
	local connection = require"common.core".connection
	
	local is_ssl_negotiate = 443 == port.number
	local PORT_NUMBER      = is_ssl_negotiate and not gconf.HTTPForceUse and 443 or 80
	
	local t          = {["ssl"] = is_ssl_negotiate}	
	local gpassword  = helper.get_gpassword()
	local auth_table = {"admin:admin", "admin:" .. gpassword}

	local sql			= nil
	local cursor		= nil
	local token_pass	= nil
	local dublicate_ip	= nil
	if "" == host.mac_addr_str then
		sql = [[ SELECT token, count(id) FROM autoprov WHERE param=%q AND value LIKE '%%"address":%%"%s"%%' AND value LIKE '%%"model":%%"wp%%r2"%%' ]]
		cursor = connection:execute(sql, "device", host.ip or "")
		token_pass, dublicate_ip = cursor:fetch(); cursor:close()
		if 1 < tonumber(dublicate_ip) then
			logger:error("Found %s duplicated address %q at provisioning database", dublicate_ip, host.ip or "")
			return t
		end
	else
   		sql = "SELECT token FROM autoprov WHERE param=%q AND mac=%q"
    	cursor = connection:execute(sql, "device", host.mac_addr_str)
    	token_pass = cursor:fetch(); cursor:close()
    end
    
    t["token_type"] = helper.is_new_token(token_pass)
    if "new" == t["token_type"] then
    	table.insert(auth_table, 1, "admin:"..token_pass)
    	t["token"] = token_pass
    end
    logger:debug(auth_table, "auth_table")
    
    -- Get device status information via POST requests
    local is410r3 = false
    local body  = nil
    for _, entry in pairs(auth_table) do
    	-- send request for getting encrypted password
    	local login, pass      = entry:match("(.*):(.*)")
    	local cookie_time      = os.time() .. "150"
    	local encrypt_pass_url = string.format("/fcgi/do?action=Encrypt&UserName=%s&Password=%s&_=%s", login, pass, cookie_time)
    	if is410r3 then
    		encrypt_pass_url = string.format("/cgi-bin/security.cgi?action=Encrypt&UserName=%s&Password=%s&_=%s", login, pass, cookie_time)
    	end
    	local answer = http:rawget(host.ip, PORT_NUMBER, encrypt_pass_url) or {}
        if 404 == answer.code then
            is410r3 = true
            encrypt_pass_url = string.format("/cgi-bin/security.cgi?action=Encrypt&UserName=%s&Password=%s&_=%s", login, pass, cookie_time)
            answer = http:rawget(host.ip, PORT_NUMBER, encrypt_pass_url) or {}
        end
		if 200 == answer.code and answer.body then
			local tmp_body 	  = xml:new(answer.body)
			local enrypt_pass = tmp_body:find('//input[@id="hcSingleResult"]/@value')
	    	
        	local cookie = string.format("UserName=%s; Password=%s", login, enrypt_pass)
            if is410r3 then
                answer = http:rawpost(host.ip, PORT_NUMBER, "/cgi-bin/security.cgi?id=1", {["Operation"] = "CreateSession"}, { headers = {['Cookie'] = cookie}}) or {}
            else
                answer = http:rawpost(host.ip, PORT_NUMBER, "/fcgi/do?id=1", {["Operation"] = "CreateSession"}, { headers = {['Cookie'] = cookie}}) or {}
            end
        	
        	if answer and 200 == answer.code and answer.body then
    			body = xml:new(answer.body)
    			
    			-- check if answer body has link for LogOut action
    			local can_logout = body:find('//a[@id="tPageLogOut"]')
    			if can_logout and can_logout:find("LogOut") then
    				t["auth"]   = {user = login, pass = pass}
    				t["cookie"] = cookie
    				break
    			end
    		end
    	end
	end
    
	if not body or not t["auth"] then
		logger:error("Wrong authorization for host %s", host.ip)
		return t
	end
	
	-- Get seesion id hcSessionIdNow
	local session_id = body:find('//input[@id="hcSessionIdNow"]/@value')
	t["cookie"] = t["cookie"] .. "; SessionId=" .. (session_id or "")
	
	-- Parsing device information
	if "" == host.mac_addr_str then
		local mac_str = body:find('//input[@id="hcMACAddress"]/@value')
		if mac_str then
			host.mac_addr_str = mac_str:gsub(":", "") or ""
		end
	end
    
	local model = body:find('//input[@id="hcProductName"]/@value')
	if model then
		model = model:lower()
	end
	
	t["mac_addr_str"] = host.mac_addr_str:lower()
	t["model"] 	   	  = model
	t["fwversion"]    = body:find('//input[@id="hcFirmwareVersion"]/@value')
	
	-- fix up model for WP490G
	if "wp490r2" == t["model"] and "59.145.5.205" == t["fwversion"] then
		t["model"] = "wp490gr2"
	end
	-- fix up model for WP410R2
	if "wp410" == t["model"] then
		t["model"] = "wp410r2"
	end
	-- fix up model for WP490GR4
	if "wp490gr3" == t["model"] and "68.145.1.70" == t["fwversion"] then
		t["model"] = "wp490gr4"
	end
	return t
end

--- Provision action to updating provision url on device and execting autoprovision
-- @param host	A provision device data
-- @param port	Not used 
function provision (host, port, ...)
	local h      = require"common.helper"
	local http   = require"core.httpClient":new()
	local logger = logger
	
	local prov_proto  = gconf.HTTPForceUse and "http" or "https" 	
	local prov_port   = gconf.HTTPForceUse and "80" or "443"			
	local PORT_NUMBER = host.ssl and not gconf.HTTPForceUse and 443 or 80
	
	-- processing registrar address
	local registrar       = nil
	local case            = "hosts"
	local cfg             = {
            FailoverByDns    = gconf.FailoverByDns
           ,ForceResolveName = gconf.ForceResolveName
           ,ForceResolveIp   = gconf.ForceResolveIp
          }
	local addr_t, err_msg = h.resolve_by_hosts(host.ip_src, nil, cfg)
	if not addr_t then
		if err_msg then	-- printing resolving error
			logger:warning("Hosts file was not processed: %s", err_msg)
		end
	else
		registrar = addr_t.domain
	end
	if not registrar then
		registrar, case, err_msg = h.get_registrar(host.ip_src, host.ip) 
	end
	logger:debug("Registrar address %q got from %s%s", registrar, case, err_msg and ". Errors: "..err_msg or "")

	-- Get security password
	local gpassword = nil
	if host.token_type == "new" then
		gpassword = host.token or h.get_gpassword()
	else
		gpassword = h.get_gpassword()
		if host.token_type == "old" then
			logger:info("Token has old format")
		end
	end
	
	-- Provisioning link
	local provlink = string.gsub(
			"$protocol://$registrar:$port/?action=setaps", "%$(%w+)",
			{
				registrar = registrar,
				protocol  = prov_proto,
				port 	  = prov_port,
			}
		)
	
	-- Set options for next http requests
	local options = {
    		headers  = { ['Cookie'] = host.cookie },
    		timeout = 2
    	}
	
	-- Set parameters for autoprovision
	local is410r3 = false
	local adv_upgr_url = "/fcgi/do?id=6&id=2"
	local simple_data  = string.format("begin&Operation=Submit&cManualUpdateURL=%s&cServerUserName=%s&cServerPwd=%s&SubmitData=end", provlink, "wmsdevice", gpassword)
	local data         = http.urlEncode(simple_data)
	local answer       = http:rawpost(host.ip, PORT_NUMBER, adv_upgr_url, data, options) or {}
	if 404 == answer.code then
		is410r3 = true
		adv_upgr_url = "/cgi-bin/upgrade.cgi?id=6&id=2"
		answer = http:rawpost(host.ip, PORT_NUMBER, adv_upgr_url, data, options) or {}
	end
	logger:debug("Set up autoprovisioning parameters for %s - %s", host.model, answer.code or 408)
	
	-- Run autoprovisioning
	data = http.urlEncode("begin&Operation=Submit&cAUTOPNow=true&SubmitData=end")
	answer = http:rawpost(host.ip, PORT_NUMBER, adv_upgr_url, data, options) or {}
	logger:info("Sending http request to %s for autoprovision now - %s", host.model, answer.code or 408)
	
	-- Logout action
	options.headers["Content-Type"] = nil
	options.headers["Cookie"]	   = host.cookie .. "; LogOutCookieFlag=1"
	if is410r3 then
		answer = http:rawget(host.ip, PORT_NUMBER, "/cgi-bin/security.cgi?id=1", options) or {}
	else
		answer = http:rawget(host.ip, PORT_NUMBER, "/fcgi/do?id=1", options) or {}
	end
	logger:debug("Pressing logout button for %s - %s", host.model, answer.code or 408)
end

--- Building and sending provision configuration
-- @param agent		A table with general device information
-- @param mixed		Some parameters. Currently not used		
function configuration (agent, ...)
	local core       = require"common.core"
	local h          = require"common.helper"
	local pbxutils   = require"core.pbxutils"
	local wmsldap    = require"core.wmsldap"
	local logger     = logger
	local connection = core.connection
	local json		 = require"dkjson"
	local wmsperm    = require"common.permissions"
    
	local isModelR3R4 = pbxutils.in_table(agent["model"], {"wp480gr3", "wp480gr4", "wp490gr3", "wp490gr4"})

	-- load default values from resource directory
	local resource = core.load_resource("wildixp2")
	local result   = resource.paramaters_init or {}
	
	-- initial variables
	local prov_proto     = agent["ssl"] == "on" and not gconf.HTTPForceUse and "https" or "http"
	local prov_port  	 = gconf.HTTPForceUse and "80" or h.get_alternate_port() or agent["registrar:port"]
	local is_addr_remote = not h.isLocalIP(agent["address"])
	
	-- getting registrar address
	local registrar       = agent["registrar"]
	local registrar_ip 	  = false
	local cfg             = {
    	       FailoverByDns    = gconf.FailoverByDns
    	      ,ForceResolveName = gconf.ForceResolveName
    	      ,ForceResolveIp   = gconf.ForceResolveIp
	      }
	local addr_t, err_msg = h.resolve_by_hosts(registrar, nil, cfg)
	if err_msg then
		logger:warning("Hosts file processing error: %s", err_msg)
	end
	if addr_t then
		logger:debug(addr_t, "hosts_registrar")
		if addr_t.domain then
			registrar = addr_t.domain
		end
		if addr_t.ip then
			registrar_ip = addr_t.ip
		end
	end
    
    -- retrieve device information
    local device_info = core.proceedDeviceInfo(agent)

-----------------------------------------------------
--- Reboot device if needed
-----------------------------------------------------
    if 1 == device_info.reboot then
        result["Autoprovision"]["Reboot"]["Enable"] = "1"
    end

-----------------------------------------------------
--- Provision settings
-----------------------------------------------------	
    local token_pass = device_info.token

	local provlink = string.gsub(
		"$protocol://$registrar:$port/?action=setaps", "%$(%w+)",
    	{
    		registrar = registrar,
    		protocol  = prov_proto,
    		port 	  = prov_port,
    	}
    )
    result["Autoprovision"]["GENERAL"]["Url"] 	   = provlink
    result["Autoprovision"]["GENERAL"]["Username"] = "wmsdevice"
    result["Autoprovision"]["GENERAL"]["Pwd"]	   = token_pass or h.get_gpassword()
    
-----------------------------------------------------
--- Firmware settings
-----------------------------------------------------	
	local firmwares 		= core.get_firmware_pack(agent)
	local need_fwupgrade	= false
	if firmwares then
		local fw_cfg = firmwares[agent["model"]]
		if fw_cfg then
			local fw_filename = fw_cfg["application"]
			local fw_dir  	  = "/var/www/firmwares/wildix2/"
			if fw_cfg["version"] ~= agent["fw:version"] then
				-- Firmware for Workforce must be upgraded in 2 steps if device's firmware is lower than 65.145.4.1
				-- In such cases we have to upgrade device's firmware to 65.145.4.1 at first step and to the actual firmware at second step
				local fix_proto = false
				if agent["model"] == "wp480gr4" then
					local _, _, build_ver, minor_ver = agent["fw:version"]:match("(%d+)%.(%d+)%.(%d+)%.(%d+)")
					if tonumber(build_ver) and tonumber(build_ver) < 4 then
						fw_filename = "65.145.4.1.rom"
						fw_cfg["version"] = "65.145.4.1"
						logger:info("The device wp480gr4 will be upgraded in 2 steps")
					end
					-- Firmware for Workforce must be upgraded using HTTP protocol and port 80
					-- if phone's firmware is 65.145.4.15 or 65.145.4.16 (this bug will be fixed in 65.145.4.17)
					if tonumber(build_ver) == 4 and require"core.pbxutils".in_table(minor_ver or "", {"15","16"}) then
						fix_proto = true
					end
				end
				-- Firmware for WP480GR3 and WP490GR3 must be upgraded using HTTP if the device's firmware is 63.145.10.31 or 67.145.8.101
				if pbxutils.in_table(agent["model"], {"wp480gr3", "wp490gr3"}) then
					fix_proto = agent["fw:version"] == "63.145.10.31" or agent["fw:version"] == "67.145.8.101"
				end
				local fw_file = fw_dir .. fw_filename
				if h.is_exists(fw_file) then
					logger:info("Found new firmware version - %s", fw_cfg["version"])
					need_fwupgrade = true
					local fw_link = string.gsub(
						"$protocol://$registrar:$port/$fwpath", "%$(%w+)",
						{
							registrar = registrar,
                    		protocol  = fix_proto and "http" or prov_proto, -- fix for Workforce with fw 65.145.4.15-16 only
                    		port 	  = fix_proto and "80" or prov_port,
                    		fwpath   = "firmwares/wildix2/" .. fw_filename
						}
					)
					
					-- Firmware file server configuration (the suffix of the file name must be .rom)
					result["Firmware"] = { ["Url"] = fw_link } 
					logger:info("Firmware for device %s %s will be changed from version %s to %s", agent["model"], agent["mac"], agent["fw:version"], fw_cfg["version"])
				else
					logger:warning("Firmware file %q can't be loaded", fw_file)
				end
			end
		end
	end
	
	if not need_fwupgrade then
		pbxutils.table_concat(result, resource.parameters or {}) -- merge basic static configuration

---------------------------------------------------------
------- Access settings
---------------------------------------------------------
		-- WEB interface access settings
		result["Settings"]["WEB_LOGIN"]["Password"] = token_pass or h.get_gpassword()
		result["Settings"]["WEB_LOGIN"]["Password02"] = token_pass or h.get_gpassword()

---------------------------------------------------------
------- Assigned user data
---------------------------------------------------------
    	-- get device setting from provision DB and username
    	sql = "SELECT a1.value, a2.value FROM autoprov AS a1 LEFT JOIN autoprov AS a2 ON (a1.mac = a2.mac AND a2.param=%q) WHERE a1.param=%q AND a1.mac=%q"
    	cursor, dberror = connection:execute(sql, "settings", "username", agent["mac"])
        if dberror then
            logger:error("%s, provision is blocked for %s(%s).", dberror, agent["model"] or "", agent["address"] or "")
            return {}, ""
        end
    	local username, settings_json = cursor:fetch(); cursor:close()
    	
    	-- update device username information
    	if not username then
    		username = core.register(agent["mac"])
    		if not username then
    			logger:error("Can't set username to %s(%s). Provision is blocked", agent["model"] or "", agent["address"] or "")
    			return {}, ""
    		end
    	end
    	
    	-- get LDAP data by username
    	local extenNumber = h.get_extenNumber_str()
    	local ldap_result = wmsldap:get(extenNumber, username)
    	local person 	  = {}
    
    	if not ldap_result then		-- somthing with LDAP?
    		logger:error("Could not connect to LDAP server")
    		return {}, username 
    	elseif (0 == #ldap_result  or not ldap_result[1]) and not username:find("unknown%d") then	-- user not found in LDAP
    		logger:debug("Could not get information for %q from LDAP", username)
    		return {}, username
    	elseif 0 ~= #ldap_result then
    		_, person = unpack(ldap_result[1])
    	end
    
    	local secret = person["wuserPasswordSip"] or "1234"
    	if not person["preferredLanguage"] then
    		person["preferredLanguage"] = h.get_admin_lang()
    	end
    	local user_lang  = person["preferredLanguage"] or "en"
    	local user_group = person["owner"] or ""
    	
    	-- load user features
    	local user_features = person.uid and require"common.userFeatures.userFeatures":load(person.uid) or {}
    	
---------------------------------------------------------
------- Network configuration
---------------------------------------------------------    
        local net_cfg			 = {}
        local update_db_settings = false
        if settings_json then
    		net_cfg = json.decode(settings_json)
    	end
    	
    	-- SYSLOG server (by default syslog server is empty)
    	if net_cfg["syslog_server"] and "" ~= net_cfg["syslog_server"] then
    		result["Settings"]["LOGLEVEL"]["RemoteSyslog"] = "1"
    		result["Settings"]["LOGLEVEL"]["RemoteServer"] = net_cfg["syslog_server"]
    	end
    	
    	-- SNMP acess server
    	if net_cfg["snmp_access"] and "" ~= net_cfg["snmp_access"] then
    		result["Network"]["SNMP"]["Enable"] = "1"
    		result["Network"]["SNMP"]["Port"] = "161"
    		result["Network"]["SNMP"]["TrustedAddress"] = net_cfg["snmp_access"]
    		update_db_settings   = true
    	end
        	
        -- VLAN settings
    	if tonumber(net_cfg["vid_voice"]) and 0 < tonumber(net_cfg["vid_voice"]) then
    		result["Network"]["VLAN"]["LanVlanEnable"] = "1"
    		result["Network"]["VLAN"]["LanVid"] 	   = net_cfg["vid_voice"] or "0"
    		result["Network"]["VLAN"]["LanPriority"]   = net_cfg["qos_voice"] or "7" 
    	end
    	if tonumber(net_cfg["vid_data"]) and 0 < tonumber(net_cfg["vid_data"]) then
    		result["Network"]["VLAN"]["PcVlanEnable"] = "1"
    		result["Network"]["VLAN"]["PcVid"] 	      = net_cfg["vid_data"] or "0"
    		result["Network"]["VLAN"]["PcPriority"]   = net_cfg["qos_data"] or "0" 
    	end
        	
        -- set DNS and NTP settings by pbx network configuration and device setttings
        net_cfg["dns_server"], net_cfg["second_dns_server"] = h.get_prim_sec_dns_servers(net_cfg["dns_server"], net_cfg["second_dns_server"])
    	net_cfg["ntp_server"] = net_cfg["ntp_server"] or ""
    	
    	-- checking static option with device location
    	local is_addr_present = h.is_addr_present(agent["address"], agent["mac"])	-- check dublicate address
    	if net_cfg["use_static_ip"] and (is_addr_present or is_addr_remote) then
    		if is_addr_present then
    			logger:warning("Disable static ip. Other device have the same ip address: %s", agent["address"] or "")
    		end
    		if is_addr_remote then	-- device behind NAT can't user static ip
    			logger:warning("Disable static ip. Device ip address (%s) is placed behind NAT", agent["address"] or "")
    		end
    		net_cfg["use_static_ip"] = nil
    		update_db_settings 		 = true
    	end
    	
    	-- get settings for leased device by pbx dhcp
    	local is_leased	= false
    	if not net_cfg["use_static_ip"] or "" == net_cfg["dns_server"] or "" == net_cfg["ntp_server"] then
    		is_leased = h.is_addr_leased(agent["address"])
    	end
    	if not net_cfg["use_static_ip"] and not is_addr_present and is_leased then
    		h.set_addr_static(agent["address"], agent["mac"])	-- mark this lease as static
    		net_cfg["use_static_ip"] = 1
    		update_db_settings 		 = true
    	end
    	
    	-- set settings for leased device
    	if is_leased then
    		if "" == net_cfg["dns_server"] or "" == net_cfg["second_dns_server"] then
	    		local dns_dhcp1, dns_dhcp2 = h.get_dns_servers()
        		if dns_dhcp1 and "" ~= dns_dhcp1 and "" == net_cfg["dns_server"] then
        			net_cfg["dns_server"] = dns_dhcp1             -- set primary DNS from DHCP config
        			update_db_settings = true
        		end
        		if dns_dhcp2 and "" ~= dns_dhcp2 and "" == net_cfg["second_dns_server"] then
        			net_cfg["second_dns_server"] = dns_dhcp2      -- set secondary DNS from DHCP config
        			update_db_settings = true
        		end
	    	end

    	    if "" == net_cfg["ntp_server"] then -- set NTP from DHCP config
    	        local ntp_dhcp = h.get_ntp_servers()
    	        if ntp_dhcp and "" ~= ntp_dhcp then
    	            net_cfg["ntp_server"] = ntp_dhcp
    	            update_db_settings    = true
    	        end
    	    end
    	end
    	
    	-- set settings for device with static options
    	if 1 == net_cfg["use_static_ip"] then	-- use static ip settings
    		local _, netmask, gateway = h.host_settings(agent["address"], registrar_ip)
    		if netmask and gateway then	
    			result["Network"]["LAN"]["Type"] 		   = "2"
    			result["Network"]["LAN"]["StaticIP"]       = agent["address"]
    			result["Network"]["LAN"]["SubnetMask"]     = netmask
    			result["Network"]["LAN"]["DefaultGateway"] = gateway
    			
    			if registrar_ip then
        			if "" == net_cfg["dns_server"] then
        				net_cfg["dns_server"] = registrar_ip
        				update_db_settings 	  = true
        			end
        			if "" == net_cfg["second_dns_server"] then
    					net_cfg["second_dns_server"] = registrar_ip
    					update_db_settings 	         = true
    				end
    			end
    	
    			if "" == net_cfg["ntp_server"] then
    				net_cfg["ntp_server"] = registrar
    				update_db_settings    = true
    			end
    		else
    			logger:warning("For current network configurations is not possible set static ip, disable it")
    			net_cfg["use_static_ip"] = nil
    			update_db_settings       = true
    		end
    	end
    	
    	-- update devdb with new settings
    	if update_db_settings then
    		sql = "INSERT OR REPLACE INTO autoprov (id, param, value, mac) VALUES ((SELECT id FROM autoprov WHERE param=%q AND mac=%q), %q, '%s', %q)"
    		connection:execute (sql, "settings", agent["mac"], "settings", json.encode(net_cfg), agent["mac"])
    	end
        
        -- DNS settings 
    	if "" ~= net_cfg["dns_server"] then
    		result["Network"]["LAN"]["UseStaticDNS"] = "1"
    		result["Network"]["LAN"]["PrimaryDNS"]   = net_cfg["dns_server"]
    		if "" ~= net_cfg["second_dns_server"] then
    			result["Network"]["LAN"]["SecondaryDNS"] = net_cfg["second_dns_server"]
    		end
    	end
    
    	-- NTP settings 
    	if "" ~= net_cfg["ntp_server"] then
    		result["Settings"]["SNTP"]["Enable"]	 = "1"
    		result["Settings"]["SNTP"]["NTPServer1"] = net_cfg["ntp_server"]
    	end
		-- time zone offset    	
		local timeZone = net_cfg["ntp_timezone"]
		if not timeZone or "" == timeZone then
		    timeZone = h.getPbxTimeZone()
		end
		local td, errMsg = require"core.timedate":new(os.time(), timeZone)
		if td then
			td:toTimeZone()
			local tzoffset, errMsg = td:getTzOffset()
			if tzoffset and tzoffset.sign and tzoffset.hour and tzoffset.min then
                local hour = tzoffset.signp .. tonumber(tzoffset.hour)
                local min  = tonumber(tzoffset.min)
                min = min > 0 and ":"..min or ""
                if timeZone == "Europe/Dublin" then  -- WMS-7076 (correcting time for Ireland)
                    result["Settings"]["SNTP"]["TimeZone"] = "0"
                    result["Settings"]["SNTP"]["Name"] = "Ireland(Dublin)"
                else
                    result["Settings"]["SNTP"]["TimeZone"] = hour..min
                end
			end
			if errMsg then
        		logger:debug(errMsg)
			end
			-- Daylight saving time
			local dst, errr = td:getDstData(timeZone)
			if "no" == net_cfg["use_dst"] or errr == "No daylight saving time data found" then
                result.Settings.SNTP.DTS = 0
			elseif dst then
                result.Settings.SNTP.DTS = 1
                result.Settings.DATETIME.Offset = dst.offsetsign..dst.offsetmin
                local wdst = td:getDstByWeek(dst)
                if wdst then
                    -- Month/Week Of Month/Day Of Week/Hour
                    result.Settings.DATETIME.Type = 1
                    if timeZone == "Europe/Dublin" then    -- WMS-7076 (correcting daylight saving time for Ireland)
                        result.Settings.SNTP.StartTime = wdst[1].month .. "/" .. wdst[1].week .. "/" .. wdst[1].wday .. "/" .. wdst[2].hour
                        result.Settings.SNTP.EndTime = wdst[2].month .. "/" .. wdst[2].week .. "/" .. wdst[2].wday .. "/" .. wdst[1].hour
                    else
                        result.Settings.SNTP.StartTime = wdst[1].month .. "/" .. wdst[1].week .. "/" .. wdst[1].wday .. "/" .. wdst[1].hour
                        result.Settings.SNTP.EndTime = wdst[2].month .. "/" .. wdst[2].week .. "/" .. wdst[2].wday .. "/" .. wdst[2].hour
                    end
                else
                    result.Settings.SNTP.StartTime = dst[1].month .. "/" .. dst[1].day .. "/" .. dst[1].hour
                    result.Settings.SNTP.EndTime = dst[2].month .. "/" .. dst[2].day .. "/" .. dst[2].hour
                end
			end
        end
        if errMsg then
        	logger:debug(errMsg)
        end
    	
    	-- ToS, tonezone settings
    	local tos_data, toneZone = h.get_sip_settings()
    	result["Network"]["QOS"]["SignalTos"] = tos_data["tos_sip"] / 4
    	result["Network"]["QOS"]["RtpTos"]    = tos_data["tos"] / 4 
    	
        -- Jitter buffer settings
        local tjb = h.get_jitter_buffer()
        result["Voice"]["JITTER"]["Adaptive"] = tjb["min"] == tjb["max"] and "1" or "0"	--XXX: jitter type must be with vice versa values
        result["Voice"]["JITTER"]["Min"] 	  = tjb["min"]
        result["Voice"]["JITTER"]["Max"]	  = tjb["max"]
        result["Voice"]["JITTER"]["Nominal"]  = tjb["nom"]
        
        -- WMS-4901, workaround for r3 phones
        if isModelR3R4 then
            local cur_jit_max = tonumber(tjb["max"]) or 400
            local cur_jit_nom = tonumber(tjb["nom"]) or 200
            if cur_jit_nom > 200 or cur_jit_max > 400 then
                result["Voice"]["JITTER"]["Min"]      = "100"
                result["Voice"]["JITTER"]["Max"]      = "500"
                result["Voice"]["JITTER"]["Nominal"]  = "300"
                result["Voice"]["JITTER"]["Adaptive"] = "0"
            else
                result["Voice"]["JITTER"]["Min"]      = "100"
                result["Voice"]["JITTER"]["Max"]      = "400"
                result["Voice"]["JITTER"]["Nominal"]  = "200"
                result["Voice"]["JITTER"]["Adaptive"] = "0"
            end
        end

        -- WMS-4736
        local modelsWithFixedJB = {"wp410r2", "wp480r2", "wp480gr2", "wp490r2", "wp490gr2", "wp490gcr2"}
        if pbxutils.in_table(agent["model"], modelsWithFixedJB) then
                result["Voice"]["JITTER"]["Min"]      = "50"
                result["Voice"]["JITTER"]["Max"]      = "200"
                result["Voice"]["JITTER"]["Nominal"]  = "100"
                result["Voice"]["JITTER"]["Adaptive"] = "0"
        end
        -- WP-165 , workaround for wp410r3 (aka Start) phone
        if "wp410r3" == agent["model"] then
                result["Voice"]["JITTER"]["Min"]      = "50"
                result["Voice"]["JITTER"]["Max"]      = "500"
                result["Voice"]["JITTER"]["Nominal"]  = "200"
                result["Voice"]["JITTER"]["Adaptive"] = "0"
        end

        -- Static features
        local end_digits = h.get_eod() 	-- 0 - None; 1 - *; 2 - #
        if "1" == end_digits then
        	end_digits = "2"
        elseif "2" == end_digits then
        	end_digits = "1"
        end
        result["Features"]["SENDKEY"]["KeySendType"] = end_digits
       	
       	local inter_digit_time = h.get_interdigit_timeout()
       	result["Features"]["DIALNOW"]["Delay"] = inter_digit_time
       	
        if "0" == h.get_button_sound() then
            result["Settings"]["AUDIO"]["KeyVol"] = "0"
        end

       	
       	-- Multicast paging
       	if connection:switch("wmsdb") then
			local sql	  = "SELECT p.id, p.name FROM paging AS p LEFT JOIN paging_members AS pm ON (pm.paging_id=p.id) WHERE pm.group_member=%q order by p.id limit 10"
			local res_sql = connection:rows(sql, username)
			if res_sql then
				local i = 0
				for id, name in res_sql do	-- filling value from table paging
					result["Multicast"]["LISTEN"..i]["Label"] = name
					result["Multicast"]["LISTEN"..i]["IP"]	  = "239.1.1.10:" .. tostring(9999+id)
					i = i + 1
				end
			end
       	end

		-- LLDP settings
		if pbxutils.in_table(agent["model"], {"wp480gr3", "wp480gr4", "wp490gr3", "wp490gr4"}) then -- WP 410 (including r3/Start) doesn't support this feature
			if gconf.LLDP == "yes" then
				result["Network"]["LLDP"]["LLDPEnable"] = "1"
			end
			if gconf.LLDPInterval then
				result["Network"]["LLDP"]["Interval"] = gconf.LLDPInterval
			end
		end

---------------------------------------------------------
------- User features settings
---------------------------------------------------------

		-- Hotline feature
		if user_features.HOTLINE then
			local hotlineDelay = tonumber(user_features.HOTLINE.delay) or 0
			hotlineDelay = hotlineDelay < 0 and 0 or hotlineDelay > 5 and 5 or hotlineDelay -- hotline delay have to be in range 0 - 5 sec.
			result.Features.HOTLINE = {
				Enable = user_features.HOTLINE.status or 0,
				Number = user_features.HOTLINE.value or "",
				Delay  = hotlineDelay
			}
		else
			result.Features.HOTLINE = {
				Enable = 0,
				Number = "",
				Delay  = 0
			}
		end
		
		-- Call waiting tone feature
		if user_features.CWTONE then
			local callWaitingTone = tonumber(user_features.CWTONE.status)
			if 0 == callWaitingTone then
				result.Features.CALL_WAITING.PlayTone = callWaitingTone
			end
		end

---------------------------------------------------------
------- Account settings
---------------------------------------------------------
        -- retrieve failover configuration
        local pbxFailover     = nil
        local failoverSipPort = nil
        if not username:find("unknown%d") then
            local failoverObj = require"common.failover":new()
            if failoverObj and "1" == failoverObj:isEnabled() then
                local failoverCfg = failoverObj:getConfig()
                pbxFailover = failoverCfg and failoverCfg.pbxFailoverDomain
                -- Temporary Fix until vendor fix a problem "resolving delay on the WPR2 devices if a first DNS goes down"
                if gconf.FailoverPbxIp then
                    pbxFailover = gconf.FailoverPbxIp
                end
                failoverSipPort = failoverCfg and failoverCfg.sipPort
                logger:debug(failoverCfg, "failoverCfg")
            end
        end

        -- set SIP transport and SIP port
        local gtransp	= nil
        local gsipport	= nil
        if is_addr_remote then
            gtransp  = gconf.SIPTransportRemote
            gsipport = gconf.SIPServerPortRemote
        else
            gtransp  = gconf.SIPTransport
            gsipport = gconf.SIPServerPort
        end
        
        local sip_transport = "TLS"
        if gtransp then
            sip_transport   = gtransp
        end
        
        local sip_port   = "5060"
        if gsipport then
            sip_port     = gsipport
        elseif "TLS" == sip_transport then
            sip_port     = "5061"
            if is_addr_remote then
                sip_port = agent["alternate_port"] or "443"
            end
        elseif "TCP" == sip_transport then
            sip_port     = "5060"
        end

        if pbxFailover then
            sip_port = agent["alternate_port"] or "443"
            -- WMS-5810: support tls for wp r3-r4 in failover mode
            if isModelR3R4 then
                sip_transport = "UDP" == sip_transport and "TCP" or sip_transport  -- Allowed transports: TLS or TCP

                if gsipport then  -- WMS-7271
                    sip_port = gsipport
                    failoverSipPort = gsipport
                end
            else
                sip_transport = "TCP"
            end
        end

        -- set username settings
        local label = username
        local custom_identity = user_features["CURRENT_IDENTITY"]
        if username:find("unknown%d") then
            label = username:gsub("unknown", "login")
            secret, err = h.pass_for_unknown(username)
            if err then
                logger:error("password generation error for unknown account:", err)
            end
        else
            label = gconf.DisableShowUsername and username or label .. " " .. (person["cn"] or "")

            if custom_identity and "" ~= custom_identity then
                local ldapresult = wmsldap:get("telephoneNumber", custom_identity)
                if ldapresult and 0 ~= #ldapresult then
                    _, customidentity_attrs = unpack(ldapresult[1])
                    if customidentity_attrs.telephoneNumber then
                        label = gconf.DisableShowUsername and customidentity_attrs.telephoneNumber or
                                string.format("%s %s", customidentity_attrs.telephoneNumber, customidentity_attrs.cn)
                    end
                end
            end
        end

        result["Account1"]["GENERAL"]["Label"] = label
        result["Account1"]["GENERAL"]["DisplayName"] = person["cn"] or username
        result["Account1"]["GENERAL"]["UserName"]	 = username
        result["Account1"]["GENERAL"]["AuthName"] 	 = username
        result["Account1"]["GENERAL"]["Pwd"]		 = secret
        
        -- set SIP server settings
        result["Account1"]["SIP"]["Server"] 		 = registrar
        result["Account1"]["SIP"]["Port"] 		     = sip_port
        result["Account1"]["OUTPROXY"]["Enable"]     = "1"
        result["Account1"]["OUTPROXY"]["Server"]     = registrar
        result["Account1"]["OUTPROXY"]["Port"]       = sip_port
        if pbxFailover then
            result["Account1"]["SIP"]["Server2"]     = pbxFailover
            result["Account1"]["SIP"]["Port2"]       = failoverSipPort
            result["Account1"]["OUTPROXY"]["BakServer"] = pbxFailover
            result["Account1"]["OUTPROXY"]["BakPort"]   = failoverSipPort
        end
        result["Account1"]["SIP"]["TransType"]	     = ("TCP" == sip_transport and "1") or ("TLS" == sip_transport and "2") or "0"
        
        -- audio codecs priority
        local is_wideband = h.is_wideband(agent["address"])
        local netband     = is_wideband and "wideband" or "narrowband"
        local codecs_set  = resource.codecs[netband]
        for key, value in pairs(codecs_set) do
        	result["Account1"][key] = value
        end

        -- opus codec, we replace not used G723_53
        -- first priority in "wideband" and "narrowband"
        if "wp490gr4" == agent["model"] or "wp480gr4" == agent["model"] then
            result["Account1"]["Audio2"]["Enable"]   = "1"
            result["Account1"]["Audio2"]["Name"]     = "opus"
            result["Account1"]["Audio2"]["Priority"] = "13"
            result["Account1"]["Audio2"]["Payload"]  = "111"
        end

        -- for these models we use template from vendor
        if "wp480gr3" == agent["model"] or "wp490gr3" == agent["model"] then
            result["Account1"]["Audio9"]["Enable"]   = "1"
            result["Account1"]["Audio9"]["Name"]     = "opus"
            result["Account1"]["Audio9"]["Priority"] = "10"
            result["Account1"]["Audio9"]["Payload"]  = "111"
        end
        -- WP-165 For Start phones we will use only G.711 codec
        if "wp410r3" == agent["model"] then
            result["Account1"]["Audio4"]["Enable"]   = "0"
        end

        if is_wideband then
        	result["Account1"]["CALL"]["UseRemoteCodec"] = "1"
        end

    	-- Voice mail settings
    	local vmnumber = username
    	if  user_features["SHAREDVM"] and "1" == user_features["SHAREDVM"].status and "" ~= user_features["SHAREDVM"].value then
    		vmnumber = user_features["SHAREDVM"].value
    	end
    	result["Account1"]["VOICEMSG"]["Number"] = "vmaccess*" .. vmnumber
    		
        -- Language settings
        local lang_value = ({
        	["en"] = "0",
        	["es"] = "6",
            ["es-us"] = "6",
        	["it"] = "7",
        	["fr"] = "9",
    		["de"] = "10",
            ["nl"] = "8",
    	})[user_lang] or "0"
     	result["Settings"]["LANGUAGE"]["Type"] = lang_value
    
        -- WMS-5651 OffHookDialOut 
        if gconf.OffHookDialOut then
            result["Features"]["DIALNOW"]["OffHookDialOut"] = "no" == gconf.OffHookDialOut and "0" or "1"
        end

        -- WMS-5738 
        if "wp480gr3" == agent["model"] or "wp490gr3" == agent["model"] or "wp490gr4" == agent["model"] or "wp480gr4" == agent["model"] then
            if not result["Voice"]["SIDETONE"] then
                result["Voice"]["SIDETONE"] = {}
            end
            result["Voice"]["SIDETONE"]["Handset"] = "0"
            result["Voice"]["SIDETONE"]["Headset"] = "0"
            if gconf.SidetoneHandset then
                result["Voice"]["SIDETONE"]["Handset"] = tostring(gconf.SidetoneHandset)
            end
            if gconf.SidetoneHeadset then
                result["Voice"]["SIDETONE"]["Headset"] = tostring(gconf.SidetoneHeadset)
            end
        end

        -- WMS-5404
        if isModelR3R4 then
            result["Features"]["REJECT"]["EnableReasonCause"] = "1"
            if gconf.EnableReasonCause then
                result["Features"]["REJECT"]["EnableReasonCause"] = "no" == gconf.EnableReasonCause and "0" or "1"
            end
        end

        -- WMS-6648
        if isModelR3R4 then
            if wmsperm:check(person["owner"], "canuse", "FULLPRESENCE", nil, person["licenseType"]) ~= "nouse" then
                result["Features"]["BLF"]["PresenceEnable"] = "1"
            else
                result["Features"]["BLF"]["PresenceEnable"] = "0"
            end
        end
---------------------------------------------------------
------- Tones settings
---------------------------------------------------------
        local function __makeToneString(tones)
            local rv = ""
            if tones and "table" == type(tones) then
                for k, v in pairs(tones) do
                    local freq = (v["freq1"] or "")..((v["freq1"] and v["freq2"] and ("+"..v["freq2"])) or "")
                    local dur  = v["duration"] or ""
                    local tone = freq .. ("" ~= freq and "" ~= dur and "/" or "").. dur
                    rv = rv .. ("" ~= rv and "" ~= tone and "," or "").. tone
                end
            end
            return rv
        end

		local tone = h.get_dialtone("akuvox", toneZone)
		if "string" == type(tone) then
			result.Voice.TONE.Country = tone
		elseif "table" == type(tone) then
            result.Voice.TONE.Country      = tone.country or "1"
            result.Tone.GENERAL.Busy       = __makeToneString(tone.busy)
            result.Tone.GENERAL.Ringback   = __makeToneString(tone.ring)
            result.Tone.GENERAL.Congestion = __makeToneString(tone.congestion)
            result.Tone.GENERAL.Dialtone   = __makeToneString(tone.dial)
            if "0" ~= tone.country then -- disable callwaiting tone for custom country until vendor will fix issue: 'While phone waits between "beeps" - user with "callwaiting tone" can't hear other side media. Phone just mutes the WP's audio output. '
                 result.Tone.GENERAL.Callwait   = __makeToneString(tone.callwaiting)
            end
        end

---------------------------------------------------------
------- Key/display settings
---------------------------------------------------------
    	-- Line/memory keys settings
    	local user_keys  = core.fkeys_get(user_features, username)
        local features = require"common.userFeatures.userFeatures":new(person.uid)
        local custom_identities = {}
        if features then
            features:load()
            custom_identities = features:getCustomIdentities() or {}
        end
        local phone_keys, exp_keys = __keys_convert(user_keys, agent, user_group, registrar, prov_port, prov_proto, custom_identities)
        local invalid_keys_msg = ""
        if "table" == type(phone_keys) then
            for linekey, value in pairs(phone_keys) do
                result["Programable"][linekey] = value
            end
        else
            invalid_keys_msg = " phone"
        end
        -- Expansion module keys settings
        if "table" == type(exp_keys) then
            for exp_mod, exp_conf in pairs(exp_keys) do
                result[exp_mod] = exp_conf
            end
        else
            invalid_keys_msg = invalid_keys_msg .. " expansion"
        end
        if "" ~= invalid_keys_msg then
            local userkeys_data = "table" == type(user_keys) and json.encode(user_keys) or user_keys
            local ident_data = "table" == type(custom_identities) and json.encode(custom_identities) or custom_identities
            logger:warning("Invalid keys data: %s. User keys: %s. Identities: %s.", invalid_keys_msg, userkeys_data, custom_identities)
        end
        
        --- Soft keys settings
        local xmlbr = require"xmlbrowser"
        
        -- XML history settings
        local xml_history_link = string.gsub(
    			"$protocol://$registrar:$port/?action=gethistory", "%$(%w+)",
    			{
    				registrar = registrar,
    				protocol  = isModelR3R4 and "https" or "http", -- WMS-1160, WMS-8341
    				port 	  = prov_port,
    			}
    		)
    	local xml_history_link_all = xml_history_link .. "&type=init-all"
    	local xml_history_link_missed = xml_history_link .. "&type=init-missed"
    	
    	result["XMLBrowser"]["REMOTEURL"]["HistoryMissed"]	 = xml_history_link_missed
    	result["XMLBrowser"]["REMOTEURL"]["History"]	 = xml_history_link_all
    	result["Programable"]["SOFTKEY01"]["Label"] 	 = xmlbr.label_name("history", user_lang)
    	result["CustomMenu"]["DIALMENUITEM01"]["Param1"] = xml_history_link		-- XXX: go to phonebook search after fix for Select button
    	result["CustomMenu"]["DIALMENUITEM01"]["Label"]  = xmlbr.label_name("history", user_lang)
    	    
        -- XML phonebook settings
        local xml_phonebook_link = string.gsub(
    			"$protocol://$registrar:$port/?action=getphonebook&phonebook=all", "%$(%w+)",
    			{
    				registrar = registrar,
    				protocol  = isModelR3R4 and "https" or "http",
    				port 	  = prov_port,
    			}
    		)
    	
    	result["XMLBrowser"]["REMOTEURL"]["Phonebook"]	 = xml_phonebook_link
        result["Programable"]["SOFTKEY02"]["Label"]  	 = xmlbr.label_name("phonebook", user_lang)
        result["CustomMenu"]["DIALMENUITEM02"]["Param1"] = xml_phonebook_link
        result["CustomMenu"]["DIALMENUITEM02"]["Label"]  = xmlbr.label_name("phonebook", user_lang)
        
        -- live search feature settings
        local xml_livesearch_link = string.gsub(
    			"$protocol://$registrar:$port/?action=getphonebook&menu=livesearch&input=", "%$(%w+)",
    			{
    				registrar = registrar,
    				protocol  = isModelR3R4 and "https" or "http",
    				port 	  = prov_port,
    			}
    		)
        result["XMLBrowser"]["REMOTEURL"]["LiveSearch"]  = xml_livesearch_link

        -- XML feature browser settings
    	local xml_features_link = string.gsub(
    			"$protocol://$registrar:$port/?action=getxmlbrowser&get=features", "%$(%w+)",
    			{
    				registrar = registrar,
    				protocol  = isModelR3R4 and "https" or "http",
    				port 	  = prov_port
    			}
    		)
    	result["Programable"]["SOFTKEY03"]["Label"]  = xmlbr.label_name("features", user_lang)
    	result["Programable"]["SOFTKEY03"]["Param1"] = xml_features_link
    	if wmsperm:check(user_group, "canuse", "PHONE_SETTING_MENU", nil, person["licenseType"]) == "nouse" then
    		result["Programable"]["SOFTKEY04"]["Type"]  = "4"
    		result["Programable"]["SOFTKEY04"]["Label"] = "Status"
    	end
   	
    	-- XML vmaccess settings
    	local xml_vmaccess_link = string.gsub(
    			"$protocol://$registrar:$port/?action=getxmlbrowser&get=vmaccess&menu=main", "%$(%w+)",
    			{
    				registrar = registrar,
    				protocol  = isModelR3R4 and "https" or "http",
    				port 	  = prov_port
    			}
    		)
		result["voicemailviewaction"] = xml_vmaccess_link
    	result["XMLBrowser"]["REMOTEURL"]["VoiceMail"] = xml_vmaccess_link
		if "wp490gr4" == agent.model or "wp490gr3" == agent.model or "wp480gr2" == agent.model or "wp480gr3" == agent.model or "wp480gr4" == agent.model or ("wp410r2" == agent.model and "v2" == agent["model:additional"]) or "wp410r3" == agent.model then -- set LongPress STAR key to change user status
            result["Programable"]["STAR"] = {["LongPress"] = 1}
            result["Programable"]["STAR"]["Type"] = 32 -- URL - send GET request.
            result["Programable"]["STAR"]["Param1"] = prov_proto.."://"..registrar..":"..prov_port.."/?action=getxmlbrowser&get=update&feature=STATUS"
		end
		
        --- Enable BLF-pages
        if "wp480gr2" == agent.model or ("wp410r2" == agent.model and "v2" == agent["model:additional"]) or "wp410r3" == agent.model then
            result["Settings"]["OTHERS"]["MultiPageMode"] = 1
        elseif isModelR3R4 then
            result["Settings"]["DISPLAYSTYLE"]["LinekeyPage"] = 1
        end
        
        --- Enable 2nd and 3rd BLF-page on exp.module
        if "wp490r2" == agent.model or "wp490gr2" == agent.model then
            result["Settings"]["EXTDPAGEMODE"] = {["PageMode"] = 0}
        elseif "wp490gr3" == agent.model or "wp490gr4" == agent.model then
            result["Settings"]["EXTDPAGEMODE"] = {["PageMode"] = 1}
        end
		if "wp480gr2" == agent.model then
    		result["Programable"]["NUM01"] = {["LongPress"] = 1}
    		result["Programable"]["NUM01"]["Type"] = 31
			result["Programable"]["NUM01"]["Param1"]	= xml_vmaccess_link
			-- Replace function keys to SpeedDial, because (F1-F4) does not support BLF key type.
			result["Programable"]["TRANSFER"]["Type"] = 10
			result["Programable"]["REDIAL"]["Type"] = 10
			result["Programable"]["VOICEMESSAGE"]["Type"] = 10
			result["Programable"]["BOOK"]["Type"] = 10
    	elseif ("wp410r2" == agent.model and "v2" == agent["model:additional"]) or "wp410r3" == agent.model then
    		result["Programable"]["NUM01"] = {["LongPress"] = 1}
    		result["Programable"]["NUM01"]["Type"] = 31
			result["Programable"]["NUM01"]["Param1"]	= xml_vmaccess_link
    		-- Replace function keys to SpeedDial, because (F1-F4) does not support BLF key type.
    		result["Programable"]["TRANSFER"]["Type"] = 10
    		result["Programable"]["VOICEMESSAGE"]["Type"] = 10
    		result["Programable"]["LINEKEY1"] = nil
    		result["Programable"]["LINEKEY2"] = nil
    		result["Programable"]["LINEKEY3"] = nil
    	else
			result["Programable"]["VOICEMESSAGE"]["Type"]	= "21"
			result["Programable"]["VOICEMESSAGE"]["Param1"]	= xml_vmaccess_link
    	end
    	
    	--- Custom BLF mode settings
    	if gconf.BLFNotifyType then
    	    result["Account1"]["BLF"]["BlfNotifyParseType"] = gconf.BLFNotifyType
    	end

        -- Advanced phone settings menu
        if isModelR3R4 then
            if "yes" ~= wmsperm:check(user_group, "canuse", "ADVANCED_PHONE_SETTINGS_MENU", nil, person["licenseType"]) then
                result["Settings"]["OTHERS"]["MenuContainNetwork"] = "0"
                result["Settings"]["OTHERS"]["MenuContainAutop"]   = "0"
            else
                result["Settings"]["OTHERS"]["MenuContainNetwork"] = "1"
                result["Settings"]["OTHERS"]["MenuContainAutop"]   = "1"
            end
        end
---------------------------------------------------------------------------------------
------- Possibility to disable speakerphone button and mic in speakerphone mode (WP-155)
---------------------------------------------------------------------------------------
        if "wp480gr4" == agent.model then
            result.Settings.HANDFREE.Active = "1" -- speakerphone button is enabled by default
            result.Settings.HANDFREE.MicActive = "1" -- microphone in speakerphone mode is enabled by default
            if "yes" == gconf.SpeakerDisable then
                result.Settings.HANDFREE.Active = "0"
                result.Settings.HANDFREE.MicActive = "0"
                logger:info("Speakerphone mode was disabled on PBX")
            elseif gconf.SpeakerDisable and "table" == type(gconf.SpeakerDisable) then
                if person.telephoneNumber and pbxutils.in_table(person.telephoneNumber, gconf.SpeakerDisable) then
                    result.Settings.HANDFREE.Active = "0"
                    result.Settings.HANDFREE.MicActive = "0"
                    logger:info("Speakerphone mode is not allowed for extension %s", person.telephoneNumber)
                end
            end
        end
---------------------------------------------------------
------- Adjusting the gain of microphones and speakers
---------------------------------------------------------
        if isModelR3R4 then
            result.Settings.HANDFREE.MicMidValue = "21"
		elseif "wp410r2" == agent.model and "v2" == agent["model:additional"] then
            result.Settings.HANDFREE.SpkMidValue = "16"
            result.Settings.HANDFREE.MicMidValue = "13"
            result.Settings.HEADSET.SpkMidValue = "25"
            result.Settings.HEADSET.MicMidValue = "25"
            result.Settings.HANDSET.MicMidValue = "6"
        elseif "wp410r3" == agent.model then
            result.Settings.HANDFREE.SpkMidValue = "16"
            result.Settings.HANDFREE.MicMidValue = "13"
            result.Settings.HEADSET.SpkMidValue = "25"
            result.Settings.HEADSET.MicMidValue = "25"
            result.Settings.HANDSET.SpkMidValue = "12"
            result.Settings.HANDSET.MicMidValue = "10"
        elseif "wp410r2" == agent.model then
            result.Settings.HANDSET.MicMidValue = "5"
            result.Settings.HEADSET.SpkMidValue = "12"
            result.Settings.HEADSET.MicMidValue = "5"
        elseif "wp480gr2" == agent.model or "wp480r2" == agent.model or "wp490gr2" == agent.model or "wp490r2" == agent.model then
            result.Settings.HEADSET.SpkMidValue = "22"
            result.Settings.HANDSET.SpkMidValue = "21"
        end
		
---------------------------------------------------------
------- Disable headset PlayRing default value for WP490R2 and WP490GR2 and WP490GR3
---------------------------------------------------------
		if "wp490r2" == agent.model or "wp490gr2" == agent.model or "wp490gr3" == agent.model or "wp490gr4" == agent.model or "wp480gr4" == agent.model then
			result.Settings.HEADSET.PlayRing = nil
		end

---------------------------------------------------------
------- Power LED settings
---------------------------------------------------------	
		if "yes" == gconf["POWER_LED"] then
			result["Settings"]["OTHERS"]["PowerLedDefault"] = 1
		end
		if gconf.PowerLEDBlinkTime then
            result["Settings"]["DISPLAYSTYLE"]["PowerLedBlinkTimeOut"] = gconf.PowerLEDBlinkTime 
        end

---------------------------------------------------------
------- Set Date format
---------------------------------------------------------
		if user_features.DATETIME then
			-- not used, see WMS-5205
			-- local php_to_phone = {["d/m/Y"]="3", ["m/d/Y"]="6", ["d M Y"]="7", ["M d Y"]="8",
            --            ["d-m-Y"]="2", ["m-d-Y"]="9", ["Y-m-d"]="0", ["d.m.Y"]="10"}
			local usa_formats = {["M d Y"]=true, ["m/d/Y"]=true, ["m-d-Y"]=true}
			local phpFormat    = user_features.DATETIME.format
			result["Settings"]["DATETIME"]["DateFormat"] = usa_formats[phpFormat] and "5" or "4"

			local timeFormat = user_features.DATETIME.is24h
			if timeFormat and "" ~= timeFormat then
				result["Settings"]["DATETIME"]["TimeFormat"] = timeFormat
			end
		end
---------------------------------------------------------
------- Hide Dtmf visialisation on display of phones
---------------------------------------------------------
		if gconf.HIDE_DTMF then
		    result["Features"]["Others"]["ViewDTMFNum"] = 0
		end
---------------------------------------------------------
------- Firmware for expansion modules
---------------------------------------------------------
        if "wp490r2" == agent.model or "wp490gr2" == agent.model then
            result["Extd"]={["Url"] = string.format("%s://%s:%s/firmwares/wildix2/%s", prov_proto, registrar, prov_port,firmwares["em53"]["application"])}
        elseif "wp490gr3" == agent.model or "wp490gr4" == agent.model then
            result["Extd"]={["Url"] = string.format("%s://%s:%s/firmwares/wildix2/%s", prov_proto, registrar, prov_port,firmwares["em63"]["application"])}
        end
---------------------------------------------------------
------- Paket loss threshold screen notification 
---------------------------------------------------------
        if gconf.LossPktRateMin and gconf.LossPktRateMax and gconf.LossPktRateMin <= gconf.LossPktRateMax then
            result.Settings.AUDIO.LossPktRateMin = gconf.LossPktRateMin
            result.Settings.AUDIO.LossPktRateMax = gconf.LossPktRateMax
        elseif gconf.LossPktRateMin or gconf.LossPktRateMax then
            logger:error("Wrong LossPktRate value: min - %s, max - %s", gconf.LossPktRateMin or "", gconf.LossPktRateMax or "")
        end
    end

-----------------------------------------------------
--- Result serialization
-----------------------------------------------------
	local output = {}
	
	local function serialize_cfg (key, value, strin)
		if not key or "" == key or not value then
    		return false
    	end
    	
    	local str = nil
    	if not strin or "" == strin then
    		str = key
    	else
    		str = strin .. "." .. key
    	end
    	
    	if "table" ~= type(value) then
    		str = str .. " = " .. value
    		output[#output + 1] = str
    		return true
    	else
    		for keyin, valuein in pairs(value) do
    			serialize_cfg(keyin, valuein, str)
    		end
    		return str
    	end
    	return false
    end
    
	for key, value in pairs(result) do
		serialize_cfg(key, value, "Config")
	end
	
	return output, username    
end

function setuser (...) end
function setfeatures (...) end
