--- PBX post upgrade device sync
-- 
-- @file    postsync.lua 
-- @author  Yuriy Zveryanskyy
-- @created 2021/09/23

module("postsync", package.seeall)

local helper = require"common.helper"
local json   = require"dkjson"

local logger = require"core.logger":init("POSTSYNC", 3)

local families_filters ={ -- by "vendor:additional"
    ["WPR2"]  = {["sql"] = '"WPR2"',  ["lua"] = 'WPR2'},
    ["W-AIR"] = {["sql"] = '"W-AIR',  ["lua"] = 'W%-AIR.*'},
    ["MGW"]   = {["sql"] = '"MGW"',   ["lua"] = 'MGW'},
    ["FXXR2"] = {["sql"] = '"FX_R_"', ["lua"] = 'FX[SO]R%d'},
    ["DWG"]   = {["sql"] = '"DWG"',   ["lua"] = 'DWG'},
    ["W-PA"]  = {["sql"] = '"W-PA',   ["lua"] = 'W%-PA.*'}
}

local function __parse_cmdline_params(clist)
    local rv = {}
    for i, param in pairs(clist) do
        if "--input-file" == param then
            rv.config = clist[i + 1]
        elseif "--from-version" == param then
            rv.version = clist[i + 1]
        elseif "--ignore-version" == param then
            rv.ignore_version = true
        end
    end
    return rv
end

local function __parse_version(version)
    local _major, _minor, date = string.match(version, "(%d+)%.(%d+)%.(%d+)%.%d+.*")
    return tonumber(date)
end

local function __parse_config_date(date)
    local y, m, d = string.match(date, "(%d%d%d%d)%-(%d%d)%-(%d%d)")
    return tonumber(y .. m .. d)
end

local function __parse_config_file(conf_file)
    local rv = {}
    if not helper.is_exists(conf_file) then
        return rv
    end
    for line in io.lines(conf_file) do
        local regex = "^%s*([^#]-)%s*[,;]%s*(.-)[,;]%s*(.-)%s*[,;]%s*(.-)%s*$"
        local date, family, model, reboot = string.match(line, regex)
        if date then
            rv[#rv + 1] = {
                ["date"] = __parse_config_date(date),
                ["family"]  = family,
                ["model"]   = model,
                ["reboot"]  = "yes" == reboot
            }
        end
    end
    return rv
end

local function __filter_by_version(conf, ver)
    local rv = {}
    local old = __parse_version(ver)
    for _, item in pairs(conf) do
        if item["date"] >= old then
            rv[#rv + 1] = item
        end
    end
    return rv
end

local function __create_filters(conf)
    local select_filter = {["any"] = nil, ["families"] = {}, ["models"] = {}}
    local reboot_filter = {["any"] = nil, ["families"] = {}, ["models"] = {}}
    for _, item in pairs(conf) do
        if "*" == item["family"] then
            select_filter["any"] = true
            if not reboot_filter["any"] then
                reboot_filter["any"] = item["reboot"]
            end
        elseif "" ~= item["family"] and "*" == item["model"] then
            select_filter["families"][item["family"]] = true
            if not reboot_filter["families"][item["family"]] then
                reboot_filter["families"][item["family"]] = item["reboot"]
            end
        elseif "" ~= item["model"] then
            select_filter["models"][item["model"]] = true
            if not reboot_filter["models"][item["model"]] then
                reboot_filter["models"][item["model"]] = item["reboot"]
            end
        end
    end
    return select_filter, reboot_filter
end

local function __create_sql_query(select_filter)
    local query = "SELECT id,value,mac FROM autoprov WHERE param='device'"
    if select_filter["any"] then
        return query
    end
    local filters = {}
    for family, _v in pairs(select_filter["families"]) do
        local ffilter = families_filters[family]
        if ffilter then
            filters[#filters + 1] = "value LIKE '%%\"vendor:additional\":%%" .. ffilter["sql"] .. "%%'"
        else
            logger:warning("Device family '%s' not found", family)
        end
    end
    for model, _v in pairs(select_filter["models"]) do
        filters[#filters + 1] = "(value LIKE '%%\"model\":%%\"" .. model .. "\"%%')"
    end
    if #filters == 0 then
        return
    end
    return query .. " AND (" .. table.concat(filters, " OR ") .. ")"
end

local function __get_family(str)
    for family, t in pairs(families_filters) do
        if string.match(str, t["lua"]) then
            return family
        end
    end
end

local function __activate_reboot_option(dbid, devtable)
    local connection = require"common.core".connection
    devtable["reboot"] = 1
    local sql = "UPDATE autoprov SET value='%s' WHERE id=%d"
    connection:execute(sql, json.encode(devtable), dbid)
end

local function __prepare_sync(dbid, devtable, reboot_filter)
    -- commands --------------------------
    local sync_cmd       = "check-sync"
    local reboot_cmd     = "reboot-phone"
    local reboot_mgw_cmd = "reboot-mgw"
    --------------------------------------

    local model = devtable["model"]
    if not model then
        return
    end

    local reboot = reboot_filter["models"][model]
    local family = __get_family(devtable["vendor:additional"])
    if family then
        reboot = reboot or reboot_filter["families"][family] or reboot_filter["any"]
    else
        if "boolean" == type(reboot) then
            logger:warning("Cannot find family for device '%s', will use typical sync command", model)
        else
            return
        end
        
    end

    local command = reboot and reboot_cmd or sync_cmd

    -- all hacks are here
    if reboot then
        if "WPR2" == family then
            __activate_reboot_option(dbid, devtable)
            command = sync_cmd
        elseif "FXXR2" == family then
            __activate_reboot_option(dbid, devtable)
        elseif "MGW" == family then
            command = reboot_mgw_cmd
        end
    end

    return command, model
end

-- from wmsbind
local function __get_exten(mac, api)
    local connection = require'common.core'.connection
    local exten      = nil
    local sql = 'SELECT value FROM autoprov WHERE param=%q and mac=%q LIMIT 1'
    local cursor = connection:execute(sql, 'username', mac:lower())
    local username = cursor:fetch(); cursor:close()
    connection:close()

    if username and '' ~= username then
        local regs = api:getRegistrationByExten(username)
        if regs then
            for reg in regs:iterate() do
                local ua = reg:getUserAgent()
                local regmac = ua:match('^Wildix%s.+%s[0-9%.]+%a?%s([0-9A-Fa-f]+)')
                regmac = regmac and regmac:upper()
                if regmac == mac:upper() then
                    exten = reg:getExten()
                    break
                end
            end
        end
    end

    if not exten then
        local regs = api:getRegistrationByUAMAC(mac)
        exten      = regs and regs:first():getExten()
    end
    return exten
end

local function __do_sync(sync_data)
    local api = require"core.sipproxy.sipproxy":new()
    local cw = require"core.astman"
    local devcnt = 0
    local models_cnt = {}
    for _, device in pairs(sync_data) do
        local exten = __get_exten(device["mac"], api)
        if exten then
            cw.send_command(string.format("sip notify %s %s", device["command"], exten), true)
            devcnt = devcnt + 1
            models_cnt[device["model"]] = models_cnt[device["model"]] and models_cnt[device["model"]] + 1 or 1
            local extra_pause = devcnt % 10 == 0 and 8 or 0
            require"socket".sleep(2 + extra_pause)
        end
    end
    return models_cnt
end

function do_postsync(carg)
    local params = __parse_cmdline_params(carg)
    if not params.config then
        print("No input file selected")
        return
    end
    if not params.version and not params.ignore_version then
        print("No version selected")
        return
    end

    local actions = __parse_config_file(params.config)
    if not params.ignore_version then
        actions = __filter_by_version(actions, params.version)
    end
    local select_filter, reboot_filter = __create_filters(actions)
    local devices_query = __create_sql_query(select_filter)
    if not devices_query then
        return
    end
    local connection = require"common.core".connection
    local sql_res = connection:rows(devices_query)
    if not sql_res then
        logger:error("devdb sql query error")
        connection:close()
        return 
    end
    local sync_data = {}
    for dbid, devdata, mac in sql_res do
        local devtable = json.decode(devdata)
        if devtable then
            local sync_command, model = __prepare_sync(dbid, devtable, reboot_filter)
            if sync_command then
                sync_data[#sync_data + 1] = {["mac"] = mac, ["command"] = sync_command, ["model"] = model}
            end
        end
    end
    connection:close()

    if #sync_data > 0 then
        logger:info("Data collected, started sync for %d devices", #sync_data)
        local synced_devices = __do_sync(sync_data)

        local function __format_counts(devdata)
            local rt = {}
            for model, cnt in pairs(devdata) do
                rt[#rt + 1] = model .. " - " .. tostring(cnt)
            end
            return table.concat(rt, ", ")
        end

        sync_counts = __format_counts(synced_devices)
        logger:info("Devices sync done: %s", sync_counts)
    end
end
