package.path    = package.path .. ";/var/www/provsion/?.lua"

local basedir   = "/var/www"
local socket    = require"socket"
local mime      = require"mime"
local helper    = require"common.helper"
local core      = require"common.core"

local env       = nil

DeviceRequestData = {} -- global

-- init logger function 
logger = require"core.logger":init("[WMS 3.0]", 3)

-----------------------------------------------------
--- Loading provision configuration
-----------------------------------------------------

--- Load provision configuration to global space
-- @param p		A pattern for getting configuration value (ex. "general.RequireAuth"
-- @param dflt	A default value used in case parameter was not found
-- @return table with provision configuration to modules and general secrions or matched by pattern value
local function __get_config(p, dflt)
	
	if not _G.provConfData then
		_G.provConfData = {}
		_G.provConfData.cfg, _G.provConfData.err = helper.get_provConfig() or {}
	end
    
	local provConf = _G.provConfData.cfg
	local rv 	   = provConf
	if "string" == type(p) then
		for param in p:gmatch("([^%.]+)%.?") do
			if rv[param] then
				rv = rv[param]
			else
				rv = dflt
				break
			end
		end
	else
		rv = provConf or dflt
	end
	
	return rv, _G.provConfData.err
end

-----------------------------------------------------
--- Response functions
-----------------------------------------------------

local header = {}

--- The main procedure to flush a response content to output buffer
-- @param content	A content string (including binary)
-- @param status	Status content
-- @return response content to output buffer
local function send_response (content, status)
    local t, header = {}, header
    local status = status or "200 OK"
    local content = content or ""

    for k, v in pairs({
        ["Content-Type"] = "text/html",
        ["Status"] = status
    }) do
        if header[k] == nil then
            header[k] = v
        end
    end

    if header["Content-Length"] == nil then
        header["Content-Length"] = string.len(content)
    end

    for k, v in pairs(header) do
        t[#t+1] = k ..": ".. v
    end
    t[#t+1] = "\r\n"
    return io.write(table.concat(t,"\r\n"), content)
end

--- Send 401 answer on the device request
-- @param agent		An agent from the device request 
-- @return response packet with 401 code 
local function send_unauthorized (agent)
	local header = header
	local method = "Basic"
	local realm = "WMSDEVICE"
	header["WWW-Authenticate"] = string.format("%s realm=%q", method, realm)
	header["Status"] = "401 Unauthorized"
	return send_response()
end

--- Send 404 answer on the device request 
-- @return response packet with 404 code 
local function send_notfound ()
	local header = header
	local realm = "WMSDEVICE"
	header["Status"] = "404 Not Found"
	return send_response()
end

local function send_bad_request ()
	local header = header
	header["Status"] = "400 Bad Request"
	return send_response()
end

--- Proxy function to flush to output buffer content of requested filename which depends on device
-- @param agent 	  A table with general device information
-- @param filename 	  A string with filename
-- @param replacement A table with parameters which can be substituted in generated content (optional, only for text mode)
-- @return response on the request 
local function send_file_stream (agent, filename, replacement)
    local h = helper
    local basedirs = {
        basedir .. "/firmwares",
        basedir .. "/images",
        basedir .. "/spoolview/sounds/00000/RingTones",
        basedir .. "/lua/phonebrowser",
    }

    -- Rewirte Wildix ATA to download firmware
    if agent["vendor"] == "Wildix" and agent["vendor:additional"] == "FXS" then
        agent["vendor"] = "wildixata"
    end

    -- Rewirte Welltech based to download firmware
    if agent["vendor"] == "Wildix" and (agent["vendor:additional"] == "FXO" or agent["vendor:additional"] == "W04FXS") then
        agent["vendor"] = "wildixfxx"
    end

    local usdir1 = filename:match("(.*)/" .. agent["vendor"]:lower() .. "/.*$") or ""
    local usdir = filename:match("(.*)/.*$") or ""
    if not h.in_table(usdir, basedirs) and not h.in_table(usdir1, basedirs) then
        return logger:error("Deny to send file %s", filename)
    end

    extension = filename:match("(%w+)$")
    local readmode, mimetype = unpack(({
        ["xml"] = {"r", "application/xml"},
        ["txt"] = {"r", "text/plain"},
        ["cfg"] = {"r", "text/plain"},
        ["zz"] = {"rb", "application/octet-stream"},
        ["gz"] = {"rb", "application/octet-stream"},
        ["dob"] = {"rb", "application/octet-stream"},
        ["rom"] = {"rb", "application/octet-stream"},
        ["ssh"] = {"rb", "application/octet-stream"},
        ["bin"] = {"rb", "application/octet-stream"},
        ["wav"] = {"rb", "application/octet-stream"},
        ["ulaw"] = {"rb", "application/octet-stream"},
        ["lpf"] = {"rb", "application/octet-stream"},
        ["jpg"] = {"rb", "application/octet-stream"}
    })[extension] or {})

    if readmode == nil then
        return logger:error("Does not support %s file extension", extension);
    end

    local fp = io.open(filename, readmode)
    if fp then
        local filesize = h.filesize(filename)
        local content = fp:read("*a")

        logger:info(string.format("File size: %d", filesize))
        header["Content-Type"] = mimetype

        if replacement ~= nil and h.in_table(extension, {"txt","xml","cfg"}) then
            content = string.gsub(content, "%$(%w+)", replacement)
            filesize = string.len(content)
        end

        header["Content-Length"] = filesize
        send_response(content)
        io.close(fp)
    end
end

-----------------------------------------------------
--- Request related functions
-----------------------------------------------------

--- Processing log file for sending to device
-- @param agent 	A table with general device information
-- @param request 	A table with parameters in request string
-- @return logo data stream
local function send_logo (agent, request)
    if request["filename"] then
        local path = string.format(basedir .. "/images/%s", request["filename"])
        return send_file_stream(agent, path)
    end
end

--- Processing wallpapper file for sending to device
-- @param agent 	A table with general device information
-- @param request 	A table with parameters in request string
-- @return wallpaper data stream
local function send_wallpaper (agent, request)
    local h = helper
    local path
    if agent["model"] == "wp600" and h.is_exists(basedir .. "/images/custom_wp600wallpaper.jpg") then
        path = string.format(basedir .. "/images/%s", 'custom_wp600wallpaper.jpg')
    else
        path = string.format(basedir .. "/images/%s", 'wp600wallpaper.jpg')
    end
    return send_file_stream(agent, path)
end

--- Processing screen saver file for sending to device
-- @param agent 	A table with general device information
-- @param request 	A table with parameters in request string
-- @return screen saver data stream
local function send_screensaver (agent, request)
    local h = helper
    local path
    if agent["model"] == "wp600" and h.is_exists(basedir .. "/images/custom_wp600screensaver.jpg") then
        path = string.format(basedir .. "/images/%s", 'custom_wp600screensaver.jpg')
    else
        path = string.format(basedir .. "/images/%s", 'wp600wallpaper.jpg')
    end
    return send_file_stream(agent, path)
end

--- Processing ringtone file for sending to device
-- @param agent 	A table with general device information
-- @param request 	A table with parameters in request string
-- @return stream with ringtone data
local function send_ringtone (agent, request)
    local path = string.format(basedir .. "/spoolview/sounds/00000/RingTones/%s", request["filename"])
    return send_file_stream(agent, path)
end

--- Processing dialnow patterns sending to device
-- @param agent 	A table with general device information
-- @param request 	A table with parameters in request string
-- @return dialnow patterns settings data stream
local function send_dialnow_patterns (agent, request)
    local h = helper
    local xml_content = nil
    if agent["model"] == "wp600" then
    	xml_content = h.dialnow_settings(agent, request)
		if not xml_content then
    		return false
    	end  
    	if "table" == type(xml_content) then
    		xml_content = table.concat(xml_content)
    		return send_response(xml_content)
    	end
    end
end

--- Processing firmware file for sending to device
-- @param agent		A table with general device information
-- @param request 	A table with parameters in request string
-- @return stream with firmware file content or special firmware string 
local function send_firmware (agent, request)
	if not request["filename"] then
		local h, core = helper, core
		
		-- Wildix ATA method
		if agent["vendor"] == "Wildix" and agent["vendor:additional"] == "FXS" then
			local firmwares = core.get_firmware_pack ({["vendor"] = "wildixata"})
			if firmwares and firmwares[agent["model"]] then
				local tcfg = firmwares[agent["model"]]
				local fmw_file = "/firmwares/wildixata/"..(tcfg["application"] or "")
				local content = "No upgrade available"
				if (tcfg["version"] ~= agent["fw:version"] or tcfg["code"] ~= agent["fw:additional"]) and h.is_exists(fmw_file) then
					if agent["model"] == "w01fxs" then
						-- Version:1008030 NAME=342 -- firmwafe_file.gz
						if tcfg["code"]:len() == 6 then -- Version must have 7 digits
							tcfg["code"] = tcfg["code"] .. "0"
						end
						content = string.format("Version:%s NAME:%s\n", tcfg["code"], string.gsub(tcfg["version"],"%.",""))
						
						if "3.44" > tcfg["version"] then  -- 3.44 version UA has correct firmware version: Wildix W01FXS V3.44 <120417>
							-- Set fw:additional to revalidate
    						local connection = core.connection 
    						local sql = "SELECT value FROM autoprov WHERE param=%q AND mac=%q"
    						local cursor = connection:execute(sql, "device", agent["mac"])
    						local device_json = cursor:fetch(); cursor:close()
    						if device_json then
    						    local json   = require"dkjson"
    							local device = json.decode(device_json) 
    			   		 		device["fw:additional"] = "revalidate"
    			   		 		sql = "UPDATE autoprov SET value = '%s' WHERE param = %q AND mac = %q"
    			   		 		connection:execute(sql, json.encode(device), "device", agent["mac"])
    						end
    					end
						
					elseif agent["model"] == "w01fxsr2" or agent["model"] == "w02fxsr2" then
						-- CPU:1309110 CNAME:351309110.ssh	-- firmwafe_file.ssh
						content = string.format("CPU:%s CNAME:%s\n", tcfg["code"], tcfg["application"])
					end
				end
				logger:info("Answer to UPGRD request: %s", content)
				return send_response(content)
			end
		end
	end
	return send_file_stream(agent, request["filename"])
end

--- Processing phone language files
-- @param agent 	A table with general device information
-- @param request 	A table with requested parameters
-- @return stream with language file content
local function send_language (agent, request)
	if "table" ~= type(agent) or "table" ~= type(request) or not request.filename then
		return false
	end
	
	local devTranslations, translationsDir = core.get_translation_pack(agent)
	if devTranslations and translationsDir then
		local lang = request["filename"]:match("lang%-(%w+)%.txt")
		if lang and devTranslations[lang] then
			logger:info("Found %s language pack ...", lang)
			local translationFile = translationsDir..devTranslations[lang]
			return send_file_stream(agent, translationFile)
		end
	end
end

--- Generates provisioning configuration for a device
-- @param agent 	A table with general device information
-- @param request 	A table with requested parameters
-- @return boolean state of the provision action
local function send_aps_configuration (agent, request)
	if not agent["vendor"] then
		return false
	end
	
	local core, h = core, helper
	local connection = core.connection
	
	local modulename = agent["vendor"]:lower()
	if agent["vendor"] == "Wildix" then
		modulename = "wildixp"
		if agent["vendor:additional"] then
			if agent["vendor:additional"] == "WPR2" then
				modulename = "wildixp2"
			elseif agent["vendor:additional"] == "MGW" then
				modulename = "wildixgw"
			elseif agent["vendor:additional"] == "FXS" then
				modulename = "wildixfxs"
			elseif agent["vendor:additional"] == "FXO" then
				modulename = "wildixfxo"
			elseif agent["vendor:additional"] == "FXOR3" then
				modulename = "wildixfxo3"
			elseif agent["vendor:additional"] == "W04FXS" then
				modulename = "wildixfxo"
			elseif agent["vendor:additional"] == "W-AIR" or agent["vendor:additional"] == "W-AIR SB" then
				modulename = "wildixair"
			-- to change this later?
			elseif agent["vendor:additional"] == "DWG" then
				if agent["model"] == "daysaver" then
					modulename = "daysaver"
				else
					modulename = "wildixgsm"
				end
			elseif "FXSR2" == agent["vendor:additional"] or "FXSR3" == agent["vendor:additional"] then
				modulename = "wildixfxs2"
			elseif "W-PA" == agent["vendor:additional"] or "W-PA2" == agent["vendor:additional"] then
				modulename = "wildixpa"
			end
		end
	end
	
	if not h.is_exists(basedir .. "/provision/modules/" .. modulename .. ".lua") then
		logger:error("Can not manage %s device (module not found)", agent["vendor"])
		return false
	end
	
	local module = require("modules." .. modulename)
	module.logger = logger -- Expand the module to make possiblity to log smth
	
	if agent["mac"] then
		-- load global provision configuration
		local gconf, gconf_err = __get_config(modulename, {})
		module.gconf = gconf
		if "table" == type(gconf_err) then
			logger:warning(gconf_err, "gconfig_errors")
		end
		if "table" == type(gconf) then
			logger:debug(gconf, "gconfig")
		end
		local result, username, err = module.configuration(agent, request)
		connection:close() --TODO: check logic

		local configuration
		if type(result) == "table" then
			logger:debug(result, "result")
			configuration = table.concat(result, "\n")
		elseif type(result) == "string" then
			logger:debug(h.split(result,"\n"), "result")
			configuration = result
		end
		
		if err then logger:error(err) end
		logger:debug("Config from <%s> to <%s>", modulename, agent["model"])
		
		agent["action"] = "setaps"
		agent["username"] = username
		send_response(configuration)
		return true
	else
		logger:error("Can't find mac address in user agent")
	end
end

--- Processing features via XML browser by phone request 
-- (like phonebook, history, user features)
-- @param agent 	A table with general device information
-- @param request 	A table with requested parameters
-- @param service	A requested service
-- @return XML content with phonebook data
local function send_xmlcontent (agent, request, service)
	if not agent or not request then
		return false
	end
	
	if not agent["vendor"] or not agent["model"] or not agent["mac"] then
		logger:error("Requested device agent information have bad format")
		return false
	end
	
	local xmlbr       = require"xmlbrowser"
    local xml_content = nil
	
	if "vmaccess" == request["get"] then
		service = "vmaccess"
	end
    
    -- try to process an authorization procedure
    local options = {["lockTimeout"] = __get_config("general.AuthLockTimeout")}
    xml_content = xmlbr.authorization_process(agent, request, service, options)
    if not xml_content then
        if "phonebook" == service then
            xml_content = xmlbr.phonebook_process(agent, request)
        elseif "history" == service then
            xml_content = xmlbr.history_process(agent, request)
        elseif "vmaccess" == service then
            xml_content = xmlbr.vmaccess_process(agent, request)
        elseif "browser" == service then
            xml_content = xmlbr.features_process(agent, request)
        elseif "callgroups" == service then
            xml_content = xmlbr.callgroup_process(agent, request)
        elseif "identities" == service then
            xml_content = xmlbr.identities_process(agent, request)
        end
    end
    if not xml_content then
        return false
    end

    if "table" == type(xml_content) then
        xml_content = table.concat(xml_content)
        return send_response(xml_content)
    elseif "string" == type(xml_content) and "" ~= xml_content then
        return send_response(xml_content)
    end
	
	logger:warning("Unknown service requested - %s", service or "")
	return false
end

local function __new_token()
	local token = require"md5".digest(core.get_random_bytes(6)..tostring(socket.gettime()*10000).."!#%dfTdfMe-GaWilDix_+_SALtt-tt8Udf")
	return token:sub(1,2).."x"..token:sub(4,6).."d"..token:sub(8,9).."l"..token:sub(11,12).."W"..token:sub(14,15)
end

-----------------------------------------------------
--- Authorization process
-----------------------------------------------------

--- Processing phone tocken in DB and request
-- @param token_pass	A token got from DB by MAC
-- @param agent			A table with general device information
-- @param request		A table with requested parameters
-- @param device_pass	A password sent by device
-- @return updated/created token and flag for skipping authorization
local function check_token (token_pass, agent, request, device_pass)
	local h 		 = helper
	local connection = core.connection
	local socket 	 = socket
	local skip_check = false
	
	local new_token   = h.is_new_token(token_pass)
	local new_devpass = h.is_new_token(device_pass)
	if new_token ~= "new" then -- generate new token or save getting from device
		local sql = "UPDATE autoprov SET token=%q WHERE mac=%q AND param=%q"
		if new_devpass == "new" then
			if new_token == "reset" then
				token_pass  = __new_token()
				logger:debug("New token after reset generated for device with mac %s", agent["mac"])
				skip_check  = true
			else
				token_pass = device_pass
				logger:debug("Token was retreived from device: %s", token_pass)
			end
			connection:execute(sql, token_pass, agent["mac"], "device")
		else
			token_pass  = __new_token()
			connection:execute(sql, token_pass, agent["mac"], "device")
			token_pass  = h.get_gpassword()
			skip_check  = true
		end
		if new_token == "old" then
			if new_devpass == "new" then
				logger:info("Auth: got token from device")
			else
				logger:info("Auth: generated token with new format")
			end
		end
	end
	return token_pass, skip_check
end

local function get_db_device(mac)
	local connection = core.connection
	local sql = "SELECT param, value, token FROM autoprov WHERE mac=%q"
	local result = {}
	local cursor = connection:execute(sql, mac)
	repeat
		local row = cursor:fetch({}, "a")
		if not row then
			break
		elseif row["param"] == "device" then
			result["device"] = {["token"] = row["token"], ["value"] = row["value"]}
		elseif row["param"] == "username" then
			result["username"] = row["value"]
		elseif row["param"] == "settings" then
			result["settings"] = row["value"]
		end
	until false
	cursor:close()
	connection:close()
	return result
end

-- Allow w-air devices to receve configuration file only once with default credentials (admin/admin)
-- @params 		json encoded string that contains data from autoprov
function fix_wair_cred (agent, params, dev_pass, wms_pass)
	local rv = false
	if "string" ~= type(params) or "table" ~= type(agent) then
		return rv
	end

	local json = require"dkjson"
	local params_unpacked = json.decode(params)

	local function update_database(params_unpacked, mac)
		local connection = core.connection
		local sql = "UPDATE autoprov SET value = '%s' WHERE mac=%q AND param=%q"
		connection:execute(sql, json.encode(params_unpacked), mac, "device")
		connection:close()
	end

	if "table" ~= type(params_unpacked) or not params_unpacked["mac"] then
		return rv
	end

	if params_unpacked["fixp"] == "true" then
		return rv
	end

	if dev_pass == wms_pass then
		params_unpacked["fixp"] = "true"
		update_database(params_unpacked, agent["mac"])
		logger:info("W-AIR base model: %s mac: %s was succesfully checked", agent["model"], params_unpacked["mac"])
	else
		if params_unpacked["fixp"] == "inprogress" then
			logger:info("W-AIR base model: %s mac: %s checking procedure was skipped", agent["model"], params_unpacked["mac"])
		else
			params_unpacked["fixp"] = "inprogress"
			update_database(params_unpacked, agent["mac"])
		end
		rv = true
	end

	return rv
end

function fix_wair_pass_after_reset (dev_pass, wms_pass, mac)
	local connection = core.connection
	local sql = "SELECT token FROM autoprov WHERE param ='device' and mac=(SELECT mac FROM wair_gateways WHERE network_id=(SELECT network_id FROM wair_gateways WHERE mac=%q) and type=1)"
	local cursor = connection:execute(sql, mac)
	local master_token = cursor:fetch(); cursor:close()
	if dev_pass == master_token then
		local sql = "UPDATE autoprov SET token=%q WHERE mac=%q AND param=%q"
		connection:execute(sql, master_token, mac, "device")
		logger:debug("Device's password will be changed to master password at pbx db: %s", master_token)
		wms_pass = dev_pass
	end
	connection:close()
	return wms_pass
end

--- Check authorization for device
-- @param agent			A table with general device information
-- @param request		A table with requested parameters
-- @param is_provision	true for provision procedure
-- @return boolean value of the authorization checking 
local function is_authorized (agent, request, is_provision)
	local h, mime = helper, mime

	-- processing device auth data
	local dev_user	= ""
	local dev_pass	= ""
	local auth 		= env["HTTP_AUTHORIZATION"] or ""
	if "" ~= auth then
		local basic 	   = auth:match("^Basic%s+(.+)$")
		dev_user, dev_pass = string.match(mime.unb64(basic) or "", "^(.+):(.+)$")
	end
	logger:debug("Auth data at device - %s/%s", dev_user or "", dev_pass or "")

	-- WMS-8094
	if "" == auth and not is_provision and agent["vendor:additional"] == "WPR2" then
		logger:info("XML content requested, unautorize immediately for WPR2 w/o token")
		send_unauthorized(agent)
		return false
	end

	local isMacFound 		           = false
	local db_device = get_db_device(agent["mac"])
	DeviceRequestData["db_device"] = db_device
	local token_pass = nil
	local params = nil
	if db_device["device"] then
		token_pass = db_device["device"]["token"]
		params = db_device["device"]["value"]
	end
	logger:debug(__get_config("general"), "gconfig_general")

	if request.localDevice and __get_config("general.TrustLocalDevice") then
		if require"common.wdmacs".get_family(agent["mac"]) then
			logger:info("Auth: skip mac checking for local Wildix device by provision config!")
			if not db_device["device"] then	-- device not found in database let's add them
				logger:debug("Device is not found in database. Let's add them!")
				connection:execute("INSERT INTO autoprov(param, mac) VALUES('device', '%s')", agent["mac"])
			end
			isMacFound = true 
		else
			logger:warning("Auth: mac is not belong to wildix devices!")
		end
	elseif db_device["device"] then
		logger:info("Auth: mac is present in database")
		isMacFound = true
	end
	if not isMacFound then
   		logger:warning("Auth: mac isn't present in database. Send unauthorized")
   		send_unauthorized(agent)
   		return false
   	end

	-- skip authentication for WP xml browser requests
	if ("getxmlbrowser" == request.action or "getphonebook" == request.action or "gethistory" == request.action) and "WP" == agent["vendor:additional"] then
		logger:debug("Auth: skip checking username and password for WP xml browser requests")
		return true
	end
	
		-- skip authentication for W-AIR xml browser requests
	if "getphonebook" == request.action and (agent["vendor:additional"] == "W-AIR" or agent["vendor:additional"] == "W-AIR SB") then
		logger:debug("Auth: skip checking username and password for W-AIR xml browser requests")
		return true
	end

	-- it's needed to allow w-air devices to receive provisioning config only once with default credentials (admin/admin) (WMS-10529, WMS-10338)
	if (agent["vendor:additional"] == "W-AIR" or agent["vendor:additional"] == "W-AIR SB") and is_provision and "" ~= auth and fix_wair_cred(agent, params, dev_pass, token_pass) then
		check_token(token_pass, agent, request)
		logger:debug("Auth: skip checking username and password for W-AIR request")
		return true
	end
	
	-- skip checking password and login for W01GSM during get firmware
	if agent["model"] == "w01gsm" and request["config"] == "xml" then
		logger:debug("Auth: skip checking username and password for getting firmware by W01GSM")
		return true
	end
	
	-- skip passowrd/login checking for FXSR2/3 in case policy file request
	if agent["model"] and agent["model"]:match("w%d+fxsr%d") and "policy" == request["config"] then
		logger:debug("Auth: skip checking username and password for getting policy file by WXXFXSR2/3")
		return true
	end
	
	-- skip passowrd/login checking for FXOR3 in case policy file request
	if agent["model"] and agent["model"]:match("w%d+fxor%d") and "policy" == request["config"] then
		logger:debug("Auth: skip checking username and password for getting policy file by WXXFXOR3")
		return true
	end

	-- skip passowrd/login checking for daysaver in case policy file request
	if agent["model"] == "daysaver" and "policy" == request["config"] then
		logger:debug("Auth: skip checking username and password for getting policy file by DaySaver")
		return true
	end

	-- device was provisioned, password cleaned in db
	if "null" == h.is_new_token(token_pass) and string.find(params or "", "address") then
		logger:debug("Password for device with mac %s cleaned from wms", agent["mac"])
		token_pass = "reset"
	end

	-- check password and login getting from device with data id database
	local wms_user 			   = "wmsdevice"
	local wms_pass, skip_check = check_token(token_pass, agent, request, dev_pass)
	if skip_check then
		logger:debug("Auth: skip checking username and password")
		request["OldPassword"] = dev_pass or ""
		return true
	else
		logger:debug("Auth data at pbx db - %s/%s", wms_user, wms_pass)
	end
	
	-- checking authorization data
	if "" == auth then
		-- @note W01FXSR2 and WELLTECH devices don't send AUTHORIZATION if auth data are empty
		if (agent["vendor:additional"] == "FXO" or agent["vendor:additional"] == "W04FXS") and require"modules.wildixfxo".is_skip_authorization(agent) then
			logger:debug("Auth: auth header isn't present. But %s in the list of the skipped devices", agent["model"])
			return true
		else
			logger:warning("Auth: auth header isn't present. Send unauthorized")
			send_unauthorized(agent)
			return false
		end
	end

	-- Allow w-air devices to receve configuration after reseting the password. We need to do so, because w-air devices can share master password over radio network.
	if (agent["vendor:additional"] == "W-AIR" or agent["vendor:additional"] == "W-AIR SB") and dev_pass ~= wms_pass then
		wms_pass = fix_wair_pass_after_reset(dev_pass, wms_pass, agent["mac"])
	end
	
	if dev_user ~= wms_user or dev_pass ~= wms_pass then
		logger:warning("Auth: device username or password is incorrect. Send unauthorized")
		send_unauthorized(agent)
		return false
	end
	
	logger:info("Auth: device username and password checked")
	return true
end
-- }}

-----------------------------------------------------
--- Getting environment variables
-----------------------------------------------------

--- Get eviroment server variable
-- @return table A table with enviroment values
local function get_env_table()
	local env_vars =
	{
		'AUTH_TYPE',
		'HTTP_AUTHORIZATION',
		'CONTENT_LENGTH',
		'CONTENT_TYPE',
		'DOCUMENT_ROOT',
		'GATEWAY_INTERFACE',
		'HTTP_ACCEPT',
		'HTTP_COOKIE',
		'HTTP_HOST',
		'HTTP_REFERRER',
		'HTTP_USER_AGENT',
		'HTTPS',
		'PATH',
		'PATH_INFO',
		'PATH_TRANSLATED',
		'REMOTE_ADDR',
		'REMOTE_HOST',
		'REMOTE_IDENT',
		'REMOTE_USER',
		'REQUEST_METHOD',
		'REQUEST_URI',
		'SCRIPT_NAME',
		'SCRIPT_FILENAME',
		'SERVER_ADMIN',
		'SERVER_NAME',
		'SERVER_PORT',
		'SERVER_ADDR',
		'SERVER_PROTOCOL',
		'SERVER_SIGNATURE',
		'SERVER_SOFTWARE',
		'HTTP_X_FORWARDED_FOR',
		'REQUEST_ID'
	}
	local defined_env = {}
	for key, value in pairs(env_vars) do
		local defined = os.getenv(value)
		if defined ~= nil then
			defined_env[value] = defined
		end
	 end
	return defined_env
end

-----------------------------------------------------
--- Parsing user agent and request
-----------------------------------------------------

--- Create request table
-- @return table with _GET paramteres and requested file
local function parse_uri ()
   local h 		 = helper
   local uri 	 = env["REQUEST_URI"] or ""
   local request = {}

   uri = require"core.httpClient".urlDecode(uri:gsub("amp;",""))
   request = h.parse_query_string(uri)
   request["requested"] = uri:match("/([^?]*)") 
   return request
end

--- Fix up UA for W01FXS (Planet) (UA of the FXS is wrong, get it from request)
-- @param agent 	A table with general device information
-- @param request	A table with requested parameters
-- @return string fixed USER_AGENT
local function fix_ata_agent (agent, request)
	if agent:match("^Wildix W01FXS V%d%..*") then	-- fix for W01FXS (old logic) --"Wildix W01FXS V3.43 <1008030>"
		local src_major_ver, src_minor_ver = agent:match("Wildix%sW01FXS%s?V(.*)%s?%<(%d+)%>")
		if not request["requested"] or request["requested"] == "" then
			request["requested"] = "/" .. env["REQUEST_URI"]:match("/?([0-9A-Fa-f]*%.dat)$")
		end
		local minor_ver, major_ver, mac, model= "", "", "", ""
		if request["requested"]:find("wfxs") then --we have provisioning in automatic mode (["REQUEST_URI"] = "/wfxs/w01fxs%20v3.43%201107220/9c75141006b6/9C75141006B6.dat")
			model, mac = string.match(request["requested"],"wfxs%/w01fxs%sv(.*)%/([0-9A-Fa-f]*)%/.*") -- i.e "3.42 1103160", "9c75141006ae"
			if model then
				major_ver, minor_ver = string.match(model, "([^%s][0-9.]+)%s?(%d*)") -- i.e "3.42", "1103160"
			end
		else --we have provisioning in semi-automatic mode (["REQUEST_URI"] = "/9C75141006B6.dat")
			major_ver = string.match(agent,"Wildix%sW01FXS%s?V(.*)%s.*")
			mac = string.match(request["requested"],"([0-9A-Fa-f]*)%.dat") -- i.e "3.42 1103160", "9c75141006ae"
		end
		if minor_ver == "" then
			minor_ver = "1107220"
		end
		if "3.44" > src_major_ver then -- 3.44 version UA has correct firmware version: Wildix W01FXS V3.44 <120417>
			agent = string.format("Wildix W01FXS V%s <%s>", major_ver, minor_ver)
		end
		agent = agent .. " " .. mac
    end
	return agent, request
end

--- Fix mx USER_AGENT string ("Wildix Wildix 3000" fix as "Wildix W02BRI") and check mx mac adress
-- @param agent 	A table with general device information
-- @param request	A table with requested parameters
-- @return string fixed USER_AGENT
local function fix_mx_agent (httpagent, request)
	--TODO: remove after 6 mount (from 20.05.13) hack for Mediatrix with empty useragent
	-- use always if UA = W02BRI/v2.0.20.301 44XX-WX-D2000-65
	if (request.useragent or ""):match("Wildix_.*_[0-9.]*_[0-9A-Fa-f]*") then
		httpagent = request.useragent:gsub("_", " ")
	end
	
	if httpagent then
		if httpagent:find ("Mediatrix") then
			return httpagent
		end
		local device, version, mac
		local family = "WMGW "
		if httpagent:find ("Wildix%sWildix%s*3000") then
			-- Wildix Wildix 3000 2.0.22.320 0090f8067bdb
			version, mac = httpagent:match ("Wildix%sWildix%s*3000%s(.*)%s([0-9A-Fa-f]*).*")
			device = "W2PRI"
		elseif httpagent:find("Wildix%sWildix%sW0%dPRI") then
		    -- Wildix Wildix W02PRI 2.0.38.712 0090f80bc264
		    local ports = nil
		    ports, version, mac = httpagent:match ("Wildix%sWildix%sW0(%d)PRI%s(.*)%s([0-9A-Fa-f]*).*")
            device = string.format("W%sPRI", ports)
            family = "WPRI "
		else
			--Wildix <DEVICE NAME> 2.0.22.320 0090f806d750
			device, version, mac = string.match (httpagent,"Wildix%s(.*)%s(.*)%s([0-9A-Fa-f]*).*")
		end
		if mac and mac:match("^0090f8.*") and version and device then -- 0090f8.* - pattern for Mediatrix gw
			httpagent = family .. device .. " v" .. version .. " " .. mac;
		end
	end
	return httpagent
end

-- Fix WPXXXR2 (Akuvox) phones user-agent string
local function __fix_wpr2_agent(httpagent, request)
	if not httpagent or not request then
		return httpagent
	end
	
	local model, firmware, mac = httpagent:match("^Wildix%s(.*)%s(.*)%s([0-9A-Fa-f]*)$")
	if not model or not firmware or not mac then
		return httpagent
	end
	
	-- fix WP490R2 59.145.5.205 => WP490GR2
	if "WP490R2" == model:upper() and "59.145.5.205" == firmware then
		httpagent = string.format("Wildix WP490GR2 %s %s", firmware, mac)
	-- fix WP410 => WP410R2
	elseif "WP410" == model:upper() then
		httpagent = string.format("Wildix WP410R2 %s %s", firmware, mac) 
	elseif "WP490GR3" == model:upper() and "68.145.1.70" == firmware then
		httpagent = string.format("Wildix WP490GR4 %s %s", firmware, mac)
	end
	
	return httpagent
end

--- Parse USER_AGENT string to discover the device
-- @return table with general device information
local function parse_agent()
	local env, h 		= env, helper
	local request 		= parse_uri()
	local httpagent 	= env["HTTP_USER_AGENT"] or ""
	httpagent			= __fix_wpr2_agent(httpagent, request)
	httpagent 			= fix_mx_agent(httpagent, request)
	httpagent, request 	= fix_ata_agent(httpagent, request)
	logger:info("http agent: %s", httpagent)
	
	local list = {
		-- Wildix WP500 2.2.10.13 00156511da2b
		["wildix wp"] = {".*(Wildix)%s([^%s]*)%s([0-9.][^%s]*)(%s?([0-9a-f:]*))", "Wildix!WP!%2!!%3!!%5"},
		
		-- Wildix W-AIR 02.00.00.05 00087b0841bc
		["w-air"] = {".*(Wildix)%s([^%s]*)%s([0-9.][^%s]*)(%s?([0-9a-f:]*))", "Wildix!W-AIR!%2!!%3!!%5"},
		
		-- Wildix MGW
		["wmgw"] = {".*(WMGW)%s(%w+)%sv([^%s]+)%s([0-9A-F]*)", "Wildix!MGW!%2!!%3!!%4"},
		
		-- Wildix PRI G7
        ["wpri"] = {".*(WPRI)%s(%w+)%sv([^%s]+)%s([0-9A-F]*)", "Wildix!MGW!%2!G7!%3!!%4"},
		
		-- Planet FXS - Wildix W01FXS V3.42 <1008030> (use space in agent name for diff from w01fxsr2
		["w01fxs%s"] = {".*(Wildix)%sW01FXS%sV(.*)%s%<([0-9]*)%>%s([0-9a-f]*)", "Wildix!FXS!w01fxs!!%2!%3!%4"},
		
		-- Welltech FXS - Wildix W01FXSR2 3.7.1404242 0001A8094A28
		["w01fxsr2"] = {".*(Wildix)%sW01FXSR2%s(%d%.%d+)%.(%d+)%s([0-9A-F]*)", "Wildix!FXS!w01fxsr2!!%2!%3!%4"},
		
		-- Welltech FXS - Wildix W02FXSR2 3.21.1404259 9C751432B5EA
		["w02fxsr2"] = {".*(Wildix)%sW02FXSR2%s(%d%.%d+)%.(%d+)%s([0-9A-F]*)", "Wildix!FXS!w02fxsr2!!%2!%3!%4"},
		
		-- Dinstar W01GSM Wildix W01GSM 27.20.01.01 9C7514330001
		["w01gsm"] = {".*(Wildix)%sW01GSM%s(.*)%s([0-9A-F]*)", "Wildix!DWG!w01gsm!!%2!!%3"},

		-- Wildix DaySaver 2.53.4.11 F8A03D59A1C6
		["daysaver"] = {".*(Wildix)%sDaySaver%s(.*)%s([0-9A-F]*)", "Wildix!DWG!daysaver!!%2!!%3"},
		
		-- Wildix W04FXO 2.07a  0001A8082572
		["fxo"] = {".*(Wildix)%s([^%s]*)%s(.*)%s([0-9A-F]*)", "Wildix!FXO!%2!!%3!!%4"},
		
		-- Wildix W04FXS 2.06a  0001A8082572
		["wildix w04fxs%s"] = {".*(Wildix)%sW04FXS%s(.*)%s([0-9A-F]*)", "Wildix!W04FXS!w04fxs!!%2!!%3"},
		
		-- Wildix W24FXSR2 27.18.01.02 9C751437001E
		["w24fxsr2"] = {".*(Wildix)%s(W%d+FXSR2)%s(.*)%s([0-9A-F]*)", "Wildix!FXSR2!%2!!%3!!%4"},
		["w08fxsr2"] = {".*(Wildix)%s(W%d+FXSR2)%s(.*)%s([0-9A-F]*)", "Wildix!FXSR2!%2!!%3!!%4"},

		-- Wildix W02FXSR3 27.19.10.05 9C751437001E
		["w02fxsr3"] = {".*(Wildix)%s(W%d+FXSR3)%s(.*)%s([0-9A-F]*)", "Wildix!FXSR3!%2!!%3!!%4"},
		["w24fxsr3"] = {".*(Wildix)%s(W%d+FXSR3)%s(.*)%s([0-9A-F]*)", "Wildix!FXSR3!%2!!%3!!%4"},
		["w04fxsr3"] = {".*(Wildix)%s(W%d+FXSR3)%s(.*)%s([0-9A-F]*)", "Wildix!FXSR3!%2!!%3!!%4"},

		-- Wildix W-PA2 2.8.3 00d84a03ecf1
		["w-pa2%s"] = {".*(Wildix)%s(W%-PA2)%s(.*)%s([0-9A-F]*)", "Wildix!W-PA2!%2!!%3!!%4"},

		-- Wildix W-PA 2.1.1.3263 0c383e1fbd06
		["w-pa%s"] = {".*(Wildix)%s(W%-PA)%s(.*)%s([0-9A-F]*)", "Wildix!W-PA!%2!!%3!!%4"},

		-- Wildix W04FXOR3
		["w04fxor3"] = {".*(Wildix)%s(W%d+FXOR3)%s(.*)%s([0-9A-F]*)", "Wildix!FXOR3!%2!!%3!!%4"},
	}
	
	local t = {}
	for agent, regexp in pairs(list) do
		if httpagent:lower():find(agent) then
			local a = httpagent:gsub(regexp[1],regexp[2])
			t["vendor"], t["vendor:additional"], -- vendor
			t["model"],  t["model:additional"],  -- model
			t["fw:version"], t["fw:additional"], -- firmware
			t["mac"] = unpack(h.split(a, "!"))
			break
		end
	end
	
	if not t["model"] then -- need to be checked
		logger:error("None valid agent string for supported devices")
		return {}, request
	end
	
	-- update "vendor:additional" to WPR2 for Akuvox phones (WP410R2/WP480R2/WP480GR2/WP490R2/WP490GR2/WP490GCR2/WP490GR4)
	if "WP" == t["vendor:additional"] and t["model"] and (t["model"]:match("WP%w+R[234]")) then
		t["vendor:additional"] = "WPR2"
	end
	
	-- update "model:additional" to v1, v2 ... for WP410R2
	if "WP410R2" == t["model"] then
		t["model:additional"] = require"common.wdmacs".get_version(t["mac"]) or "v1"
	end

	-- WMS-8517 new BRI C7, W04BRI and maybe analogs
	if t["model"]:match("W%d%dBRI") then
		t["model:additional"] = require"common.wdmacs".get_version(t["mac"]) or ""
	end

	-- update "vendor:additional" to W-AIR SB for W-AIR base small business
	if "W-AIR" == t["vendor:additional"] then
		local family = require"common.wdmacs".get_family(t["mac"])
		if family then
			t["vendor:additional"] = family
		end
		-- fixup model for W-AIR SB
		if "W-AIR-SB" == t["model"]:upper() then
			t["model"] = "W-AIR SB"
		elseif "W-AIR SB" == family and "W-AIR" == t["model"] then
			t["model"] = "W-AIR SB"
		end
	end
	
	-- Use forwarded address instead of the portal proxy address
	local forwardedAddr = h.split(env["HTTP_X_FORWARDED_FOR"], ",")
	forwardedAddr = forwardedAddr[#forwardedAddr] 
	
	t["address"] = forwardedAddr or env["REMOTE_ADDR"]	-- add ip address
	t["model"] = t["model"]:lower()		-- model set to lower
	
	-- fix for planet
	if t["mac"] and t["mac"] == "001122334455" then
		t["mac"] = ""
	end
	-- fix mac address
	if t["mac"] and t["mac"] ~= "" then
		t["mac"] = string.gsub(t["mac"], "([0-9A-Fa-f]+)([\.\:\-]*)", "%1"):lower() -- normalize
	else
		logger:error("MAC for device '%s' not found!", httpagent)
		return {}, request
	end
	
	local result = {}
	for k, v in pairs(t) do
		if v ~= "" and v ~= "; " then
			result[k] = v:match("(.-)%s*$")
		end
	end
	
	result["registrar"]      = h.get_registrar(env)
	result["ssl"]            = env["HTTPS"] or "off"
	result["registrar:port"] = env["HTTP_HOST"]:match("\:([0-9]*)") or (env["HTTPS"] and "443" or "80")
	result["alternate_port"] = h.get_alternate_port()
	
	return result, request
end

--- Fix request table depends on device
-- @return table with fixed requested parameters and table with general device information
local function fix_request()
	local agent, request = parse_agent()
	if not request then
		request = {}
	end
	if not request["requested"] then
		request["requested"] = ""
	end
	
	if agent["vendor"] == "Wildix" then
        -- redirect default request from Wildix phone
		if request["requested"]:find ("y" .. string.rep("0",11)) ~= nil then
			request["mode"] = "basic"
			request["action"] = "setaps"
		elseif (agent["model"] == "w-pa" or agent["model"] == "w-pa2") and not request["action"] then
			send_notfound() -- Skip default W-PA request "f0W-PAhw1.100.cfg"
		elseif request["requested"]:find(".cfg") then
			local mac = request["requested"]:match("([0-9a-f]*)\.cfg")
			if mac and not agent["mac"] then
				agent["mac"] = mac
			end
			request["action"] = "setaps"
			
		-- redirect default request from Wildix ATA
		elseif agent["vendor:additional"] == "FXS" then
			if request["requested"]:find ("UPGRD%_ver%.dat") then
				request["action"] = "getfirmware"
		  	elseif request["requested"]:find("([0-9A-Fa-f]*)%.dat$") then
		  		request["action"] = "setaps"
		  		if not agent["mac"] then
		  			local mac = request["requested"]:match("%/([0-9A-Fa-f]*)%.dat")
		  			if mac then
		  				agent["mac"] = mac
		  			end
		  		end
		  	elseif request["requested"]:find("%.ssh$") then
		  		request["action"] = "getfirmware"
		  		request["filename"] = basedir .. "/firmwares/wildixata/" .. request["requested"]:match(".*%/(.*)$")
		  	elseif request["requested"]:find("%.gz$") then
		  		request["action"] = "getfirmware"
		  		request["filename"] = basedir .. "/firmwares/wildixata/" .. request["requested"]:match(".*%/(.*)$")
		  	end
		 
		-- redirect default request from Wildix FXO
		elseif agent["vendor:additional"] == "FXO" or agent["vendor:additional"] == "W04FXS" then
			if request["action"] and request["action"]:match("^(setaps)%/?[0-9A-Fa-f]+%.cfg$") then
				request["action"] = "setaps"	-- fix for uri: actions=setaps/9c751432001d.cfg
			elseif request["requested"]:find("%.bin$") then
				request["action"] = "getfirmware"
				request["filename"] = basedir .. "/firmwares/wildixfxx/" .. request["requested"]:match("(.*)$")
			end
		
		-- fix up request of the policy file
		elseif "FXSR2" == agent["vendor:additional"] or "FXSR3" == agent["vendor:additional"] or "FXOR3" == agent["vendor:additional"] then
			--- XXX fixup incorrect request string "actions=setaps"
			if request.actions then
				request.action  = request.actions
				request.actions = nil
			end
			for _, id in pairs({"18", "19", "81"}) do -- 18 - FXSR2, 19, 81 - FXSR3, FXOR3
				local policy_ptrn = id .. "%/.*%.xml$" 
				if request["action"] and request["action"]:match("^setaps%/" .. policy_ptrn) or request["requested"]:match("^" .. policy_ptrn) then
					request["action"]    = "setaps"
					request["config"]    = "policy"
					request["requested"] = ""
					break
				end
			end
		elseif "daysaver" == agent["model"] then
			request["action"]    = "setaps"
			request["config"]    = "policy"
			request["requested"] = ""
		end
	end
	
	-- check device location by local interfaces
	if agent["address"] then
		request["localDevice"] = helper.host_settings(agent["address"]) and true or false
	end

	return request, agent
end

-----------------------------------------------------
--- Provision implementation
-----------------------------------------------------

logger:info(">> Start")

env = get_env_table()
if env["REQUEST_URI"] and string.len(env["REQUEST_URI"]) > 500 then
    logger:info("REQUEST_URI is too long. Canceling processing.")
    os.exit()
end
logger:debug(env, "environment")

-- Enviroment and request string correcting
local user_agent = env["HTTP_USER_AGENT"]
if user_agent and user_agent:find("WP410") or user_agent:find("WP600") or user_agent:find("WP%w+R[234]") or user_agent:find("WP490G") then
    env["REQUEST_URI"]  = "/" .. helper.split(env["REQUEST_URI"], "/")[1]
end

local request, agent = fix_request()
-- WMS-11802 ISSUE #4
if not agent["address"] then
    local envlist = {"HTTP_USER_AGENT", "HTTP_X_FORWARDED_FOR", "REMOTE_ADDR", "REMOTE_HOST"}
    local envstr = ""
    for _, var in pairs(envlist) do
        envstr = envstr .. string.format(" %s '%s'", var, tostring(env[var]))
    end
    logger:error("Cannot get agent address:%s", envstr)
    send_bad_request()
    os.exit()
end

request["uri"]       = env["REQUEST_URI"]
logger:info(request, "request")
logger:info(agent, "agent")
DeviceRequestData = request

local handler = ({
    ["getfirmware"] = function()
        send_firmware(agent, request)
    end;

    ["setaps"] = function()
        if is_authorized(agent, request, true) then
            send_aps_configuration(agent, request)
        end
    end;

    ["getlang"] = function()
        send_language(agent, request)
    end;

    ["getringtone"] = function()
        send_ringtone(agent, request)
    end;

    ["getxmlbrowser"] = function()
    	if is_authorized(agent, request) then
	        send_xmlcontent(agent, request, "browser")
		end
    end;

    ["getphonebook"] = function()
    	if is_authorized(agent, request) then
        	send_xmlcontent(agent, request, "phonebook")
        end
    end;
    
    ["gethistory"] = function()
    	if is_authorized(agent, request) then
    		send_xmlcontent(agent, request, "history")
    	end
    end;

    ["getlogo"] = function()
        send_logo(agent, request)
    end;

    ["getwallpaper"] = function()
        send_wallpaper(agent, request)
    end;

    ["getscreensaver"] = function()
        send_screensaver(agent, request)
    end;

    ["getdialnowpatters"] = function()
        send_dialnow_patterns(agent, request)
    end;

    ["getcallgroups"] = function()
        if is_authorized(agent, request) then
            send_xmlcontent(agent, request, "callgroups")
        end
    end;
    ["getidentities"] = function()
        if is_authorized(agent, request) then
            send_xmlcontent(agent, request, "identities")
        end
    end;
 })[request["action"]]

if handler ~= nil then
    handler()
    core.connection:close()
else
    send_bad_request()
end
-- }}

os.exit()
