--- XML browser interface
-- 
-- @file	xmlbrowser.lua 
-- @author	Vyacheslav Khudan
-- @created 2014/11/04

module("xmlbrowser", package.seeall)

-----------------------------------------------------
--- Common implemenation details
-----------------------------------------------------
local base_xml_str = "<?xml version=\"1.0\" encoding=\"utf-8\" ?>"

--- Retrieve device user by MAC-address
local function __get_user_by_mac(mac)
    if not mac then
        return nil
    end

    local username = nil
    local request = DeviceRequestData
    if request["db_device"]["device"] then
        username = request["db_device"]["username"]
    else
        local connection = require"common.core".connection
        local sql        = "SELECT value FROM autoprov WHERE param=%q AND mac=%q"
        local cursor     = connection:execute(sql, "username", mac)
        username   = cursor:fetch(); cursor:close()
        connection:close()
    end

    return username
end

--- Get user data (id and language) from LDAP by username
local function __get_user_data(mac, username)
	if not username then
        username = __get_user_by_mac(mac)
    end
	
	if not username or "" == username then
		logger:error("Not found user name")
		return false
	end
	
	if username:find("^unknown") then
		logger:info("Unknown user is found")
		return false
	end
	
	-- get user data from LDAP
	local attrs    = {}
	local ldap_res = require"core.wmsldap":get("telephoneNumber", username)
	if 0 ~= #ldap_res then
		attrs = ldap_res[1][2]
	end
	
	if not attrs["uid"] then
		logger:error("Not found user in LDAP")
		return false
	end
	
	if not attrs["preferredLanguage"] then
		attrs["preferredLanguage"] = require"common.helper".get_admin_lang()
	end
	
	return attrs
end

local function __getUserFeatures(uid, feature)
    if not uid or "" == uid then
        return false
    end
    local rv = false
    
    local userFeatures = require"common.userFeatures.userFeatures":new(uid)
    if not userFeatures then
        logger:info("Error loading user features.")
        return false
    end
    userFeatures:load()

    if feature and "" ~= feature then
        rv = userFeatures:get(feature)
    else
        rv = userFeatures
    end
    
    return rv
end

--- Get user device data (time zone) from wmsdb
local function __getDevData(mac)
	if not mac then
		return false
	end
	local json        = require"dkjson"

    local devSettings = nil
    local request = DeviceRequestData

    if request["db_device"]["device"] then
        devSettings = request["db_device"]["settings"]
    else
        local connection  = require"common.core".connection
        local sql = "SELECT value FROM autoprov WHERE param=%q AND mac=%q"
        local cursor = connection:execute(sql, "settings", mac)
        devSettings = cursor:fetch(); cursor:close()
        connection:close()
    end

	if not devSettings or "" == devSettings then
		logger:debug("Not found device settings")
		return false
	end
	
	return (json.decode(devSettings) or {})["ntp_timezone"]
end

--- Find correct image name for presence field
local function __image_convert(name, family_dev)
	local res_name = name
	if name and "WPR2" == family_dev then 
		name = name:lower()
		if "incoming" == name or "ringing" == name then
			res_name = "RINGING"
		elseif "outgoing" == name then
			res_name = "OUTGOING"
		elseif "missed" == name then
			res_name = "MISSCALL"			
		elseif "offline" == name then
			res_name = "OFFLINE"
		elseif "online" == name then
			res_name = "ONLINE"
		elseif "inuse" == name then
			res_name = "BUSY"
		elseif "ringinginuse" == name then
			res_name = "BUSYRINGING"		
		elseif "donotdistrub" == name then
			res_name = "DND"
		elseif "away" == name then
			res_name = "AWAY"
		end
	end
	
	return res_name
end

local function __getScreenType(model)
    local rv = 43 -- 4.3" 480*272 display - default value
    if string.match(model, "wp410r[23]") or "wp480r2" == model or "wp480gr2" == model then
        rv = 23 -- 2.3" 132*64 display
    elseif "wp480gr3" == model then
        rv = 28 -- 2.8" 320 *240 pixel display
    end
    
    return rv
end

--- Get display width by screen and data type
local function __getScreenWidth(screenType, dataType)
    local rv = 40 --default value
    if "BLF_LABEL" == dataType then
        rv = 23 == screenType and 40 or rv
        rv = 28 == screenType and 75 or rv
        rv = 43 == screenType and 107 or rv
    elseif "HISTORY" == dataType or "VOICEMAIL" == dataType then
        rv = 23 == screenType and 100 or rv
        rv = 28 == screenType and 256 or rv
        rv = 43 == screenType and 420 or rv -- experimental adjusted value
    end

    return rv
end

--- Fill the dictionary with characters in length in pixels
-- @param  model    A device model
-- @return characters length dictionary, screen type
local function __getCharPixDict(screenType)
    local rv = {}
    local utf8 = require"utf8"
    if 23 == screenType then -- 2.3" 132*64 display
        for k, v in pairs({
            [2] = "il",
            [3] = " rÌÎÍìîïí",
            [4] = "ftIÏ()*",
            [5] = "+1234567890EFJLabcdeghjknopqsu-_ßäéöüàèòùâçêëôûÿåáóúñãõþðµ",
            [6] = "vwxyzABCDGHKMNOPQRSTUVWXYZÄÈÉÖÜÒÙÇÊËÔÛŸÓÚÝÑÕýÞÐ",
            [7] = "ÀÂÅÁÃ",
            [8] = "æœm%",
            [9] = "ÆŒ"}) do
            for _, code in utf8.codes(v) do rv[code] = k end
        end
        rv["undefined"] = 7
    elseif 28 == screenType then -- 2.8" 320 *240 pixel display
        for k, v in pairs({
            [4] = "ijl",
            [5] = " []íÍ",
            [6] = "-.,Ifrtìîï()",
            [7] = "{}",
            [8] = "*",
            [9] = "Jkuvxyzþðµ",
            [10] = "+aäcghnsÞ",
            [11] = "1234567890JLbdeopqvçùûüÿúýñÑ",
            [12] = "_FPTYZÌÎÏvàâèéêëòôöåáãóõ",
            [13] = "BCEHKNSUVXÀÄÂßwÅÁÃÐ",
            [14] = "ADPRÈÉÊËÙÛÜÚŸÝ",
            [15] = "GCMOQÇ",
            [16] = "ÒÔÖÓÕm",
            [17] = "",
            [19] = "Wæ%",
            [20] = "œ",
            [21] = "ÆŒ"}) do
            for _, code in utf8.codes(v) do rv[code] = k end
        end
        rv["undefined"] = 17
    else -- 4.3" 480*272 display
        for k, v in pairs({
            [4] = "jl",
            [5] = "i[]",
            [6] = " .,Iftìîïí()ÌÎÏÍ",
            [7] = "-r{}",
            [8] = "*",
            [9] = "zþðµ",
            [10] = "kxÞ",
            [11] = "JbcdghnpqsuvyçùûüÿúýñÑ",
            [12] = "+_1234567890LTaeovàâäèéêëòôöåáãóõ",
            [13] = "AFVZÀÄÂßÅÁÃÐ",
            [14] = "BEHKNPSUXYÈÉÊËÙÛÜÚŸÝ",
            [15] = "CDRÇw",
            [16] = "GOQÒÔÖÓÕm",
            [17] = "M",
            [19] = "æ%",
            [20] = "œ",
            [21] = "WÆŒ"}) do
            for _, code in utf8.codes(v) do rv[code] = k end
        end
        rv["undefined"] = 17
    end

    return rv
end

--- Get string length in pixels
local function __getStringLengthInPixels(strArg, charPixDict, screenType)
    local rv = 0
    if not strArg or "" == strArg then
        return 0
    end
    if not charPixDict or "table" ~= type(charPixDict) then
        charPixDict =  __getCharPixDict(screenType)
    end
    for _, char in require"utf8".codes(strArg) do
        rv = rv + (charPixDict[char] or charPixDict["undefined"])
    end
    return rv
end

--- Fit string to display text in XML Brwoser
-- @param data      A table with elements that should be adjusted
-- @param dataType  A type of the input data
-- @param model           A device model
-- @return fixed string
function fitString(data, dataType, model)
    local rv = ""
    if "table" ~= type(data) or 0 == #data then
        return rv
    end
    
    local utf8 = require"utf8"
    local function _removeLastChar(str)
        if str then
            return utf8.char(utf8.codepoint(str, 1, utf8.offset(str, utf8.len(str) - 1)))
        end
    end

    local screenType  = __getScreenType(model)
    local fitInWidth  = __getScreenWidth(screenType, dataType)
    local charPixDict =  __getCharPixDict(screenType)

    -- processing input data for cutting
    data = "table" == type(data[1]) and data or {data}
    for idx, dataCell in ipairs(data) do
        if dataCell[1] and "" ~= dataCell[1] then   -- skip data with empty values
            local actWidth = fitInWidth
            if dataCell.size then
                actWidth = data[idx].size > 0 and data[idx].size or (fitInWidth - math.abs(data[idx].size))
            elseif 1 < idx and data[idx-1].size then
                actWidth =  data[idx-1].size > 0 and (fitInWidth - data[idx-1].size) or math.abs(data[idx-1].size)
            end

            local strToDisplay   = table.concat(dataCell)
            local strLenInPixels = __getStringLengthInPixels(strToDisplay, charPixDict)
            while strLenInPixels > actWidth do -- remove extra chars
                local strIdxToCut = 1
                while 1 >= string.len(dataCell[strIdxToCut]) do
                    strIdxToCut = strIdxToCut + 1
                end
                dataCell[strIdxToCut] = _removeLastChar(dataCell[strIdxToCut])
                strToDisplay          = table.concat(dataCell)
                strLenInPixels        = __getStringLengthInPixels(strToDisplay, charPixDict)
            end
            if strLenInPixels < actWidth then   -- add spaces to fill string
                local spacesChar  = " "
                local numOfSpaces = (actWidth - strLenInPixels)/__getStringLengthInPixels(spacesChar, charPixDict)
                strToDisplay      = strToDisplay .. string.rep(spacesChar, numOfSpaces)
            end
            rv = rv .. strToDisplay
        end
    end

    return rv
end

--- Formatting date string
local function __formatDate(date, trobj, options)
	local rv = ""
	if not date then
		return rv
	end
	local timeStamp = nil
		  options   = "table" == type(options) and options or {}
	local year, month, day, hour, min = nil
	if "number" == type(date) then
		timeStamp = date
	elseif "string" == type(date) then
		year, month, day, hour, min = date:match("(%d+)%-(%d+)%-(%d+)%s*(%d+):(%d+)")
	end
	if year and month and day and hour and min then
		timeStamp = os.time({year = year, month = month, day = day, hour = hour, min = min})
	end
	
	if not timeStamp then
		return rv
	end

	local dtSettings = options.dtSettings or {}
	local timeZone = dtSettings.timeZone
	local td = require"core.timedate":new(tonumber(timeStamp), timeZone, dtSettings)
	td:toTimeZone() --set time to device local time
	local adaptT = {
	     hideYear      = false
	    ,hideTime      = false
	}

	if "main" == options.menu then
	    adaptT.hideYear = true
	    if string.match(options.model, "wp410r[23]") or "wp480r2" == options.model or "wp480gr2" == options.model or "wp480gr3" == options.model then
	        adaptT.hideTime = true
	    end
	end

    rv = td:formatBySpecOpt(trobj, adaptT)

	return rv
end

--- Get column width in percentage
-- @param model           A device model
-- @param options  An options of the input data
-- @return table with column percentage
local function __getColumnsCfg(model, options)
    local rv = {[1] = 64, [2] = 36}
    if "wp480gr3" == model or"wp490gr3" == model or "wp490gr4" == model then
        rv = {[1] = 67, [2] = 33}
    end
    local err = nil
    repeat
        if not model or "table" ~= type(options) then
            err = "Invalid input data"
            break
        end
        --24h format, leave default settings
        if "1" == options.is24h then
            break
        end
        --AM/PM format set custom columns width
        if "wp480gr3" == model or"wp490gr3" == model or "wp490gr4" == model then
            rv[1] = 60
            rv[2] = 40
        else
            rv[1] = 55
            rv[2] = 45
        end
    until true

    return rv
end

-----------------------------------------------------
--- Phonebook implemenation details
-----------------------------------------------------
--- Build XML message for phonebooks with error message
local function __pb_error_content(err_msg, dev_family, tr)
	local err_msg = err_msg or "pbempty"
	
	local t = {}
	if dev_family:match("^W%-AIR") then
		t.IPPhoneDirectory = {
			DirectoryEntry = {Name = {v = tr:_t("xml." .. err_msg)}},
			Total          = {v = "1"},
			RequestID      = {v = "0"},
			SoftKey        = {["!index"] = "1", Label = {v = tr:_t("btn.Back")}, URI = {v = "SoftKey:Exit"}}
		}
        elseif "WPR2" == dev_family then
                t.IPPhoneDirectory = {
			Title   = {v = tr:_t("xml." .. err_msg)},
			SoftKey = {["!index"] = "4", Label = {v = tr:_t("btn.Back")}, URI = {v = "SoftKey:Exit"}}
		}
        else
		t.IPPhoneDirectory = {
			Title   = {v = tr:_t("xml." .. err_msg)},
			SoftKey = {["!index"] = "1", Label = {v = tr:_t("btn.Back")}, URI = {v = "SoftKey:Exit"}}
		}
	end
	
	--Result serialization
	local wxml_obj = require"core.wxml":new()
	local content  = wxml_obj:generate(t, base_xml_str)
	
	return content	
end

--- Get phonebook name by id
local function __get_pb_name(pb_id, pb_names_set, tr)
	if not pb_id then
		return ""
	end
	
	local pb_name = ""
	if "searchall" == pb_id or "Directory" == pb_id then
		pb_name = tr:_t("xml.pb" .. pb_id:lower()) or ""
	else
		local pb_names_set = pb_names_set or require"common.core".get_phonebook_names({pb_id}) or {}
		pb_name = pb_names_set[pb_id] or ""
	end
	
	return pb_name
end

-- invalid utf8 can break XML serialization
local function __sanitize_utf8(str)
	if "string" ~= type(str) then
		return ""
	end
	return require"common.helper".utf8_validate(str) and str or string.gsub(str, "[\128-\255]", "?")
end

-- decode phonebook names in UTF-8 to W-AIR devices built-in code table
-- table is non-standard (modified Windows-1252), guide page 56 "5.11.3 Character supported"
local function __utf8_decode_wair_name(content)

	local function __utf8_decode_wair(utf8_str)
		local decode_table = {
			[0x011e] = 0x7f, [0x01e6] = 0x7f, [0x20ac] = 0x80, [0x201a] = 0x82,
			[0x0192] = 0x83, [0x017f] = 0x83, [0x201e] = 0x84, [0x2026] = 0x85,
			[0x2020] = 0x86, [0x2021] = 0x87, [0x02c6] = 0x88, [0x0158] = 0x89,
			[0x0160] = 0x8a, [0x2039] = 0x8b, [0x0152] = 0x8c, [0x015e] = 0x8d,
			[0x017d] = 0x8e, [0x011f] = 0x8f, [0x01e7] = 0x8f, [0x00a1] = 0x90, 
			[0x2018] = 0x91, [0x2019] = 0x92, [0x201c] = 0x93, [0x201d] = 0x94,
			[0x25aa] = 0x95, [0x2022] = 0x95, [0x2013] = 0x96, [0x2014] = 0x97,
			[0x02dc] = 0x98, [0x0159] = 0x99, [0x0161] = 0x9a, [0x203a] = 0x9b,
			[0x02c3] = 0x9b, [0x0153] = 0x9c, [0x015f] = 0x9d, [0x017c] = 0x9e,
			[0x0178] = 0x9f, [0x00a1] = 0xa1, [0x00a2] = 0xa2, [0x00a3] = 0xa3,
			[0x00a4] = 0xa4, [0x00a5] = 0xa5, [0x00a6] = 0xa6, [0x00a7] = 0xa7,
			[0x00a8] = 0xa8, [0x016e] = 0xa9, [0x016f] = 0xaa, [0x00ab] = 0xab,
			[0x0114] = 0xac, [0x0115] = 0xad, [0x0147] = 0xae, [0x00af] = 0xaf,
			[0x00b0] = 0xb0, [0x00b1] = 0xb1, [0x010c] = 0xb2, [0x010d] = 0xb3,
			[0x00b4] = 0xb4, [0x00b5] = 0xb5, [0x00b6] = 0xb6, [0x00b7] = 0xb7,
			[0x00b8] = 0xb8, [0x010e] = 0xb9, [0x010f] = 0xba, [0x00bb] = 0xbb,
			[0x0164] = 0xbc, [0x0165] = 0xbd, [0x0148] = 0xbe
		}

		local decoded_str = ""
		for _, code in require"utf8".codes(utf8_str) do
			-- up to 0x7f is standard, 0xbf-0xff is standard Latin-1 Supplement code table
			if code < 0x7f or (code > 0xbe and code < 0x100) then
				decoded_str = decoded_str .. string.char(code)
			else
				decoded_str = decoded_str .. (decode_table[code] and string.char(decode_table[code]) or string.char(0xb0))
			end
		end

		return "<Name>" .. decoded_str .. "</Name>"
	end

	return string.gsub(content, "<Name>(.+)</Name>", __utf8_decode_wair)
end

--- Build XML content for phonebook records to W-AIR devices:
	-- <Name>Support</Name>
	-- <Telephone>444</Telephone>
	-- <Telephone>+380679992211</Telephone>
	-- <Presence>online</Presence>
local function __phonebook_records_wair(params, set_presence)
	local max_numbers    = 3	-- W-AIR phonebook supports only 3 contacts
	local entries_cnt    = 0
	local t              = {IPPhoneDirectory = {}}
	local output 	     = {}
	local total_records  = params["total_records"]
	
	for k, record in ipairs(params["records"]) do
		local num_cnt 		= 0
		t.IPPhoneDirectory["DirectoryEntry_" .. k] = {Name = {v = record.username}}

		if record["extension"] and "" ~= record["extension"] and max_numbers > num_cnt then
			t.IPPhoneDirectory["DirectoryEntry_" .. k]["Telephone_" .. 1] = {v = record.extension}
			num_cnt = 1 + num_cnt
		end
		if record["phone"] and "" ~= record["phone"] and max_numbers > num_cnt then
			t.IPPhoneDirectory["DirectoryEntry_" .. k]["Telephone_" .. 2] = {v = record.phone}
			num_cnt = 1 + num_cnt
		end
		if record["mobile"] and "" ~= record["mobile"] and max_numbers > num_cnt then
			t.IPPhoneDirectory["DirectoryEntry_" .. k]["Telephone_" .. 3] = {v = record.mobile}
			num_cnt = 1 + num_cnt
		end
		if record["office"] and "" ~= record["office"] and max_numbers > num_cnt then
            t.IPPhoneDirectory["DirectoryEntry_" .. k]["Telephone_" .. 4] = {v = record.office}
            num_cnt = 1 + num_cnt
        end
        if record["home"] and "" ~= record["home"] and max_numbers > num_cnt then
            t.IPPhoneDirectory["DirectoryEntry_" .. k]["Telephone_" ..  5]= {v = record.home}
            num_cnt = 1 + num_cnt
        end
        if record["home_mobile"] and "" ~= record["home_mobile"] and max_numbers > num_cnt then
            t.IPPhoneDirectory["DirectoryEntry_" .. k]["Telephone_" .. 6] = {v = record.home_mobile}
            num_cnt = 1 + num_cnt
        end
        
		if record["presence"] and "" ~= record["presence"] then
			if set_presence then
				t.IPPhoneDirectory["DirectoryEntry_" .. k].Presence = {v = record.presence}
			end
		end

		entries_cnt = entries_cnt + 1
	end
	t.IPPhoneDirectory["Total_" .. #params.records + 1] = {v = total_records}
	if params["request_id"] then
	t.IPPhoneDirectory["RequestID_" .. #params.records + 2] = {v = params.request_id}	
	end

	--Result serialization
	local wxml_obj = require"core.wxml":new()
	local content  = wxml_obj:generate(t, base_xml_str)
	
	return __utf8_decode_wair_name(content)
end

--- Build XML content for phonebook records to WP phones:
	-- <Name>Support</Name>
	-- <Telephone>444</Telephone>
	-- <Telephone>+380679992211</Telephone>
local function __phonebook_records_wp(params)
	local t  = {IPPhoneDirectory = {}}
	local tr = params["trobj"]

	-- make XML table
	for k, record in ipairs(params["records"]) do
		t.IPPhoneDirectory["DirectoryEntry_" .. k] = {Name = {v = record.username}}

		if record["extension"] and "" ~= record["extension"] then
			t.IPPhoneDirectory["DirectoryEntry_" .. k].Telephone_1 = {v = record.extension}
		end
		if record["phone"] and "" ~= record["phone"] then
			t.IPPhoneDirectory["DirectoryEntry_" .. k].Telephone_2 = {v = record.phone}
        end
        if record["office"] and "" ~= record["office"] then
        	t.IPPhoneDirectory["DirectoryEntry_" .. k].Telephone_3 = {v = record.office}
        end
		if record["mobile"] and "" ~= record["mobile"] then
			t.IPPhoneDirectory["DirectoryEntry_" .. k].Telephone_4 = {v = record.mobile}
		end
		if record["home"] and "" ~= record["home"] then
			t.IPPhoneDirectory["DirectoryEntry_" .. k].Telephone_5 = {v = record.home}
        end
        if record["home_mobile"] and "" ~= record["home_mobile"] then
        	t.IPPhoneDirectory["DirectoryEntry_" .. k].Telephone_6 = {v = record.home_mobile}
        end

		if k >= params["page"].finish then
			t.IPPhoneDirectory["DirectoryEntry_" .. (k + 1)] = {
				Name      = {v = tr:_t("xml.searchlimit")},
				Telephone = {v = ""}
			}
			break
		end
	end

	-- Result serialization
	local wxml_obj = require"core.wxml":new()
	local content  = wxml_obj:generate(t, base_xml_str)
	
	return content	
end

--- Build XML content for phonebook records to WPR2 phones
	-- <Name>Support</Name>
	-- <Telephone>444</Telephone>
	-- <Telephone>+380679992211</Telephone>
	-- <Presence>online</Presence>
local function __phonebook_records_wpr2(params, set_presence)
	local t  = {}
	local tr = params["trobj"]
	
	local init_url 	   	  = string.format("%s://%s:%s/?action=getphonebook&", params["rproto"], params["raddress"], params["rport"])
	local pb_detail_url   = init_url .. "menu=detl"
	local pb_adv_dial_url = init_url .. "menu=dial"
	local pb_filter_url   = init_url .. "menu=filter"
	local pb_scroll_url   = init_url .. "phonebook="..params.search_filter.."&"

	if params.input ~= "" then
		pb_scroll_url = pb_scroll_url .. "input=" .. params.input .. "&"
	end
	
	local scrollSize = params["page"].size or 200
	local prevIndex  = params["page"].start - scrollSize - 1
	      prevIndex  = prevIndex > 0 and prevIndex or 0
    local nextIndex  = params["page"].finish
          nextIndex  = nextIndex < params["total_records"] and nextIndex or 0
	local pb_scroll_prev  = string.format("paginationIndex=%d&paginationCount=%d", prevIndex, scrollSize) 
	local pb_scroll_next  = string.format("paginationIndex=%d&paginationCount=%d", nextIndex, scrollSize)
    
	-- make XML table
	t.IPPhoneDirectory = { 
         ["!style"]      = "none"
        ,["!Selection"]  = pb_detail_url
        ,["!Selection1"] = pb_scroll_url
        ,["!Next"]       = pb_filter_url
        ,["!Keypad"]     = pb_scroll_url
        ,["!Exit"]       = "Idle"
    }
	
	for k, record in ipairs(params["records"]) do
	    local user_info = ""
	    if record.pbookname and "" ~= record.pbookname then
	       user_info  = user_info .. "&phonebook=" .. (record.pbookname or "")
	    end
	    if record.entryuid and "" ~= record.entryuid then
	       user_info  = user_info .. "&entryuid=" .. (record.entryuid or "")
	    end
        
        t.IPPhoneDirectory["DirectoryEntry_" .. k] = {Name = {v = record.username}}
		
		if record["simple_ext"] and "" ~= record["simple_ext"] then
			t.IPPhoneDirectory["DirectoryEntry_" .. k].Telephone = {v = record.simple_ext}
		else
			t.IPPhoneDirectory["DirectoryEntry_" .. k].URI = {v = pb_adv_dial_url .. user_info}
		end
		if record["presence"] and "" ~= record["presence"] then
			if set_presence then
				t.IPPhoneDirectory["DirectoryEntry_" .. k].Presence = {v = __image_convert(record.presence, "WPR2")}
			end
		end
		t.IPPhoneDirectory["DirectoryEntry_" .. k].Selection = {v = user_info}
		if 1 ~= params["page"].start and 1 == k then
		    t.IPPhoneDirectory["DirectoryEntry_" .. k].Selection1 = {v = pb_scroll_prev}
		end
		if #params["records"] == k and nextIndex ~= 0 then
		    t.IPPhoneDirectory["DirectoryEntry_" .. k].Selection1 = {v = pb_scroll_next}
		end 
	end
	
	-- Soft keys for phonebook content directory
    t.IPPhoneDirectory.SoftKey_1 = {["!index"] = "1", Label = {v = tr:_t("btn.Dial")},   URI = {v = "SoftKey:Dial"}}
    t.IPPhoneDirectory.SoftKey_2 = {["!index"] = "2", Label = {v = tr:_t("btn.Detail")}, URI = {v = "SoftKey:Select"}}
    t.IPPhoneDirectory.SoftKey_3 = {["!index"] = "3", Label = {v = tr:_t("btn.Filter")}, URI = {v = "SoftKey:Next"}}
    t.IPPhoneDirectory.SoftKey_4 = {["!index"] = "4", Label = {v = tr:_t("btn.Exit")},   URI = {v = "SoftKey:Exit"}}
    t.IPPhoneDirectory.SoftKey_5 = {["!index"] = "UP",       URI = {v = "SoftKey:Select1"}}
    t.IPPhoneDirectory.SoftKey_6 = {["!index"] = "UPLONG",   URI = {v = "SoftKey:Select1"}}
    t.IPPhoneDirectory.SoftKey_7 = {["!index"] = "DOWN",     URI = {v = "SoftKey:Select1"}}
    t.IPPhoneDirectory.SoftKey_8 = {["!index"] = "DOWNLONG", URI = {v = "SoftKey:Select1"}}

	--Result serialization
	local wxml_obj = require"core.wxml":new()
	local content  = wxml_obj:generate(t, base_xml_str)
	
	return content
end

--- Phonebook switcher by vendor family in phonebook records
local function __phonebook_records(pb_params, dev_family, set_presence)
    local rv = false
	if "table" ~= type(pb_params) or not dev_family then
		return rv
	end
	
	if "WP" == dev_family then
		rv = __phonebook_records_wp(pb_params)
	elseif "WPR2" == dev_family then
		rv = __phonebook_records_wpr2(pb_params, set_presence)
	elseif dev_family:match("^W%-AIR") then
		rv = __phonebook_records_wair(pb_params, set_presence)
	end
    return rv
end

--- Build XML content for live search records to WPR2 phones
	-- <Name>Support</Name>
	-- <Telephone>444</Telephone>
local function __phonebook_livesearch_records_wpr2(params)
	local t  = {}
	
	-- make XML table
	t.IPPhoneDirectory = {["!style"] = "none"}
    
	for k, record in ipairs(params["records"]) do
		if record["unique_ext"] and "" ~= record["unique_ext"] then
            t.IPPhoneDirectory["DirectoryEntry_" .. k] = {
                 Name      = {v = record["username"]}
                ,Telephone = {v = record["unique_ext"]}
                ,Presence  = {v = __image_convert(record["presence"], "WPR2")}
            }
		end

		if k >= params["page"].finish then
			break
		end
	end
	
	-- Result serialization
	local wxml_obj = require"core.wxml":new()
	local content  = wxml_obj:generate(t, base_xml_str)
	
	return content
end

--- Live search menu
local function __phonebook_livesearch_records(pb_params, dev_family)
    local rv = false
    if "table" ~= type(pb_params) or not dev_family then
		return rv
	end
    
    if "WPR2" == dev_family then
        rv = __phonebook_livesearch_records_wpr2(pb_params)
    end
    return rv
end

--- Build phonebook menu
local function __phonebook_make_pblist_menu(params, user_id)
	if not user_id or not params then
		return false
	end

	local core     = require"common.core"
	local tr   	   = params["trobj"]
	local init_url = string.format("%s://%s:%s/?action=getphonebook", params["rproto"], params["raddress"], params["rport"])
	local srch_url = init_url .. "&menu=search"
	local back_url = init_url .. "&phonebook=all" 
	
	-- make XML table
	local t  = {}
	t.IPPhoneDirectory = {
		["!defaultIndex"] = "1",
		["!style"]        = "none",
		["!Beep"]         = "no",
		["!wrapList"]     = "no",
		["!Timeout"]      = "70",
		["!LockIn"]       = "no",
		["!Selection"]    = init_url,
		["!Selection1"]   = srch_url,
		["!Exit"]         = back_url
	}
	-- get user phonebooks ids and names
	local pb_set = core.get_list_phonebook(user_id) or {}
	table.insert(pb_set, 1, "searchall")
	table.insert(pb_set, 2, "Directory")
	local pb_set_names = core.get_phonebook_names(pb_set)
	
	-- iterate in phonebooks set and create menu items
	for i, pb_id in pairs(pb_set) do
		local pb_name  = __get_pb_name(pb_id, pb_set_names, tr)
		local pb_param = string.format("&phonebook=%s", require"core.httpClient".urlEncode(string.lower(pb_id)))
		t.IPPhoneDirectory["DirectoryEntry_" .. i] = {
			   Name       = {v = pb_name}
			  ,Selection  = {v = pb_param}
			  ,Selection1 = {v = pb_param}
		}
	end

	-- soft keys for phonebook main menu
    t.IPPhoneDirectory.SoftKey_1 = {["!index"] = "1", Label = {v = tr:_t("btn.Enter")},  URI = {v = "SoftKey:Select"}}
    t.IPPhoneDirectory.SoftKey_2 = {["!index"] = "2", Label = {v = tr:_t("btn.Search")}, URI = {v = "SoftKey:Select1"}}
    t.IPPhoneDirectory.SoftKey_4 = {["!index"] = "4", Label = {v = tr:_t("btn.Back")},   URI = {v = "SoftKey:Exit"}}

	--Result serialization
	local wxml_obj = require"core.wxml":new()
	local content  = wxml_obj:generate(t, base_xml_str)
	
	return content
end

--- Creating search menu for phonebook feature
-- @note parameter InputType = "abc"  "123"  "ABC"  "Abc" "1aB"
local function __phonebook_make_search_menu(params, phonebook)
	if not params or not phonebook then
		return false
	end
	
	local tr = params["trobj"]
	local search_val_url = string.format("%s://%s:%s/?action=getphonebook&phonebook=%s", params["rproto"], params["raddress"], params["rport"], require"core.httpClient".urlEncode(phonebook))
	
	-- make XML table
	local t       = {}
	t.IPPhoneInputScreen = {
		["!Timeout"]   = "70",
		["!InputType"] = "abc",
		Title          = {["!wrap"] = "no", v = tr:_t("xml.pbsearchtitle")},
		Prompt         = {v = tr:_t("xml.pbsearch") .. ":"},
		Parameter      = {v = "input"},
		URL            = {v = search_val_url}
	}
    
    t.IPPhoneInputScreen.SoftKey_1 = {["!index"] = "1", Label = {v = tr:_t("btn.Enter")},           URI = {v = "SoftKey:Submit"}}
    t.IPPhoneInputScreen.SoftKey_3 = {["!index"] = "3", Label = {v = tr:_t("btn.Backspace.short")}, URI = {v = "SoftKey:BackSpace"}}
    t.IPPhoneInputScreen.SoftKey_4 = {["!index"] = "4", Label = {v = tr:_t("btn.Back")},            URI = {v = "SoftKey:Exit"}}
                        
    return require"core.wxml":new():generate(t, base_xml_str)
end

--- Common content of the phonebook information about specified record
local function __phonebook_record_info(params, request, menu)
	if not params or not request then
		return false
	end
	
	local max_numbers = 6
	local num_cnt = 0
	local tr	       = params["trobj"]
	local pb_record    = params["records"][1]
	local name         = pb_record["username"] or ""
	local extension    = pb_record["extension"] or ""
	local phone        = pb_record["phone"] or ""
	local mobile       = pb_record["mobile"] or ""
	local office       = pb_record["office"] or ""
	local home         = pb_record["home"] or ""
	local home_mobile  = pb_record["home_mobile"] or ""
	
	-- make XML table
	local t = {}
	t.IPPhoneDirectory = {["!style"] = "none", Title = {["!wrap"] = "no", v = name}}
	local output = {}
	output[#output+1] = '<?xml version="1.0" encoding="utf-8" ?>'
	output[#output+1] = '<IPPhoneDirectory style="none">'
	output[#output+1] = '<Title wrap="no">' .. name .. '</Title>' 
	
	-- name field
	if "advdial" ~= menu then
    	t.IPPhoneDirectory["DirectoryEntry_" .. 1] = {
    		Name = {v = tr:_t("xml.pbdetname") .. ': ' .. name},
        	Telephone = {v = extension}
    	}
    end
	
	-- extension field
    if "" ~= extension and max_numbers > num_cnt then
        t.IPPhoneDirectory["DirectoryEntry_" .. 2] = {
        	Name      = {v = tr:_t("xml.pbdetextension") .. ': ' .. extension},
        	Telephone = {v = extension}
        }
        num_cnt    = 1 + num_cnt
    end
	
	-- phone field
	if "" ~= phone and max_numbers > num_cnt then
		t.IPPhoneDirectory["DirectoryEntry_" .. 3] = {
        	Name      = {v = tr:_t("xml.pbdetphone") .. ': ' .. phone},
        	Telephone = {v = phone}
        }
    	num_cnt    = 1 + num_cnt
    end
    
    -- office field
    if "" ~= office and max_numbers > num_cnt then
		t.IPPhoneDirectory["DirectoryEntry_" .. 4] = {
        	Name      = {v = tr:_t("xml.pbdetoffice") .. ': ' .. office},
        	Telephone = {v = office}
        }
        num_cnt    = 1 + num_cnt
    end
	
	-- mobile field
	if "" ~= mobile and max_numbers > num_cnt then
    	t.IPPhoneDirectory["DirectoryEntry_" .. 5] = {
        	Name      = {v = tr:_t("xml.pbdetmobile") .. ': ' .. mobile},
        	Telephone = {v = mobile}
        }
    	num_cnt    = 1 + num_cnt
    end
    
    -- home field
    if "" ~= home and max_numbers > num_cnt then
    	t.IPPhoneDirectory["DirectoryEntry_" .. 6] = {
        	Name      = {v = tr:_t("xml.pbdethome") .. ': ' .. home},
        	Telephone = {v = home}
        }
        num_cnt    = 1 + num_cnt
    end
    
    -- home_mobile field
    if "" ~= home_mobile and max_numbers > num_cnt then
    	t.IPPhoneDirectory["DirectoryEntry_" .. 6] = {
        	Name      = {v = tr:_t("xml.pbdethome_mobile") .. ': ' .. home_mobile},
        	Telephone = {v = home_mobile}
        }
        num_cnt    = 1 + num_cnt
    end
	
	-- Soft keys for phonebook content directory
        t.IPPhoneDirectory.SoftKey_1 = {["!index"] = "1", Label = {v = tr:_t("btn.Dial")}, URI = {v = "SoftKey:Dial"}}
	t.IPPhoneDirectory.SoftKey_4 = {["!index"] = "4", Label = {v = tr:_t("btn.Back")}, URI = {v = "SoftKey:Exit"}}

	--Result serialization
	local wxml_obj = require"core.wxml":new()
	local content  = wxml_obj:generate(t, base_xml_str)
	
	return content
end

--- Creating content of the phonebook menu with user details
local function __phonebook_make_detail_menu(params, request)
	return __phonebook_record_info(params, request)
end

--- Advansed dial menu in case phonebook user has two numbers
local function __phonebook_make_advdial_menu(params, request)
	return __phonebook_record_info(params, request, "advdial")
end

-----------------------------------------------------
--- History implemenation details
-----------------------------------------------------
--- Update history in CDR backend
local function __history_set_cdr_info(in_data, action)
    if "table" ~= type(in_data) or not action then
        return false
    end
    
    local user_num = in_data["user_num"]
    local row_id   = in_data["row_id"]
    local direct   = in_data["call_direct"]
    if not user_num or not row_id then
        return false
    end
    
    require"common.cdrconnection"
    local cdr_conn = require"core.pbxutils".tcopy(cdrconnection)
    local cdr_backend, cdr_table = cdr_conn:init()
    if not cdr_backend or not cdr_table then
        logger:error("CDR backend or table not found")
        return false
    end
    
    local sql_query = ""
    if "archive" == action then
        local archived_field = "out" == direct and "from_archived" or "to_archived"
        if "tds" == cdr_backend then
            sql_query = string.format("UPDATE %s SET %s='1' WHERE rowid='%d'", cdr_table, archived_field, row_id)
        else
            sql_query = string.format("UPDATE %s SET %s='1' WHERE rowid='%d'", cdr_table, archived_field, row_id)
        end
    end
    
    -- executing query
    local sql_rv = cdr_conn:execute(sql_query)
    if not sql_rv then
        logger:error("SQL executing error")
    end
    
    return sql_rv
end

--- Get history from CDR backend
local function __history_get_cdr_info(in_data, filter)
	if not in_data or "table" ~= type(in_data) then
		return false
	end
	
	local user_num = in_data["user_num"]
	local row_id   = in_data["row_id"] or 0
	if not user_num then
		return false
	end
	
	-- check and load CDR backend
	os.setlocale('en_US.UTF-8', 'ctype') -- to get correct special symbols

	require"common.cdrconnection"
	local pbxutils = require"core.pbxutils"
	local cdr_conn = pbxutils.tcopy(cdrconnection)
	local cdr_backend, cdr_table = cdr_conn:init()
	if not cdr_backend or not cdr_table then
		logger:error("CDR backend or table not found")
		return false, "histnotactive"
	end
	
	local sql_query = ""
	local sql_where = ""
	if "detail" == filter then	-- select history info about one call
		sql_where = string.format("WHERE rowid=%d AND ((c_from='%s' AND from_archived='0') OR (c_to='%s' AND to_archived='0'))", row_id, user_num, user_num)
		if "tds" == cdr_backend then
			sql_query = string.format("SELECT [src], [dst], [c_from], [c_to], CONVERT(NVARCHAR, [from_name]) AS [from_name], CONVERT(NVARCHAR, [to_name]) AS [to_name], CONVERT(VARCHAR, [start], 120) AS [start], [duration], [dest_type], [lastapp], [organization] FROM %s %s", cdr_table, sql_where)
		else
			sql_query = string.format("SELECT src, dst, c_from, c_to, from_name, to_name, start, duration, dest_type, lastapp, organization FROM %s %s", cdr_table, sql_where)
		end
	else
		-- prepare sql query for user history
		local sql_limit	= 100	-- 100 records will be selected
		local pbx_serial = pbxutils.pbxSerial()
    	if "all" == filter then
    		sql_where = sql_where .. string.format(" WHERE (((c_to='%s' AND callclass<>'5' AND to_archived='0') OR (c_from='%s' AND from_archived='0'))) AND pbxid='%s'", user_num, user_num, pbx_serial)
    	elseif "missed" == filter then
    		sql_where = sql_where .. string.format(" WHERE c_to='%s' AND to_archived='0' AND c_from<>'%s' AND callclass<>'5' AND (disposition='NO ANSWER' OR disposition='BUSY') AND faxdata='' AND pbxid='%s'", user_num, user_num, pbx_serial)
    	end
    	
    	if "tds" == cdr_backend then
    		sql_query = string.format("SELECT TOP %d [rowid], [src], [dst], [c_from], [c_to], CONVERT(NVARCHAR, [from_name]) AS [from_name], CONVERT(NVARCHAR, [to_name]) AS [to_name], CONVERT(VARCHAR, [start], 120) AS [start], [disposition], [lastapp], [organization] FROM %s %s ORDER BY rowid DESC", sql_limit, cdr_table, sql_where)
    	else
    		sql_query = string.format("SELECT rowid, src, dst, c_from, c_to, from_name, to_name, start, disposition, lastapp, organization FROM %s %s ORDER BY rowid DESC LIMIT %d", cdr_table, sql_where, sql_limit)
    	end
    end

	-- executing query
	local cursor = cdr_conn:execute(sql_query)
	if not cursor or type(cursor) == "boolean" then
		cdr_conn:close()
		logger:error("SQL executing error")
		return false, "histnotaccess"
	end
	
	-- processing result data
	local fromCharSet = "mysql" == cdr_backend and "iso-8859-1" or "utf8"
	local codeSet     = require"core.codeSet":new(fromCharSet, "utf8")
	
	local result = {}
	-- WMS-5497 use positional result
	local mode = ("sqlite3" == cdr_backend and "detail" ~= filter) and "n" or "a"
	local numbered_fields = {"rowid", "src", "dst", "c_from", "c_to", "from_name", "to_name", "start", "disposition", "lastapp", "organization"}
	local row 	 = cursor:fetch({}, mode)
	while row do
		if "n" == mode then
			for i, fname in pairs(numbered_fields) do
				row[fname] = row[i]
			end
		end
		
		-- set call type  
		if user_num == row["c_from"] then
			row["type"] = "outgoing"
		elseif "NO ANSWER" == row["disposition"] or ("Queue" == row["lastapp"] and "BUSY" == row["disposition"]) then
			row["type"] = "missed"
		else
			row["type"] = "incoming"
		end
        
        row.src       = codeSet:conv(row.src)
        row.c_from    = codeSet:conv(row.c_from)
        row.from_name = __sanitize_utf8(codeSet:conv(row.from_name))
        row.dst       = __sanitize_utf8(codeSet:conv(row.dst))
        row.c_to      = __sanitize_utf8(codeSet:conv(row.c_to))
        row.to_name   = __sanitize_utf8(codeSet:conv(row.to_name))
        row.organization = __sanitize_utf8(codeSet:conv(row.organization))
		
		result[#result + 1] = row
		row = cursor:fetch({}, mode)
	end

	cursor:close()
	cdr_conn:close()
	
	if 0 == #result then
		return false, "histemptylist"
	end
	
	return result
end

--- Build XML error conter for history request
local function __history_error_content(err_msg, tr)
	local err_msg = err_msg or "histempty"
	
	local t = {}
	if "histemptylist" == err_msg then
		t.IPPhoneDirectory  = {Title = {["!wrap"] = "no", v = tr:_t("xml.histempty")}}
	else
		t.IPPhoneTextScreen = {Title = {v = tr:_t("features.Error")},
		                       Text  = {["!wrap"] = "yes", v = tr:_t("xml." .. err_msg)}}
	end
	
	--Result serialization
	local wxml_obj = require"core.wxml":new()
	local content  = wxml_obj:generate(t, base_xml_str)
	
	return content	
end

--- Build XML content with user history
local function __history_make_content(history_t, params)
	if not history_t or "table" ~= type(history_t) or not params or "table" ~= type(params) then
		return false
	end
	
	local tr = params["trobj"]
	local hist_next_type = "all" == params["htype"] and "missed" or "all"
	local opts = {
        model      = params.model,
        menu       = "main",
        dtSettings = params.dtSettings
    }

	-- set urls for other requests
	local init_url      = string.format("%s://%s:%s/", params["rproto"], params["raddress"], params["rport"])
	local hist_type_url = init_url .. "?action=gethistory&type=" .. hist_next_type
--	local hist_opt_url  = init_url .. "?action=gethistory&menu=option&data=" --TODO: with using __history_make_option_menu
	local hist_det_url  = init_url .. "?action=gethistory&menu=detail&type=" .. params["htype"] .."&rid="
	
	-- make XML table
	local t = {}
	local columns = __getColumnsCfg(params.model, params.dtSettings)
	t.IPPhoneDirectory = {["!style"] = "none", ["!Next"] = hist_type_url, ["!Selection"] = hist_det_url 
	                     ,["!Exit"] = "Idle", ["!Col1"] = columns[1], ["!Col2"] = columns[2]}
	
	for k, row in pairs(history_t) do
		local rid	 = row["rowid"]
		local name   = nil
		local number = nil
		if "outgoing" == row["type"] then
			number = row["c_to"]
			name   = row["to_name"]
			if not name or "" == name then
				name = row["c_to"]
			end
		else
			number = row["c_from"]
			name   = row["from_name"]
			if not name or "" == name then
				name = row["c_from"]
			end
			if row["organization"] and "" ~= row["organization"] then
				name = string.format("%s (%s)", name, row["organization"])
			end
		end

		local history_presence = __image_convert(row["type"], params["dfamily"])		
		t.IPPhoneDirectory["DirectoryEntry_" .. k] = {
             Text1      = {v = name}
            ,Text2      = {v = __formatDate(row.start, tr, opts)}
		    ,Telephone  = {v = number}
		    ,Presence   = {v = history_presence}
		    ,Selection  = {v = rid}
		}
	end

	-- soft keys for history content 
    t.IPPhoneDirectory.SoftKey_1 = {["!index"] = "1", Label = {v = tr:_t("btn.Dial")}, URI = {v = "SoftKey:Dial"}}
    t.IPPhoneDirectory.SoftKey_2 = {["!index"] = "2", Label = {v = tr:_t("btn.Detail")}, URI = {v = "SoftKey:Select"}}
    t.IPPhoneDirectory.SoftKey_3 = {["!index"] = "3", Label = {v = tr:_t("btn." .. ("all" == hist_next_type and "All" or "Missed"))}, URI = {v = "SoftKey:Next"}}
    t.IPPhoneDirectory.SoftKey_4 = {["!index"] = "4", Label = {v = tr:_t("btn.Back")}, URI = {v = "SoftKey:Exit"}}
	
	--Result serialization
	local wxml_obj = require"core.wxml":new()
	local content  = wxml_obj:generate(t, base_xml_str)
	
	return content
end

--- Build XML for option menu in History feature
-- @note will be used when system can modify phonebook records
local function __history_make_option_menu(opt_data, params)
	if not opt_data or not params then
		return false
	end
	
	local rid, pbnum = opt_data:match("(.*)_(.*)")
	if not rid or not pbnum then
		logger:debug("Invalid option data")
		return false, "histempty"
	end
	
	local tr = params["trobj"]
	local init_url = string.format("%s://%s:%s/", params["rproto"], params["raddress"], params["rport"])	
	
	-- make XML table
	local t = {}
	t.IPPhoneTextMenu = {["!defaultIndex"] = "1", ["!style"] = "none", ["!Beep"] = "no", ["!wrapList"] = "no", ["!Timeout"] = "70", ["!LockIn"] = "no",
						 ["Title"] = {["!wrap"] = "no", v = tr:_t("btn.Option")}}
	-- prompt for detail
	local hist_detail_url = init_url .. "?action=gethistory&menu=detail&rid=" .. rid
	t.IPPhoneTextMenu.MenuItem = {
		Prompt = {v = tr:_t("btn.Detail")},
		URI    = {v = hist_detail_url}
	}
	 
	-- prompt for phonebook actions
	--TODO: make search in PB. If present then view in PB otherwise add to PB
--	local hist_pbact_url = init_url .. "?action=gethistory&menu=pbshow&num=" .. pbnum --TODO: url
--	t.IPPhoneTextMenu.MenuItem = {
--		Prompt = {v = "Add to Phonebook"}, --TODO: translation
--		URI    = {v = hist_pbact_url}
--	}
	
        -- soft keys for history option menu
        t.IPPhoneTextMenu.SoftKey_1 = {["!index"] = "1", Label = {v = tr:_t("btn.Enter")}, URI = {v = "SoftKey:Select"}}
        t.IPPhoneTextMenu.SoftKey_4 = {["!index"] = "4", Label = {v = tr:_t("btn.Back")},  URI = {v = "SoftKey:Exit"}}
	
	--Result serialization
	local wxml_obj = require"core.wxml":new()
	local content  = wxml_obj:generate(t, base_xml_str)
	
	return content
end

--- create XML content with detail call information
local function __history_make_detail_menu(rid, user_num, params)
	if not rid or not rid:find("%d+") or not user_num or not params or "table" ~= type(params) then
		return false
	end
	
	if not params["trobj"] then
		return false
	end
	
	local tr = params["trobj"]
	local opts = {
           model      = params.model,
           menu       = "detail",
           dtSettings = params.dtSettings
    }
    
	-- translave direction digit to string
	local function _convert_direction(code)
		local dt = "other"
		code = tostring(code)
		if "2" == code then
			dt = "inc"
		elseif "3" == code then
			dt = "out"
		elseif "4" == code then
			dt = "local"
		end
		
		return tr:_t("xml.histdirect" .. dt)
	end
	
	-- get information from CDR
	local in_data = {
		["row_id"]   = rid,
		["user_num"] = user_num
	}
	local hist_info, err = __history_get_cdr_info(in_data, "detail")
	if not hist_info or not hist_info[1] then
		return false, err
	end
	
	-- parsing CDR information and creating detail information
	local hist_info = hist_info[1]
	local number    = nil
	local name      = nil
	local direct    = nil
	if "outgoing" == hist_info["type"] then
	    direct = "out"
		number = hist_info["c_to"]
		name   = hist_info["to_name"]
		if not name or "" == name then
			name = hist_info["c_to"]
		end
	else
	    direct = "inc"
		number = hist_info["c_from"]
		name   = hist_info["from_name"]
		if not name or "" == name then
			name = hist_info["c_from"]
		end
	end
	local direction = _convert_direction(hist_info["dest_type"])

	local info_data = {}
	table.insert(info_data, string.format("%s: %s", tr:_t("xml.histdetname"), name and "" ~= name and name or "Unknown name"))
	if hist_info["organization"] and "" ~= hist_info["organization"] then
		table.insert(info_data, string.format("%s: %s", tr:_t("faxcover.Company"), hist_info["organization"]))
	end
	table.insert(info_data, string.format("%s: %s", tr:_t("xml.histdetnumber"), number or ""))
	table.insert(info_data, string.format("%s: %s", tr:_t("xml.histdettime"), __formatDate(hist_info["start"], tr, opts)))
	table.insert(info_data, string.format("%s: %s", tr:_t("xml.histdetduration"), os.date("!%X", tonumber(hist_info["duration"]) or 0)))
	table.insert(info_data, string.format("%s: %s", tr:_t("xml.histdetdirect"), direction))

	-- make XML table
	local t     = {}
	local title = name and "" ~= name and name or number
	
	local init_url         = string.format("%s://%s:%s/", params["rproto"], params["raddress"], params["rport"])
    local hist_archive_url = init_url .. "?action=gethistory&menu=archive&type="..params["htype"].."&rid="..rid.."&direct="..direct
	
	t.IPPhoneTextMenu = {["!defaultIndex"] = "1", ["!style"] = "none", ["!Beep"] = "no", ["!wrapList"] = "no", ["!Timeout"] = "70", ["!LockIn"] = "no",
						 ["Title"] = {["!wrap"] = "no", v = title}, ["!Next"] = hist_archive_url}
	
	-- making items for detail information
	for k, value in ipairs(info_data) do
		t.IPPhoneTextMenu["MenuItem_" .. k] = {Prompt = {v = value}, Dial  = {v = number}}
	end
	
    -- soft keys for history detail menu
    t.IPPhoneTextMenu.SoftKey_1 = {["!index"] = "1", Label = {v = tr:_t("btn.Dial")},    URI = {v = "SoftKey:Dial"}}
    t.IPPhoneTextMenu.SoftKey_3 = {["!index"] = "3", Label = {v = tr:_t("btn.Archive")}, URI = {v = "SoftKey:Next"}}
    t.IPPhoneTextMenu.SoftKey_4 = {["!index"] = "4", Label = {v = tr:_t("btn.Back")},    URI = {v = "SoftKey:Exit"}}
	
	-- Result serialization
	local wxml_obj = require"core.wxml":new()
	local content  = wxml_obj:generate(t, base_xml_str)
	
	return content
end

-----------------------------------------------------
--- Voicemail implemenation details
-----------------------------------------------------
--- Build XML error conter for vmaccess request
local function __vm_error_content(err_msg, tr)
	local err_msg = err_msg or "vmemptylist"
	local t = {}
	if "vmemptylist" == err_msg then
		t.IPPhoneDirectory = {
			["!style"] = "none", ["!Exit"] = "Idle",
			Title = {["!wrap"] = "no", v = tr:_t("xml.no_messages")},
			SoftKey = {["!index"] = "4", Label = {v = tr:_t("btn.Back")}, URI = {v = "SoftKey:Exit"}}
		}
	else
		t.IPPhoneTextScreen = {
			Title = {v = tr:_t("features.Error")},
		    Text  = {["!wrap"] = "yes", v = tr:_t("xml." .. err_msg)},
		    SoftKey = {["!index"] = "4", Label = {v = tr:_t("btn.Back")}, URI = {v = "SoftKey:Exit"}}
		}
	end

	--Result serialization
	local wxml_obj = require"core.wxml":new()
	local content  = wxml_obj:generate(t, base_xml_str)

	return content
end

--- Build XML content with voicemail
local function __vm_make_content(params, cont)
	if not params or "table" ~= type(params) or not cont or "table" ~= type(cont)then
		return false
	end
	
	local h        = require"common.helper"
	local tr       = params.trobj
	local options  = {format="short", model=params.model}
	local opts = {
        model      = params.model,
        menu       = "main",
        dtSettings = params.dtSettings
    }
	
	-- set urls for other requests
	local init_url		= string.format("%s://%s:%s/", params["rproto"], params["raddress"], params["rport"])
	local vm_detail_url
	local cg_manager = false
	if params["user"] then
		vm_detail_url	= string.format("%s?action=getxmlbrowser&get=vmaccess&menu=detail&user=%s&id=", init_url, params["user"])
		local cg_managers = h.get_cg_managers_cache()
		cg_manager = cg_managers[params["user"]]
	else
		vm_detail_url	= string.format("%s?action=getxmlbrowser&get=vmaccess&menu=detail&id=", init_url)
	end
	
	-- make XML table
	local t = {}
	local columns = __getColumnsCfg(params.model, params.dtSettings)
	t.IPPhoneDirectory = {["!style"] = "none", ["!Selection"] = vm_detail_url, ["!Exit"] = "Idle", ["!Dial1"] = "Telephone1", ["!Col1"] = columns[1], ["!Col2"] = columns[2]}

	if cont then
		for k, msg in ipairs(cont) do
			if "table" == type(msg) then
				local timeStamp    = tonumber(msg.origtime)
                local VMStr        = ("new" == msg.type and "*" or " ") .. " " .. (msg.name or msg.number)
				t.IPPhoneDirectory["DirectoryEntry_" .. k] = {
					Text1      = {v = VMStr},
					Text2      = {v = __formatDate(timeStamp, tr, opts)},
					Telephone  = {v = msg.number},
					Telephone1 = not cg_manager and {v = "vmplay" .. msg.mailbox .. msg.type .. msg.msg_file} or nil,
					Selection  = {v = msg.msgId},
					Presence = cg_manager and {v = msg.missed and __image_convert("missed", "WPR2") or ""} or nil
				}
			end
		end
	end	
	
	-- soft keys for vmaccess content 
	if not cg_manager then
		t.IPPhoneDirectory.SoftKey_1 = {["!index"] = "1", Label = {v = tr:_t("btn.Play")}, URI = {v = "SoftKey:Dial1"}}
	end
	t.IPPhoneDirectory.SoftKey_2 = {["!index"] = "2", Label = {v = tr:_t("btn.Dial")}, URI = {v = "SoftKey:Dial"}}
	t.IPPhoneDirectory.SoftKey_3 = {["!index"] = "3", Label = {v = "Info"},            URI = {v = "SoftKey:Select"}}
	t.IPPhoneDirectory.SoftKey_4 = {["!index"] = "4", Label = {v = tr:_t("btn.Back")}, URI = {v = "SoftKey:Exit"}}
	
	--Result serialization
	local wxml_obj = require"core.wxml":new()
	local content  = wxml_obj:generate(t, base_xml_str)
	
	return content
end

--- create XML content with detail call information
local function __vm_make_detail_menu(params, cont)
	if not params or "table" ~= type(params) or not cont or "table" ~= type(cont) then
		return false
	end
	
	local tr        = params["trobj"]
	local timeZone  = params.timeZone
	local timeStamp = cont.origtime
	local opts = {
           model      = params.model,
           menu       = "detail",
           dtSettings = params.dtSettings
    }
	
	-- make XML table
	local voice_msg = not cont.missed
	local t = {}
	local init_url		= string.format("%s://%s:%s/", params["rproto"], params["raddress"], params["rport"])
	local vm_selection_url, vm_next_type_url, vm_previous_url
	if params["user"] then 
		vm_selection_url = string.format("%s?action=getxmlbrowser&get=vmaccess&menu=detail&user=%s&id=", init_url, params["user"])
		vm_next_type_url = string.format("%s?action=getxmlbrowser&get=vmaccess&type=%s&menu=delete&user=%s&id=", init_url, cont["type"], params["user"])
		vm_previous_url = string.format("%s?action=getxmlbrowser&get=vmaccess&menu=main&user=%s", init_url, params["user"])
	else
		vm_selection_url = string.format("%s?action=getxmlbrowser&get=vmaccess&menu=detail&id=", init_url)
		vm_next_type_url = string.format("%s?action=getxmlbrowser&get=vmaccess&type=%s&menu=delete&id=", init_url, cont["type"])
		vm_previous_url = string.format("%s?action=getxmlbrowser&get=vmaccess&menu=main", init_url)
	end

	t.IPPhoneDirectory = {["!style"] = "none", ["!Next"] = vm_next_type_url .. cont.msgId, ["!Selection"] = vm_selection_url, ["!Previous"] = vm_previous_url, ["!Exit"] = "Idle"}
	if voice_msg then
		t.IPPhoneDirectory["!Dial1"] = "Telephone1"
	elseif cont["type"] == "new" then
		local proceed_url = string.format("%s?action=getxmlbrowser&get=vmaccess&type=%s&menu=proceed&user=%s&id=", init_url, cont["type"], params["user"])
		t.IPPhoneDirectory["!Selection1"] = proceed_url .. cont.msgId
	end

	-- making items for detail information
	t.IPPhoneDirectory.DirectoryEntry_1 = {
		Name       = {v = tr:_t("xml.histdetname") .. ": " .. (cont.name or cont.number)},
		Telephone  = {v = cont.number},
		Telephone1 = voice_msg and {v = "vmplay" .. cont.mailbox .. cont.type .. cont.msg_file} or nil,
		Selection  = {v = cont.msgId}
	}
	t.IPPhoneDirectory.DirectoryEntry_2 = {
		Name       = {v = tr:_t("xml.histdetnumber") .. ": " .. cont.number},
		Telephone  = {v = cont.number},
		Telephone1 = voice_msg and {v = "vmplay" .. cont.mailbox .. cont.type .. cont.msg_file} or nil,
		Selection  = {v = cont.msgId}
	}
	t.IPPhoneDirectory.DirectoryEntry_3 = {
		Name       = {v = tr:_t("str.Date") .. ": " .. __formatDate(timeStamp, tr, opts)},
		Telephone  = {v = cont.number},
		Telephone1 = voice_msg and {v = "vmplay" .. cont.mailbox .. cont.type .. cont.msg_file} or nil,
		Selection  = {v = cont.msgId}
	}
	if voice_msg then
		t.IPPhoneDirectory.DirectoryEntry_4 = {
			Name       = {v = tr:_t("xml.histdetduration") .. ": " .. cont["duration"] .. " " .. tr:_t("str.seconds")},
			Telephone  = {v = cont.number},
			Telephone1 = {v = "vmplay" .. cont.mailbox .. cont.type .. cont.msg_file},
			Selection  = {v = cont.msgId}
		}
		t.IPPhoneDirectory.DirectoryEntry_5 = {
			Name       = {v = tr:_t("xml.vmFwdDirectoryEntry")},
			Telephone  = {v = cont.number},
			Telephone1 = {v = "vmplay" .. cont.mailbox .. cont.type .. cont.msg_file},
			Selection  = {v = cont.msgId .. "&forward=1"}
		}
	end

	-- soft keys for voicemail detail menu
	if voice_msg then
		t.IPPhoneDirectory.SoftKey_1 = {["!index"] = "1", Label = {v = tr:_t("btn.Play")},   URI = {v = "SoftKey:Dial1"}}
	elseif cont["type"] == "new" then
		t.IPPhoneDirectory.SoftKey_1 = {["!index"] = "1", Label = {v = tr:_t("btn.Read")}, URI = {v = "SoftKey:Select1"}}
	end
	t.IPPhoneDirectory.SoftKey_2 = {["!index"] = "2", Label = {v = tr:_t("btn.Dial")},   URI = {v = "SoftKey:Dial"}}
	t.IPPhoneDirectory.SoftKey_3 = {["!index"] = "3", Label = {v = tr:_t("btn.Delete")}, URI = {v = "SoftKey:Next"}}
	t.IPPhoneDirectory.SoftKey_4 = {["!index"] = "4", Label = {v = tr:_t("btn.Back")},   URI = {v = "SoftKey:Previous"}}	

	--Result serialization
	local wxml_obj = require"core.wxml":new()
	local content  = wxml_obj:generate(t, base_xml_str)

	return content
end

--- Build XML menu for forward voicemails
local function __vm_make_forward_menu(params, request_id)
    if not params or not request_id then
        return false
    end

    local tr = params["trobj"]
    local search_val_url
    if params["user"] then
        search_val_url = string.format("%s://%s:%s/?action=getxmlbrowser&get=vmaccess&menu=detail&user=%s&forward=1&id=%s", params["rproto"], params["raddress"], params["rport"], params["user"], request_id)
    else
        search_val_url = string.format("%s://%s:%s/?action=getxmlbrowser&get=vmaccess&menu=detail&forward=1&id=%s", params["rproto"], params["raddress"], params["rport"], request_id)
    end
    -- make XML table
    local t       = {}
    t.IPPhoneInputScreen = {
        ["!Timeout"]   = "70",
        ["!InputType"] = "123",
        Title          = {["!wrap"] = "no", v = tr:_t("xml.vmFwdTitle")},
        Prompt         = {v = tr:_t("xml.vmFwdPrompt") .. ":"},
        Parameter      = {v = "input"},
        URL            = {v = search_val_url}
    }

    t.IPPhoneInputScreen.SoftKey_1 = {["!index"] = "1", Label = {v = tr:_t("btn.Enter")},           URI = {v = "SoftKey:Submit"}}
    t.IPPhoneInputScreen.SoftKey_3 = {["!index"] = "3", Label = {v = tr:_t("btn.Backspace.short")}, URI = {v = "SoftKey:BackSpace"}}
    t.IPPhoneInputScreen.SoftKey_4 = {["!index"] = "4", Label = {v = tr:_t("btn.Back")},            URI = {v = "SoftKey:Exit"}}

    return require"core.wxml":new():generate(t, base_xml_str)
end

--- Build XML content with forward result
local function __vmforward_result_screen(trobj, success_list, fail_list)
    if not trobj or not success_list or not fail_list then
        logger:debug("no params")
        return false
    end
    local tr = trobj

    local t = {}
    if 0 < #success_list and 0 == #fail_list then
        t.IPPhoneTextScreen = {
            ["!Exit"]       = "Idle",
            Title   = {v = tr:_t("xml.vmFwdResult")},
            Text    = {["!wrap"] = "yes", v = tr:_t("xml.vmFwdSuccess")},
            SoftKey = {["!index"] = "4", Label = {v = tr:_t("btn.Back")}, URI = {v = "SoftKey:Exit"}},
        }
    elseif 0 == #success_list and 0 <= #fail_list then
        t.IPPhoneTextScreen = {
            Title   = {v = tr:_t("xml.vmFwdResult")},
            SoftKey = {["!index"] = "4", Label = {v = tr:_t("btn.Back")}, URI = {v = "SoftKey:Exit"}},
        }
        if 0 == #fail_list then
            t.IPPhoneTextScreen["Text"] = {["!wrap"] = "yes", v = tr:_t("xml.vmFwdError")}
        else
            t.IPPhoneTextScreen["Text"] = {["!wrap"] = "yes", v = tr:_t("xml.vmFwdError") .. ": " .. table.concat(fail_list, ", ") .. " does not exist" }
        end
    else
        t.IPPhoneDirectory = {["!style"] = "none", Title = {v = tr:_t("xml.vmFwdResult")}}
        t.IPPhoneDirectory.DirectoryEntry_1 = { Name = {v = tr:_t("xml.vmFwdSuccess") .. ": " .. table.concat(success_list, ", ")} }
        t.IPPhoneDirectory.DirectoryEntry_2 = { Name = {v = tr:_t("xml.vmFwdError") .. ": " .. table.concat(fail_list, ", ") .. " does not exist"} }
        t.IPPhoneDirectory.SoftKey_4 = {["!index"] = "4", Label = {v = tr:_t("btn.Back")},   URI = {v = "SoftKey:Exit"}}
    end

    --Result serialization
    local wxml_obj = require"core.wxml":new()
    local content  = wxml_obj:generate(t, base_xml_str)
    return content
end

--- Processing forward request from the phone
local function __vm_process_forward(vm_params, vm_cont, request, vmApp)
    logger:info("VmFwd. Voicemail forwarding feature has started")

    if not request or not vm_cont or not vm_params then
        return false, "wrong args"
    end

    local function parseMailboxes(mailboxes)
        local rv = {}
        if not mailboxes or "" == mailboxes then
            return rv
        end

        for _, vmbox in pairs(require"common.helper".split(mailboxes, "*")) do
            if "" ~= vmbox and tonumber(vmbox) then
                vmbox = vmbox:gsub("%s", "")
                rv[vmbox] = true
            end
        end
        return rv
    end

    local vmBoxes = parseMailboxes(request["input"])
    local sList = {}
    local fList = {}
    for fwdVMBox, _ in pairs(vmBoxes) do
        rv, err = vmApp.fwdMsg(vm_cont[request["id"]], fwdVMBox)
        if rv then
            table.insert(sList, fwdVMBox)
            vmApp.updateVM(fwdVMBox)
        else
            logger:warning("VmFwd. Forwarding voicemail to %s failed: %s", fwdVMBox, err)
            table.insert(fList, fwdVMBox)
        end
    end
    logger:info("VmFwd. Request string: %q. Success: %s. Fail: %s.", request["input"], table.concat(sList, ", "), table.concat(fList, ", "))
    return __vmforward_result_screen(vm_params["trobj"], sList, fList)
end

-----------------------------------------------------
--- Features implemenation details
-----------------------------------------------------
--- Build XML error content for features request
local function __features_error_content(params, err_msg)
    if "table" ~= type(params) then
        return false
    end
    
    local err_msg = err_msg or "PermissionsDeniedToModifyThisFeature"
    local tr      = params.trobj
    local is_wpr2 = "WPR2" == params.dfamily
    
    local t = {}
    t.IPPhoneTextScreen       = {["!Beep"] = "no", ["!Timeout"] = "70", ["!LockIn"] = "no"}
    t.IPPhoneTextScreen.Title = {["!wrap"] = "yes", v = tr:_t("features.Error")}
    t.IPPhoneTextScreen.Text  = {["!wrap"] = "yes", v = tr:_t("features."..err_msg)}
    
    if is_wpr2 then
        t.IPPhoneTextScreen.SoftKey_4 = {["!index"] = "4", Label = {v = tr:_t("btn.Back")}, URI = {v = "SoftKey:Exit"}}
    else
        t.IPPhoneTextScreen.SoftKey_1 = {["!index"] = "1", Label = {v = tr:_t("btn.Back")}, URI = {v = "SoftKey:Exit"}}
    end
        
	return require"core.wxml":new():generate(t, base_xml_str)
end

--- Build XML content with features list
local function __features_make_list(features_t, params)
    if "table" ~= type(features_t) or "table" ~= type(params) then
        return false
    end
    
    local tr       = params.trobj
    local is_wpr2  = "WPR2" == params.dfamily
    local lbl_add  = params.shortTag and ".short" or ""
    
    -- build itmes formatted uri
    local fmtUri   = string.format("%s?action=getxmlbrowser&get=%%s&feature=%%s", params.rlink)
    local callType = features_t.callType
    if callType and "internal" ~= callType then
        fmtUri = fmtUri .. "&type=" .. callType
    end

    -- types list menu
    local callTypesUri = string.format("%s?action=getxmlbrowser&get=types", params.rlink)
    
    local t = { IPPhoneTextMenu = {["!defaultIndex"] = "1", ["!style"] = "none", ["!Beep"] = "no", ["!Timeout"] = "70", ["!LockIn"] = "no", ["!wrapList"] = "yes"} }
    if is_wpr2 then 
        t.IPPhoneTextMenu["!Exit"]  = "Idle"
        t.IPPhoneTextMenu.SoftKey_1 = {["!index"] = "1", Label = {v = tr:_t("btn.Modify")}, URI = {v = "SoftKey:Select"}}
        t.IPPhoneTextMenu.SoftKey_4 = {["!index"] = "4", Label = {v = tr:_t("btn.Back")}, URI = {v = "SoftKey:Exit"}}
    else
        t.IPPhoneTextMenu.Title     = {["!wrap"] = "yes", v = tr:_t("features.Title")}
        t.IPPhoneTextMenu.SoftKey_1 = {["!index"] = "1", Label = {v = ""}, URI = {v = ""}}
        t.IPPhoneTextMenu.SoftKey_2 = {["!index"] = "2", Label = {v = ""}, URI = {v = ""}}
        t.IPPhoneTextMenu.SoftKey_3 = {["!index"] = "3", Label = {v = ""}, URI = {v = ""}}
        t.IPPhoneTextMenu.SoftKey_4 = {["!index"] = "4", Label = {v = tr:_t("btn.Modify")}, URI = {v = "SoftKey:Select"}}
    end
    
    -- processing user features
    local idxShift   = 0
    local fwdAddItem = false
    for i, item in ipairs(features_t) do
        local value = item.v
        if nil ~= value then
            local idx   = i + idxShift

            local name   = item.n
            local lable  = lbl_add
            local uriReq = "update"

            if "boolean" == type(value) then
                value = tr:_t("feature.state." .. (value and "ON" or "OFF"))
            end

            -- processing Ring Tone feature
            if "RING" == name then
                if value then
                    local menu_str     = string.format("%s: ", tr:_t("feature." .. name .. lbl_add)) 
                    local menu_str_len = menu_str:len()
                    if (menu_str_len + value:len()) > 22 then
                        value = value:sub(1,(22 - menu_str_len))
                        value = value .. "..." 
                    end
                end
                uriReq = "list"
            end

            -- processing additional procedures for forward features
            if "FCF" == name or "BCF" == name or "UCF" == name then
                if is_wpr2 then
                    if not fwdAddItem then   -- add additional string
                        t.IPPhoneTextMenu["MenuItem_" .. idx] = {Prompt = {v = string.format("%s:", tr:_t("features.Forwards"))}}
                        idxShift                              = 1
                        idx                                   = idx + 1
                        fwdAddItem                            = true
                    end
                    lable = ".wp2"
                else                         -- update forward string for WP family devices
                    if features_t[i]["v"] then
                        value = string.format("%s,\n%s: %s", tr:_t("feature.state.ON"), tr:_t("features.Destination.short"), value)
                    end
                end
                if features_t[i]["v"] then
                    uriReq = "submenu"
                else
                    uriReq = "value"
                    -- way to go '__feature_input_value' from main menu
                end
            end
            
            -- processing user status feature
            if "STATUS" == name then
                uriReq = is_wpr2 and "list&val=s" .. value or "update"
                value  = tr:_t("status." .. value:upper())
                lable  = ".short"
            end
            
            if "STATUS_UNTIL" == name then
                if is_wpr2 then
                    local reqVal = value:match("^(%d%d:%d%d)")
                    uriReq       = "list&val=t" .. reqVal
                    if value:find("na$") then
                        value = tr:_t("str.NotSet")
                    else
                        local time, wDay, day, month = value:match("([^%s]+) (%w+) (%d+) (%w+)")
                        if time and wDay and day and month then
                            value = string.format("%s %s %s %s", time, tr:_t("weekday." .. wDay), day, tr:_t("month." .. month))
                        end
                    end
                else
                    value  = nil
                    uriReq = nil
                end
            end

            if value and uriReq then
                local uri = nil
                if "CALLTYPE" == name then
                    value = tr:_t("type." .. value)
                    uri = callTypesUri
                else
                    uri = string.format(fmtUri, uriReq, name)
                end
                t.IPPhoneTextMenu["MenuItem_" .. idx] = {
                    Prompt = {v = string.format("%s: %s", tr:_t("feature." .. name .. lable), value or "")},
                    URI    = {v = uri}
                }	
            end
            
            -- block features for deactivated class
            if "ACTIVATED" == name and not features_t[i]["v"] then
                break
            end
        end
    end
    
    return require"core.wxml":new():generate(t, base_xml_str)
end

--- Build XML content with call types list
local function __features_select_calltype(params)
    if "table" ~= type(params) then
        return false
    end

    local tr       = params.trobj
    local is_wpr2  = "WPR2" == params.dfamily
            
    local callTypesUri   = string.format("%s?action=getxmlbrowser&get=update&feature=CALLTYPE", params.rlink)

    local t = { IPPhoneTextMenu = {["!defaultIndex"] = "1", ["!style"] = "none", ["!Beep"] = "no", ["!Timeout"] = "70", ["!wrapList"] = "yes", ["!LockIn"] = "no"} }
                                        
    if is_wpr2 then
        t.IPPhoneTextMenu.SoftKey_1 = {["!index"]          = "1", Label = {v = tr:_t("btn.Enter")}, URI = {v = "SoftKey:Select"}}
        t.IPPhoneTextMenu.SoftKey_4 = {["!index"]          = "4", Label = {v = tr:_t("btn.Back")}, URI = {v = "SoftKey:Exit"}}
    else
        t.IPPhoneTextMenu.SoftKey_1 = {["!index"] = "1", Label = {v = ""}, URI = {v = ""}}
        t.IPPhoneTextMenu.SoftKey_2 = {["!index"] = "2", Label = {v = ""}, URI = {v = ""}}
        t.IPPhoneTextMenu.SoftKey_3 = {["!index"] = "3", Label = {v = ""}, URI = {v = ""}}
        t.IPPhoneTextMenu.SoftKey_4 = {["!index"] = "4", Label = {v = tr:_t("btn.Enter")}, URI = {v = "SoftKey:Select"}}
    end
    
    local types = {"internal", "external", "blacklist", "whitelist"}

    for idx, c_type in pairs(types) do
    	t.IPPhoneTextMenu["MenuItem_" .. idx] = {
        	 Prompt = {v = tr:_t("type." .. c_type)}
        	,URI    = {v = callTypesUri .. "&type=" .. c_type}
    	}
    end

    return require"core.wxml":new():generate(t, base_xml_str)
end

--- Build XML content with list of the available ring tones
local function __features_make_ringtones_list(data, params)
    if "table" ~= type(data) or "table" ~= type(params) then
        return false
    end
    
    if not data.actRingTone or "table" ~= type(data.ringTonesList) then
        return false
    end
    
    local tr       = params.trobj
    local is_wpr2  = "WPR2" == params.dfamily
    
    -- build itmes formatted uri
    local fmtUri   = string.format("%s?action=getxmlbrowser&get=update&val=%%s&feature=RING", params.rlink)
    local callType = data.callType
    if callType and "internal" ~= callType then
        fmtUri = fmtUri .. "&type=" .. callType
    end
    
    local t = { IPPhoneTextMenu = {["!defaultIndex"] = "1", ["!style"] = "none", ["!Beep"] = "no", ["!Timeout"] = "70", ["!LockIn"] = "no", ["!wrapList"] = "yes"} }
    if is_wpr2 then 
        t.IPPhoneTextMenu.SoftKey_1 = {["!index"] = "1", Label = {v = tr:_t("btn.Modify")}, URI = {v = "SoftKey:Select"}}
        t.IPPhoneTextMenu.SoftKey_4 = {["!index"] = "4", Label = {v = tr:_t("btn.Back")}, URI = {v = "SoftKey:Exit"}}
    else
        t.IPPhoneTextMenu.Title     = {["!wrap"] = "yes", v = tr:_t("feature.RING")}
        t.IPPhoneTextMenu.SoftKey_1 = {["!index"] = "1", Label = {v = ""}, URI = {v = ""}}
        t.IPPhoneTextMenu.SoftKey_2 = {["!index"] = "2", Label = {v = ""}, URI = {v = ""}}
        t.IPPhoneTextMenu.SoftKey_4 = {["!index"] = "4", Label = {v = tr:_t("btn.Modify")}, URI = {v = "SoftKey:Select"}}
    end
    t.IPPhoneTextMenu.SoftKey_3 = {["!index"] = "3", Label = {v = tr:_t("btn.Play")}, URI = {v = "SoftKey:Dial"}}
    
    local ringTonesList = data.ringTonesList
    for idx, val in pairs(ringTonesList) do
        if 30 < idx then
            t.IPPhoneTextMenu["MenuItem_" .. idx] = { Prompt = {v = tr:_t("xml.searchlimit") .. " ..."} }
            break
        else
            local tag = ""  -- set tag for used ring tone
            if val == data.actRingTone then
                tag = " (*)"
            end
            
            -- value normalization
            if val:len() > 22 then
                val = val:sub(1,(18 - tag:len())) .. "..."
            end
            
            t.IPPhoneTextMenu["MenuItem_" .. idx] = { 
                 Prompt = {v = val .. tag} 
                ,URI    = {v = string.format(fmtUri, ringTonesList[idx])}
                ,Dial   = {v = "ringtone" .. ringTonesList[idx]}
            }
        end
    end
    
    return require"core.wxml":new():generate(t, base_xml_str)
end

--- Build XML content for input feature value (ex. FCF)
local function __feature_select_action(data, params)
    if "table" ~= type(data) or "table" ~= type(params) then
        return false
    end
    
    if not data.value or not data.name then
        return false
    end
    
    local tr       = params.trobj
    local is_wpr2  = "WPR2" == params.dfamily
    
    local featValue  = data.value
    local featName   = data.name
    
    -- build itmes formatted uri
    local fmtUri   = string.format("%s?action=getxmlbrowser&feature=%%s", params.rlink)
    local callType = data.callType
    if callType and "internal" ~= callType then
        fmtUri = fmtUri .. "&type=" .. callType
    end

    local t = { IPPhoneTextMenu = {["!defaultIndex"] = "1", ["!style"] = "none", ["!Beep"] = "no", ["!Timeout"] = "70", ["!wrapList"] = "yes", ["!LockIn"] = "no"} }
                                        
    if is_wpr2 then
        t.IPPhoneTextMenu.Title     = {["!wrap"] = "yes", v = string.format("%s\n  %s", tr:_t("feature." .. featName), featValue)}
        t.IPPhoneTextMenu.SoftKey_1 = {["!index"]          = "1", Label = {v = tr:_t("btn.Modify")}, URI = {v = "SoftKey:Select"}}
        t.IPPhoneTextMenu.SoftKey_4 = {["!index"]          = "4", Label = {v = tr:_t("btn.Back")}, URI = {v = "SoftKey:Exit"}}
    else
        t.IPPhoneTextMenu.Title     = {["!wrap"] = "yes", v = string.format("%s\n%s: %s, %s: %s", tr:_t("feature." .. featName), tr:_t("str.Status"), 
                                            tr:_t("feature.state.ON"), tr:_t("features.Destination.short"), featValue)}
        t.IPPhoneTextMenu.SoftKey_1 = {["!index"] = "1", Label = {v = ""}, URI = {v = ""}}
        t.IPPhoneTextMenu.SoftKey_2 = {["!index"] = "2", Label = {v = ""}, URI = {v = ""}}
        t.IPPhoneTextMenu.SoftKey_3 = {["!index"] = "3", Label = {v = ""}, URI = {v = ""}}
        t.IPPhoneTextMenu.SoftKey_4 = {["!index"] = "4", Label = {v = tr:_t("btn.Modify")}, URI = {v = "SoftKey:Select"}}
    end
    
    t.IPPhoneTextMenu.MenuItem_1 = {
         Prompt = {v = tr:_t("action.setOFF")}
        ,URI    = {v = string.format(fmtUri .. "&get=update", featName)}
    }
    
    t.IPPhoneTextMenu.MenuItem_2 = {
         Prompt = {v = tr:_t("action.setDestination")}
        ,URI    = {v = string.format(fmtUri .. "&get=value&val=" .. featValue, featName)}
    }
        
    return require"core.wxml":new():generate(t, base_xml_str)
end

--- Build XML content for feature value actions (input, deactivate)
local function __feature_input_value(data, params)
    if "table" ~= type(data) or "table" ~= type(params) then
        return false
    end
    
    if not data.value or not data.name then
        return false
    end
    
    local tr       = params.trobj
    local is_wpr2  = "WPR2" == params.dfamily
    local lbl_add  = params.shortTag and ".short" or ""
    
    local featValue  = data.value
    local featName   = data.name
    
    -- build itmes formatted uri
    local fmtUri   = string.format("%s?action=getxmlbrowser&get=update&feature=%%s", params.rlink)
    local callType = data.callType
    if callType and "internal" ~= callType then
        fmtUri = fmtUri .. "&type=" .. callType
    end
    
    local t = { IPPhoneInputScreen = {["!defaultIndex"] = "1", ["!style"] = "none", ["!Beep"] = "no", ["!Timeout"] = "70", ["!wrapList"] = "yes",
                                        ["!type"] = "number", ["!password"] = "no", ["!editable"] = "yes", ["!LockIn"] = "no"} }
                                    
    t.IPPhoneInputScreen.Title = {["!wrap"] = "yes", v = tr:_t("feature." .. featName)}
    
    if is_wpr2 then
        t.IPPhoneInputScreen.SoftKey_1 = {["!index"] = "1", Label = {v = tr:_t("btn.Save")},                 URI = {v = "SoftKey:Submit"}}
        t.IPPhoneInputScreen.SoftKey_3 = {["!index"] = "3", Label = {v = tr:_t("btn.Backspace" .. lbl_add)}, URI = {v = "SoftKey:BackSpace"}}
        t.IPPhoneInputScreen.SoftKey_4 = {["!index"] = "4", Label = {v = tr:_t("btn.Back")},                 URI = {v = "SoftKey:Exit"}}
    else
        t.IPPhoneInputScreen.SoftKey_1 = {["!index"] = "1", Label = {v = ""}, URI = {v = ""}}
        t.IPPhoneInputScreen.SoftKey_2 = {["!index"] = "2", Label = {v = ""}, URI = {v = ""}}
        t.IPPhoneInputScreen.SoftKey_3 = {["!index"] = "3", Label = {v = tr:_t("btn.Backspace" .. lbl_add)}, URI = {v = "SoftKey:BackSpace"}}
        t.IPPhoneInputScreen.SoftKey_4 = {["!index"] = "4", Label = {v = tr:_t("btn.Save")}, URI = {v = "SoftKey:Submit"}}
    end
    
    t.IPPhoneInputScreen.Prompt    = {v = tr:_t("features.Destination") .. ":"} 
    t.IPPhoneInputScreen.Default   = {v = featValue}
    t.IPPhoneInputScreen.URL       = {v = string.format(fmtUri, featName)} 
    t.IPPhoneInputScreen.Parameter = {v = "val"}
    if is_wpr2 then
        t.IPPhoneInputScreen.Parameter["!emptyable"] = "yes"
    end
    
    return require"core.wxml":new():generate(t, base_xml_str)
end

--- Build XML content with list of the available ring tones (not supported for WP devices)
local function __features_make_status_data_list(data, params)
    if "table" ~= type(data) or "table" ~= type(params) then
        return false
    end
    
    if not data.tag or "table" ~= type(data.data) then
        return false
    end
    
    local tr       = params.trobj
    local lbl_add  = params.shortTag and ".short" or ""
    
    -- build itmes formatted uri
    local fmtUri   = string.format("%s?action=getxmlbrowser&feature=STATUS", params.rlink)
    local callType = data.callType
    if callType and "internal" ~= callType then
        fmtUri = fmtUri .. "&type=" .. callType
    end
    
    local t = { IPPhoneTextMenu = {["!defaultIndex"] = "1", ["!style"] = "none", ["!Beep"] = "no", ["!Timeout"] = "70", ["!LockIn"] = "no", ["!wrapList"] = "yes"} }
    t.IPPhoneTextMenu["!Selection"]    = fmtUri
    t.IPPhoneTextMenu["!defaultIndex"] = data.dfltField
    t.IPPhoneTextMenu.Title            = {["!wrap"] = "yes", v = tr:_t("feature.STATUS")}
    if "status" == data.tag then
        for statusName, idx in pairs(data.data) do
            local value = tr:_t("status."..statusName:upper())
            local url   = "&get=update&val=s" .. statusName
            t.IPPhoneTextMenu["MenuItem_" .. idx] = {
                 Prompt    = {v = value}
                ,Selection = {v = url}
            }
        end
    elseif "time" == data.tag then
        t.IPPhoneTextMenu.Title.v          = tr:_t("features.UntilTime")
        for idx, timeStr in ipairs(data.data) do
            t.IPPhoneTextMenu["MenuItem_" .. idx] = {
                 Prompt    = {v = timeStr}
                ,Selection = {v = "&get=list&val=dt" .. timeStr}
            }
        end
    elseif "date" == data.tag then
        t.IPPhoneTextMenu.Title.v = tr:_t("features.UntilDate")
        for idx, dateStr in ipairs(data.data) do
            local promptStr, reqVal   = dateStr:match("(.+)%-(.+)")
            local pWDay, pDay, pMonth = promptStr:match("(%w+) (%d+) (%w+)")
            if pWDay and pDay and pMonth then
                promptStr = string.format("%s %s %s", tr:_t("weekday." .. pWDay), pDay, tr:_t("month." .. pMonth))
            end
            t.IPPhoneTextMenu["MenuItem_" .. idx] = {
                 Prompt    = {v = promptStr}
                ,Selection = {v = "&get=update&val=dt" .. require"socket.url".escape(data.val.."-".. reqVal)}
            }
        end
    end
    
    t.IPPhoneTextMenu.SoftKey_1 = {["!index"] = "1", Label = {v = tr:_t("btn." .. ("time" == data.tag and "Submit" or "Save"))}, URI = {v = "SoftKey:Select"}}
    t.IPPhoneTextMenu.SoftKey_4 = {["!index"] = "4", Label = {v = tr:_t("btn.Back")}, URI = {v = "SoftKey:Exit"}}
    
    return require"core.wxml":new():generate(t, base_xml_str)
end

-----------------------------------------------------
--- Contact center implemenation details
-----------------------------------------------------
--- Build XML error content for contact center request
local function __callgroup_error(params, err_msg)
    if not params then
        logger:debug("no params")
        return false
    end

    local tr = params["trobj"]

    t = {}
    t.IPPhoneTextScreen = {
        ["!Exit"]       = "Idle",
        Title   = {v = tr:_t("features.Error")},
        Text    = {["!wrap"] = "yes", v = tr:_t(err_msg)},
        SoftKey = {["!index"] = "4", Label = {v = tr:_t("btn.Back")}, URI = {v = "SoftKey:Exit"}},
    }

    --Result serialization
    local wxml_obj = require"core.wxml":new()
    local content  = wxml_obj:generate(t, base_xml_str)
    return content
end

--- Build XML content with list of callgroups
local function __callgroups_wpr2(params)
    local t  = {}
    local tr = params["trobj"]
    local init_url        = string.format("%s://%s:%s/?action=getcallgroups&", params["rproto"], params["raddress"], params["rport"])

    local cg_signin_url = init_url .. "menu=toogle"
    local cg_signall_url = init_url .. "menu=signall"

    -- make XML table
    t.IPPhoneDirectory = {
         ["!style"]      = "none"
        ,["!Selection"]  = cg_signin_url
        ,["!Selection1"] = cg_signall_url
        ,["!Exit"]       = "Idle"
    }

    -- Soft keys for phonebook content directory
    for k, record in ipairs(params["records"]) do
        local user_info = ""
        if record.entryuid and "" ~= record.entryuid then
           user_info  = user_info .. "&entryuid=" .. (record.entryuid or "")
        end

        t.IPPhoneDirectory["DirectoryEntry_" .. k] = {Name = {v = record.username}}

        if record["presence"] and "" ~= record["presence"] then
            t.IPPhoneDirectory["DirectoryEntry_" .. k].Presence = {v = __image_convert(record.presence, "WPR2")}
        end
        if params["sign_status"] == "1" then
            if record.membership ~= "static" then
                t.IPPhoneDirectory["DirectoryEntry_" .. k].Selection = {v = user_info}
            else
                t.IPPhoneDirectory["DirectoryEntry_" .. k].Selection = {v = "&error=true"}
            end
        end
        t.IPPhoneDirectory["DirectoryEntry_" .. k].Selection1 = {v = user_info}
    end
    if params["sign_status"] == "1" then
        t.IPPhoneDirectory.SoftKey_1 = {["!index"] = "1", Label = {v = tr:_t("xml.switch")},   URI = {v = "SoftKey:Select"}}
        if string.match(params["dmodel"]:lower(), "wp410r[23]") or "wp480r2" == params["dmodel"]:lower() or "wp480gr2" == params["dmodel"]:lower() then
            t.IPPhoneDirectory.SoftKey_2 = {["!index"] = "2", Label = {v = tr:_t("S-out")}, URI = {v = "SoftKey:Select1"}}
        else
            t.IPPhoneDirectory.SoftKey_2 = {["!index"] = "2", Label = {v = tr:_t("xml.signout")}, URI = {v = "SoftKey:Select1"}}
        end
        t.IPPhoneDirectory.SoftKey_4 = {["!index"] = "4", Label = {v = tr:_t("btn.Exit")},   URI = {v = "SoftKey:Exit"}}
    else
        if string.match(params["dmodel"]:lower(), "wp410r[23]") or "wp480r2" == params["dmodel"]:lower() or "wp480gr2" == params["dmodel"]:lower() then
            t.IPPhoneDirectory.SoftKey_1 = {["!index"] = "1", Label = {v = tr:_t("S-on")}, URI = {v = "SoftKey:Select1"}}
        else
            t.IPPhoneDirectory.SoftKey_1 = {["!index"] = "1", Label = {v = tr:_t("xml.signon")}, URI = {v = "SoftKey:Select1"}}
        end
        t.IPPhoneDirectory.SoftKey_4 = {["!index"] = "4", Label = {v = tr:_t("btn.Exit")},   URI = {v = "SoftKey:Exit"}}
        t.IPPhoneDirectory.SoftKey_5 = {["!index"] = "OK", URI = {v = "SoftKey:Select1"}}
    end

    --Result serialization
    local wxml_obj = require"core.wxml":new()
    local content  = wxml_obj:generate(t, base_xml_str)
    return content
end

-----------------------------------------------------
--- Changing identity implemenation details
-----------------------------------------------------
--- Build XML error content for identities request
local function __identity_error(params, err_msg)
    if not params then
        logger:debug("no params")
        return false
    end

    local tr = params["trobj"]

    t = {}
    t.IPPhoneTextScreen = {
        ["!Exit"]       = "Idle",
        Title   = {v = tr:_t("features.Error")},
        Text    = {["!wrap"] = "yes", v = tr:_t(err_msg)},
        SoftKey = {["!index"] = "4", Label = {v = tr:_t("btn.Back")}, URI = {v = "SoftKey:Exit"}},
    }

    --Result serialization
    local wxml_obj = require"core.wxml":new()
    local content  = wxml_obj:generate(t, base_xml_str)
    return content
end

--- Build XML content with list of available identities
local function __identities_wpr2(params)
    local t  = {}
    local tr = params["trobj"]
    local init_url        = string.format("%s://%s:%s/?action=getidentities", params["rproto"], params["raddress"], params["rport"])

    -- make XML table
    t.IPPhoneDirectory = {
         ["!style"]      = "none"
        ,["!Selection"]  = init_url
        ,["!Exit"]       = "Idle"
    }

    -- Soft keys for phonebook content directory
    for k, record in ipairs(params["records"]) do
        local user_info = ""
        user_info = user_info .. "&identity=" .. (record.identity or "")
        t.IPPhoneDirectory["DirectoryEntry_" .. k] = {Name = {v = record.username}}
        if record["presence"] and "online" == record["presence"] then
            t.IPPhoneDirectory["DirectoryEntry_" .. k].Presence = {v = __image_convert("offline", "WPR2")}
        end
        t.IPPhoneDirectory["DirectoryEntry_" .. k].Selection = {v = user_info}

    end
    t.IPPhoneDirectory.SoftKey_1 = {["!index"] = "1", Label = {v = tr:_t("btn.Enter")},   URI = {v = "SoftKey:Select"}}
    t.IPPhoneDirectory.SoftKey_4 = {["!index"] = "4", Label = {v = tr:_t("btn.Exit")},   URI = {v = "SoftKey:Exit"}}

    --Result serialization
    local wxml_obj = require"core.wxml":new()
    local content  = wxml_obj:generate(t, base_xml_str)
    return content
end

-----------------------------------------------------
--- Authorization implemenation details
-----------------------------------------------------
--- BUild XML content to user entering pin value 
local function __auth_make_content(params, pinLen)
    if "table" ~= type(params) then
        return false
    end
    
    local tr      = params.trobj
    local authUrl = params.rlink .. params.rquery
    local is_wpr2 = "WPR2" == params.dfamily
    
    local t = { IPPhoneInputScreen = { ["!defaultIndex"] = "1", ["!style"] = "none", ["!Beep"] = "no", ["!Timeout"] = "70", ["!wrapList"] = "yes", 
                                       ["!type"] = "number", ["!password"] = "yes", ["!editable"] = "yes", ["!LockIn"] = "no"}}
                                       
    if is_wpr2 then
        t.IPPhoneInputScreen["!Exit"]          = "Idle"
        t.IPPhoneInputScreen["!autoAcceptLen"] = pinLen
        t.IPPhoneInputScreen["!InputType"]     = "*123"
    else
        t.IPPhoneInputScreen["!InputType"]     = "123"
    end
        
    t.IPPhoneInputScreen.Title     = {["!wrap"] = "no", v = tr:_t("xml.enterpin")}
    t.IPPhoneInputScreen.Parameter = {v = "pin"}
    t.IPPhoneInputScreen.URL       = {v = authUrl}
    t.IPPhoneInputScreen.Prompt    = {v = ""}
    
    t.IPPhoneInputScreen.SoftKey_1 = {["!index"] = "1", Label = {v = tr:_t("btn.Enter")},           URI = {v = "SoftKey:Submit"}}
    t.IPPhoneInputScreen.SoftKey_3 = {["!index"] = "3", Label = {v = tr:_t("btn.Backspace.short")}, URI = {v = "SoftKey:BackSpace"}}
    t.IPPhoneInputScreen.SoftKey_4 = {["!index"] = "4", Label = {v = tr:_t("btn.Exit")},            URI = {v = "SoftKey:Exit"}}
    
    return require"core.wxml":new():generate(t, base_xml_str)
end

--- Build a XML content with timeout PIN locking
local function __auth_lock_error(params, timeout)
    local timeout    = timeout or "-1"
    local dev_family = params.dfamily
    local tr         = params.trobj
    
    local t = {}
    if dev_family:match("^W%-AIR") then
        t.IPPhoneDirectory = {
            DirectoryEntry = {Name = {v = tr:_t("features.Error")}},
            Total          = {v = "1"},
            RequestID      = {v = "0"},
            SoftKey        = {["!index"] = "1", Label = {v = tr:_t("btn.Back")}, URI = {v = "SoftKey:Exit"}}
        }
    elseif "WPR2" == dev_family then
        t.IPPhoneTextScreen = {
            ["!Exit"]       = "Idle",
            Title   = {v = tr:_t("features.Error")},
            Text    = {["!wrap"] = "yes", v = tr:_t("xml.deviceLocked", timeout)},
            SoftKey = {["!index"] = "4", Label = {v = tr:_t("btn.Back")}, URI = {v = "SoftKey:Exit"}}
        }
    else
        t.IPPhoneTextScreen = {
            Title   = {v = tr:_t("features.Error")},
            Text    = {["!wrap"] = "yes", v = tr:_t("xml.deviceLocked", timeout)},
            SoftKey = {["!index"] = "1", Label = {v = tr:_t("btn.Back")}, URI = {v = "SoftKey:Exit"}}
        }
    end
    
    --Result serialization
    local wxml_obj = require"core.wxml":new()
    local content  = wxml_obj:generate(t, base_xml_str)
    
    return content  
end
-----------------------------------------------------
--- Public interface
-----------------------------------------------------

--- Get label translation for XML broswer feature
-- @param feature	A feature name used via XML broswer (supported feature: phonebook, features, history)
-- @param lang		An user language 
function label_name(feature, lang)
	if not feature then
		return ""
	end
	
	local tr = require"common.translation".translation
	tr:set_language(lang or "en")
	
	return tr:_t("xml." ..  feature)
end

--- Processing phonebook request from the phone
-- @param agent		A table with general device information
-- @param request	A table with requested parameters
-- @return XML content with phonebook data
-- @note used for W-AIR, WP and WPR2 devices
function phonebook_process(agent, request)
	if not agent or not request then
		return false
	end
	
	local core   = require"common.core"
	local h 	 = require"common.helper"
	local http   = require"core.httpClient"
	local logger = logger
	
	if "Wildix" ~= agent["vendor"] then
		logger:error("Request got from not supported device")
		return false
	end
	
	-- load translation object
	local trobj = require"common.translation".translation
	
	-- get device user information
	local dev_family = agent["vendor:additional"] or ""
	local user_info = __get_user_data(agent["mac"], dev_family:match("^W%-AIR") and request["extension"])
	if not user_info then
		return __pb_error_content("pbempty", dev_family, trobj)
	end
	local user_id   = user_info["uid"]
	local user_lang = user_info["preferredLanguage"]
	local username  = user_info["telephoneNumber"]
	local userKind	= user_info["userKind"]
	logger:debug("User %q(%s) with uid: %s", username, user_lang, user_id)
	
	-- set translation object language	
	trobj:set_language(user_lang or "en")
	
	-- table of the used parameters
	local pb_params = {
                ["trobj"]    = trobj,
                ["rproto"]   = agent["ssl"] == "on" and "https" or "http",  -- used by WPR2
                ["raddress"] = agent["registrar"] or "",                    -- used by WPR2
                ["rport"]    = agent["registrar:port"] or "",               -- used by WPR2
                ["dfamily"]  = agent["vendor:additional"],
	}

	-- preparing phonebook request parameters
	local pbookname = request["phonebook"] or "all"
	if pbookname and pbookname == "all" then
		pbookname =  "searchall"
	end

	-- show menu with available phonebooks
	if "filter" == request["menu"] then
		return __phonebook_make_pblist_menu(pb_params, user_id)
	elseif "search" == request["menu"] then
		return __phonebook_make_search_menu(pb_params, pbookname)
	end
	
	-- check supported devices
	local fw_ver = h.parse_firmware(agent["fw:version"])
	if "WP" == dev_family and  "wp600" ~= agent["model"] then
		-- checking firmware for WP family (exeption WP600)
		if fw_ver and 50 > fw_ver.major then
			fw_ver = false 		-- mark as unsupported firmware
		end
	elseif dev_family:match("^W%-AIR") then
		-- checking firmware for W-AIR devices
		if 303 > fw_ver.sht_v1 then
			fw_ver = false		-- mark as unsupported firmware
		end
	end
	
	if not fw_ver then
		logger:warning("Device firmware is old. Need upgrading ...")
		return __pb_error_content("oldfw", dev_family, trobj)
	end

	-- try detect variable for searching
	local input = request["input"] or ""
	if input == "" then
		input = request["search"] or ""
	end
	if input == "#SEARCH" then
		input = ""
	end
	
	pb_params["input"] = input

	local search_filter = {false, input}
	if request["entryuid"] and "" ~= request["entryuid"] then
		search_filter = {"entryuid", request["entryuid"]}
	end
    if "livesearch" == request["menu"] then
        -- WMS-5582 skip first digit/letter
        local ilen = #(request["input"] or "")
        if ilen < 2 then
            return __phonebook_livesearch_records({["records"] = {}}, dev_family)
        end
        search_filter[3] = "unique"
    end
    search_filter[4] = "username"
	
	local max_entries_count = 200
	if dev_family:match("^W%-AIR") then
		max_entries_count = 2500
	elseif "WPR2" == dev_family then
	    max_entries_count          = 200000
	    request["paginationCount"] = 100
	    -- do sort later
	    search_filter[4] = false
	end
    
	-- getting records
	local pb_records = {}
	if user_id and "" ~= user_id then
		if "directory" == pbookname then
			if "room" == userKind then 
				pb_records = core.get_callable_ldap_entries({user_id, userKind}, search_filter, max_entries_count + 1)
			else
				pb_records = core.get_ldap_entries_from_ctidb({user_id, user_info["owner"], user_info["licenseType"]}, search_filter, max_entries_count + 1)
			end
		elseif "searchall" == pbookname then
			if "room" == userKind then 
				pb_records = core.get_callable_ldap_entries({user_id, userKind}, search_filter, max_entries_count + 1)
			else
				pb_records = core.get_ldap_entries_from_ctidb({user_id, user_info["owner"], user_info["licenseType"]}, search_filter, max_entries_count + 1)
			end
			local pbookst = core.get_list_phonebook(user_id);
			pbookname = http.urlDecode(pbookname)	-- firmware >= X.50.10.X encodes twice the phonebook url, so encoded "space" %20 converted to %2520
			local tmp_records = core.get_phonebook_entries(pbookst, search_filter, max_entries_count + 1)
			if tmp_records then
				for _, v in ipairs(tmp_records) do
					pb_records[#pb_records + 1] = v
				end
			end
		elseif "" ~= pbookname then
			pbookname = http.urlDecode(pbookname)	-- firmware >= X.50.10.X encodes twice the phonebook url, so encoded "space" %20 converted to %2520
			pbookname = tonumber(pbookname)
			local pbookst = core.get_list_phonebook(user_id)
			
			local availabe_pbook = false
			for _,v in pairs(pbookst) do
				if v == pbookname then
					availabe_pbook = true
					break
				end
			end
	
			if availabe_pbook then
				pb_records = core.get_phonebook_entries(pbookname, search_filter, max_entries_count + 1)
			end
		end
	end

	if not pb_records or 0 == #pb_records then
		local err_msg = "pbempty"
		if "" ~= input then
			err_msg = "pbnouser"
			logger:warning("User not found")
		else
			logger:warning("Phonebook doesn't have records")
		end
		return __pb_error_content(err_msg, dev_family, trobj)
	end
	
	table.sort(pb_records, function(first, second) return first.username < second.username end)
	local keys_table = {["2"] = {"a", "b", "c"},
						["3"] = {"d", "e", "f"},
						["4"] = {"g", "h", "i"},
						["5"] = {"j", "k", "l"},
						["6"] = {"m", "n", "o"},
						["7"] = {"p", "q", "r", "s"},
						["8"] = {"t", "u", "v"},
						["9"] = {"w", "x", "y", "z"}}
	local find_keys = request["keypad"] and keys_table[request["keypad"]]

	local function _binary_search(records, letter)
		local istart = 1
		local iend   = #records
		local found  = false
		while istart <= iend do
			local imid = math.floor((istart + iend) / 2)
			repeat
				local record = records[imid]
				local first  = record and string.sub(record.username, 1, 1):lower()
				if letter == first then
					found = true
					imid  = imid - 1
				end
				if found and letter ~= first then
					return imid + 1
				end
				if letter < first then
					iend = imid - 1
					break
				end
				if letter > first then
					istart = imid + 1
					break
				end
			until false
		end
	end

	local function _onkey_find(records, keys)
		for _, letter in pairs(keys) do
			local index = _binary_search(records, letter)
			if index then
				return index
			end
		end
	end

	-- preparing parameter for phonebook specifications
	local pagination_cnt = tonumber(request["paginationCount"]) or max_entries_count
	local pagination_idx = 0
	if not find_keys then
		pagination_idx = request["paginationIndex"] and request["paginationIndex"] + 1 or 1
	else
		pagination_idx = _onkey_find(pb_records, find_keys) or 1
	end

	local page_data = {
	            start  = tonumber(pagination_idx)
	           ,finish = tonumber(pagination_idx + pagination_cnt - 1)
	           ,size   = pagination_cnt
          }

	local records = h.table_slice(pb_records, page_data.start, page_data.finish) or {}
	for _, record in pairs(records) do
		record.username = __sanitize_utf8(record.username)
	end

	pb_params["page"]           = page_data
	pb_params["records"]        = records
	pb_params["total_records"]  = #pb_records
	pb_params["request_id"]     = request["requestId"]
	pb_params["details"]        = request["details"] or "short"
	pb_params["search_filter"]  = request["phonebook"] or "all"
	
	-- processing users presence
    if (dev_family:match("^W%-AIR") and "full" == request["details"]) or "WPR2" == dev_family then
        pb_params["records"] = require"common.core".get_jabber_presence_statuses(pb_params["records"])
    end
	
	local output = nil
	if "detl" == request["menu"] then
		output = __phonebook_make_detail_menu(pb_params, request)
	elseif "dial" == request["menu"] then
		output = __phonebook_make_advdial_menu(pb_params, request)
    elseif "livesearch" == request["menu"] then
        output = __phonebook_livesearch_records(pb_params, dev_family)
	else
		local set_presence = "yes" == require"common.permissions":check(user_info["owner"] or "", "canuse", "PHONEBOOKPRESENCE", nil, user_info["licenseType"])
		output = __phonebook_records(pb_params, dev_family, set_presence)
		if output and 0 ~= #output then
			logger:info("Sending phonebook to %s(%s)", agent["model"], dev_family)
		else
			logger:error("Some error found in phonebook processing")
		end
	end
	return output
end

--- Processing history request from the phone
-- @param agent		A table with general device information
-- @param request	A table with requested parameters
-- @return XML content with history data
function history_process(agent, request)
	if not agent or not request then
		return false
	end
	
	local core   = require"common.core"
    local h      = require"common.helper"
    local logger = logger
	
	if "Wildix" ~= agent["vendor"] or "WPR2" ~= agent["vendor:additional"] then
		logger:error("Request got from not supported device")
		return false
	end
	
	-- load translation object
	local trobj = require"common.translation".translation
	
	-- get device user information
	local user_info = __get_user_data(agent["mac"])
	if not user_info then
		return __history_error_content("histempty", trobj)
	end
	local user_id   = user_info["uid"]
	local user_lang = user_info["preferredLanguage"]
	local user_num	= user_info["telephoneNumber"]
	logger:debug("User %q(%s) with uid: %s", user_num, user_lang, user_id)
	
	-- set translation object language	
	trobj:set_language(user_lang or "en")

    --Can't use history by ACL
    local wmsperm = require"common.permissions"
    if wmsperm:check(user_info["owner"] or "", "canuse", "CTI_CDR", nil, user_info["licenseType"]) ~= "yes" then
        return __history_error_content("permissionDenied", trobj)
    end

	-- get user dateTime settings
    local dtSettings = __getUserFeatures(user_id, "DATETIME") or {}
    if dtSettings.format then
        dtSettings.type = "php"
    end

    -- get local time zone for device
    dtSettings.timeZone = __getDevData(agent.mac)
	
	-- table for used params
	local hist_params = {
               ["trobj"]      = trobj,
               ["rproto"]     = agent["ssl"] == "on" and "https" or "http",
               ["raddress"]   = agent["registrar"] or "",
               ["rport"]      = agent["registrar:port"] or "",
               ["dfamily"]    = agent["vendor:additional"],
               ["model"]      = agent["model"],
               ["dtSettings"] = dtSettings
	}
	
	--- processing additional menus	
	if "option" == request["menu"] then
		--- processing Option menu
		local xml_cont, err = __history_make_option_menu(request["data"], hist_params)
		if not xml_cont then
			return __history_error_content(err, trobj)
		end
		logger:info("Sending history option menu to %s(%s)", agent["model"], agent["vendor:additional"])
		return xml_cont
	elseif "detail" == request["menu"] then
		--- processing Detail menu by showing call information
		hist_params["htype"] = request.type 
		local xml_cont, err  = __history_make_detail_menu(request["rid"], user_num, hist_params)
		if not xml_cont then
			return __history_error_content(err, trobj)
		end
		logger:info("Sending history option menu to %s(%s)", agent["model"], agent["vendor:additional"])
		return xml_cont
	elseif "archive" == request["menu"] then
	   --- make selected call raw as archive
	   local set_data = {user_num = user_num, row_id = request.rid, call_direct = request.direct}
	   __history_set_cdr_info(set_data, "archive")
	end
	
	--- processing Main menu by showing All/Missed history calls
	-- detecting history type
	local hist_type		= "all"
	local hist_flag		= nil
	local hType			= nil
	if request["type"] then
		hist_flag, hType = request["type"]:match("(init)%-(%w+)")
		hist_type        = hType or request["type"] or hist_type
	end 
	
	-- get CDR records with user call history
	local hist_cont, err = __history_get_cdr_info({["user_num"] = user_num}, hist_type)
	if not hist_cont then
		return __history_error_content(err, trobj)
	end
	
	-- build XML content with history
	hist_params["htype"] = hist_type
	local xml_cont, err  = __history_make_content(hist_cont, hist_params)
	if not xml_cont or "" == xml_cont then
		logger:error("Some error found in history processing")
		return __history_error_content("histemptylist", trobj)
	end
	
	-- send notify for clearing missed calls
	if hist_flag then
    	local rs, err = require"core.astman".send_command("sip notify clear-missed-call " .. user_num)
    	if not rs then -- show logs from astman module
    		logger:error("Notify with clearing missed calls was not sent: %s", err)
    	else
    		logger:debug("Notify with clearing missed calls was sent")
    	end
    end
	
	logger:info("Sending history to %s(%s)", agent["model"], agent["vendor:additional"])
	return xml_cont
end

--- Processing features request from the phone
-- @param agent		A table with general device information
-- @param request	A table with requested parameters
-- @return updated device information with answer content
function features_process(agent, request)
    if not agent or not request then
        return false
    end
    
    local base64 = require"base64"
    local json   = require"dkjson"
        
    -- buile table with user features
    local function _collectFeaturesData(userFeatures, custList)
        local rv = {}
        if "table" ~= type(userFeatures) then
            return rv
        end
        
        local defList  = {"STATUS", "CALLTYPE", "ACTIVATED", "REJECT", "BCF", "UCF", "FCF", "CW", "MEXT", "RING"}
        local fixList  = {}
        for idx, name in ipairs(defList) do
            fixList[idx]  = name
            fixList[name] = idx
        end
        local featList = defList
        if "table" == type(custList) and custList[1] then
            featList = {}
            for _, name in ipairs(custList) do
                if fixList[name] then
                    featList[#featList + 1] = name
                end
            end
            if 0 == #featList then
                featList = defList
            end
        end
        
        for idx, name in ipairs(featList) do
            local value = nil           -- field with 'nil' value will not be shown
            local data  = {}
            if "STATUS" == name then
                local status = userFeatures:getUserStatus()
                value        = "" == status and "available" or status
                
                -- create additional field for until time
                if "available" ~= value then
                    local untilTime = userFeatures:getUserStatusData("untilTime") or ""
                    local timeZone  = nil   -- TODO: processing from devdb
                    local tdObj = require"core.timedate":new(untilTime, timeZone)
                    if not tdObj then   -- prepare until time to not set with current time prefix
                        tdObj     = require"core.timedate":new(os.time(), timeZone)
                        tdObj:toTimeZone()
                        untilTime = tdObj:format("%H:%M ") .. "na"
                    else                -- make until time from jabber data
                        tdObj:toTimeZone()
                        untilTime = tdObj:format("%H:%M %a %d %b")
                    end
                    data[2] = {n = "STATUS_UNTIL", v = untilTime}
                end
            elseif "CALLTYPE" == name then
                value = userFeatures:getActiveClass()
                -- save direct link to call type
                if "internal" ~= value then
                    rv.callType = value
                end
            elseif "ACTIVATED" == name then
                local userStatus = userFeatures:getUserStatus()
                userStatus       = "" == userStatus and "available" or userStatus
                if not ("internal" == userFeatures:getActiveClass() and "available" == userStatus) then
                    value = userFeatures:isActivatedClassValues() or false
                end
            elseif "MEXT" == name then
                value = userFeatures:get(name) and true or false
            else
                value = userFeatures:get(name) or false
            end
            data[1] = {n = name, v = value}
            for _, val in ipairs(data) do
                table.insert(rv, val)
            end
        end
        
        return rv
    end
    
    -- set up features data
    local function _setFeaturesData(exten, userFeatures, featName, featData)
        local rv = false
        if "table" ~= type(userFeatures) or not featName or not exten then
            return rv
        end
        
        repeat
            if "CALLTYPE" == featName then  -- set up call type feature
                rv = userFeatures:getActiveClass()
                break
            end
            
            if "RING" == featName then      -- set up ring tone feature
                if featData and "" ~= featData then
                    userFeatures:setValueData(featName, {
                         status = "1"
                        ,value  = featData
                    })
                end
                rv = userFeatures:saveData()
                break
            end
            
            if "ACTIVATED" == featName then -- set up activated class flag
               if not ("internal" == userFeatures:getActiveClass() and "" == userFeatures:getUserStatus()) then
                    userFeatures:setUpActivatedClassValue()
                    rv = userFeatures:saveData()
                end
                break
            end
            
            if "STATUS" == featName then
                local statusData = {}
                local statusList = {
                    [1]       = "available"     
                   ,[2]       = "dnd"
                   ,[""]      = 1
                   ,available = 1
                   ,dnd       = 2
                   ,away      = 2

                }
                if featData and featData:find("^s") then
                    if featData then -- update statuses list
                        statusList[3]      = "away"
                        statusList["away"] = 3
                        if "full" == featData then
                            featData =  nil
                        end
                    end

                    featData = featData:match("^s(.+)")
                    if featData then
                        statusData.val = statusList[statusList[featData]]
                    end
                elseif featData and featData:find("^dt") then
                    statusData.val                   = userFeatures:getUserStatus()
                    statusData.time, statusData.date = featData:match("^dt(.+)%-(.+)")
                    statusData.tz                    = nil -- TODO: processing from devdb
                else    -- switch to next status value
                    local statusVal = userFeatures:getUserStatus()
                    local statusIdx = statusList[statusVal]
                    statusIdx       = #statusList <= statusIdx and 1 or statusIdx + 1
                    statusData.val  = statusList[statusIdx]
                end
                if statusData.val then
                    rv = require"common.userFeatures.userStatus":setDataToJabber(exten, {
                         status = statusData.val
                        ,time   = statusData.time
                        ,date   = statusData.date
                        ,tz     = statusData.tz
                    })
                    userFeatures:load()
                end
                break
            end
            
            -- set up other features
            local featStatus = true
            if featData then
                featStatus = false
            else
                featStatus = userFeatures:getStatus(featName)
            end
            local featValue = featData
            if not featValue and not featStatus then
                if "FCF" == featName or "BCF" == featName or "UCF" == featName then
                    featValue  = "VOICEMAIL"
                end
            end
            
            userFeatures:setValueData(featName, {
                 status = featStatus and "0" or "1"
                ,value  = featValue
            })
            rv = userFeatures:saveData()
        until true
        
        if rv then  -- send notifications
            if "internal" == userFeatures:getActiveClass() then
                if "FCF" == featName or "ACTIVATED" == featName then
                    local syncUserFeatures = require"common.userFeatures.syncUserFeaturesViaSip"
                    if "ACTIVATED" == featName then
                        featName = syncUserFeatures:getDefaultSyncFeatures()
                        featName[featName.STATUS] = nil
                    end
                    syncUserFeatures:byExten(exten, featName, userFeatures)
                end
            end
        end
        
        return rv
    end
    
    --- preparing status data list values
    local function _buildStatusDataListValues(value, userFeatures)
        local idx = 1
        local rv  = {}
        local tag = nil
        local val = nil
         if "table" ~= type(userFeatures) or not value then
            return rv
        end
        
        if value:find("^s") then
            local actStatus = value:match("^s(.+)")
            rv              = userFeatures:getUserStatusList()
            idx             = rv[actStatus] + 1
            tag             = "status"
        elseif value:find("^t") then
            tag             = "time"
            --  processing actual time list index
            local indexTime = nil
            if value then
                local hour, min = value:match("(%d%d):(%d%d)")
                indexTime = os.time({
                     hour = hour
                    ,min  = min
                    ,sec = 0
                    ,year  = 1970
                    ,month = 1
                    ,day   = 1
                })
                idx = 1
            end
            local unixTime = os.time({
                 hour  = 0
                ,min   = 0
                ,sec   = 0
                ,year  = 1970
                ,month = 1
                ,day   = 1
            })
            local timeStep = 60 * 60            -- 60 minutes
            repeat
                unixTime    = unixTime + timeStep
                rv[#rv + 1] = os.date("%H:%M", unixTime)
                if indexTime and indexTime >= unixTime then
                    idx = idx + 1
                end
            until "02" == os.date("%d", unixTime)
        elseif value:find("^d") then
            tag            = "date"
            val            = value:match("^dt(.+)")
            local unixTime = os.time()
            local dateStep = 60 * 60 * 24     -- one day
            for i = 1, 30 do
                if 1 ~= i then
                    unixTime    = unixTime + dateStep
                end
                rv[#rv + 1] = os.date("%a %d %b-%d/%m/%Y", unixTime)
            end
        end
        
        return rv, idx, tag, val
    end
    
    
    local tr          = require"common.translation".translation
    local pbxutils    = require"core.pbxutils"
    
    local registrar   = agent["registrar"]
    local req_proto   = "on" == agent["ssl"] and "https" or "http"
    local req_port    = agent["registrar:port"]
    local return_link = string.format("%s://%s:%s/", req_proto, registrar, req_port)
    
    local respBody   = nil
	local featParams = {    -- table for used params
        ["trobj"]    = tr,
        ["rlink"]    = return_link,
        ["dfamily"]  = agent["vendor:additional"],
        ["shortTag"] = "wp480" == agent["model"] or "wp490" == agent["model"] or string.match(agent["model"], "wp410r[23]") or "wp480r2" == agent["model"] or "wp480r3" == agent["model"]
	}
    
    -- get device user information
    local attrs = __get_user_data(agent["mac"])
    if not attrs then
        respBody = __features_error_content(featParams)
        return respBody
    end
    local userLang  = attrs["preferredLanguage"]
    local userKind  = attrs["userKind"]
    local userExten = attrs["telephoneNumber"]
    tr:set_language(userLang)
    logger:debug("User %q(%s) with uid: %s", userExten, userLang, attrs["uid"])
    
    -- check user settings directory
    if pbxutils.get_users_settings_dir() == false then
        logger:error("Users settings Share is not mounted")
        respBody = __features_error_content(featParams, "UsersShareNotMounted")
        return respBody
    end

    -- loading user features
    local userFeatures = require"common.userFeatures.userFeatures":new(attrs["uid"])
    if not userFeatures then
        respBody = __features_error_content(featParams)
        return respBody
    end
    userFeatures:load()
    
    -- parsing request fields
    local reqType     = request["get"]
    local reqCallType = request["type"]
    local reqFeature  = request["feature"] 
    local reqValue    = request["val"]

    -- permission check does not required for show features
    if "features" ~= reqType then
        if "yes" ~= require"common.permissions":check(attrs["owner"] or "", "canset", reqFeature, nil, attrs["licenseType"]) then
            respBody = __features_error_content(featParams)
            return respBody
        end
    end
    
    -- set up call class for user features
    userFeatures:setUpActiveClass(reqCallType or "internal")
    
    local custListFeat = "room" == attrs.userKind and {"STATUS"}
    if "features" == reqType then           -- show user features list
        local features_t = _collectFeaturesData(userFeatures, custListFeat)
        respBody         = __features_make_list(features_t, featParams)
    elseif "update" == reqType then  -- update only feature status
        if _setFeaturesData(userExten, userFeatures, reqFeature, reqValue) then
            local features_t = _collectFeaturesData(userFeatures, custListFeat)
            respBody         = __features_make_list(features_t, featParams)
        end
    elseif "list" == reqType then   -- show list of the feature values
        local data = { callType = reqCallType }
        if "RING" == reqFeature then
            local ringTonesList, errMsg = require"core.fsutils".readdir("/var/lib/callweaver/sounds/00000/RingTones/", "^(.*)%.ulaw")
            if ringTonesList then
                data.actRingTone   = userFeatures:get("RING")
                data.ringTonesList = ringTonesList
                respBody           = __features_make_ringtones_list(data, featParams)
            else
                logger:debug(errMsg)
                respBody = __features_error_content(featParams, "UsersShareNotMounted")
            end
        elseif "STATUS" == reqFeature or "STATUS_UNTIL" == reqFeature then
            local list, idx, tag, val = _buildStatusDataListValues(reqValue, userFeatures)
            data.data      = list
            data.dfltField = idx
            data.tag       = tag
            data.val       = val
            respBody       = __features_make_status_data_list(data, featParams)
        end
    elseif "value" == reqType then  -- menus to manage feature value
        local data = { callType = reqCallType }
        if "FCF" == reqFeature or "UCF" == reqFeature or "BCF" == reqFeature then
            data.name  = reqFeature
            data.value = reqValue or userFeatures:getValue(reqFeature) 
            if not data.value or "" == data.value then
                data.value = "VOICEMAIL"
            end
        end
        respBody = __feature_input_value(data, featParams)
    elseif "submenu" == reqType then
        local data = { callType = reqCallType }
        if "FCF" == reqFeature or "UCF" == reqFeature or "BCF" == reqFeature then
            data.name  = reqFeature
            data.value = userFeatures:getValue(reqFeature)
            respBody   = __feature_select_action(data, featParams)
        end
    elseif "types" == reqType then
	    respBody = __features_select_calltype(featParams)
    end
    
    return respBody
end

--- Processing vmaccess request from the phone
-- @param agent		A table with general device information
-- @param request	A table with requested parameters
-- @return XML content with voicemailbox data
function vmaccess_process(agent, request)
	if not agent or not request then
		return false
	end
    
    local logger = logger	
	local core   = require"common.core"
    local h      = require"common.helper"
	local vmApp	 = require"common.voiceMailApplication"
	
	if "Wildix" ~= agent["vendor"] or "WPR2" ~= agent["vendor:additional"] then
		logger:error("Request got from not supported device")
		return false
	end
		
	-- load translation object
	local trobj = require"common.translation".translation
	
	local foreign_vm = false
	if request["user"] then
		if "" ~= request["user"] and tonumber(request["user"]) then
			foreign_vm = true
		else
			return __vm_error_content("permissionDenied", trobj)
		end
	end

	-- get voicemail box user information
	local user_info = __get_user_data(agent["mac"])
	if request["user"] == user_info["telephoneNumber"] then
		foreign_vm = false
	end
    local sharedVM = __getUserFeatures(user_info and user_info.uid or "", "SHAREDVM")
    if sharedVM and not foreign_vm then
        user_info = __get_user_data(nil, sharedVM)
    end

	if not user_info then
		return __vm_error_content("vmemptylist", trobj)
	end

	-- check user permissions
	if "yes" ~= require"common.permissions":check(user_info["owner"] or "", "canuse", "VOICEMAIL", nil, user_info["licenseType"]) then
		return __vm_error_content("permissionDenied", trobj)
	end

	local permit, user_num
	if foreign_vm then
		local dst_attrs = __get_user_data(nil, request["user"])
		if not dst_attrs then 
			return __vm_error_content("pbnouser", trobj)
		end
		permit_view_group = "yes" == require"common.permissions":check(user_info["owner"] or "", "yesno", "SEE", dst_attrs["owner"], user_info["licenseType"])
		permit_see_voicemail = "yes" == require"common.permissions":check(user_info["owner"] or "", "yesno", "SEE_VOICEMAIL", dst_attrs["owner"], user_info["licenseType"])
		permit = permit_view_group and permit_see_voicemail
		user_num = request["user"]
	else
		user_num = user_info["telephoneNumber"]
		permit = true
	end
	if not permit then
		return __vm_error_content("permissionDenied", trobj)
	end

	local user_id   = user_info["uid"]
	local user_lang = user_info["preferredLanguage"]
	if foreign_vm then
		logger:debug("User %q(%s) with uid %s to voicemail of user %s", user_info["telephoneNumber"], user_lang, user_id, user_num)
	else
		logger:debug("User %q(%s) with uid: %s", user_num, user_lang, user_id)
	end

	-- set translation object language	
	trobj:set_language(user_lang or "en")

	-- get user dateTime settings
    local dtSettings = __getUserFeatures(user_id, "DATETIME") or {}
    if dtSettings.format then
        dtSettings.type = "php"
    end

    -- get local time zone for device
    dtSettings.timeZone = __getDevData(agent.mac)
	
	-- table for used params
	local vm_params	= {
		["trobj"]      = trobj,
		["rproto"]     = agent["ssl"] == "on" and "https" or "http",
		["raddress"]   = agent["registrar"] or "",
		["rport"]      = agent["registrar:port"] or "",
		["model"]      = agent["model"],
		["dfamily"]    = agent["vendor:additional"],
		["dtSettings"] = dtSettings,
		["user"]       = foreign_vm and request["user"] or nil
	}

	-- get vm messages
	local xml_cont = {}
	local err      = nil
	local app_params = {["mailbox"] = user_num, ["limit"] = 200}
	local vm_cont, err = vmApp.getMsgInfo(app_params)
	if not vm_cont then
		return __vm_error_content(err, trobj)
	end

	if request["id"] then
	--- processing delete request
		if "delete" == request["menu"] then
			if vm_cont[request["id"]] then
				vmApp.deleteMsg(vm_cont[request["id"]])
				vmApp.updateVM(user_num)
				vm_cont, err = vmApp.getMsgInfo(app_params)
				if not vm_cont then
					return __vm_error_content(err, trobj)
				end
			end
			request["menu"] = "main"
	--- process missed call
		elseif "proceed" == request["menu"] then
			if vm_cont[request["id"]] then
				vmApp.markAsReadMsg({["mailbox"] = user_num, ["msgId"] = request["id"]})
				vmApp.updateVM(user_num)
				vm_cont, err = vmApp.getMsgInfo(app_params)
				if not vm_cont then
					return __vm_error_content(err, trobj)
				end
			end
			request["menu"] = "main"
	--- processing Detail menu
		elseif "detail" == request["menu"] then
			if vm_cont[request["id"]] then
				if request["forward"]  and "1" == request["forward"] then
					if request["input"] then
						xml_cont, err = __vm_process_forward(vm_params, vm_cont, request, vmApp)
					else
						xml_cont, err = __vm_make_forward_menu(vm_params, vm_cont[request["id"]]["msgId"])
					end
				else
					xml_cont, err = __vm_make_detail_menu(vm_params, vm_cont[request["id"]])
				end
			else
				request["menu"] = "main"
			end	
		end
	end
	--- processing Main menu
	if "main" == request["menu"] then
		-- get voicemail messages info
		xml_cont, err = __vm_make_content(vm_params, vm_cont)
	end
	
	if not xml_cont or 0 == #xml_cont then
		logger:error("Some error found in vmaccess processing:", err)
		return __vm_error_content("vmemptylist", trobj)
	end

	logger:info("Sending vmaccess menu to %s", agent.model)

	return xml_cont
end

--- Authorization procedure for some requests
function authorization_process(agent, request, service, options)
    local authContent = nil     -- XML body content if required
    local passLen     = 5      -- predefined password length for XML browser authorization
    
    
    --- check autorization for service
    if "vmaccess" ~= service then
        -- check requested service: apply only for voicemail menu
        return authContent
    end

    local user_info = __get_user_data(agent["mac"])
    if not user_info then
        local trobj = require"common.translation".translation
        return __vm_error_content("pbnouser", trobj)
    end
    if "yes" == require"common.permissions":check(user_info["owner"] or "", "canuse", "VOICEMAIL_WITHOUT_PIN", nil, user_info["licenseType"]) then
        return authContent
    end
    
    --- an authorization main block
    local lock_options = {passLen = passLen, lockTimeout = options.lockTimeout}
    repeat
        -- initialize auth lock object
        local reqEnv       = {
             remoteIP  = agent.address
            ,ua        = string.format("%s %s %s %s", agent.vendor, string.upper(agent.model), agent["fw:version"], agent.mac)
            ,uri       = request.uri
            ,source    = __get_user_by_mac(agent.mac)
        } 
        local authLockObj, errMsg = require"common.userAuth.userAuthDevLock":new(reqEnv, lock_options)
        if not authLockObj then
            logger:error(errMsg)
            break
        end
        
        local user            = reqEnv.source
        local pass            = request.pin
        local authRes, errMsg = authLockObj:authenticate(user, pass) 
        if not authRes then
            logger:warning("Authentification failed for service (%s): %s.", service, errMsg)
            
            -- init return link
            local registrar   = agent["registrar"]
            local req_proto   = "on" == agent["ssl"] and "https" or "http"
            local req_port    = agent["registrar:port"]
            local return_link = string.format("%s://%s:%s", req_proto, registrar, req_port)
            
            -- table for used params
            local auth_params = {
                 ["trobj"]    = require"common.translation".translation
                ,["rlink"]    = return_link
                ,["rquery"]   = request["uri"]
                ,["dfamily"]  = agent["vendor:additional"]
            }
            -- return query normalization
            auth_params.rquery = auth_params.rquery:gsub("%&pin=%d+", "")
            
            local authLockTimeout = authLockObj:getSessionFreezeTime()
            if 0 < authLockTimeout then
                logger:debug("Processing an authorization lock page for a service")
                authContent = __auth_lock_error(auth_params, authLockTimeout)
            else
                logger:debug("Processing an authorization menu for the service (%s)...", service)
                authContent = __auth_make_content(auth_params, passLen)
            end
            break
        end
        
        local cachedTime = authLockObj:getCredentialsTime() 
        if 0 < cachedTime then
            logger:debug("Authentification success by an actual cached password for %d seconds", cachedTime)
        end
        
    until true
    
    return authContent
end

--- Processing contect center request from the phone
-- @param agent		A table with general device information
-- @param request	A table with requested parameters
function callgroup_process(agent, request)
    if not agent or not request then
        return false
    end
    local core   = require"common.core"
    local h      = require"common.helper"
    local http   = require"core.httpClient"
    local utils  = require"core.pbxutils"
    local logger = logger
    
    if "Wildix" ~= agent["vendor"] then
        logger:error("Request got from not supported device")
        return false
    end
    
    -- load translation object
    local trobj = require"common.translation".translation
    
    -- get device user information
    local dev_family = agent["vendor:additional"] or ""
    local user_info = __get_user_data(agent["mac"], dev_family:match("^W%-AIR") and request["extension"])
    if not user_info then
        logger:error("User info not found")
        return false
    end
    local user_id   = user_info["uid"]
    local user_lang = user_info["preferredLanguage"]
    local username  = user_info["telephoneNumber"]

    logger:debug("User %q(%s) with uid: %s", username, user_lang, user_id)
    
    -- set translation object language  
    trobj:set_language(user_lang or "en")
    
    -- table of the used parameters
    local cg_params = {
                ["trobj"]    = trobj,
                ["rproto"]   = agent["ssl"] == "on" and "https" or "http",  -- used by WPR2
                ["raddress"] = agent["registrar"] or "",                    -- used by WPR2
                ["rport"]    = agent["registrar:port"] or "",               -- used by WPR2
                ["dfamily"]  = agent["vendor:additional"],
                ["dmodel"]  = agent["model"],
    }

    -- check supported devices
    local fw_ver = h.parse_firmware(agent["fw:version"])
    if "WP" == dev_family and  "wp600" ~= agent["model"] then
        -- checking firmware for WP family (exeption WP600)
        if fw_ver and 50 > fw_ver.major then
            fw_ver = false      -- mark as unsupported firmware
        end
    end
    
    if not fw_ver then
        logger:warning("Device firmware is old. Need upgrading ...")
        return __pb_error_content("oldfw", dev_family, trobj)
    end

    local userFeatures = require"common.userFeatures.userFeatures":new(user_id)

    if userFeatures then
        userFeatures:load()
    else
        return false
    end

    local connection = require"core.wsqlite":new("wmsdb")
    local sql = "SELECT id, call_groupname FROM call_groups;"
    local pmIterator = connection:rows(sql)
    if "function" ~= type(pmIterator) then
        logger:error("DB error")
        return __callgroup_error(cg_params, "features.Error")
    end
    local cgroups = {}
    local cg_wl_enable, allowed_callgroups = userFeatures:getAllowedCallgroups()
    for cgroup_id, cgroup_name in pmIterator do
        if not cg_wl_enable or utils.in_table(tostring(cgroup_id), allowed_callgroups) then
            cgroups[cgroup_id] = cgroup_name
        end
    end
    connection:close()

    function getMemberStatus()
        local connection = require"core.wsqlite":new("wmsdb")
        local sql = string.format("SELECT call_group_id FROM call_groups_members WHERE group_member='%s@internalcalls';", username)
        local pmIterator = connection:rows(sql)
        if "function" ~= type(pmIterator) then
            logger:error("DB error")
            return false
        end
        local membership = {}
        for cgroup_id in pmIterator do
           membership[cgroup_id] = "static"
        end
        connection:close()

        userFeatures:load()
        local u_callcenter = userFeatures:get("CALLCENTER")
        if "1" == u_callcenter["enable"] then
            for k, v in pairs(u_callcenter['callGroups']) do 
                if tonumber(v['id']) then
                    membership[tonumber(v['id'])] = "dynamic"
                end
            end
        end

        return membership, u_callcenter["enable"]
    end


    function build_records(membership, cgroups)
        local params = {}
        for cgroup_id, cgroup_name in pairs(cgroups) do
            local param = {}
            if "static" == membership[cgroup_id] then
                param["username"] = string.format("%s(%s)", cgroup_name, trobj:_t("xml.static"))
                param["presence"] = "online"
                param["membership"] = "static"
            elseif "dynamic" == membership[cgroup_id] then
                param["username"] = cgroup_name
                param["membership"] = "dynamic"
                param["presence"] = "online"
            else
                param["username"] = cgroup_name
                param["membership"] = "dynamic"
                param["presence"] = "offline"
            end
            param["entryuid"] = tostring(cgroup_id)
            param["extension"] = tostring(cgroup_id)
            table.insert(params, param)
        end

        local sortf = function(first, second)
            if first.membership ~= second.membership then
                return first.membership < second.membership
            else
                if first.presence ~= second.presence then
                    return first.presence > second.presence
                else
                    return first.username < second.username
                end
            end
        end

        table.sort(params, sortf)

        return params
    end


    function sign_switch()
        local u_callcenter = userFeatures:get("CALLCENTER")
        u_callcenter["enable"] = u_callcenter["enable"] == "1" and "0" or "1"
        userFeatures:set("CALLCENTER", u_callcenter)
        userFeatures:saveData()
    end


    function membership_update(membership)
        local i = 1
        local tmp = {}
        for cgroup_id, membership_type in pairs(membership) do
            if "dynamic" == membership_type then
                tmp[tostring(i)] = {id = cgroup_id}
                i = i + 1
            end
        end
        local u_callcenter = userFeatures:get("CALLCENTER")
        u_callcenter['callGroups'] = tmp
        userFeatures:set("CALLCENTER", u_callcenter)
        userFeatures:saveData()
    end

    local membership
    membership, cg_params["sign_status"] = getMemberStatus()
    if not membership then
        logger:error("No results")
        return false
    end
    
    cg_params["records"] = build_records(membership, cgroups)


    local output = nil
    if "toogle" == request["menu"] then
        if "true" == request["error"] then
            logger:debug("Error screen. Static groups cannot be deselected")
            output = __callgroup_error(cg_params, "xml.staticgrouperr")
        else
            if request['entryuid'] and "" ~= request['entryuid'] then
                ami = require"core.AMIClient":new()
                if membership[tonumber(request['entryuid'])] and "dynamic" == membership[tonumber(request['entryuid'])] then
                    logger:debug(string.format("Remove %s from queue %s", username, request['entryuid']))
                    ami:removeQueueMember(request['entryuid'], username)
                    membership[tonumber(request['entryuid'])] = nil
                else
                    logger:debug(string.format("Add %s to queue %s", username, request['entryuid']))
                    ami:addQueueMember(request['entryuid'], username)
                    membership[tonumber(request['entryuid'])] = "dynamic"
                end
                membership_update(membership)
                membership, cg_params["sign_status"] = getMemberStatus()
                if not membership then
                    logger:error("No results")
                    return false
                end
                cg_params["records"] = build_records(membership, cgroups)
            end
            output = __callgroups_wpr2(cg_params, dev_family)
        end
    elseif "signall" == request["menu"] then
        ami = require"core.AMIClient":new()
        if "0" == cg_params["sign_status"] then
            logger:debug(string.format("Sign-on %s", username)) 
            sign_switch()
            membership, _ = getMemberStatus()
            for cgroup_id, membership_type in pairs(membership) do
                if "dynamic" == membership_type then
                    ami:addQueueMember(cgroup_id, username)
                end
            end
            cg_params["sign_status"] = "1"
        elseif "1" == cg_params["sign_status"] then
            logger:debug(string.format("Sign-out %s", username)) 
            for cgroup_id, membership_type in pairs(membership) do
                if "dynamic" == membership_type then
                    ami:removeQueueMember(cgroup_id, username)
                end
            end
            sign_switch()
            cg_params["sign_status"] = "0"
        end

        membership, _ = getMemberStatus()
        if not membership then
            logger:error("No results")
            return false
        end
        cg_params["records"] = build_records(membership, cgroups)
        output = __callgroups_wpr2(cg_params, dev_family)
    else
        logger:info("Sending callgroups to %s (%s - %s)", username, agent["model"], agent["vendor:additional"])
        output = __callgroups_wpr2(cg_params, dev_family)                    
    end

    return output
end

--- Processing identities request from the phone
-- @param agent		A table with general device information
-- @param request	A table with requested parameters
function identities_process(agent, request)
    if not agent or not request then
        return false
    end
    local core   = require"common.core"
    local h      = require"common.helper"
    local http   = require"core.httpClient"
    local logger = logger

    if "Wildix" ~= agent["vendor"] then
        logger:error("Request got from not supported device")
        return false
    end

    -- load translation object
    local trobj = require"common.translation".translation

    -- get device user information
    local dev_family = agent["vendor:additional"] or ""
    local user_info = __get_user_data(agent["mac"])
    if not user_info then
        logger:error("User info not found")
        return false
    end
    local user_id   = user_info["uid"]
    local user_lang = user_info["preferredLanguage"]
    local username  = user_info["telephoneNumber"]

    logger:debug("User %q(%s) with uid: %s", username, user_lang, user_id)

    -- set translation object language
    trobj:set_language(user_lang or "en")

    -- table of the used parameters
    local cg_params = {
                ["trobj"]    = trobj,
                ["rproto"]   = agent["ssl"] == "on" and "https" or "http",  -- used by WPR2
                ["raddress"] = agent["registrar"] or "",                    -- used by WPR2
                ["rport"]    = agent["registrar:port"] or "",               -- used by WPR2
                ["dfamily"]  = agent["vendor:additional"],
                ["dmodel"]  = agent["model"],
    }

    -- check supported devices
    local fw_ver = h.parse_firmware(agent["fw:version"])
    if "WP" == dev_family and  "wp600" ~= agent["model"] then
        -- checking firmware for WP family (exeption WP600)
        if fw_ver and 50 > fw_ver.major then
            fw_ver = false      -- mark as unsupported firmware
        end
    end

    if not fw_ver then
        logger:warning("Device firmware is old. Need upgrading ...")
        return __pb_error_content("oldfw", dev_family, trobj)
    end


    local userFeatures = require"common.userFeatures.userFeatures":new(user_id)

    if userFeatures then
        userFeatures:load()
    else
        return false
    end

    function build_records(identities, current_identity, user)
        local connection = require"core.wsqlite":new("phonebook")

        local function getIdentityName(num)
            local sql = string.format("SELECT username FROM cti_contacts WHERE phonebook_id IN (SELECT rid FROM cti_phonebooks WHERE type='3') AND extension='%s'", num)
            local cursor = connection:execute(sql)
            if not cursor or type(cursor) == "boolean" then
                logger:error("DB query failed")
                return ""
            end
            local name = cursor:fetch(); cursor:close()
            return name
        end

        local params = { [1] = { username = string.format("%s %s", user["telephoneNumber"], user["cn"] or ""),
                                 identity = user["telephoneNumber"], presence = "" == current_identity and "online" or "offline"} }
        for entryuid, identity in pairs(identities) do
            local param = {}
            param["username"] = string.format("%s %s", identity, getIdentityName(identity) or "")
            param["identity"] = identity
            param["presence"] = identity == current_identity and "online" or "offline"
            table.insert(params, param)
        end
        connection:close()

        return params
    end

    local available_identities = userFeatures:getCustomIdentities() or {}
    local current_identity = userFeatures:getValue("CURRENT_IDENTITY") or ""

    cg_params["records"] = build_records(available_identities, current_identity, user_info)

    local output = nil
    if request['identity'] then
        if username == request['identity'] then
            userFeatures:setValueData("CURRENT_IDENTITY", "")
        elseif require"core.pbxutils".in_table(request['identity'], available_identities) then
            if not __get_user_data(nil, request['identity']) then
                output = __identity_error(cg_params, "xml.pbnouser")
                return output
            end
            userFeatures:setValueData("CURRENT_IDENTITY", request['identity'])
        else
            logger:error("Identities list of the user doesn't contain the requested identity:", request['identity'])
            output = __identity_error(cg_params, "xml.vmFwdError")
            return output
        end

        local rv, err = userFeatures:saveData()
        if rv then
            package.path = package.path .. ";/var/lib/callweaver/dialplan/?.lua"
            require"utils":notify_peer(username, 0)
        else
            logger.error("Identity has not changed:", err or "")
            output = __identity_error(cg_params, "xml.vmFwdError")
            return output
        end

        available_identities = userFeatures:getCustomIdentities() or {}
        current_identity = userFeatures:getValue("CURRENT_IDENTITY") or ""
        cg_params["records"] = build_records(available_identities, current_identity, user_info)
        output = __identities_wpr2(cg_params, dev_family)
    else
        logger:info("Sending identities list to %s (%s - %s)", username, agent["model"], agent["vendor:additional"])
        output = __identities_wpr2(cg_params, dev_family)
    end

    return output
end
